/*
 * Animate! Animation module for AfterStep
 * 
 * Copyright (c) 1996 Alfredo Kengi Kojima (kojima@inf.ufrgs.br)
 * 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
/*
 * changes needed in afterstep
 * broadcast(M_DEICONIFY) must include the icon position>
 */

#include <stdlib.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <signal.h>
#include "../../configure.h"
#include "../../afterstep/module.h"
#include "../../lib/aftersteplib.h"

/* Animation granularity. How many iterations use for the animation.  */
#define ANIM_ITERATIONS 24

/* delay for each iteration of the animation in ms */
#define ANIM_DELAY	1

/* delay for each iteration of the close animation in ms */
#define ANIM_DELAY2	20

Display *dpy;
Window root;
GC DrawGC;
char *ProgName;
int Channel[2];

void Loop();

/*
 * This makes a iconify/deiconify animation for a window.
 * Parameters specify the position and the size of the initial window
 * and the final window
 */
void AnimateResize(int x, int y, int w, int h, int fx, int fy, int fw, int fh)
{
    float cx, cy, cw, ch;
    float xstep, ystep, wstep, hstep;
    int i;
    
    xstep = (float)(fx-x)/ANIM_ITERATIONS;
    ystep = (float)(fy-y)/ANIM_ITERATIONS;
    wstep = (float)(fw-w)/ANIM_ITERATIONS;
    hstep = (float)(fh-h)/ANIM_ITERATIONS;

    cx = (float)x;
    cy = (float)y;
    cw = (float)w;
    ch = (float)h;
    for (i=0; i<ANIM_ITERATIONS; i++) {
	XDrawRectangle(dpy, root, DrawGC, (int)cx, (int)cy, (int)cw, (int)ch);
	XFlush(dpy);
	usleep(ANIM_DELAY*1000);
	XDrawRectangle(dpy, root, DrawGC, (int)cx, (int)cy, (int)cw, (int)ch);
	cx+=xstep;
	cy+=ystep;
	cw+=wstep;
	ch+=hstep;
    }
    XFlush(dpy);    
}

#if 0
/*
 * This makes a animation that looks like that light effect
 * when you turn off an old TV.
 * Used for window destruction
 */
void AnimateClose(int x, int y, int w, int h)
{
    int i, step;
    
    if (h>4) {
	step = h*4/ANIM_ITERATIONS;
	if (step==0) {
	    step = 2;
	}
	for (i=h; i>=2; i-=step) {
	    XDrawRectangle(dpy, root, DrawGC, x, y, w, i);
	    XFlush(dpy);
	    usleep(ANIM_DELAY2*600);
	    XDrawRectangle(dpy, root, DrawGC, x, y, w, i);
	    y+=step/2;
	}
    }
    if (w<2) return;
    step = w*4/ANIM_ITERATIONS;
    if (step==0) {
	step = 2;
    }
    for (i=w; i>=0; i-=step) {
	XDrawRectangle(dpy, root, DrawGC, x, y, i, 2);
	XFlush(dpy);
	usleep(ANIM_DELAY2*1000);
	XDrawRectangle(dpy, root, DrawGC, x, y, i, 2);
	x+=step/2;
    }
    usleep(100000);    
    XFlush(dpy);
}
#endif

void DeadPipe(int foo)
{
    exit(0);
}

int main(int argc, char **argv)
{
    char *temp;
    XGCValues gcv;

    ProgName = argv[0];
    if ((argc != 6)&&(argc != 7)) {
	fprintf(stderr,"%s should only be executed by afterstep!\n",ProgName);
	exit(1);
    }

    /* Dead pipe == AS died */
    signal (SIGPIPE, DeadPipe);
    
    Channel[0] = atoi(argv[1]);
    Channel[1] = atoi(argv[2]);
    
    dpy = XOpenDisplay("");
    if (dpy==NULL) {
	fprintf(stderr,"%s: could not open display\n",ProgName);
	exit(1);
    }
    root = DefaultRootWindow(dpy);
    gcv.line_width = 2;
    gcv.function = GXequiv;
    gcv.foreground = BlackPixel(dpy,DefaultScreen(dpy));
    DrawGC=XCreateGC(dpy, root, GCFunction|GCForeground|GCLineWidth, &gcv);
    SendText(Channel,"Nop",0);
    Loop();
}

void GetWindowSize(Window win, int *x, int *y, int *w, int *h)
{
    Window root_r;
    int bw_r, d_r;
    
    XGetGeometry(dpy, win, &root_r, x, y, w, h, &bw_r, &d_r);
}

/*
 * Wait for some event like iconify, deiconify and stuff.
 * 
 */
void Loop()
{
    unsigned long header[3], *body;
    int c, x0, y0, w0, h0, xf, yf, wf, hf;
    int iconify;
    Window win;
    
    while (1) {
	c=ReadASPacket(Channel[1], header, &body);
	if (c>0) {
	    switch (header[1]) {
	     case M_DEICONIFY:
		win = body[1];
		x0 = body[3];
		y0 = body[4];

		GetWindowSize(win, &xf, &yf, &wf, &hf);		
		AnimateResize(x0, y0, 64, 64, xf, yf, wf, hf);
		break;

	     case M_ICONIFY:
		xf = body[3];
		yf = body[4];
		if (xf<=-10000) break;
		wf = 64;
		hf = 64;
		free(body);

		c=ReadASPacket(Channel[1], header, &body);
		if (c<=0) break;
		if (header[1]!=M_CONFIGURE_WINDOW) break;
		x0 = body[3];
		y0 = body[4];
		w0 = body[5];
		h0 = body[6];
		AnimateResize(x0, y0, w0, h0, xf, yf, wf, hf);
		break;
	    }
	    free(body);	    
	}
    }
}
