#!
# Save AfterStep configuration file 
#
# Copyright (c) 1996 Alfredo K. Kojima
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

#---------------------------------------------------------------
# Saves configuration for AfterStep
#---------------------------------------------------------------
proc save_asconfig {} {
	global ASConfig Windows KeyBind Wharf MouseBind Functions ASMenus
	if [file exists $ASConfig(file)] {
		set fname [glob $ASConfig(file)]
		exec /bin/mv $fname $fname~
	} else {
		set fname $ASConfig(file)
	}
	proc eputs {f name arg} {
		if {$arg!=""} {
			puts $f "$name $arg"
		}
	}	
	proc bputs {f name var} {
		if {$var} {
			puts $f $name
		}
	}
	proc nbputs {f name var} {
		if {! $var} {
			puts $f $name
		}
	}	
	set fl [open $fname w+]
	puts $fl "########## AfterStep Configuration File ##########"
	puts $fl "#\n# This file was automatically generated by ascp.\n#"
	puts $fl "# Report problems to kojima@inf.ufrgs.br"

	puts $fl "\n#------------------==  Paths Section  ==-----------------#"
	puts $fl "IconPath $ASConfig(icoPath)"
	puts $fl "PixmapPath $ASConfig(pixPath)"
	puts $fl "ModulePath $ASConfig(modPath)"
	
	puts $fl "\n#------------== Color and Font Section ==----------------#"
	puts $fl "StdForeColor $ASConfig(UFScolor)"
	puts $fl "StdBackColor $ASConfig(UBScolor)"
	puts $fl "HiForeColor  $ASConfig(FFScolor)"
	puts $fl "HiBackColor  $ASConfig(FBScolor)"
	puts $fl "MenuForeColor $ASConfig(MFScolor)"
	puts $fl "MenuBackColor $ASConfig(MBScolor)"
	puts $fl "ButtonBgColor $ASConfig(CBScolor)\n"	
	puts $fl "Font 		$ASConfig(Ifont)"
	puts $fl "WindowFont 	$ASConfig(Tfont)"
	puts $fl "IconFont 	$ASConfig(Cfont)"

	puts $fl "\n#----------------== Gradient Section ==------------------#"
	puts $fl "TextureTypes $ASConfig(FTexture) $ASConfig(UTexture) $ASConfig(STexture) $ASConfig(MTexture) $ASConfig(ITexture)"
	puts $fl "TextureMaxColors $ASConfig(Fmaxcols) $ASConfig(Umaxcols) $ASConfig(Smaxcols) $ASConfig(Mmaxcols) $ASConfig(Imaxcols)"
	puts $fl "TitleTextureColor $ASConfig(FFcolor) $ASConfig(FTcolor)"
	puts $fl "UTitleTextureColor $ASConfig(UFcolor) $ASConfig(UTcolor)"
	puts $fl "STitleTextureColor $ASConfig(SFcolor) $ASConfig(STcolor)"
	puts $fl "MTitleTextureColor $ASConfig(MFcolor) $ASConfig(MTcolor)"
	puts $fl "MenuTextureColor $ASConfig(IFcolor) $ASConfig(ITcolor)\n"
	
	eputs $fl "TitlePixmap" $ASConfig(Fpixmap)
	eputs $fl "UTitlePixmap" $ASConfig(Upixmap)
	eputs $fl "STitlePixmap" $ASConfig(Spixmap)
#	eputs $fl "MTitlePixmap" $ASConfig(Mpixmap)
#	eputs $fl "MenuPixmap" $ASConfig(Ipixmap)

	if {$ASConfig(GText)} {
		puts $fl "GradientText"
		puts $fl "TextGradientColor $ASConfig(GTFcolor) $ASConfig(GTTcolor)"
	}
	
	puts $fl "ButtonTextureType $ASConfig(CTexture)"
	puts $fl "ButtonMaxColors $ASConfig(Cmaxcols)"
	eputs $fl "ButtonPixmap" $ASConfig(Cpixmap)
	puts $fl "ButtonTextureColor $ASConfig(CFcolor) $ASConfig(CTcolor)"
	
	puts $fl "\n#----------------==  Desktop Section  ==-----------------#"
	switch  $ASConfig(IconBox) {
		0 {puts $fl "IconBox 0 -64 -1 -1"}
		1 {puts $fl "IconBox 0 0 -1 64"}
		2 {puts $fl "IconBox 0 0 64 -1"}
		{puts $fl "IconBox -64 0 0 -1"}
	}
	bputs $fl "IconTitle" $ASConfig(IconT)
	bputs $fl "StickyIcons" $ASConfig(SticI)
	bputs $fl "StubbornIcons" $ASConfig(StubI)
	bputs $fl "StubbornIconPlacement" $ASConfig(StubIP)
	set hesc $ASConfig(EScroll,h)
	set vesc $ASConfig(EScroll,v)
	if {$ASConfig(EScroll,hw)} {set hesc [expr $hesc*1000]}
	if {$ASConfig(EScroll,vw)} {set vesc [expr $vesc*1000]}
	puts $fl "EdgeScroll  $hesc $vesc"
	puts $fl "EdgeResistance $ASConfig(ERes,s) $ASConfig(ERes,m)"
	puts $fl "DesktopSize $ASConfig(DTSize,v)x$ASConfig(DTSize,h)"
	puts $fl "DesktopScale $ASConfig(DTScale)\n"
	puts $fl "NoBoundaryWidth 0"
	puts $fl "BoundaryWidth 8"
	
	puts $fl "\n#--------------== General Style Section ==---------------#"
	bputs $fl "TexturedHandle" $ASConfig(THandle)
	bputs $fl "TitlebarNoPush" $ASConfig(TBnpush)
	bputs $fl "IconTitle" $ASConfig(IconT)
	puts $fl "TitleTextAlign $ASConfig(talig)"
	
	puts $fl "\n#-------------== Window Handling Section ==--------------#"
	bputs $fl "RandomPlacement" $ASConfig(RandomP)
	bputs $fl "SmartPlacement" $ASConfig(SmartP)
	bputs $fl "ClickToFocus" $ASConfig(ClickF)
	bputs $fl "SloppyFocus" $ASConfig(SlopF)
	bputs $fl "CirculateSkipIcons" $ASConfig(CSkipI)	
	bputs $fl "AppsBackingStore" $ASConfig(ABackS)
	bputs $fl "BackingStore" $ASConfig(BackS)
	bputs $fl "SaveUnders" $ASConfig(SaveU)
	bputs $fl "NoPPosition" $ASConfig(NoPos)
	bputs $fl "DecorateTransients" $ASConfig(DTrans)
	puts $fl "ClickTime $ASConfig(CTime)"
	puts $fl "OpaqueMove $ASConfig(OMove)"
	bputs $fl "OpaqueResize" $ASConfig(ORes)
	if {$ASConfig(AutoOn)>0} {puts $fl "Module Auto $ASConfig(Auto)"}
	
	puts $fl "\n#--------------== Window Style Section ==----------------#"
	foreach win $Windows(idx) {
		set line ""
		
		if {$Windows($win,icon)!={}} {
			set line "Icon $Windows($win,icon),"
		}
		if {$Windows($win,sticky)} {
			set line "$line Sticky,"
		}
		if {$Windows($win,notitle)} {
			set line "$line NoTitle,"
		}
		if {$Windows($win,nohandle)} {
			set line "$line NoHandles,"
		}
		if {$Windows($win,ontop)} {
			set line "$line StaysOnTop,"
		}
		if {$Windows($win,skip)} {
			set line "$line WindowListSkip,"
		}
		if {$Windows($win,nofocus)} {
			set line "$line NoFocus,"
		}
		if {$Windows($win,startico)} {
			set line "$line StartIconic,"
		}
		if {$line==""} continue

		
		puts $fl "Style \"$win\" [string trimright $line ,]"
	}
	
	puts $fl "\n#-----------== Function Definition Section ==------------#"
	foreach i $Functions(idx) {
		puts $fl "Function \"$i\""
		foreach j [split [string trim $Functions($i)] \n] {
			puts $fl "\t$j"
		}
		puts $fl "EndFunction\n"
	}
	
	puts $fl "\n#-------------== Menu Definition Section ==--------------#"
	foreach i $ASMenus(idx) {
		puts $fl "PopUp $i"
		if {$ASMenus($i,0,t)!={}} {
			puts $fl "\tTitle \"$ASMenus($i,0,t)\""
		}
		foreach j $ASMenus($i,idx) {
		    	set j1 ""
			set j2 ""
			set j3 ""
			set tmp ""
			scan $ASMenus($i,$j,c) "%s %\[^\n\]" j1 tmp
			puts $fl "\t$j1 \"$ASMenus($i,$j,t)\" $tmp"
		}
		puts $fl "EndPopUp\n"
	}
	
	puts $fl "\n#------------------== Audio Section ==-------------------#"
	puts $fl "*AudioDir $ASConfig(AuDir)"	
	if $ASConfig(AuOn) {
		puts $fl "Module Audio"
		puts $fl "*AudioPlayCmd $ASConfig(AuCmd)"
		puts $fl "*AudioDelay $ASConfig(AuDelay)"		
		for {set i 0} {$i < 18} {incr i} {
			if {"$ASConfig(AuSnd,$i)"!=""} {
				puts $fl "*Audio $ASConfig(AuEv,$i) \
					$ASConfig(AuSnd,$i)"
			}
		}
	}	
	puts $fl "\n#----------------== Wharf Section ==---------------------#"
	nbputs $fl "Module Wharf"  $ASConfig(whdisable)
	bputs $fl "*WharfAnimate" $ASConfig(whanim)
	bputs $fl "*WharfAnimateMain" $ASConfig(whanimm)
	nbputs $fl "*WharfNoBorder" $ASConfig(whborder)
	bputs $fl "*WharfFullPush" $ASConfig(whfpush)
	if {!$ASConfig(WTexture)} {
		puts $fl "*WharfTextureType 255"
	} else {
		puts $fl "*WharfTextureType $ASConfig(WTexture)"
	}
	if {$ASConfig(whdir)==0} {
		puts $fl "*WharfColumns $ASConfig(wsize)"
	} else {
		puts $fl "*WharfRows $ASConfig(wsize)"
	} 
	switch $ASConfig(whcorner) {
		ne {set x "-1"; set y "+1"}
		nw {set x "+1"; set y "+1"}
		sw {set x "+1"; set y "-1"}
		se {set x "-1"; set y "-1"}
	}
	puts $fl "*WharfGeometry $x$y"
	puts $fl "*WharfPixmap $ASConfig(Wpixmap)"
	puts $fl "*WharfTextureColor  $ASConfig(WFcolor) $ASConfig(WTcolor)"
	puts $fl "*WharfBgColor $ASConfig(WBScolor)"
	puts $fl "*WharfMaxColors $ASConfig(Wmaxcols)"
	
	foreach i $Wharf(idx) {
		if {[lindex $Wharf(cmd,$i) 0]=="EndFolder"} {
			puts $fl "*Wharf ~Folder\n"
			continue
		} 
		if {[lindex $Wharf(cmd,$i) 0]=="Folder"} {
			puts $fl "\n*Wharf \"$i\" $Wharf(ico,$i) Folder"
			continue
		}
		puts $fl "*Wharf \"$i\" $Wharf(ico,$i) $Wharf(cmd,$i)"
	}
	puts $fl "\n#--------------== Mouse Binding Section ==---------------#"
	foreach i $MouseBind(idx) {
		puts $fl "Mouse\t$MouseBind($i,but)\t$MouseBind($i,ctx)\t$MouseBind($i,mod)\t$MouseBind($i,cmd)"
	}
	puts $fl "\n#------------== Keyboard Binding Section ==--------------#"
	foreach i $KeyBind(idx) {
		puts $fl "Key\t$KeyBind($i,key)\t$KeyBind($i,ctx)\t$KeyBind($i,mod)\t$KeyBind($i,cmd)"
	}
	puts $fl "\n#-------------== Unknown Stuff Section ==----------------#"
	foreach i $ASConfig(MiscMod) {puts $fl "Module $i"}
	puts $fl $ASConfig(dunno)
	
	puts $fl "#EOF"
	close $fl
}
