#include <stdio.h>
#include <string.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>

Display	*display;
int	screen;
Window	root, mainwindow, filewindow;
Pixel	bgpixel, fgpixel, popupfgpixel, popupbgpixel, popupborderpixel;
GC	mainwindowgc, flipgc, thingc, flipthingc, popupgc;
XGCValues	gcv;
unsigned	long	gcm;
int	interval, poppedup=0;
char	bgcolor[100], fgcolor[100];
char	popupbgcolor[100], popupfgcolor[100], popupbordercolor[100];
Pixmap	bgpixmap, canvas;
XFontStruct	*popupfont1, *popupfont2;

struct	fsrecord {
			char	mountdir[11];
			char	filesystem[21];
			char	blocks[11];
			char	used[11];
			char	free[11];
			char	percent[4];
			int	x1, y1, x2, y2;
		};

struct	fsrecord	fs[24];

Pixel getcolor(char *colorname) {

	XColor	color;
	XWindowAttributes	attributes;
	
	XGetWindowAttributes(display, root, &attributes);
	color.pixel=0;

	XParseColor(display, attributes.colormap, colorname, &color);
	XAllocColor(display, attributes.colormap, &color);

	return color.pixel;
}

void	defaults() {
	strcpy(bgcolor, "DarkSlateGray");
	strcpy(fgcolor, "LightSlateGray");
	strcpy(popupbgcolor, "DarkSlateGray");
	strcpy(popupfgcolor, "#20B2AA");
	strcpy(popupbordercolor, "LightGray");
	interval=3;
}

void	parsecmdline(int argc, char *argv[]) {

	char	*argument;
	int	i;

	for (i=1; i<argc; i++) {
		argument=argv[i];
		if (argument[0]=='-') {
			switch(argument[1]) {
				case 'b':
					strcpy(bgcolor, argv[++i]);
					break;
				case 'f':
					strcpy(fgcolor, argv[++i]);
					break;
				case 'i':
					interval=atoi(argv[++i]);
					break;
				default:
					break;
			}
		}
	}
}

void	setuppixmaps() {

	bgpixmap=XCreatePixmap(display, mainwindow, 48, 48, 
				DefaultDepth(display, screen));
	XFillRectangle(display, bgpixmap, flipgc, 0, 0, 48, 48);

	canvas=XCreatePixmap(display, mainwindow, 48, 48, 
				DefaultDepth(display, screen));

}

void	setupfonts() {

	popupfont1=XLoadQueryFont(display, 
			"-*-helvetica-*-*-*-*-10-240-*-*-*-*-*-*");
	popupfont2=XLoadQueryFont(display, 
			"-*-helvetica-medium-r-*-*-10-240-*-*-*-*-*-*");
}

void	readstatfile(int *fscounter) {
	FILE	*tempfile;
	int	counter;

	tempfile=fopen("/usr/tmp/statfs", "r");
        if (tempfile!=NULL) {
		/* clear header */
		fscanf(tempfile,"%s%s%s%s%s%s%s",
				fs[0].mountdir,
				fs[0].mountdir,
				fs[0].filesystem,
				fs[0].blocks,
				fs[0].used,
				fs[0].free,
				fs[0].percent);
		/* read list */
		counter=0;
		while (1) {
			strcpy(fs[counter].mountdir,"");
			fscanf(tempfile,"%s%s%s%s%s%s",
				fs[counter].mountdir,
				fs[counter].filesystem,
				fs[counter].blocks,
				fs[counter].used,
				fs[counter].free,
				fs[counter].percent);
			if (strlen(fs[counter].mountdir)>0) {	
				counter++;
			} else {
				break;
			}
		}
		fclose(tempfile);
	
		*fscounter=counter;
	}
}

int	calclineheight(int num) {

	return ((48-num)/num);
}

int	calclinewidth(struct fsrecord *fs, int y, int lineheight) {

	float	blocks, used, linewidth;

	blocks=(float)atoi(fs->blocks);
	used=(float)atoi(fs->used);
	linewidth=(int)((used/blocks)*48.0);

	/* set up fsrecord x,y's */
	fs->x1=0;
	fs->y1=y-(lineheight/2);
	fs->x2=linewidth;
	fs->y2=y+(lineheight/2);

	return linewidth;
}

void	checkusage() {

	int	i, lineheight, y, linewidth, fscounter;

	while(1) {

		/* get a list of filesystems */
		system ("df -v > /usr/tmp/statfs");

		readstatfile(&fscounter);

		/* restore canvas pixmap */
		XCopyArea(display, bgpixmap, canvas, mainwindowgc, 
				0, 0, 48, 48, 0, 0);

		/* calculate line height */
		lineheight=calclineheight(fscounter);

		XSetLineAttributes(display, mainwindowgc, lineheight, 
					LineSolid, CapButt, JoinRound);

		/* draw lines */
		y=1+(lineheight/2);
		for (i=0; i<fscounter; i++) {

			/* calculate line width */
			linewidth=calclinewidth(&fs[i], y, lineheight);

			/* draw new lines */
			XDrawLine(display, canvas, mainwindowgc, 
					0, y, linewidth, y);

			if (lineheight>2 && linewidth>2) {
				XDrawLine(display, canvas, thingc,
					0, y+lineheight/2, 
					linewidth, y+lineheight/2);
				XDrawLine(display, canvas, flipthingc,
					0, y-lineheight/2, 
					linewidth, y-lineheight/2);
				XDrawLine(display, canvas, flipthingc,
					0, y+lineheight/2, 
					0, y-lineheight/2);
				XDrawLine(display, canvas, thingc,
					linewidth, y+lineheight/2, 
					linewidth, y-lineheight/2);
			}

			y=y+lineheight+1;
		}

		XCopyArea(display, canvas, mainwindow, mainwindowgc, 
				0, 0, 48, 48, 0, 0);

		/* get my stuff drawn, harmlessly */
		XPending(display);

		sleep(interval);
	}
}

void	popupfilewindow(struct fsrecord fs, int x, int y) {

	XSetWindowAttributes setwinattr;

	filewindow=XCreateSimpleWindow(display, root, x, y, 177, 76, 2, 
					popupborderpixel, popupbgpixel);

	setwinattr.override_redirect=True;
	XChangeWindowAttributes(display, filewindow, CWOverrideRedirect, 
				&setwinattr);
	XSetFont(display, popupgc, popupfont1->fid);

	XMapWindow(display, filewindow);

	XDrawString(display, filewindow, popupgc, 2, 12, "Mount Dir", 9);
	XDrawString(display, filewindow, popupgc, 2, 27, "File System", 11);
	XDrawString(display, filewindow, popupgc, 2, 42, "Blocks", 6);
	XDrawString(display, filewindow, popupgc, 2, 57, "Free", 4);
	XDrawString(display, filewindow, popupgc, 2, 72, "Percent Used", 12);

	XSetFont(display, popupgc, popupfont2->fid);

	XDrawString(display, filewindow, popupgc, 77, 12, fs.mountdir, 
			strlen(fs.mountdir));
	XDrawString(display, filewindow, popupgc, 77, 27, fs.filesystem, 
			strlen(fs.filesystem));
	XDrawString(display, filewindow, popupgc, 77, 42, fs.blocks,
			strlen(fs.blocks));
	XDrawString(display, filewindow, popupgc, 77, 57, fs.free,
			strlen(fs.free));
	XDrawString(display, filewindow, popupgc, 77, 72, fs.percent,
			strlen(fs.percent));

	/* get my stuff drawn, harmlessly */
	XFlush(display);

	poppedup=1;
}

void	popdownfilewindow() {

	if (poppedup==1) {
		XDestroyWindow(display, filewindow);
		poppedup=0;
	}
}

void	fsinfo() {

	Window	dummy;
	int	rootx, rooty, winx, winy, i, fscounter;
	unsigned	int	dummyui;
	int	lineheight, y;
	unsigned	int	mask;

	readstatfile(&fscounter);

	XQueryPointer(display, mainwindow, &dummy, &dummy,
			&rootx, &rooty, &winx, &winy, &mask);

	lineheight=calclineheight(fscounter);
	y=1+(lineheight/2);

	for (i=0; i<fscounter; i++) {

		calclinewidth(&fs[i], y, lineheight);

		if (winx>=fs[i].x1 && winx<=fs[i].x2 && 	
			winy>=fs[i].y1 && winy<=fs[i].y2) {

			popupfilewindow(fs[i], rootx+12, rooty);
		}

		y=y+lineheight+1;
	}
			
}

main (int argc, char *argv[]) {

	char	*displayname=NULL;
	XEvent	report;
	int	i, exposenum=0;
	XTextProperty	name;
	char	*winname="asfsm";

	defaults();
	parsecmdline(argc, argv);

	display=XOpenDisplay(&displayname);

	screen=DefaultScreen(display);

	root=RootWindow(display, screen);

	bgpixel=getcolor(bgcolor);
	fgpixel=getcolor(fgcolor);
	popupbgpixel=getcolor(popupbgcolor);
	popupfgpixel=getcolor(popupfgcolor);
	popupborderpixel=getcolor(popupbordercolor);

	mainwindow=XCreateSimpleWindow(display, root, 0, 0, 48, 48, 0, 
					fgpixel, bgpixel);

	gcm=GCForeground | GCBackground | GCGraphicsExposures;
	gcv.foreground=fgpixel;
	gcv.background=bgpixel;
	gcv.graphics_exposures=True;
	mainwindowgc=XCreateGC(display, root, gcm, &gcv);
	gcm=GCForeground | GCBackground;
	gcv.background=fgpixel;
	gcv.foreground=bgpixel;
	flipgc=XCreateGC(display, root, gcm, &gcv);
	gcv.foreground=popupfgpixel;
	gcv.background=popupbgpixel;
	popupgc=XCreateGC(display, root, gcm, &gcv);
	gcm=GCForeground;
	gcv.foreground=getcolor("black");
	thingc=XCreateGC(display, root, gcm, &gcv);
	XSetLineAttributes(display, thingc, 1, 
				LineSolid, CapButt, JoinRound);
	gcv.foreground=getcolor("white");
	flipthingc=XCreateGC(display, root, gcm, &gcv);
	XSetLineAttributes(display, flipthingc, 1, 
				LineSolid, CapButt, JoinRound);

	XSelectInput(display, mainwindow, ExposureMask | ButtonPressMask | 
						ButtonReleaseMask | 
						StructureNotifyMask );

	XStringListToTextProperty(&winname, 1, &name);
	XSetWMName(display, mainwindow, &name);

	XMapWindow(display, mainwindow);

	setuppixmaps();
	setupfonts();

	while (1) {
		
		XNextEvent(display, &report);
		switch (report.type) {
			case ButtonPress:
				fsinfo();
				break;
			case ButtonRelease:
				popdownfilewindow();
				break;
			case Expose:
				if (exposenum==0) {
					if (fork()) {
						checkusage();
					}
					exposenum=1;
				}
				break;
			case DestroyNotify:
				XFreeGC(display, mainwindowgc);
				XFreeGC(display, flipgc);
				XFreeGC(display, thingc);
				XFreeGC(display, flipthingc);
				XFreeGC(display, popupgc);
				XDestroyWindow(mainwindow);
 				XCloseDisplay(display);
				exit(0);
				break;
			default:
				break;
		}
	}
}
