#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Intrinsic.h>
#include <X11/xpm.h>
#include "surreal.xpm"
#include "purplestoneswirl.xpm"

struct XpmIcon {
	Pixmap		pixmap;
	Pixmap		mask;
	XpmAttributes	xpmattributes;
};

Display	*display;
int	screen, screenwidth, screenheight;
Window	root, mainwindow, filewindow;
Pixel	bgpixel, fgpixel, zeropixel, popupfgpixel, 
		popupbgpixel, popupborderpixel;
GC	mainwindowgc, zerogc, flipgc, thingc, flipthingc, popupgc;
GC	translucentgc;
XGCValues	gcv;
unsigned	long	gcm;
int	interval, poppedup=0;
char	bgcolor[100], fgcolor[100], zerocolor[100];
char	popupbgcolor[100], popupfgcolor[100], popupbordercolor[100];
char	bgpix[100], popupbgpix[100];
Pixmap	bgpixmap, canvas;
struct	XpmIcon	bgtexture, popupbgpixmap;
XFontStruct	*popupfont1, *popupfont2;
int	fscounter;
struct	timeb	lasttime, thistime;

struct	fsrecord {
			char	mountdir[11];
			char	filesystem[21];
			char	blocks[11];
			char	used[11];
			char	free[11];
			char	percent[4];
			int	x1, y1, x2, y2;
		};

struct	fsrecord	fs[24];

Pixel getcolor(char *colorname) {

	XColor	color;
	XWindowAttributes	attributes;
	
	XGetWindowAttributes(display, root, &attributes);
	color.pixel=0;

	XParseColor(display, attributes.colormap, colorname, &color);
	XAllocColor(display, attributes.colormap, &color);

	return color.pixel;
}

void	defaults1() {
	strcpy(bgcolor, "DarkSlateGray");
	strcpy(fgcolor, "LightSlateGray");
	strcpy(zerocolor, "IndianRed");
	strcpy(popupbgcolor, "DarkSlateGray");
	strcpy(popupfgcolor, "#20B2AA");
	strcpy(popupbordercolor, "LightGray");
	strcpy(bgpix, "default");
	strcpy(popupbgpix, "default");
	interval=3;
}

void	defaults2() {

	if (strcmp(bgpix,"None")) {
		bgtexture.xpmattributes.valuemask |= (XpmReturnPixels | 
							XpmReturnExtensions);
		if (!strcmp(bgpix,"default")) {
			XpmCreatePixmapFromData(display, root, surreal_xpm,
					&bgtexture.pixmap, 
					&bgtexture.mask, 
					&bgtexture.xpmattributes);
		} else {
			XpmReadFileToPixmap(display, root, bgpix, 
					&bgtexture.pixmap, 
					&bgtexture.mask, 
					&bgtexture.xpmattributes);
		}
	}

	if (strcmp(popupbgpix,"None")) {
		popupbgpixmap.xpmattributes.valuemask |= (XpmReturnPixels | 
							XpmReturnExtensions);
		if (!strcmp(popupbgpix,"default")) {
			XpmCreatePixmapFromData(display, root, 
					purplestoneswirl_xpm,
					&popupbgpixmap.pixmap, 
					&popupbgpixmap.mask, 
					&popupbgpixmap.xpmattributes);
		} else {
			if (!strcmp(popupbgpix,"Transparent") ||
				!strcmp(popupbgpix,"Translucent")) {
				popupbgpixmap.pixmap=NULL;
			} else {
				XpmReadFileToPixmap(display, root, 
					popupbgpix, 
					&popupbgpixmap.pixmap, 
					&popupbgpixmap.mask, 
					&popupbgpixmap.xpmattributes);
			}
		}
	}
}

void	parsecmdline(int argc, char *argv[]) {

	char	*argument;
	int	i;

	for (i=1; i<argc; i++) {
		argument=argv[i];
		if (argument[0]=='-') {
			if (!strncmp(argument,"-bg",3)) {
				strcpy(bgcolor, argv[++i]);
			}
			if (!strncmp(argument,"-bp",3)) {
				strcpy(bgpix, argv[++i]);
			}
			if (!strncmp(argument,"-fg",3)) {
				strcpy(fgcolor, argv[++i]);
			}
			if (!strncmp(argument,"-pbg",4)) {
				strcpy(popupbgcolor, argv[++i]);
			}
			if (!strncmp(argument,"-pbp",4)) {
				strcpy(popupbgpix, argv[++i]);
			}
			if (!strncmp(argument,"-pfg",4)) {
				strcpy(popupfgcolor, argv[++i]);
			}
			if (!strncmp(argument,"-pbd",4)) {
				strcpy(popupbordercolor, argv[++i]);
			}
			if (!strncmp(argument,"-z",2)) {
				strcpy(zerocolor, argv[++i]);
			}
			if (!strncmp(argument,"-i",2)) {
				interval=atoi(argv[++i]);
			}
		}
	}
}

void	setuppixmaps() {

	if (strcmp(bgpix,"None")) {
		bgpixmap=bgtexture.pixmap;
	} else {
		bgpixmap=XCreatePixmap(display, mainwindow, 48, 48, 
				DefaultDepth(display, screen));
		XFillRectangle(display, bgpixmap, flipgc, 0, 0, 48, 48);
	}

	canvas=XCreatePixmap(display, mainwindow, 48, 48, 
				DefaultDepth(display, screen));

}

void	setupfonts() {

	popupfont1=XLoadQueryFont(display, 
			"-*-helvetica-*-*-*-*-10-240-*-*-*-*-*-*");
	popupfont2=XLoadQueryFont(display, 
			"-*-helvetica-medium-r-*-*-10-240-*-*-*-*-*-*");
}

int	itstime() {

	ftime(&thistime);
	
	if (thistime.time-lasttime.time>=interval) {
	
		lasttime.time=thistime.time;

		return 1;
	} else {
		return 0;
	}
}

void	readstatfile() {
	FILE	*tempfile;
	int	counter;
	char	dummy[21];

	tempfile=fopen("/usr/tmp/statfs", "r");
        if (tempfile!=NULL) {
		/* clear header */
		fscanf(tempfile,"%s%s%s%s%s%s%s",
				dummy, dummy, dummy, dummy, 
				dummy, dummy, dummy);
		/* read list */
		counter=0;
		while (1) {
			strcpy(fs[counter].mountdir,"");
#ifdef SCO
			fscanf(tempfile,"%s%s%s%s%s%s",
				fs[counter].mountdir,
				fs[counter].filesystem,
				fs[counter].blocks,
				fs[counter].used,
				fs[counter].free,
				fs[counter].percent);
#endif
#ifdef SUN 
			fscanf(tempfile,"%s%s%s%s%s%s",
				fs[counter].filesystem,
				fs[counter].blocks,
				fs[counter].used,
				fs[counter].free,
				fs[counter].percent,
				fs[counter].mountdir);
#endif
#ifdef LINUX
			fscanf(tempfile,"%s%s%s%s%s%s",
				fs[counter].filesystem,
				fs[counter].blocks,
				fs[counter].used,
				fs[counter].free,
				fs[counter].percent,
				fs[counter].mountdir);
#endif
			if (strlen(fs[counter].mountdir)>0 && counter<23) {	
				counter++;
			} else {
				break;
			}
		}
		fclose(tempfile);
	
		fscounter=counter;
	}
}

int	calclineheight(int num) {

	return ((48-num)/num);
}

int	calclinewidth(struct fsrecord *fs, int y, int lineheight) {

	float	blocks, used, linewidth;

	blocks=(float)atoi(fs->blocks);
	if (blocks==0.0) {
		linewidth=48;
	} else {
		used=(float)atoi(fs->used);
		linewidth=(int)((used/blocks)*48.0);
	}

	/* set up fsrecord x,y's */
	fs->x1=0;
	fs->y1=y-(lineheight/2);
	fs->x2=linewidth;
	fs->y2=y+(lineheight/2);

	return linewidth;
}

void	checkusage() {

	int	i, lineheight, y, linewidth;


		/* get a list of filesystems */
#ifdef SCO
		system("df -v > /usr/tmp/statfs");
#endif
#ifdef LINUX
		system("df -v > /usr/tmp/statfs");
#endif
#ifdef SUN
		system("df -a > /usr/tmp/statfs");
#endif

		readstatfile();

		/* restore canvas pixmap */
		XCopyArea(display, bgpixmap, canvas, mainwindowgc, 
				0, 0, 48, 48, 0, 0);

		/* calculate line height */
		lineheight=calclineheight(fscounter);

		XSetLineAttributes(display, mainwindowgc, lineheight, 
					LineSolid, CapButt, JoinRound);

		/* draw lines */
		y=1+(lineheight/2);
		for (i=0; i<fscounter; i++) {

			/* calculate line width */
			linewidth=calclinewidth(&fs[i], y, lineheight);

			/* draw new lines */
			if (atoi(fs[i].blocks)>0) {
				XDrawLine(display, canvas, mainwindowgc, 
					0, y, linewidth, y);
			} else {
				XDrawLine(display, canvas, zerogc, 
					0, y, linewidth, y);
			}

			if (lineheight>2 && linewidth>2) {
				if (atoi(fs[i].blocks)>0) {
					/* raised */
					XDrawLine(display, canvas, thingc,
						0, y+lineheight/2, 
						linewidth, y+lineheight/2);
					XDrawLine(display, canvas, flipthingc,
						0, y-lineheight/2, 
						linewidth, y-lineheight/2);
					XDrawLine(display, canvas, flipthingc,
						0, y+lineheight/2, 
						0, y-lineheight/2);
					XDrawLine(display, canvas, thingc,
						linewidth, y+lineheight/2, 
						linewidth, y-lineheight/2);
				} else {
					/* lowered */
					XDrawLine(display, canvas, flipthingc,
						0, y+lineheight/2, 
						linewidth, y+lineheight/2);
					XDrawLine(display, canvas, thingc,
						0, y-lineheight/2, 
						linewidth, y-lineheight/2);
					XDrawLine(display, canvas, thingc,
						0, y+lineheight/2, 
						0, y-lineheight/2);
					XDrawLine(display, canvas, flipthingc,
						linewidth, y+lineheight/2, 
						linewidth, y-lineheight/2);
				}
			}

			y=y+lineheight+1;
		}

		XCopyArea(display, canvas, mainwindow, mainwindowgc, 
				0, 0, 48, 48, 0, 0);

		/* get my stuff drawn, harmlessly */
		XPending(display);

}

void	popupfilewindow(struct fsrecord fs, int x, int y) {

	XSetWindowAttributes	setwinattr;
	int			winx, winy, i;
	unsigned long		pixel;
	static	unsigned char		dotted[2]={ 1, 1 };

	if (poppedup==0) {

		poppedup=2;

		if (x>=screenwidth/2) { 
			winx=x-12-177;
		} else {
			winx=x+12;
		}
	
		if (y>=screenheight/2) {
			winy=y-76;
		} else {
			winy=y;
		}
	 
		filewindow=XCreateSimpleWindow(display, root, winx, winy, 
						177, 76, 2, 
						popupborderpixel, popupbgpixel);
	
		setwinattr.override_redirect=True;
		if (strcmp(popupbgpix,"None")) {
			setwinattr.background_pixmap=popupbgpixmap.pixmap;
			XChangeWindowAttributes(display, filewindow, 
					CWOverrideRedirect | CWBackPixmap, 
					&setwinattr);
		} else {
			XChangeWindowAttributes(display, filewindow, 
					CWOverrideRedirect,
					&setwinattr);
		}
	
		XSetFont(display, popupgc, popupfont1->fid);
	
		XMapWindow(display, filewindow);
	
		if (!strcmp(popupbgpix,"Translucent")) {
			gcm=GCForeground | GCLineStyle | GCCapStyle | 
				GCLineWidth;
			gcv.foreground=bgpixel;
			gcv.line_style=LineOnOffDash;
			gcv.cap_style=CapButt;
			gcv.line_width=1;
			translucentgc=XCreateGC(display, root, gcm, &gcv);
			XSetDashes(display, translucentgc, 0, dotted, 2);
			for (i=0; i<76; i++) {
				if (i%2==0) {
					XDrawLine(display, filewindow, 
							translucentgc,
							0, i, 177, i);
				} else {
					XDrawLine(display, filewindow, 
							translucentgc,
							1, i, 177, i);
				}
			}
			XFreeGC(display, translucentgc);
		}
	
		XDrawString(display, filewindow, popupgc, 2, 12, 
				"Mount Dir", 9);
		XDrawString(display, filewindow, popupgc, 2, 27, 
				"File System", 11);
		XDrawString(display, filewindow, popupgc, 2, 42, "Blocks", 6);
		XDrawString(display, filewindow, popupgc, 2, 57, "Free", 4);
		XDrawString(display, filewindow, popupgc, 2, 72, 
				"Percent Used", 12);
	
		XSetFont(display, popupgc, popupfont2->fid);
	
		XDrawString(display, filewindow, popupgc, 77, 12, fs.mountdir, 
				strlen(fs.mountdir));
		XDrawString(display, filewindow, popupgc, 77, 27, 
				fs.filesystem, 
				strlen(fs.filesystem));
		XDrawString(display, filewindow, popupgc, 77, 42, fs.blocks,
				strlen(fs.blocks));
		XDrawString(display, filewindow, popupgc, 77, 57, fs.free,
				strlen(fs.free));
		XDrawString(display, filewindow, popupgc, 77, 72, fs.percent,
				strlen(fs.percent));
	
		/* get my stuff drawn, harmlessly */
		XFlush(display);
	
		poppedup=1;
	}

}

void	popdownfilewindow() {

	if (poppedup==1) {
		XDestroyWindow(display, filewindow);
		poppedup=0;
	}
}

void	fsinfo() {

	Window	dummy;
	int	rootx, rooty, winx, winy, i;
	unsigned	int	dummyui;
	int	lineheight, y;
	unsigned	int	mask;

	XQueryPointer(display, mainwindow, &dummy, &dummy,
			&rootx, &rooty, &winx, &winy, &mask);

	lineheight=calclineheight(fscounter);
	y=1+(lineheight/2);

	for (i=0; i<fscounter; i++) {

		calclinewidth(&fs[i], y, lineheight);

		if (winx>=fs[i].x1 && winx<=fs[i].x2 && 	
			winy>=fs[i].y1 && winy<=fs[i].y2) {

			popupfilewindow(fs[i], rootx, rooty);
		}

		y=y+lineheight+1;
	}
			
}

main (int argc, char *argv[]) {

	char	*displayname=NULL;
	XEvent	report;
	int	i, exposenum=0;
	XTextProperty	name;
	char	*winname="asfsm";

	defaults1();
	parsecmdline(argc, argv);

	display=XOpenDisplay(displayname);

	screen=DefaultScreen(display);
	screenwidth=DisplayWidth(display, screen);
	screenheight=DisplayHeight(display, screen);

	root=RootWindow(display, screen);

	defaults2();

	bgpixel=getcolor(bgcolor);
	fgpixel=getcolor(fgcolor);
	zeropixel=getcolor(zerocolor);
	popupbgpixel=getcolor(popupbgcolor);
	popupfgpixel=getcolor(popupfgcolor);
	popupborderpixel=getcolor(popupbordercolor);

	mainwindow=XCreateSimpleWindow(display, root, 0, 0, 48, 48, 0, 
					fgpixel, bgpixel);

	gcm=GCForeground | GCBackground | GCGraphicsExposures;
	gcv.foreground=fgpixel;
	gcv.background=bgpixel;
	gcv.graphics_exposures=True;
	mainwindowgc=XCreateGC(display, root, gcm, &gcv);
	gcm=GCForeground;
	gcv.foreground=zeropixel;
	zerogc=XCreateGC(display, root, gcm, &gcv);
	gcm=GCForeground | GCBackground;
	gcv.background=fgpixel;
	gcv.foreground=bgpixel;
	flipgc=XCreateGC(display, root, gcm, &gcv);
	gcv.foreground=popupfgpixel;
	gcv.background=popupbgpixel;
	popupgc=XCreateGC(display, root, gcm, &gcv);
	gcm=GCForeground;
	gcv.foreground=getcolor("black");
	thingc=XCreateGC(display, root, gcm, &gcv);
	XSetLineAttributes(display, thingc, 1, 
				LineSolid, CapButt, JoinRound);
	gcv.foreground=getcolor("white");
	flipthingc=XCreateGC(display, root, gcm, &gcv);
	XSetLineAttributes(display, flipthingc, 1, 
				LineSolid, CapButt, JoinRound);

	XSelectInput(display, mainwindow, ExposureMask | ButtonPressMask | 
						ButtonReleaseMask | 
						StructureNotifyMask );

	XStringListToTextProperty(&winname, 1, &name);
	XSetWMName(display, mainwindow, &name);

	XMapWindow(display, mainwindow);

	setuppixmaps();
	setupfonts();
	lasttime.time=0;

	while (1) {
		
		if (XCheckWindowEvent(display, mainwindow, ButtonPressMask, 
					&report)==True) {
				fsinfo();
		}
		if (XCheckWindowEvent(display, mainwindow, ButtonReleaseMask, 
					&report)==True) {
				popdownfilewindow();
		}
		if (XCheckWindowEvent(display, mainwindow, ExposureMask, 
					&report)==True) {
			if (exposenum==0) {
				exposenum=1;
			}
		}
		if (XCheckWindowEvent(display, mainwindow, StructureNotifyMask, 
					&report)==True) {
			if (report.type==DestroyNotify) {
				XFreeGC(display, mainwindowgc);
				XFreeGC(display, flipgc);
				XFreeGC(display, thingc);
				XFreeGC(display, flipthingc);
				XFreeGC(display, popupgc);
				XDestroyWindow(display, mainwindow);
 				XCloseDisplay(display);
				exit(0);
			}
		}
		if (itstime()==1 && exposenum==1) {
			checkusage();
		}
	}
}
