/*****************************************************************************/
/*	astuner - AfterStep Tuner for RadioTrack and RadioReveal cards.      */
/*	Version 0.1						             */
/*	By David Muench.						     */		
/*	cc@spork.neonexus.com						     */
/*	http://www.neonexus.com/cc/					     */
/*									     */
/*	This is an 'AfterStep Look & Feel' Wharf style applet that can be    */
/*	used to control one of the before mentioned radio tuner cards.       */
/*									     */
/*   Based on ascd by Rob Malda and radiotrack-1.1 by Gideon J. le Grange.   */
/*****************************************************************************/

/* Change this if your card is set to a different I/O base (like 20C) */
#define RADIOBASE 0x30C
/* Set this to the station you want as the default on startup. */
/* Remove the decimal place and specify it out to the thousandths place. */
/* For example, 90.5 would be 90500 and 101.7 would be 101700. */
#define DEFSTATION 90500		/* 90.5 FM WBER Rochester, NY! */

/********/

#include <stdio.h>
#include <stdlib.h>       
#include <string.h>

#include <X11/Xlib.h>
#include <X11/xpm.h>
#include <X11/extensions/shape.h>

#include <errno.h>
#include <unistd.h>

#include "priv.h"
#include "card.h"

/* XPM struct and icons ******************************************************/
typedef struct _XpmIcon {
	Pixmap pixmap;
	Pixmap mask;
	XpmAttributes attributes;
} XpmIcon;

XpmIcon radiobgXPM, ledXPM;

#include "XPM/led.xpm"
#include "XPM/radiobg.xpm"

/* Functions *****************************************************************/
void	Help(void);
void	CreateWindow(void);
void	ParseCmdLine(int argc, char *argv[]);
void	MainLoop();
void	GetXPM(void);
int	FlushExpose(Window w);
void	RedrawWindow( XpmIcon *v);
Pixel	GetColor(char *name);



/* Global stuff **************************************************************/
#define TRUE 1
#define FALSE 0

const long STATION_HI = 107900;
const long STATION_LOW = 87900;

int	lasttime=-1;
Display	*Disp;	 
Window	Root;
Window	Iconwin;
Window	Win;
char	*Geometry = 0;

GC	WinGC;

long	currstation = DEFSTATION;
float	currstationf = 0.0;
char	radioon = 0;
int	err = 0;

/*****************************************************************************/
int main(int argc,char *argv[])
{       	
	RadioBase = RADIOBASE;

	ParseCmdLine(argc, argv); 
	
	err=get_io_privs(RadioBase);
	if (err!=0)
	{
		printf("%s : %s needs IO permissions. Run %s setuid root.\n",argv[0], argv[0], argv[0]);
		exit(-1);
	}
	/* Give up root to be safe. */
	give_up_root();                          
	
        CreateWindow();
	MainLoop();
	return 0;
}

/*****************************************************************************/
void Help()
{       
	fprintf(stderr,"astuner - Version 0.1\n");
	fprintf(stderr,"Written by David Muench - cc@spork.neonexus.com\n");
	fprintf(stderr,"http://www.neonexus.com/cc/\n");
	fprintf(stderr,"usage:  astuner [-options ...] \n");
	fprintf(stderr,"options:\n");
	fprintf(stderr,"\n");       
	fprintf(stderr,"  Command                 Default   Why\n");
	fprintf(stderr,"g Geometry                none      Standard X Location\n");
	fprintf(stderr,"h Help                              Duh.\n");
	fprintf(stderr,"\n\n");
	exit(1); 
}

/****************************************************************************/
void CreateWindow(void)
{
	int i;
	unsigned int borderwidth ;
	char *display_name = NULL; 
	char *wname = "astuner";
	XGCValues gcv;
	unsigned long gcm;
	XTextProperty name;
	Pixel back_pix, fore_pix;
	Pixmap pixmask;
	int screen;	
	int x_fd;
	int d_depth;
	int ScreenWidth, ScreenHeight;
	XSizeHints SizeHints;
	XWMHints WmHints;
	
	/* Open display */
	if (!(Disp = XOpenDisplay(display_name)))  
	{ 
		fprintf(stderr,"astuner: can't open display %s\n", XDisplayName(display_name));    
		exit (1); 
	} 
	
	screen = DefaultScreen(Disp);
	Root = RootWindow(Disp, screen);
	d_depth = DefaultDepth(Disp, screen);
	x_fd = XConnectionNumber(Disp);	
	ScreenHeight = DisplayHeight(Disp,screen);
	ScreenWidth = DisplayWidth(Disp,screen);
	       
	GetXPM();
		
	SizeHints.flags= USSize|USPosition;
	SizeHints.x = 0;
	SizeHints.y = 0;	
	back_pix = GetColor("white");
	fore_pix = GetColor("black");
	
	XWMGeometry(Disp, screen, Geometry, NULL, (borderwidth =1), &SizeHints,
		    &SizeHints.x,&SizeHints.y,&SizeHints.width,
		    &SizeHints.height, &i); 	
	
	SizeHints.width = radiobgXPM.attributes.width;
	SizeHints.height= radiobgXPM.attributes.height;	
	Win = XCreateSimpleWindow(Disp,Root,SizeHints.x,SizeHints.y,
				  SizeHints.width,SizeHints.height,
				  borderwidth,fore_pix,back_pix);
	Iconwin = XCreateSimpleWindow(Disp,Win,SizeHints.x,SizeHints.y,
				      SizeHints.width,SizeHints.height,
				      borderwidth,fore_pix,back_pix);	     
	XSetWMNormalHints(Disp, Win, &SizeHints);
	XSelectInput(Disp, Win, (ExposureMask | ButtonPressMask | 
				 StructureNotifyMask));
	XSelectInput(Disp, Iconwin, (ExposureMask | ButtonPressMask | 
				     StructureNotifyMask));
	
	if (XStringListToTextProperty(&wname, 1, &name) ==0) 
	{
		fprintf(stderr, "astuner: can't allocate window name\n");
		exit(-1);
	}
	XSetWMName(Disp, Win, &name);
	
	/* Create WinGC */
	gcm = GCForeground|GCBackground|GCGraphicsExposures;
	gcv.foreground = fore_pix;
	gcv.background = back_pix;
	gcv.graphics_exposures = False;
	WinGC = XCreateGC(Disp, Root, gcm, &gcv);  

	 
		
		XShapeCombineMask(Disp, Win, ShapeBounding, 0, 0, 
				  radiobgXPM.mask, ShapeSet);
		XShapeCombineMask(Disp, Iconwin, ShapeBounding, 0, 0, 
				  radiobgXPM.mask, ShapeSet);
		
	WmHints.initial_state = NormalState;
	WmHints.icon_window = Iconwin;
	WmHints.icon_x = SizeHints.x;
	WmHints.icon_y = SizeHints.y;
	WmHints.flags = StateHint | IconWindowHint | IconPositionHint;
	XSetWMHints(Disp, Win, &WmHints); 	
	XMapWindow(Disp,Win);
	RedrawWindow(&radiobgXPM);
}

/****************************************************************************/
void ParseCmdLine(int argc, char *argv[])
{

	char *Argument;
	int i;       
	
	for(i = 1; i < argc; i++) 
	{
		Argument = argv[i];
		
		if (Argument[0] == '-') 
		{
			switch(Argument[1]) 
			{
			 case 'g': /* Geometry */
				if(++i >= argc) Help();
				Geometry = argv[i];
				continue;
			case 'h':  /* Help */
				if(++i >= argc) Help();
				continue;
			
			 default:
				Help();
			}
		}
		else
			Help();
	}

}

/****************************************************************************/
void MainLoop()
{
	XEvent Event;            
			      
	/* Main loop */
	while(1)
	{		
		/* Check events */
		while (XPending(Disp))
		{
			XNextEvent(Disp,&Event);
			switch(Event.type)
			{		     
			 case Expose:		/* Redraw window */
				if(Event.xexpose.count == 0)
				{
					lasttime=01;
					XFlush(Disp);
					RedrawWindow(&radiobgXPM);
				} 
				break;		     
			 case ButtonPress:	/* Mouseclick */
				if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 32) && (Event.xbutton.x <= 14))
				{
					if(!(currstation <= STATION_LOW))
					{
						currstation = currstation - 200;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
					}
				}
				else if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 32) && (Event.xbutton.x <= 31) && (Event.xbutton.x >= 18))
				{
					if(!(currstation >= STATION_HI))
					{
						currstation = currstation + 200;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
					}
				}
				else if((Event.xbutton.y >= 34) && (Event.xbutton.y <= 48) && (Event.xbutton.x <= 22))
				{
					if(radioon == 0)
					{
						radioon = 1;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
					}
					else
					{
						radioon = 0;
						CardOff();
					}
				}
				if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 24) && (Event.xbutton.x >= 36))
				{
					VolumeUp();
				}
				if((Event.xbutton.y >= 26) && (Event.xbutton.y <= 32) && (Event.xbutton.x >= 36))
				{
					VolumeDown();
				}
				XFlush(Disp);
				RedrawWindow(&radiobgXPM);
				break;			 
			 case DestroyNotify:	/* Destroy window */
				XFreeGC(Disp, WinGC);
				XDestroyWindow(Disp, Win);
				XDestroyWindow(Disp, Iconwin);
				XCloseDisplay(Disp);				
				exit(0); 
				break;			
			}
		
		}
	usleep(50000L);
	}
}

/****************************************************************************/
void GetXPM(void)
{
	XWindowAttributes Attributes;
	int Ret;

	
	XGetWindowAttributes(Disp,Root,&Attributes);		
	radiobgXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
	Ret = XpmCreatePixmapFromData(Disp, Root, radiobg, &radiobgXPM.pixmap, 
				      &radiobgXPM.mask, &radiobgXPM.attributes);

        ledXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
        Ret = XpmCreatePixmapFromData(Disp, Root, led, &ledXPM.pixmap,
                                      &ledXPM.mask, &ledXPM.attributes);
      

	if(Ret != XpmSuccess)
        {
                fprintf(stderr, "astuner: not enough free color cells\n");
                exit(1);
        }          
}

/****************************************************************************/
int FlushExpose(Window w)
{
	XEvent dummy;
	int i=0;
	
	while (XCheckTypedWindowEvent (Disp, w, Expose, &dummy))i++;
	return i;
}

/****************************************************************************/
void RedrawWindow(XpmIcon *Icon)
{	
	int station[4];
	char ts[5];

	sprintf(ts, "%dl", currstation);
	if(currstation >= 100000)
	{
		station[0] = 1;
		station[1] = ts[1] - 48;
		station[2] = ts[2] - 48;
		station[3] = ts[3] - 48;
	}
	else
	{
		station[0] = 0;
		station[1] = ts[0] - 48;
		station[2] = ts[1] - 48;
		station[3] = ts[2] - 48;

	}

	FlushExpose(Win);
	FlushExpose(Iconwin);
	
	XShapeCombineMask(Disp, Win, ShapeBounding, 0, 0, 
			  radiobgXPM.mask, ShapeSet);
	XShapeCombineMask(Disp, Iconwin, ShapeBounding, 0, 0, 
			  radiobgXPM.mask, ShapeSet);

	XCopyArea(Disp,radiobgXPM.pixmap,Win,WinGC,
		  0,0,radiobgXPM.attributes.width, radiobgXPM.attributes.height,0,0);	
	XCopyArea(Disp,radiobgXPM.pixmap,Iconwin,WinGC,
       	          0,0,radiobgXPM.attributes.width, radiobgXPM.attributes.height,0,0);

	if(currstation >= 100000)
	{
		XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
			9*10+4,0, 6,11, 1,2);

		XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
			9*10+4,0, 6,11, 1,2);
	}
	else
	{
		XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
			9*10+10,0, 6,11, 1,2);

		XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
			9*10+10,0, 6,11, 1,2);		
	}
	
        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[1],0, 9,11, 7,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[1],0, 9,11, 7,2);               

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[2],0, 9,11, 16,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[2],0, 9,11, 16,2);               

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*10,0, 4,11, 25,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*10,0, 4,11, 25,2);    

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[3],0, 9,11, 28,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[3],0, 9,11, 28,2);    
}

/****************************************************************************/
Pixel GetColor(char *ColorName)
{
	XColor Color;
	XWindowAttributes Attributes;
	
	XGetWindowAttributes(Disp,Root,&Attributes);
	Color.pixel = 0;
	
	if (!XParseColor (Disp, Attributes.colormap, ColorName, &Color)) 
		fprintf(stderr,"astuner: can't parse %s\n", ColorName);
	else if(!XAllocColor (Disp, Attributes.colormap, &Color)) 
		fprintf(stderr,"astuner: can't allocate %s\n", ColorName);       
	
	return Color.pixel;
}
