/*****************************************************************************/
/*	astuner - AfterStep Tuner for RadioTrack and RadioReveal cards.      */
/*	Version 1.1						             */
/*	By David Muench.						     */		
/*	cc@spork.neonexus.com						     */
/*	http://www.neonexus.com/cc/					     */
/*									     */
/*	This is an 'AfterStep Look & Feel' Wharf style applet that can be    */
/*	used to control one of the before mentioned radio tuner cards.       */
/*									     */
/*   Based on ascd by Rob Malda and radiotrack-1.1 by Gideon J. le Grange.   */
/*****************************************************************************/

/********/

#include "astuner.h"

/* XPM struct and icons ******************************************************/

XpmIcon radiobg_onXPM, radiobg_offXPM, ledXPM, radiobgXPM, ledtunedXPM, leduntunedXPM;

#include "XPM/leduntuned.xpm"
#include "XPM/radiobg_on.xpm"
#include "XPM/radiobg_off.xpm"
#include "XPM/ledtuned.xpm"

/* Functions *****************************************************************/
void    Help(void);
void    CreateWindow(void);
void    ParseCmdLine(int argc, char *argv[]);
void    MainLoop();
void    GetXPM(void);
int     FlushExpose(Window w);
void    RedrawWindow( XpmIcon *v);
Pixel   GetColor(char *name);      

/* Globals *******************************************************************/
int     lasttime=-1;
Display *Disp;
Window  Root;
Window  Iconwin;
Window  Win;
char    *Geometry = 0;

GC      WinGC;

long    currstation = 90500;
float   currstationf = 0.0;
char    radioon = 0;
int     err = 0;
int	used = 0;
int	presetcount = 0;
int	intune = FALSE;

/*****************************************************************************/
int main(int argc,char *argv[])
{
	int count;
	       	
	myname = argv[0];

	ParseCmdLine(argc, argv); 

	yyin=fopen(RCFILE,"r");
	if (yyin==NULL)
	{
		printf("%s : Error opening configuration file %s\n",argv[0], RCFILE);
        	exit(-1);
	}
	yylex();                         
	fclose(yyin);

	currstationf = currstation;
	currstationf = currstationf / 1000;

	err=get_io_privs(RadioBase);
	if (err!=0)
	{
		printf("%s : %s needs IO permissions. Run %s setuid root.\n",argv[0], argv[0], argv[0]);
		exit(-1);
	}
	/* Give up root to be safe. */
	give_up_root();                          
	
        CreateWindow();
	MainLoop();
	return 0;
}

/*****************************************************************************/
void Help()
{       
	fprintf(stderr,"astuner - Version 0.1\n");
	fprintf(stderr,"Written by David Muench - cc@spork.neonexus.com\n");
	fprintf(stderr,"http://www.neonexus.com/cc/\n");
	fprintf(stderr,"usage:  astuner [-options ...] \n");
	fprintf(stderr,"options:\n");
	fprintf(stderr,"\n");       
	fprintf(stderr,"  Command                 Default   Why\n");
	fprintf(stderr,"g Geometry                none      Standard X Location\n");
	fprintf(stderr,"h Help                              Duh.\n");
	fprintf(stderr,"\n\n");
	exit(1); 
}

/****************************************************************************/
void CreateWindow(void)
{
	int i;
	unsigned int borderwidth ;
	char *display_name = NULL; 
	char *wname = "astuner";
	XGCValues gcv;
	unsigned long gcm;
	XTextProperty name;
	Pixel back_pix, fore_pix;
	Pixmap pixmask;
	int screen;	
	int x_fd;
	int d_depth;
	int ScreenWidth, ScreenHeight;
	XSizeHints SizeHints;
	XWMHints WmHints;
	
	/* Open display */
	if (!(Disp = XOpenDisplay(display_name)))  
	{ 
		fprintf(stderr,"astuner: can't open display %s\n", XDisplayName(display_name));    
		exit (1); 
	} 
	
	screen = DefaultScreen(Disp);
	Root = RootWindow(Disp, screen);
	d_depth = DefaultDepth(Disp, screen);
	x_fd = XConnectionNumber(Disp);	
	ScreenHeight = DisplayHeight(Disp,screen);
	ScreenWidth = DisplayWidth(Disp,screen);
	       
	GetXPM();
		
	SizeHints.flags= USSize|USPosition;
	SizeHints.x = 0;
	SizeHints.y = 0;	
	back_pix = GetColor("white");
	fore_pix = GetColor("black");
	
	XWMGeometry(Disp, screen, Geometry, NULL, (borderwidth =1), &SizeHints,
		    &SizeHints.x,&SizeHints.y,&SizeHints.width,
		    &SizeHints.height, &i); 	
	
	SizeHints.width = radiobgXPM.attributes.width;
	SizeHints.height= radiobgXPM.attributes.height;	
	Win = XCreateSimpleWindow(Disp,Root,SizeHints.x,SizeHints.y,
				  SizeHints.width,SizeHints.height,
				  borderwidth,fore_pix,back_pix);
	Iconwin = XCreateSimpleWindow(Disp,Win,SizeHints.x,SizeHints.y,
				      SizeHints.width,SizeHints.height,
				      borderwidth,fore_pix,back_pix);	     
	XSetWMNormalHints(Disp, Win, &SizeHints);
	XSelectInput(Disp, Win, (ExposureMask | ButtonPressMask | 
				 StructureNotifyMask));
	XSelectInput(Disp, Iconwin, (ExposureMask | ButtonPressMask | 
				     StructureNotifyMask));
	
	if (XStringListToTextProperty(&wname, 1, &name) ==0) 
	{
		fprintf(stderr, "astuner: can't allocate window name\n");
		exit(-1);
	}
	XSetWMName(Disp, Win, &name);
	
	/* Create WinGC */
	gcm = GCForeground|GCBackground|GCGraphicsExposures;
	gcv.foreground = fore_pix;
	gcv.background = back_pix;
	gcv.graphics_exposures = False;
	WinGC = XCreateGC(Disp, Root, gcm, &gcv);  

	 
		
		XShapeCombineMask(Disp, Win, ShapeBounding, 0, 0, 
				  radiobgXPM.mask, ShapeSet);
		XShapeCombineMask(Disp, Iconwin, ShapeBounding, 0, 0, 
				  radiobgXPM.mask, ShapeSet);
		
	WmHints.initial_state = NormalState;
	WmHints.icon_window = Iconwin;
	WmHints.icon_x = SizeHints.x;
	WmHints.icon_y = SizeHints.y;
	WmHints.flags = StateHint | IconWindowHint | IconPositionHint;
	XSetWMHints(Disp, Win, &WmHints); 	
	XMapWindow(Disp,Win);
	RedrawWindow(&radiobgXPM);
}

/****************************************************************************/
void ParseCmdLine(int argc, char *argv[])
{

	char *Argument;
	int i;       
	
	for(i = 1; i < argc; i++) 
	{
		Argument = argv[i];
		
		if (Argument[0] == '-') 
		{
			switch(Argument[1]) 
			{
			 case 'g': /* Geometry */
				if(++i >= argc) Help();
				Geometry = argv[i];
				continue;
			case 'h':  /* Help */
				if(++i >= argc) Help();
				continue;
			
			 default:
				Help();
			}
		}
		else
			Help();
	}

}

/****************************************************************************/
void MainLoop()
{
	XEvent Event;            
			      
	/* Main loop */
	while(1)
	{		
		/* Check events */
		while (XPending(Disp))
		{
			XNextEvent(Disp,&Event);
			switch(Event.type)
			{		     
			 case Expose:		/* Redraw window */
				if(Event.xexpose.count == 0)
				{
					lasttime=01;
					XFlush(Disp);
					RedrawWindow(&radiobgXPM);
				} 
				break;		     
			 case ButtonPress:	/* Mouseclick */
				if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 32) && (Event.xbutton.x <= 14))
				{
					if(!(currstation <= STATION_LOW))
					{
						currstation = currstation - 200;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
						if(radioon == 0) radioon = 1;
						intune = IsTuned();
					}
				}
				else if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 32) && (Event.xbutton.x <= 31) && (Event.xbutton.x >= 18))
				{
					if(!(currstation >= STATION_HI))
					{
						currstation = currstation + 200;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
						if(radioon == 0) radioon = 1;
						intune = IsTuned();
					}
				}
				else if((Event.xbutton.y >= 34) && (Event.xbutton.y <= 48) && (Event.xbutton.x <= 22))
				{
					if(radioon == 0)
					{
						radioon = 1;
						currstationf = currstation;
						currstationf = currstationf / 1000;
						Tune(currstationf);
						intune = IsTuned();
					}
					else
					{
						radioon = 0;
						CardOff();
						intune = FALSE;
					}
				}
				else if((Event.xbutton.y >= 34) && (Event.xbutton.y <= 48) && (Event.xbutton.x >= 27))
				{
					currstation = stations[presetcount].freq;
					currstationf = currstation;
					currstationf = currstationf / 1000;
					Tune(currstationf);
					presetcount++;
					if(presetcount >= used)
						presetcount = 0;
					if(radioon == 0) radioon = 1;
					intune = IsTuned();
				}
				else if((Event.xbutton.y >= 17) && (Event.xbutton.y <= 24) && (Event.xbutton.x >= 36))
				{
					VolumeUp();
				}
				else if((Event.xbutton.y >= 26) && (Event.xbutton.y <= 32) && (Event.xbutton.x >= 36))
				{
					VolumeDown();
				}
				XFlush(Disp);
				RedrawWindow(&radiobgXPM);
				break;			 
			 case DestroyNotify:	/* Destroy window */
				XFreeGC(Disp, WinGC);
				XDestroyWindow(Disp, Win);
				XDestroyWindow(Disp, Iconwin);
				XCloseDisplay(Disp);				
				exit(0); 
				break;			
			}
		
		}
	usleep(50000L);
	}
}

/****************************************************************************/
void GetXPM(void)
{
	XWindowAttributes Attributes;
	int Ret;

	
	XGetWindowAttributes(Disp,Root,&Attributes);		
	radiobg_onXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
	Ret = XpmCreatePixmapFromData(Disp, Root, radiobg_on, &radiobg_onXPM.pixmap, 
				      &radiobg_onXPM.mask, &radiobg_onXPM.attributes);

	radiobg_offXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
	Ret = XpmCreatePixmapFromData(Disp, Root, radiobg_off, &radiobg_offXPM.pixmap, 
				      &radiobg_offXPM.mask, &radiobg_offXPM.attributes);

        ledtunedXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
        Ret = XpmCreatePixmapFromData(Disp, Root, ledtuned, &ledtunedXPM.pixmap,
                                      &ledtunedXPM.mask, &ledtunedXPM.attributes);
      
        leduntunedXPM.attributes.valuemask |= (XpmReturnPixels | XpmReturnExtensions);
        Ret = XpmCreatePixmapFromData(Disp, Root, leduntuned, &leduntunedXPM.pixmap,
                                      &leduntunedXPM.mask, &leduntunedXPM.attributes);

	radiobgXPM = radiobg_offXPM;
	ledXPM = leduntunedXPM;
	if(Ret != XpmSuccess)
        {
                fprintf(stderr, "astuner: not enough free color cells\n");
                exit(1);
        }          
}

/****************************************************************************/
int FlushExpose(Window w)
{
	XEvent dummy;
	int i=0;
	
	while (XCheckTypedWindowEvent (Disp, w, Expose, &dummy))i++;
	return i;
}

/****************************************************************************/
void RedrawWindow(XpmIcon *Icon)
{	
	int station[4];
	char ts[5];

	sprintf(ts, "%dl", currstation);
	if(currstation >= 100000)
	{
		station[0] = 1;
		station[1] = ts[1] - 48;
		station[2] = ts[2] - 48;
		station[3] = ts[3] - 48;
	}
	else
	{
		station[0] = 0;
		station[1] = ts[0] - 48;
		station[2] = ts[1] - 48;
		station[3] = ts[2] - 48;

	}

	FlushExpose(Win);
	FlushExpose(Iconwin);

	if(radioon == 0) radiobgXPM = radiobg_offXPM;
	else radiobgXPM = radiobg_onXPM;
	
	if(intune == TRUE) ledXPM = ledtunedXPM;
	else ledXPM = leduntunedXPM;
	
	XShapeCombineMask(Disp, Win, ShapeBounding, 0, 0, 
			  radiobgXPM.mask, ShapeSet);
	XShapeCombineMask(Disp, Iconwin, ShapeBounding, 0, 0, 
			  radiobgXPM.mask, ShapeSet);

	XCopyArea(Disp,radiobgXPM.pixmap,Win,WinGC,
		  0,0,radiobgXPM.attributes.width, radiobgXPM.attributes.height,0,0);	
	XCopyArea(Disp,radiobgXPM.pixmap,Iconwin,WinGC,
       	          0,0,radiobgXPM.attributes.width, radiobgXPM.attributes.height,0,0);

	if(currstation >= 100000)
	{
		XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
			9*10+4,0, 6,11, 1,2);

		XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
			9*10+4,0, 6,11, 1,2);
	}
	else
	{
		XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
			9*10+10,0, 6,11, 1,2);

		XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
			9*10+10,0, 6,11, 1,2);		
	}
	
        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[1],0, 9,11, 7,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[1],0, 9,11, 7,2);               

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[2],0, 9,11, 16,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[2],0, 9,11, 16,2);               

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*10,0, 4,11, 25,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*10,0, 4,11, 25,2);    

        XCopyArea(Disp, ledXPM.pixmap, Win, WinGC,
                  9*station[3],0, 9,11, 28,2);

        XCopyArea(Disp, ledXPM.pixmap, Iconwin, WinGC,
                  9*station[3],0, 9,11, 28,2);
                  
}

/****************************************************************************/
Pixel GetColor(char *ColorName)
{
	XColor Color;
	XWindowAttributes Attributes;
	
	XGetWindowAttributes(Disp,Root,&Attributes);
	Color.pixel = 0;
	
	if (!XParseColor (Disp, Attributes.colormap, ColorName, &Color)) 
		fprintf(stderr,"astuner: can't parse %s\n", ColorName);
	else if(!XAllocColor (Disp, Attributes.colormap, &Color)) 
		fprintf(stderr,"astuner: can't allocate %s\n", ColorName);       
	
	return Color.pixel;
}
