/* preferences.h */ 

#ifndef __ML_PREFERENCES_H__
#define __ML_PREFERENCES_H__


typedef struct _PREFERENCES_ {

  char        *defaultConfig;
  char        *smtp_server;
  char        *nntp_server;
  char        *default_domain;
  char        *default_mailbox;
  char        *default_savebox;
  char        *default_mailbox_pattern;
  char        *default_file_pattern;
  char        *default_cc;
  char        *default_bcc;
  char        *default_action;
  char        *help_directory;
  char        *header_set;
  char        *viewline_format;
  char        *sendlog;
  char        *mime_directory;
  char        *reply_address;
  char        *reply_prefix;
  char        *print_command;
  char        *signature_file;
  char        *folder_path;
  char        *url_command;
  char        *newMailProg;
  char        *mail_sort;
  char        *news_sort;
  char        *charset;
  char        *language;
  char        *ispellCommand;
  char        *smart_reply;
  char        *netCheckCommand;
  char        *netStartCommand;
  char        *netStopCommand;
  int          netCheckDelay;
  int          check_interval;
  int          mail_prefetch;
  int          news_prefetch;
  int          mail_background;
  int          news_background;
  int          mail_fetch;
  int          news_fetch;
  int          note_save;
  Boolean      confirmDestroy;
  Boolean      useIspell;
  Boolean      imap_debug;
  Boolean      smtp_debug;
  Boolean      auto_open;
  Boolean      word_wrap;
  Boolean      mime_ask;
  Boolean      logit;
  Boolean      log_full;
  Boolean      tear_off;
  Boolean      decode_all;
  Boolean      hide_empty;
  Boolean      new_news;
  Boolean      send_eight;
  Boolean      developer_debug;
  Boolean      alert_new;
  Boolean      keep_open;
  Boolean      newMailBeep;
  Boolean      keepOnServer;
  Boolean      autoPlace;
  Boolean      showLog;

  /* Don't mess with these. */
  Boolean      local_defs;
  Boolean      defaults_installed;
  char         *defaults_version;
} Preferences;

#define PREF_TYPE_INT   0
#define PREF_TYPE_STR   1
#define PREF_TYPE_BOOL  2
#define PREF_TYPE_FLOAT 3



/* Language string resources. Class supplied automatically. */

#define XtNmsgReplySender       "msgReplySender"
#define MsgReplySender          "Reply"
#define XtNmsgReplyAll          "msgReplyAll"
#define MsgReplyAll             "Reply To ALL"
#define XtNmsgForward           "msgForward"
#define MsgForward              "Forward Message(s)"

#define XtNmsgRemail            "msgRemail"
#define MsgRemail               "Remail"
#define XtNmsgComposeNew        "msgComposeNew"
#define MsgComposeNew           "New Message"

#define XtNmsgMailboxOpFailed   "msgMailboxOpFailed"
#define MsgMailboxOpFailed  "A mailbox operation failed. See transcript above."
#define XtNmsgMailboxOpSuccess  "msgMailboxOpSuccess"
#define MsgMailboxOpSuccess     "Mailbox manage operation successful."


#define XtNmsgNotConnected      "msgNotConnected"
#define MsgNotConnected         "Network not connected."
#define XtNmsgUnsupportedConfig "msgUnsupportedConfig"
#define MsgUnsupportedConfig    "Configuration does not support copy/move."
#define XtNmsgYouConnected      "msgYouConnected"
#define MsgYouConnected         "Is your network connection running? "
#define XtNmsgCannotQualify     "msgCannotQualify"
#define MsgCannotQualify  "Cannot lookup '%s'. Ignore if fully qualified."
#define XtNmsgDestroyMailbox    "msgDestroyMailbox"
#define MsgDestroyMailbox       "Destroy Mailbox Contents? "
#define XtNmsgDestroyEdit       "msgDestroyEdit"
#define MsgDestroyEdit          "Destroy This Message? "
#define XtNmsgDestroyAddress    "msgDestroyAddress"
#define MsgDestroyAddress       "Destroy This Address? "
#define XtNmsgDestroyConfig     "msgDestroyConfig"
#define MsgDestroyConfig        "Destroy This Configuration? "
#define XtNmsgFilterParseError  "msgFilterParseError"
#define MsgFilterParseError     "Filter Parse: ??? "
#define XtNmsgNoFilterName      "msgNoFilterName"
#define MsgNoFilterName         "No filter name supplied."
#define XtNmsgNoFilterDef       "msgNoFilterDef"
#define MsgNoFilterDef          "No filter definition supplied."
#define XtNmsgDestroyFilter     "msgDestroyFilter"
#define MsgDestroyFilter        "Destroy this filter? "
#define XtNmsgNoDefaults        "msgNoDefaults"
#define MsgNoDefaults           "No application defaults found."
#define XtNmsgWrongDefaults     "msgWrongDefaults"
#define MsgWrongDefaults        "Application defaults file version incorrect."
#define XtNmsgUnknownConfig     "msgUnknownConfig"
#define MsgUnknownConfig        "Unknown Configuration name '%s'."
#define XtNmsgOpeningMailbox    "msgOpeningMailbox"
#define MsgOpeningMailbox       "Opening mailbox '%s'."
#define XtNmsgOpenFailed        "msgOpenFailed"
#define MsgOpenFailed           "Mailbox open failed."
#define XtNmsgEmptyMailbox      "msgEmptyMailbox"
#define MsgEmptyMailbox         "Mailbox is empty."
#define XtNmsgClosingMailbox    "msgClosingMailbox"
#define MsgClosingMailbox       "Closing mailbox '%s'."
#define XtNmsgClosingAll        "msgClosingAll"
#define MsgClosingAll           "Closing all mailboxes."
#define XtNmsgLostConnection    "msgLostConnection"
#define MsgLostConnection       "Lost connection to server '%s'."
#define XtNmsgUpdatingNewsrc    "XtNmsgUpdatingNewsrc"
#define MsgUpdatingNewsrc       "Updating NEWSRC File."
#define XtNmsgBackgroundStop    "msgBackgroundStop"
#define MsgBackgroundStop       "Background fetch paused."
#define XtNmsgBackgroundResume  "msgBackgroundResume"
#define MsgBackgroundResume     "Background fetching resumed."
#define XtNmsgBackgrounding     "msgBackgrounding"
#define MsgBackgrounding        "Background fetching %lu messages."
#define XtNmsgBackgroundDone    "msgBackgroundDone"
#define MsgBackgroundDone       "Background fetch complete."
#define XtNmsgPruning           "msgPruning"
#define MsgPruning              "Pruning mailbox to remove old messages."
#define XtNmsgNotYetLoaded      "msgNotYetLoaded"
#define MsgNotYetLoaded         "[Message Not Yet Loaded]"
#define XtNmsgOpenNNTPFailed    "msgOpenNNTPFailed"
#define MsgOpenNNTPFailed       "Failed to connect to NNTP delivery service."
#define XtNmsgSendNNTPFailed    "msgSendNNTPFailed"
#define MsgSendNNTPFailed       "News delivery failed."
#define XtNmsgNewsSent          "msgNewsSent"
#define MsgNewsSent             "News sent."
#define XtNmsgOpenSMTPFailed    "msgOpenSMTPFailed"
#define MsgOpenSMTPFailed       "Failed to connect to SMTP delivery service."
#define XtNmsgSendEightFailure  "msgSendEightFailure"
#define MsgSendEightFailure     "8-bit transport unavailable. Message encoded."
#define XtNmsgSendSMTPFailed    "msgSendSMTPFailed"
#define MsgSendSMTPFailed       "Mail delivery failed."
#define XtNmsgMailSent          "msgMailSent"
#define MsgMailSent             "Mail sent."
#define XtNmsgLoadingFilters    "msgLoadingFilters"
#define MsgLoadingFilters       "Loading Default Filters."
#define XtNmsgEmptyMultipart    "msgEmptyMultipart"
#define MsgEmptyMultipart       "Empty multipart attachment fixed."
#define XtNmsgCannotWriteFile   "msgCannotWriteFile"
#define MsgCannotWriteFile      "Unable to write to file '%s'."
#define XtNmsgCannotReadFile    "msgCannotReadFile"
#define MsgCannotReadFile       "Unable to read from file '%s'."
#define XtNmsgAppendFailed      "msgAppendFailed"
#define MsgAppendFailed         "Message log (append) failed."
#define XtNmsgInvalidParams     "msgInvalidParams"
#define MsgInvalidParams        "Invalid parameter setting."
#define XtNmsgNonEmptyMulti     "msgNonEmptyMulti"
#define MsgNonEmptyMulti        "Multipart not empty. Unable to remove."
#define XtNmsgForwardExpunged   "msgForwardExpunged"
#define MsgForwardExpunged      "Message unavailable. (Expunged)?"
#define XtNmsgUnknownAction     "msgUnknownAction"
#define MsgUnknownAction        "Action not recognized: '%s'"
#define XtNmsgActionRequiresArg "msgActionRequiresArg"
#define MsgActionRequiresArg    "Argument required for action '%s'."
#define XtNmsgMissingQuote      "msgMissingQuote"
#define MsgMissingQuote         "Missing quotation mark."
#define XtNmsgNewMailboxName    "msgNewMailboxName"
#define MsgNewMailboxName       "New Mailbox Name"
#define XtNmsgNoMailboxSelected "msgNoMailboxSelected"
#define MsgNoMailboxSelected    "No mailbox selected."


#define XtNmsgInputCommand      "msgInputCommand"
#define MsgInputCommand         "Command to Execute"
#define XtNmsgFilterDefinition  "msgFilterDefinition"
#define MsgFilterDefinition     "Filter Definition"
#define XtNmsgAddressName       "msgAddressName"
#define MsgAddressName          "Name for Address Book"



#define XtNmsgSpellNoRead       "msgSpellNoRead"
#define MsgSpellNoRead          "Unknown spell sub-process error."
#define XtNmsgNCorrections      "msgNCorrections"
#define MsgNCorrections         "%d corrections found."
#define XtNmsgPromptParams      "msgPromptParams"
#define MsgPromptParams         "Parameters"
#define XtNmsgExecuting         "msgExecuting"
#define MsgExecuting            "Executing '%s'."
#define XtNmsgSubProcessFailure "msgSubProcessFailure"
#define MsgSubProcessFailure    "External command failed."
#define XtNmsgNoteSaveFail      "msgNoteSaveFail"
#define MsgNoteSaveFail         "Notebook save failed."
#define XtNmsgNoteSaved         "msgNoteSaved"
#define MsgNoteSaved            "Notebook saved."
#define XtNmsgPrefSaveFail      "msgPrefSaveFail"
#define MsgPrefSaveFail         "Failed to save preferences."
#define XtNmsgPrefSaved         "msgPrefSaved"
#define MsgPrefSaved            "Preferences saved."



#define XtNmsgUndetermined      "msgUndetermined"
#define MsgUndetermined         "Undetermined Message Information"
#define XtNmsgNotRegularFile    "msgNotRegularFile"
#define MsgNotRegularFile       "Not a regular file."
#define XtNmsgFileGoUp          "msgFileGoUp"
#define MsgFileGoUp             "[ Up one level ]"
#define XtNmsgFileGoHome        "msgFileGoHome"
#define MsgFileGoHome           "[Home directory]"
#define XtNmsgChdirFailed       "msgChdirFailed"
#define MsgChdirFailed          "Directory not accessible."
#define XtNmsgChdirHomeFailed   "msgChdirHomeFailed"
#define MsgChdirHomeFailed      "Home directory not accessible."
#define XtNmsgTCPPort           "msgTCPPort"
#define MsgTCPPort              "TCP Port Number"
#define XtNmsgDuplicateFilter   "msgDuplicateFilter"
#define MsgDuplicateFilter      "Duplicate filter name not allowed."
#define XtNmsgDuplicate         "msgDuplicate"
#define MsgDuplicate            "Duplicate name not allowed."
#define XtNmsgFilterNotSaved    "msgFilterNotSaved"
#define MsgFilterNotSaved       "Filter hasn't been saved."
#define XtNmsgNoWebBrowser      "msgNoWebBrowser"
#define MsgNoWebBrowser         "No web browser application has been defined."
#define XtNmsgWebCommand        "msgWebCommand"
#define MsgWebCommand           "Web browser command syntax unsupported."
#define XtNmsgShowURL           "msgShowURL"
#define MsgShowURL              "Show URL"
#define XtNmsgBinaryMessage     "msgBinaryMessage"
#define MsgBinaryMessage    "[Potential non-text message contents suppressed.]"
#define XtNmsgNoDataDir         "msgNoDataDir"
#define MsgNoDataDir            "\007Unable to create/access  '%s'.\n"
#define XtNmsgNoParentView      "msgNoParentView"
#define MsgNoParentView         "(Internal error). No parent view found."
#define XtNmsgAuthenticateTo    "msgAuthenticateTo"
#define MsgAuthenticateTo       "Authenticate to: %s"
#define XtNmsgUpdatingStopFile  "msgUpdatingStopFile"
#define MsgUpdatingStopFile     "Updating spell checker word list."
#define XtNmsgAttachingFilters  "msgAttachingFilters"
#define MsgAttachingFilters     "Attaching session filters."
#define XtNmsgDetachSessionFilter "msgDetachSessionFilter"
#define MsgDetachSessionFilter  "Detach filter from session?"
#define XtNmsgNewMessageFrom    "msgNewMessageFrom"
#define MsgNewMessageFrom       "'%s':\nFrom:    %s\nSubject: %s"
#define XtNmsgFetchFTP          "msgFetchFTP"
#define MsgFetchFTP             "Fetching message via FTP."
#define XtNmsgFTPFail           "msgFTPFail"
#define MsgFTPFail              "FTP transfer failed."
#define XtNmsgFTPSaved          "msgFTPSaved"
#define MsgFTPSaved             "FTP: %lu bytes written to '%s'."
#define XtNmsgFetchMail         "msgFetchMail"
#define MsgFetchMail            "Fetching message from mail server '%s'"
#define XtNmsgFetchMailFail     "msgFetchMailFail"
#define MsgFetchMailFail        "External mail fetch failed."
#define XtNmsgNotDraft          "msgNotDraft"
#define MsgNotDraft             "File does not appear to be a message draft."
#define XtNmsgCopySuccess       "msgCopySuccess"
#define MsgCopySuccess          "Copy successful."
#define XtNmsgCopyFail          "msgCopyFail"
#define MsgCopyFail             "Copy failed."
#define XtNmsgMoveSuccess       "msgMoveSuccess"
#define MsgMoveSuccess          "Move successful."
#define XtNmsgMoveFail          "msgMoveFail"
#define MsgMoveFail             "Move failed."
#define XtNmsgSaveSuccess       "msgSaveSuccess"
#define MsgSaveSuccess          "Save successful."
#define XtNmsgSaveFail          "msgSaveFail"
#define MsgSaveFail             "Save failed."
#define XtNmsgPipeFail          "msgPipeFail"
#define MsgPipeFail             "External command failed."
#define XtNmsgPrintSuccess      "msgPrintSuccess"
#define MsgPrintSuccess         "Print successful."
#define XtNmsgPrintFail         "msgPrintFail"
#define MsgPrintFail            "Print failed."
#define XtNmsgNoPrintCommand    "msgNoPrintCommand"
#define MsgNoPrintCommand       "No print command configured."
#define XtNmsgNewMessage        "msgNewMessage"
#define MsgNewMessage           "%u New message%s"
#define XtNmsgMessageSingular   "msgMessageSingular"
#define MsgMessageSingular      ""
#define XtNmsgMessagePlural     "msgMessagePlural"
#define MsgMessagePlural        "s"
#define XtNmsgFoundMessage      "msgFoundMessage"
#define MsgFoundMessage         "%lu Message%s found."
#define XtNmsgInvalidWrap       "msgInvalidWrap"
#define MsgInvalidWrap          "Invalid text wrap length."
#define XtNmsgEntireUniverse    "msgEntireUniverse"
#define MsgEntireUniverse       "Show all files in the known universe?"



#define XtNstrFilters           "strFilters"
#define StrFilters              "Filters"
#define XtNstrDraft             "strDraft"
#define StrDraft                "Draft"
#define XtNstrDefault           "strDefault"
#define StrDefault              "default"
#define XtNstrNA                "strNA"
#define StrNA                   "N/A"
#define XtNstrWarn              "strWarn"
#define StrWarn                 "Warning: "
#define XtNstrError             "strError"
#define StrError                "ERROR: "
#define XtNstrAll               "strAll"
#define StrAll                  "ALL"
#define XtNstrBucket            "strBucket"
#define StrBucket               "UNFILTERED"
#define XtNstrTrue              "strTrue"
#define StrTrue                 "true"
#define XtNstrFalse             "strFalse"
#define StrFalse                "false"
#define XtNstrUnknown           "strUnknown"
#define StrUnknown              "Unknown"
#define XtNstrSubscribed        "strSubscribed"
#define StrSubscribed           "Subscribed"
#define XtNstrUnSubscribed      "strUnSubscribed"
#define StrUnSubscribed         "Unsubscribed"
#define XtNstrMailboxes         "strMailboxes"
#define StrMailboxes            "Mailboxes"
#define XtNstrNewsgroups        "strNewsgroups"
#define StrNewsgroups           "Newsgroups"
#define XtNstrInternalViewer    "strInternalViewer"
#define StrInternalViewer       "MLVIEWER"
#define XtNstrForwardSubject    "strForwardSubject"
#define StrForwardSubject       "Forwarded: [%s]"
#define XtNstrMultiForwardSubject "strMultiForwardSubject"
#define StrMultiForwardSubject  "[Multiple Messages Forwarded]"
#define XtNstrMultiForwardHeader "strMultiForwardHeader"
#define StrMultiForwardHeader   "------- Forwarded Messages:\n\n"
#define XtNstrBeginForward      "strBeginForward"
#define StrBeginForward         "\n------- Begin Forwarded Message -------\n"
#define XtNstrEndForward        "strEndForward"
#define StrEndForward           "\n-------  End Forwarded Message  -------\n"
#define XtNmsgNoReplySubject    "msgNoReplySubject"
#define MsgNoReplySubject       "Re: (No Subject)"
#define XtNstrRFC822            "strRFC822"
#define StrRFC822               " [RFC822]"
#define XtNstrFakePart          "strFakePart"
#define StrFakePart             "EMPTY ATTACHMENT"
#define XtNstrLogicalViews      "strLogicalViews"
#define StrLogicalViews         "Virtual Mailboxes"



#define XtNpresentNormal         "presentNormal"
#define XtNpresentReverse        "presentReverse"
#define XtNpresentSender         "presentSender"
#define XtNpresentSubject        "presentSubject"
#define XtNpresentThread         "presentThread"



#define XtNflagDeleted           "flagDeleted"
#define XtNflagSeen              "flagSeen"
#define XtNflagFlagged           "flagFlagged"
#define XtNflagAnswered          "flagAnswered"
#define XtNflagDraft             "flagDraft"
#define XtNflagRecent            "flagRecent"






#define XtNactionNoHeader       "actionNoHeader"
#define XtNactionPartHeader     "actionPartHeader"
#define XtNactionFullHeader     "actionFullHeader"
#define XtNactionDelete         "actionDelete"
#define XtNactionExpunge        "actionExpunge"
#define XtNactionCopy           "actionCopy"
#define XtNactionMove           "actionMove"
#define XtNactionSave           "actionSave"
#define XtNactionShell          "actionShell"
#define XtNactionFlag           "actionFlag"
#define XtNactionSet            "actionSet"
#define XtNactionUnset          "actionUnset"
#define XtNactionReply          "actionReply"
#define XtNactionPrint          "actionPrint"
#define XtNactionSelect         "actionSelect"
#define XtNactionNew            "actionNew"
#define XtNactionUnseen         "actionUnseen"
#define XtNactionUnselect       "actionUnselect"
#define XtNactionRead           "actionRead"


#define XtNfilterAnd            "filterAnd"
#define XtNfilterOr             "filterOr"
#define XtNfilterNot            "filterNot"
#define XtNfilterCorrespondent  "filterCorrespondent"
#define XtNfilterConversant     "filterConversant"
#define XtNfilterRecipient      "filterRecipient"
#define XtNfilterFrom           "filterFrom"
#define XtNfilterTo             "filterTo"
#define XtNfilterCc             "filterCc"
#define XtNfilterSubject        "filterSubject"
#define XtNfilterNewsgroups     "filterNewsgroups"
#define XtNfilterMessageId      "filterMessageId"
#define XtNfilterSender         "filterSender"
#define XtNfilterReplyTo        "filterReplyTo"
#define XtNfilterInReplyTo      "filterInReplyTo"

#define XtNfilterText           "filterText"
#define XtNfilterHeader         "filterHeader"
#define XtNfilterBody           "filterBody"
#define XtNfilterBefore         "filterBefore"
#define XtNfilterSince          "filterSince"
#define XtNfilterOn             "filterOn"
#define XtNfilterLarger         "filterLarger"
#define XtNfilterSmaller        "filterSmaller"
#define XtNfilterNew            "filterNew"
#define XtNfilterOld            "filterOld"
#define XtNfilterDeleted        "filterDeleted"
#define XtNfilterUndeleted      "filterUndeleted"
#define XtNfilterSeen           "filterSeen"
#define XtNfilterUnseen         "filterUnseen"
#define XtNfilterRecent         "filterRecent"
#define XtNfilterAnswered       "filterAnswered"
#define XtNfilterUnanswered     "filterUnanswered"
#define XtNfilterFlagged        "filterFlagged"
#define XtNfilterUnflagged      "filterUnflagged"




#define XtNfilterToday          "filterToday"
#define XtNfilterYesterday      "filterYesterday"
#define XtNfilterLastMonth      "filterLastMonth"
#define XtNfilterThisMonth      "filterThisMonth"
#define XtNfilterJanuary        "filterJanuary"
#define XtNfilterFebruary       "filterFebruary"
#define XtNfilterMarch          "filterMarch"
#define XtNfilterApril          "filterApril"
#define XtNfilterMay            "filterMay"
#define XtNfilterJune           "filterJune"
#define XtNfilterJuly           "filterJuly"               
#define XtNfilterAugust         "filterAugust"
#define XtNfilterSeptember      "filterSeptember"
#define XtNfilterOctober        "filterOctober"
#define XtNfilterNovember       "filterNovember"
#define XtNfilterDecember       "filterDecember"
#define XtNfilterSunday         "filterSunday"
#define XtNfilterMonday         "filterMonday"
#define XtNfilterTuesday        "filterTuesday"
#define XtNfilterWednesday      "filterWednesday"
#define XtNfilterThursday       "filterThursday"
#define XtNfilterFriday         "filterFriday"
#define XtNfilterSaturday       "filterSaturday"
#define XtNfilterLastSunday     "filterLastSunday"
#define XtNfilterLastMonday     "filterLastMonday"
#define XtNfilterLastTuesday    "filterLastTuesday"
#define XtNfilterLastWednesday  "filterLastWednesday"
#define XtNfilterLastThursday   "filterLastThursday"
#define XtNfilterLastFriday     "filterLastFriday"
#define XtNfilterLastSaturday   "filterLastSaturday"



#define XtNheaderFrom            "headerFrom"
#define HeaderFrom               HEADER_FROM
#define XtNheaderSender          "headerSender"
#define HeaderSender             HEADER_SENDER
#define XtNheaderReplyTo         "headerReplyTo"
#define HeaderReplyTo            HEADER_REPLY_TO
#define XtNheaderTo              "headerTo"
#define HeaderTo                 HEADER_TO
#define XtNheaderNewsgroups      "headerNewsgroups"
#define HeaderNewsgroups         HEADER_NEWSGROUPS
#define XtNheaderSubject         "headerSubject"
#define HeaderSubject            HEADER_SUBJECT
#define XtNheaderCc              "headerCc"
#define HeaderCc                 HEADER_CC
#define XtNheaderBcc             "headerBcc"
#define HeaderBcc                HEADER_BCC
#define XtNheaderInReplyTo       "headerInReplyTo"
#define HeaderInReplyTo          HEADER_IN_REPLY_TO
#define XtNheaderMessageId       "headerMessagesId"
#define HeaderMessageId          HEADER_MESSAGE_ID
#define XtNheaderDate            "headerDate"
#define HeaderDate               HEADER_DATE
#define XtNheaderReferences      "headerReferences"
#define HeaderReferences         HEADER_REFERENCES


/* Not yet used... */


#define XtNbuttonAddressAdd     "buttonAddressAdd"
#define ButtonAddressAdd        "Add Entry to Address Book"
#define XtNbuttonAddressRemove  "buttonAddressRemove"
#define ButtonAddressRemove     "Remove Entry from Address Book"
#define XtNbuttonAddressModify  "buttonAddressModify"
#define ButtonAddressModify     "Modify Address Book Entry"
#define XtNbuttonAddressClose   "buttonAddressClose"
#define ButtonAddressClose      "Close Address Book Window"
#define XtNbuttonAddressFile    "buttonAddressFile"
#define ButtonAddressFile       "File Menu: Add, Remove, Modify, Close"
#define XtNbuttonAddressSearch  "buttonAddressSearch"
#define ButtonAddressSearch     "Search Address Book for these items"
#define XtNbuttonAddressReset   "buttonAddressReset"
#define ButtonAddressReset      "Reset Fields, return to Add mode"
#define XtNbuttonAddressHelp    "buttonAddressHelp"
#define ButtonAddressHelp       "Address Book Window Help"

#define XtNbuttonLoginDone      "buttonLoginDone"
#define ButtonLoginDone         "Login using this name/password"
#define XtNbuttonLoginAbort     "buttonLoginAbort"
#define ButtonLoginAbort        "Cancel login, abort current operation"
#define XtNbuttonLoginHelp      "buttonLoginHelp"
#define ButtonLoginHelp         "Login Window Help"



  


#endif /* __ML_PREFERENCES_H__ */









