/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/*	 $Id: fmail.c,v 1.165 1996/06/24 13:11:00 gena Exp $
 */
#include <fmail.h>
#include <umail.h>
#include <math.h>

char    lockname[MAXPATHLEN];
int     locked = 0;

struct _mail_folder *mailbox[MAX_FOLDERS];
int folders_num;
/* system folders */
struct _mail_folder *trash, *inbox, *outbox, *sentm, *draft, *template, *ftemp = NULL;

char mailbox_path[255];
u_int sort_type, folder_sort;
u_int offline;

char *stripheaders[]= { STATUS_FIELD, "X-RDate", "X-Forwarded", "X-Replied",
			"X-Bounced" , NULL };

char *months[12] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug",
		  "Sep", "Oct", "Nov", "Dec" };

#ifdef MMAP_MSG		/* mmap message to speed up parsing */

unsigned char *mmsg = NULL;	/* mmapped message */
long mmpos = 0;			/* position in msg */
long mmlen = 0;			/* length of mmapped block */

long
mmtell(stream)			/* for compatibility with ftell */
FILE *stream;
{
 return mmpos;
}

int
mmseek(stream ,offset, whence)	/* for compatibility with fseek */
FILE *stream;
long offset;
int whence;
{
long tpos;

 switch (whence) {
   case SEEK_SET:
	if ((offset > mmlen) || (offset < 0))
		return -1;
	else {
		mmpos = offset;
		return 0;
	     }
   break;

   case SEEK_CUR:
	tpos = mmpos + offset;
	if ((tpos > mmlen) || (tpos < 0))
		return -1;
	else {
		mmpos = tpos;
		return 0;
	     }
   break;

   case SEEK_END:
	tpos = mmlen + offset;
	if ((tpos > mmlen) || (tpos < 0))
		return -1;
	else {
		mmpos = tpos;
		return 0;
	     }
   break;

   default:
	return -1;
   break;
		}

   return -1;
}

char *
mmgets(str, size, stream)	/* for compatibility with fgets */
char *str;
register size_t size;
register FILE *stream;
{
register size_t len;
register unsigned char *t;

 if ((mmsg == NULL) || (mmpos >= mmlen))
	return NULL;

 if (size == 0)
	return NULL;

 if ((mmlen - mmpos) > size)
	len = size;
 else
	len = mmlen - mmpos;

 if ((t = memchr((void *)(mmsg + mmpos), '\n', len)) != NULL) {
	len = ++t - (mmsg + mmpos);
	(void)memcpy((void *)str, (void *)(mmsg + mmpos), len);
	str[len] = '\0';
	mmpos += len;
	return str; } else {
	(void)memcpy((void *)str, (void *)(mmsg + mmpos), len);
	str[len] = '\0';
	mmpos += len;
	return str; 	   }

}

#endif

int
get_date_offt()
{
static int off = -1;
register struct tm *lt;
time_t t;
struct tm gmt;

	if (off != -1)
		return off;

	(void) time(&t);
        gmt = *gmtime(&t);
        lt = localtime(&t);

        off = (lt->tm_hour - gmt.tm_hour) * 60 + lt->tm_min - gmt.tm_min;

        if (lt->tm_year < gmt.tm_year)
                off -= 24 * 60;
        else if (lt->tm_year > gmt.tm_year)
                off += 24 * 60;
        else if (lt->tm_yday < gmt.tm_yday)
                off -= 24 * 60;
        else if (lt->tm_yday > gmt.tm_yday)
                off += 24 * 60;

        if (off >= 24*60)
		off = 23*60+59;

	return off;
}

char *
get_arpa_date(t)
time_t t;
{
static char arpadate[60];
char buf[50];
int off;

off = get_date_offt();

strftime(buf, 49, "%a, %d %h %Y %T %%c%%04d (%Z)", localtime(&t));
sprintf(arpadate, buf, off > 0 ? '+' : '-', abs((off/60*100) + (off % 60)));

return arpadate;
}

time_t
get_date(str)
char *str;
{
int year, mon, day, hour, min, sec;
int i, offt;
struct tm tmm;
char month[5], hbuf[6];
char *p;

 if (strlen(str) < 24)
	return 0L;

 mon = year = hour = min = sec = -1;
 month[0] = '\0';
 offt = 0;

 if ((p = strchr(str, ',')) != NULL)
	p++;
 else
	p = str;

 while (*p == ' ')
	p++;

 sscanf(p, "%d%3s%d%d:%d:%d%d", &day, month, &year, &hour, &min, &sec, &offt);
 for(i=0;i<12;i++) {
    if(!strncasecmp(month, months[i], 3)) {
	mon = i;
	break; }
		}

 if ((mon == -1) || (year == -1) || (hour == -1))
	return 0L;

 if (offt)
	offt = ((offt / 100) * 60 + (offt % 100)) * 60;

 if (year > 1900)
	year -= 1900;

 if ((hour > 23) || (min < 0)) {
	sprintf(hbuf, "%04d", hour);
	p = hbuf + 2;
	min = atoi(p);
	*p = '\0';
	hour = atoi(hbuf);
	sec = 0L; }

 if (sec < 0)
	sec = 0L;

 tmm.tm_sec = sec;
 tmm.tm_min = min;
 tmm.tm_hour = hour;
 tmm.tm_mday = day;
 tmm.tm_mon = mon;
 tmm.tm_year = year;

 tmm.tm_yday = 0;
 tmm.tm_wday = 0;
 tmm.tm_isdst = -1;
#if !defined(__linux__) && !defined(SOLARIS) && !defined(HP) && !defined(_AIX) && !defined(SGI)
 tmm.tm_gmtoff = offt;
 tmm.tm_zone = NULL;
#endif
 return mktime(&tmm);
}

struct _mail_addr
	*get_address(str)
char *str;
{
mail_addr *ma , *ma1, *ma2;
char addr[256];
char name[256];
char comment[256];
char *cur, *p, c;
int len, curlen;
#define	ABUF_LEN	200

   if (!str)
	return NULL;

   ma1 = NULL;
   ma2 = NULL;
   addr[0] = name[0] = comment[0] = '\0';
   curlen = 0;

   cur = addr;

   while (1) {
	if ((p = strpbrk(str, "<(,\'\"")) != NULL) {
		c = *p;
		*p = '\0';
		len = ABUF_LEN - curlen;
		if (len < 0)
			len = 0;
		strncpy(cur, str, len);
		cur[len] = '\0';
		len = strlen(cur);
		cur = cur + len;
		curlen += len;
		*p = c;
		str = p;
						   }
	else    {
		len = ABUF_LEN - curlen;
		if (len < 0)
			len = 0;
		strncpy(cur, str, len);
		cur[len] = '\0';
		c = '\0';
		}

    switch(c) {

	case '(':
	  str++;
	  if ((p = strchr(str, ')')) == NULL) {
		*cur = c;
		cur++;
		*cur = '\0';
		break;  }

	  if (comment[0] == '\0') {
	  c = *p;
	  *p = '\0';
	  strcat(comment, str);
	  *p = c;
				  }

	  str = p;
	  str++;
	  if (addr[0] != '\0') {
		cur = name + strlen(name);
		curlen = strlen(name);
				}
	  else	{
	        cur = addr + strlen(addr);
		curlen = strlen(addr);
		}
	break;

	case '<':
	  str++;
	  if ((p = strchr(str, '>')) == NULL) {
		*cur = c;
		cur++;
		*cur = '\0';
		break; }

	  if (addr[0] != '\0') {
		if (name[0] == '\0')
                	strcat(name, addr);
		addr[0] = '\0';
				}

	  c = *p;
	  *p = '\0';
	  strcat(addr, str);
	  *p = c;

	  p++;
	  str = p;
	  cur = name + strlen(name);
	  curlen = strlen(name);
	break;

	case '\'':
	case '"':
	  str++;
	  *cur = c;
	  cur++;
	  *cur = '\0';

	  if ((p = strchr(str, c)) == NULL)
		break;

	  p++;
	  c = *p;
	  *p = '\0';
	  len = ABUF_LEN - curlen;
	  if (len < 0)
		len = 0;
	  strncpy(cur, str, len);
	  cur[len] = '\0';

	  *p = c;
	  curlen += strlen(cur);
	  cur = cur + strlen(cur);
	  str = p;
	break;

	case ',':
	case '\0':
	 if (c == ',')
		str++;

	 p = rem_tr_space(addr);
	 if (p[0] == '\0') {
		if (c == '\0')
			return ma1;
		else
			break;
				}

	 ma = (struct _mail_addr *)malloc(sizeof(struct _mail_addr));
	 ma->next_addr = NULL;
	 ma->num = 0;

	 ma->addr = strdup(p);
	 
	 p = rem_tr_space(name);

	 if (p[0] != '\0')
	 	ma->name = strdup(p);
	 else
		ma->name = NULL;
	 
	 p = rem_tr_space(comment);

	 if (comment[0] != '\0')
	 	ma->comment = strdup(p);
	 else
		ma->comment = NULL;
	 

	 if (ma1)
		ma2->next_addr = ma;
	 else
		ma1 = ma;

	 ma2 = ma;
	 ma1->num++;
		
	 if ((c == '\0') || (ma1->num > 80))
		return ma1;

	 addr[0] = name[0] = comment[0] = '\0';

	 cur = addr;
	 curlen = 0;

	break;

	}

	}

	return NULL;

}

int
graph_str(str)
char *str;
{
char *p = str;
int changed = 0;

 if (!str)
	return 0;

 while (*p) {
	if ((*p < 32) && (*p != '\n') && (*p != 0x09)) 	{
		p[0] = '_';
		changed = 1;
							}
	p++; }

 return changed;
}

char *
rem_tr_space(str)
char *str;
{
char *p;

 if (!str)
	return "";

 while((*str == ' ') ||	
	(*str == 0x09) )
	str++;

 if(*str == '\0')
	return "";

 p = str + strlen(str) - 1;
 while((p != str) && ((*p == ' ') || (*p == 0x09))) {
	*p = '\0';
	p--; }

 if(*str == '\0')
	return "";

 return str;
}

void
replace_field(msg, name, str)
struct _mail_msg *msg;
char *str;
char *name;
{
struct _head_field *hf;

 if (!msg || !str || !name)
	return;

 if (!msg->header)
	return;

 if ((hf = find_field(msg, name)) == NULL) {
	add_field(msg, name, str);
	return; }

 free(hf->f_line);
 hf->f_line = strdup(str);

 if (hf->f_line == NULL) {
	display_msg(MSG_FATAL, "malloc", "strdup failed");
	return; }

 return;
}

struct _head_field *
copy_field(fld)
struct _head_field *fld;
{
struct _head_field *hf;

 if (!fld)
	return NULL;

 hf = (struct _head_field *)malloc(sizeof(struct _head_field));
 hf->f_line = strdup(fld->f_line);
 strcpy(hf->f_name, fld->f_name);
 hf->num_fields = 0;
 hf->next_head_field = NULL;

 return hf;
}

void
add_field(msg, name, str)
struct _mail_msg *msg;
char *str;
char *name;
{
struct _head_field *hf;

 if (!msg || !str || !name)
	return;

 if (!msg->header)
	return;

 if ((strlen(name) < 1) || (strlen(name) >= MAX_FIELD_NAME_LEN))
	return;

 hf = (struct _head_field *)malloc(sizeof(struct _head_field));
 hf->f_line = strdup(str);
 strcpy(hf->f_name, name);
 hf->num_fields = 0;
 hf->next_head_field = msg->header->other_fields;
 msg->header->other_fields = hf;

 if (msg->header->other_fields)
	msg->header->other_fields->num_fields++;

 return;

}

void
delete_field(msg, field)
struct _mail_msg *msg;
struct _head_field *field;
{
struct _head_field *hf;

 if (!msg || !field)
	return;

 if (!msg->header)
	return;

 hf = msg->header->other_fields;
 if (hf == field) 
	msg->header->other_fields = field->next_head_field;
 else 	{
	while (hf && (hf->next_head_field != field))
		hf = hf->next_head_field;

	if (!hf)
		return;

	hf->next_head_field = field->next_head_field;
	}

 if (field->f_line)
	free(field->f_line);

 free(field);

 return;

}

struct _head_field *find_field(msg, str)
struct _mail_msg *msg;
char *str;
{
struct _head_field *hf;

 if (!msg)
	return NULL;

 if (!msg->header)
	return NULL;

 if (!str)
	return NULL;

 if ((strlen(str) < 1) || (strlen(str) > MAX_FIELD_NAME_LEN))
	return NULL;

 hf = msg->header->other_fields;
 while (hf) {
 if (!strcasecmp(hf->f_name, str))
	return hf;
 hf = hf->next_head_field;
	}

 return NULL;
}

void
free_field_content(msg, p, fr)
struct _mail_msg *msg;
char *p;
int fr;
{
 switch(fr) {
	case 0:
		return;
	break;

	case 1:
		msg->free_text(msg);
	break;

	case 2:
		if (p)
		  free(p);
		msg->free_text(msg);
	break;

	default:
		return;
	return;
		}
}

char *
get_field_content(msg, str, fr)
struct _mail_msg *msg;
char *str;
int *fr;
{
static char fld[512];
char *mbuf;
struct _head_field *hf;

 if (!fr)
	return NULL;

 *fr = 0;

 if (!msg || !str)
	return NULL;

 if (!msg->header)
	return NULL;

 if (!strncasecmp(str, "Message", 7) && (msg->get_text(msg) == 0)) {
	*fr = 1;
	return msg->msg_body; }

 if (!strncasecmp(str, "Body", 4) && (msg->get_text(msg) == 0)) {
	*fr = 1;
	return (msg->msg_body + msg->header->header_len); }

 if (!strncasecmp(str, "Header", 6) && (msg->get_text(msg) == 0)) {
	*fr = 1;

	if ((mbuf = malloc(msg->header->header_len + 1)) == NULL) {
	  display_msg(MSG_WARN, "malloc", "malloc failed");
	  return NULL; }

	*fr = 2;

	memcpy(mbuf, msg->msg_body, msg->header->header_len);
	mbuf[msg->header->header_len] = '\0';
	return mbuf;	}

 if (!strncasecmp(str, "Subject", 7)) {
	strncpy(fld, msg->header->Subject ? msg->header->Subject : "nosubject", 255);
	fld[255] = '\0';
	return fld;	}

 if (!(hf = find_field(msg, str)))
	return NULL;

 if (!hf->f_line)
	return NULL;

 strcpy(fld, hf->f_line);
 return fld;
}

struct _head_field *get_field(str)
char *str;
{
head_field *nhf;
char *p;
int len;

 if (!str)
	return NULL;

 if ((nhf = (struct _head_field *)malloc(sizeof(struct _head_field))) == NULL) {
	display_msg(MSG_FATAL, "field parse", "malloc failed");
	return NULL;
									}	
 nhf->next_head_field = NULL;
 nhf->num_fields = 0;

 if ((p = strchr(str, ':')) == NULL) {
	free(nhf);
	return NULL;		     }

 *p = '\0';
 len = p - str;

 p++;
 while ((*p == ' ') || (*p == 0x09))
	p++;

 if ((len >= MAX_FIELD_NAME_LEN) || (len < 1)) {
	free(nhf);
	return NULL;
						}

 strcpy(nhf->f_name, str);

 len = strlen(p);

 if (len > MAX_FIELD_LEN)
	len = MAX_FIELD_LEN;

 while ((len > 0) && ((p[len - 1] == ' ') ||
	(p[len - 1] == 0x09)))
	len--;

 p[len] = '\0';

 nhf->f_line = strdup(p);

 return nhf;
}

struct _msg_header
	*get_msg_header(mfd)
FILE *mfd;
{
 struct _mail_addr *addr;
 char buf[MAX_FIELD_LEN + 1], buf1[127];
 head_field *cur_fld, *end_fld;
 msg_header *mh;
 char *p;
 long pos;
 u_int fldnum;

  if (!(mh = (msg_header *)malloc(sizeof(msg_header)))) {
	display_msg(MSG_FATAL, "malloc", "Can not allocate memory!");
	exit(1); }

  fldnum = 0;
  mh->other_fields = NULL;
  mh->flags = 0;
  mh->From = mh->To = mh->Sender = mh->Cc = mh->Bcc = NULL;
  end_fld = mh->other_fields;
  mh->snt_time = 0L;
  mh->rcv_time = 0L;
  mh->Subject = NULL;

  while(msg_gets(buf, MAX_FIELD_LEN, mfd)) {
    buf[MAX_FIELD_LEN - 1] = '\0';

    if (strlen(buf) < 2)
	break;

    if (fldnum > MAX_FIELDS_IN_HDR)
	break;

    if (!strncasecmp(buf, "From ", 5))
	continue;

    if ((p = strrchr(buf, '\n')) != NULL)
	*p = '\0';
    else	{
	while(msg_gets(buf1, sizeof(buf1) - 1, mfd)) 	{
	   if ((p = strrchr(buf1, '\n')) != NULL)
		break;
							}
		}

    if ((cur_fld = get_field(buf))) {
nsp:	pos = msg_tell(mfd);
	if (msg_gets(buf, MAX_FIELD_LEN, mfd)) {
    	  buf[strlen(buf) - 1] = '\0';
	  if ((buf[0] == ' ') || (buf[0] == 0x09)) {
	    if (strlen(cur_fld->f_line) >= MAX_FIELD_LEN)
		goto nsp;

	    p = buf;
	    while ((p[1] == ' ') || (p[1] == 0x09))
		p++;

	    *p = ' ';	
	    if (!(cur_fld->f_line = realloc(cur_fld->f_line, strlen(p) + strlen(cur_fld->f_line) + 1))) {
		display_msg(MSG_FATAL, "realloc", "Can not allocate memory!");
		exit(1); }

	    strcat(cur_fld->f_line, p);
	    goto nsp;
			     }
	  else
	    msg_seek(mfd, pos, SEEK_SET);
				  }
	else
	  msg_seek(mfd, pos, SEEK_SET);

	fldnum++;

	if (!strncasecmp(cur_fld->f_name, STATUS_FIELD, 9)) {
	   sscanf(cur_fld->f_line, "%4x", &mh->flags);
	   mh->flags = mh->flags & 0xFFFF;
	   goto dsc_fld; }

	if (!mh->Subject && !strncasecmp(cur_fld->f_name, "Subject", 7) &&
		(strlen(cur_fld->f_line) > 1) ) {
	    mh->Subject = strdup(cur_fld->f_line);
	    goto dsc_fld; }

	if (!mh->From && !strncasecmp(cur_fld->f_name, "From", 4)) {
	    mh->From = get_address(cur_fld->f_line);
	    goto dsc_fld; }

	if (!strncasecmp(cur_fld->f_name, "To", 2)) {
	    if (mh->To == NULL)
	    	mh->To = get_address(cur_fld->f_line);
	    else {
		addr = mh->To;
		while (addr->next_addr)
			addr = addr->next_addr;
		addr->next_addr = get_address(cur_fld->f_line);
		 }
	    goto dsc_fld; }

	if (!mh->Sender && !strncasecmp(cur_fld->f_name, "Sender", 6)) {
	    mh->Sender = get_address(cur_fld->f_line);
	    goto dsc_fld; }

	if (!mh->Cc && !strncasecmp(cur_fld->f_name, "Cc", 2)) {
	    mh->Cc = get_address(cur_fld->f_line);
	    goto dsc_fld; }

	if (!mh->Bcc && !strncasecmp(cur_fld->f_name, "Bcc", 4)) {
	    mh->Bcc = get_address(cur_fld->f_line);
	    goto dsc_fld; }

	if (!strncasecmp(cur_fld->f_name, "Date", 4))
	   mh->snt_time = get_date(cur_fld->f_line);
	else
	if (!strncasecmp(cur_fld->f_name, "X-RDate", 4))
	   mh->rcv_time = get_date(cur_fld->f_line);

	if (mh->other_fields == NULL)
		mh->other_fields = cur_fld;
	   else
		end_fld->next_head_field = cur_fld;
	   end_fld = cur_fld;
	continue;

dsc_fld:   free(cur_fld->f_line);
	   free(cur_fld);

				}

				}

  if (!mh->rcv_time && mh->snt_time)
	mh->rcv_time = mh->snt_time;
  else
  if (mh->rcv_time && !mh->snt_time)
	mh->snt_time = mh->rcv_time;

  mh->header_len = msg_tell(mfd);

  return mh;
}

int
compare_folders(folder1, folder2)
struct _mail_folder **folder1, **folder2;
{
int cmpres = 0;

  if (!*folder1 || !*folder2 || (*folder1 == *folder2))
	return 0;

  switch (folder_sort & 0x0f) {
   case BY_NAME:
	cmpres = strcmp((*folder1)->name(*folder1) , (*folder2)->name(*folder2));
   break;

   case BY_PRIOR:
	cmpres = ( (*folder1)->flags - (*folder2)->flags );
   break;

   case BY_MSGNUM:
	cmpres = ((*folder1)->num_msg - (*folder2)->num_msg);
   break;

   case BY_UNREAD:
	cmpres = ( (*folder1)->unread_num - (*folder2)->unread_num);
   break;
		}

   return (folder_sort & MSG_ASCEND) ? cmpres : (cmpres * -1);
}

void sort_folders()
{
int i = 0;

 if ((folder_sort & 0x0f) == NO_SORT)
	return;

 while ((i < folders_num) && (mailbox[i]->flags & SYSTEM))
	i++;

 qsort(&mailbox[i], folders_num - i, sizeof(struct _mail_folder *),
	(int(*)(const void*, const void*))compare_folders);

 folder_sort |= FLD_SORTED;
 return;
}

int
compare_msgs(msg1, msg2)
struct _mail_msg **msg1, **msg2;
{
int cmpres = 0;
struct _mail_msg *m1 = *msg1, *m2 = *msg2;

 if (!*msg1 || !*msg2 || (*msg1 == *msg2))
	return 0;

 if (!(*msg1)->header || !(*msg2)->header)
	return 0;

 if (sort_type & MSG_THREAD) 	{
  struct _mail_msg **msgs1 = malloc(m1->refs * sizeof(struct _mail_msg*));
  struct _mail_msg **msgs2 = malloc(m2->refs * sizeof(struct _mail_msg*));

   while(m1->ref) {
    msgs1[m1->refs - 1] = m1;
    m1 = m1->ref;
    /* msg1 > msg2 because of a reference */
    if(m1 == *msg2) 	{
      free(msgs1);
      free(msgs2);
      return 1;
			}
		  } 
  
  while(m2->ref)  {
    msgs2[m2->refs - 1] = m2;
    m2 = m2->ref;
    /* here msg1 < msg2 */
    if(m2 == *msg1)	{
      free(msgs1);
      free(msgs2);
      return -1;
			}
		 } 

  if(m1 == m2) { /* the thread starters are identical */
   int i = 0;

    /* => find first referenced message, where they differ */
    while(msgs1[i] == msgs2[i])
	i++;
    m1 = msgs1[i];
    m2 = msgs2[i];
  }

  free(msgs1);
  free(msgs2);
				}

 switch(sort_type & 0x0f)	{
  case NO_SORT:
	return 0;
  break;

  case BY_SNT_TIME:
   cmpres = m1->header->snt_time - m2->header->snt_time;
  break;

  case BY_RCV_TIME:
   cmpres = m1->header->rcv_time - m2->header->rcv_time;
  break;

  case BY_SUBJECT:
   if (!m1->header->Subject) {
	cmpres = -1;
	break; }
   else
   if (!m2->header->Subject) {
	cmpres = 1;
	break; }

   cmpres = strcmp(m1->header->Subject, m2->header->Subject);
  break;

  case BY_AUTHOR:
   if (!m1->header->From) {
	cmpres = -1;
	break; }
   else
   if (!m2->header->From) {
	cmpres = 1;
	break; }

   cmpres = strcmp(get_short_addr_line(m1->header->From), get_short_addr_line(m2->header->From) );
  break;

  case BY_FLAGS:
   cmpres = m1->flags - m2->flags;
  break;

  case BY_SIZE:
   cmpres = m1->msg_len - m2->msg_len;
  break;

  default:
	return 0;
  break;
		}

  return (sort_type & MSG_ASCEND) ? cmpres : (cmpres * -1);
}

unsigned long
hash(name)
const char *name;
{
unsigned long key = 0;
unsigned char c;

  while ((c = *name++) && (c != '>')) 	{
	key += c;
	key ^= (key << 24) | (key >> 8 );
  					}

  return key;
}

void
make_entry(struct _ht *ht, unsigned long key, int htsize,
          char *s, struct _mail_msg *msg)
{
unsigned long nkey = key;

  while(ht[nkey].id) { /* solve key collision */
    nkey++;
    if(nkey >= htsize)
      nkey = 0;
  		     }

  if(nkey != key) {
    while(ht[key].next < htsize)
      key = ht[key].next;
    ht[key].next = nkey;
    key = nkey;
  		  }    

  ht[key].id = s;
  ht[key].msg = msg;

  return;
}

struct _mail_msg *
find_entry(struct _ht *ht, unsigned long key,
          int htsize, char *s)
{
int cnt = strchr(s,'>') - s + 1;

  while((key < htsize) && ht[key].id)
    if(strncmp(ht[key].id, s, cnt)) /* we have a key collision */
      key = ht[key].next;
    else                          /* found the key */
      break;

  return key < htsize ? ht[key].msg : NULL;
}

char *
remove_lead_trail_blanks(name)
char *name;
{
int len;

  if(!strncasecmp(name,"Re:",3))
    name += 3;

  while(*name == ' ')
    name++;

  len = strlen(name);
  while(*(name+len-1) == ' ')
    *(name+(--len)) = '\0';

  return name;
}

void
sort_folder(folder)
struct _mail_folder *folder;
{
struct _mail_msg *msg, **marr;
int total, i;

 if (!folder)
	return;

 if (!folder->messages) {
 	folder->flags |= SORTED;
	return;		}

 if ((sort_type & 0x0f) == NO_SORT)
	return;

 for (total = 0,msg = folder->messages; msg; msg = msg->next, total++) ;

 if ((marr = (struct _mail_msg **)malloc(sizeof(struct _mail_msg *) * total)) == NULL) {
	display_msg(MSG_WARN, "sort", "Failed to allocate memory");
	return; }

 for (total = 0,msg = folder->messages; msg; msg = msg->next, total++)
  {
	marr[total] = msg;
	msg->flags &= ~S_THREAD;
  }

 if (sort_type & MSG_THREAD) 	{
  /* should be enough for the hash table */
  int htsize = 2 << ((int)(log(total) / M_LN2));
  struct _ht *ht;

  if ((ht = (struct _ht *)malloc(htsize * sizeof(struct _ht))) == NULL)
  {
      display_msg(MSG_FATAL, "sort", "Malloc failed");
      return;
  }

  for (total = 0; total < htsize; total++) {
    ht[total].id = NULL;
    ht[total].msg = NULL;
    ht[total].next = htsize;
					   }

  for (total = 0,msg = folder->messages; msg; msg = msg->next, total++) {
   struct _head_field *m_id = find_field(msg,"Message-ID");
   char *s;

    if (m_id &&
	((s = strchr(m_id->f_line,'<')) != NULL)) {
      unsigned long key = hash(s) % htsize;
      make_entry(ht, key, htsize, s, msg);
						  }
    msg->ref = NULL;
    msg->refs = 0;

  									}

  for (total = 0,msg = folder->messages; msg; msg = msg->next, total++) {
   struct _head_field *m_id;
   struct _mail_msg *ref;
   unsigned long key;
   char *s;

    m_id = find_field(msg,"In-Reply-To");
    if((m_id == NULL) || (strchr(m_id->f_line,'<') == NULL))
	m_id=find_field(msg,"References");

    /* we have a "References" or a "In-Reply-To" field */
    if(m_id && (s=strchr(m_id->f_line,'<')))	{
	key = hash(s) % htsize;

	/* there actually is a "parent" message ... */
	if((ref = find_entry(ht, key, htsize, s)) != NULL) {
	 char *s1, *s2, *subj1, *subj2;

         if (msg->header->Subject)
           subj1 = strdup(msg->header->Subject);
	 else
	   subj1 = strdup("");

         if (ref->header->Subject)
           subj2 = strdup(ref->header->Subject);
	 else
	   subj2 = strdup("");

         s1 = remove_lead_trail_blanks(subj1);
         s2 = remove_lead_trail_blanks(subj2);

	  if(!strcmp(s1,s2))   {
		msg->ref=ref;
		msg->flags |= S_THREAD;
			       }

	  free(subj1);
	  free(subj2);
					      }
		}
									}

	free(ht);
	for (total = 0,msg = folder->messages; msg; msg = msg->next, total++) {
	 struct _mail_msg *m = msg;
	 int i = 0;

	  while (m->ref) {
		m = m->ref;
		i++;
			 }
	  msg->refs=i;
									      } 

				}

 qsort(marr, total, sizeof(struct _mail_msg *),
	(int(*)(const void*, const void*))compare_msgs);

 folder->messages = marr[0];

 total--;

 for (i = 0; i < total; i++)
	marr[i]->next = marr[i + 1];

 marr[total]->next = NULL;

 free(marr);

 folder->flags |= SORTED;
 return;
}

struct _mail_msg
	*create_message(folder)
struct _mail_folder *folder;
{
int new_num;
#ifdef FACES
char *face;
#endif
struct _mail_msg *message;
msg_header *mh;
FILE *mfd;
char buf[255];

 if (!folder)
	folder = ftemp;

 if (folder->num_msg >= MAX_MSG_IN_FOLDER) {
	display_msg(MSG_WARN, "Folder is full", "%s", folder->name(folder));
	return NULL;		}

 if ((new_num = get_new_name(folder)) == -1) {
	display_msg(MSG_WARN, "Can not choose file name for a new message in", "%s", folder->name(folder));
	return NULL; }

  message = (struct _mail_msg *)malloc(sizeof(struct _mail_msg));
  if (!message) {
	display_msg(MSG_FATAL, "malloc", "Can not allocate memory");
	return NULL; }

  local_message(message);

  message->msg_len = 0;
  message->folder = folder;
  message->num = new_num;
  message->data = 0;
  message->pdata = NULL;
  message->ref = NULL;
  message->refs = 0;

  if (!(mh = (msg_header *)malloc(sizeof(msg_header)))) {  
        display_msg(MSG_FATAL, "malloc", "Can not allocate memory!");
        return NULL; }

  mh->other_fields = NULL;
  mh->flags = 0;
  mh->To = mh->Sender = mh->Cc = mh->Bcc = NULL;
  mh->From = get_address(b_getcfg_str(conf_name, "from", ""));

  if (b_getcfg_int(conf_name, "smtpbcc", 0))
   mh->Bcc = get_address(b_getcfg_str(conf_name, "from", ""));

  mh->snt_time = 0L;
  mh->rcv_time = 0L;
  mh->Subject = NULL;

  message->header = mh;
  message->flags = 0;
  message->msg_body = NULL;
  message->mime = NULL;
  message->next = folder->messages;
  folder->num_msg++;
  folder->messages = message;
  
  if (cfg_exist("organization"))
   add_field(message, "Organization", b_getcfg_str(conf_name, "organization", ""));

#ifdef FACES
  if ((face = xbm_to_face()) != NULL)
   add_field(message, FACEFIELD, face);
#endif

/*
  add_field(message, "Reply-To", mh->From->addr);
*/

  add_field(message, MIME_C_ENCR, supp_encodings[def_encoding].encoding_name);

  sprintf(buf, "%s/%s; charset=%s", mailcap[DEFAULT_MAILCAP].type_text, mailcap[DEFAULT_MAILCAP].subtype_text, supp_charsets[def_charset].charset_name);
  add_field(message, MIME_C_TYPE, buf);

  sprintf(buf, "%02d", MIME_VERS_SUPP);
  buf[2] = buf[1];
  buf[1] = '.';
  buf[3] = '\0';
  add_field(message, MIME_VERSION, buf);

  message->header->snt_time = message->header->rcv_time =  time(NULL);
  add_field(message, "Date", get_arpa_date(time(NULL)));

  if ((mfd = fopen(message->get_file(message), "w")) == NULL) {
     display_msg(MSG_WARN, "update", "Can not open file %s", message->get_file(message));
     unlink(message->get_file(message));
     discard_message(message);
     return NULL; }

  if (fchmod(fileno(mfd), 00600) == -1)
        display_msg(MSG_WARN, "Can not change mod of", "%s", message->get_file(message));


  print_message_header(message, mfd);

  fflush(mfd);
  if ((message->header->header_len = ftell(mfd)) == -1) {
        display_msg(MSG_WARN, "Can not determine file length", "%s", message->get_file(message));
	unlink(message->get_file(message));
	discard_message(message);
	return NULL;
							}

  add_signature(message, mfd, 0);
  fflush(mfd);

  if ((message->msg_len = ftell(mfd)) == -1) 		{
        display_msg(MSG_WARN, "Can not determine file length", "%s", message->get_file(message));
	unlink(message->get_file(message));
	discard_message(message);
	return NULL;
							}

  fclose(mfd);

  return message;
}

int
dump_field(hf, dfd)
struct _head_field *hf;
FILE *dfd;
{
struct _head_field hfld;

 if (!hf) {
	hfld.num_fields = 0;
	hfld.f_name[0] = '\0';
	hfld.f_line = NULL;
	hfld.next_head_field = NULL;
	hf = &hfld; }

 if (fwrite(hf, sizeof(struct _head_field), 1, dfd) != 1)
	return -1;

 if (hf->f_line) {
    if (*hf->f_line)
     if (fwrite(hf->f_line, strlen(hf->f_line), 1, dfd) != 1)
	return -1; 

    fputc('\n', dfd);
	}

 return 0;
}

int
dump_addr(addr, dfd)
struct _mail_addr *addr;
FILE *dfd;
{
struct _mail_addr dmp_addr;

 if (!addr) {
	dmp_addr.num = 0;
	dmp_addr.addr = NULL;
	dmp_addr.name = NULL;
	dmp_addr.comment = NULL;
	dmp_addr.next_addr = NULL;
	addr = &dmp_addr; }

 if (fwrite(addr, sizeof(struct _mail_addr), 1, dfd) != 1)
	return -1;

 if (addr->addr) {
    if (fwrite(addr->addr, strlen(addr->addr), 1, dfd) != 1)
	return -1; 

    fputc('\n', dfd);
	}

 if (addr->name) {
    if (fwrite(addr->name, strlen(addr->name), 1, dfd) != 1)
	return -1; 

    fputc('\n', dfd);
	}

 if (addr->comment) {
    if (fwrite(addr->comment, strlen(addr->comment), 1, dfd) != 1)
	return -1; 

    fputc('\n', dfd);
	}

 return 0;

}

int
dump_msg(msg)
struct _mail_msg *msg;
{
char dfname[255];
FILE *dfd;
struct _mail_addr *addr;
struct _head_field *hf;

 if (!msg)
	return -1;

 if (!msg->header)
	return -1;

 sprintf(dfname, "%s/%s/%d", msg->folder->fold_path, CACHE_DIR, msg->num);


 if ((dfd = fopen(dfname, "w")) == NULL) {
	display_msg(MSG_WARN, "cache", "Can not open %s", dfname);
	return -1; }

 if (fwrite(msg, sizeof(struct _mail_msg), 1, dfd) != 1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 if (fwrite(msg->header, sizeof(struct _msg_header), 1, dfd) != 1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 if (msg->header->Subject) {
    fwrite(msg->header->Subject, strlen(msg->header->Subject), 1, dfd);
    fputc('\n', dfd);
	}

 addr = msg->header->From;
 while(addr) {
  if (dump_addr(addr, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  addr = addr->next_addr;
	}

 if (dump_addr(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 addr = msg->header->To;
 while(addr) {
  if (dump_addr(addr, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  addr = addr->next_addr;
	}

 if (dump_addr(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 addr = msg->header->Sender;
 while(addr) {
  if (dump_addr(addr, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  addr = addr->next_addr;
	}

 if (dump_addr(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 addr = msg->header->Cc;
 while(addr) {
  if (dump_addr(addr, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  addr = addr->next_addr;
	}

 if (dump_addr(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 addr = msg->header->Bcc;
 while(addr) {
  if (dump_addr(addr, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  addr = addr->next_addr;
	}

 if (dump_addr(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 hf = msg->header->other_fields;
 while (hf) {
  if (dump_field(hf, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

  hf = hf->next_head_field;
	}

 if (dump_field(NULL, dfd) == -1) {
	display_msg(MSG_WARN, "cache", "Can not write to %s", dfname);
	fclose(dfd);
	return -1; }

 fclose(dfd);
 return 0;
}

struct _head_field
	*restore_field(dfd)
FILE *dfd;
{
struct _head_field hfld, *hf;
char hline[MAX_FIELD_LEN];

 if (fread(&hfld, sizeof(struct _head_field), 1, dfd) != 1)
	return NULL;

 if (hfld.f_name[0] == '\0')
	return NULL;

 if ((hf = (struct _head_field *)malloc(sizeof(struct _head_field))) == NULL) {
	display_msg(MSG_WARN, "restore", "malloc failed");
	return NULL; }

 memcpy((void *)hf, &hfld, sizeof(struct _head_field));

 if (hf->f_line) {
    if (fgets(hline, MAX_FIELD_LEN, dfd) == NULL) {
	free(hf);
	return NULL; }

    hline[strlen(hline) - 1] = '\0';
    hf->f_line = strdup(hline);
	}

 hf->num_fields = 0;
 hf->next_head_field = NULL;

 return hf;
}

struct _mail_addr
	*restore_addr(dfd)
FILE *dfd;
{
struct _mail_addr addr, *ad;
char aline[255];

 if (fread(&addr, sizeof(struct _mail_addr), 1, dfd) != 1)
	return NULL;

 if (addr.addr == NULL)
	return NULL;

 if ((ad = (struct _mail_addr *)malloc(sizeof(struct _mail_addr))) == NULL) {
	display_msg(MSG_WARN, "restore", "malloc failed");
	return NULL; }

 memcpy((void *)ad, &addr, sizeof(struct _mail_addr));

 if (fgets(aline, 254, dfd) == NULL) {
	free(ad);
	return NULL; }

 aline[strlen(aline) - 1] = '\0';
 ad->addr = strdup(aline);

 if (ad->name) {
   if (fgets(aline, 254, dfd) == NULL) {
	free(ad->addr);
	free(ad);
	return NULL; }

   aline[strlen(aline) - 1] = '\0';
   ad->name = strdup(aline);
	}

 if (ad->comment) {
   if (fgets(aline, 254, dfd) == NULL) {
	free(ad->addr);
	free(ad);
	return NULL; }

   aline[strlen(aline) - 1] = '\0';
   ad->comment = strdup(aline);
	}

 ad->num = 0;
 ad->next_addr = NULL;

 return ad;
}

struct _mail_msg *restore_msg(num, folder)
int num;
struct _mail_folder *folder;
{
struct _mail_msg msg, *nmsg;
struct _msg_header header, *nheader;
struct _mail_addr *addr;
struct _head_field *fld;
char dfname[255], msg_file[255];
char hline[MAX_FIELD_LEN];
FILE *dfd;
struct stat st1, st2;

 sprintf(dfname, "%s/%s/%d", folder->fold_path, CACHE_DIR, num);
 sprintf(msg_file, "%s/%d", folder->fold_path, num);

 if (lstat(dfname, &st2) == -1) 
	return NULL;

 if (lstat(msg_file, &st1) == -1) {
	display_msg(MSG_WARN, "restore", "Can not stat %s", msg_file);
	return NULL; }

 if (st1.st_mtime > st2.st_mtime)
	return NULL;

 if (st1.st_ctime > st2.st_ctime) 
	return NULL;

 if ((dfd = fopen(dfname, "r")) == NULL)
	return NULL;

 if (fread(&msg, sizeof(struct _mail_msg), 1, dfd) != 1) {
	display_msg(MSG_WARN, "restore", "Can not read from %s", dfname);
	fclose(dfd);
	return NULL; }

 if ((nmsg = (struct _mail_msg *)malloc(sizeof(struct _mail_msg))) == NULL) {
	display_msg(MSG_WARN, "restore", "malloc failed");
	fclose(dfd);
	return NULL; }

 memcpy((void *)nmsg, &msg, sizeof(struct _mail_msg));

 local_message(nmsg);

 if (fread(&header, sizeof(struct _msg_header), 1, dfd) != 1) {
	free(nmsg);
	fclose(dfd);
	return NULL; }

 if ((nheader = (struct _msg_header *)malloc(sizeof(struct _msg_header))) == NULL) {
	display_msg(MSG_WARN, "restore", "malloc failed");
	free(nmsg);
	fclose(dfd);
	return NULL; }

 memcpy((void *)nheader, &header, sizeof(struct _msg_header));

 nmsg->header = nheader;
 nmsg->folder = NULL;
 nmsg->next = NULL;
 nmsg->msg_body = NULL; 
 nmsg->mime = NULL; 
 nmsg->pdata = NULL; 
 nmsg->ref = NULL; 
 nmsg->refs = 0; 
 nmsg->data = 0; 

 if (nheader->Subject) {
   if (fgets(hline, MAX_FIELD_LEN - 1, dfd) != NULL) {
	hline[strlen(hline) - 1] = '\0';
	nheader->Subject = strdup(hline); }
	}

 nheader->From = NULL;
 while((addr = restore_addr(dfd)) != NULL) {
	addr->next_addr = nheader->From;
	nheader->From = addr; 
	}

 nheader->To = NULL;
 while((addr = restore_addr(dfd)) != NULL) {
	addr->next_addr = nheader->To;
	nheader->To = addr; 
	}

 nheader->Sender = NULL;
 while((addr = restore_addr(dfd)) != NULL) {
	addr->next_addr = nheader->Sender;
	nheader->Sender = addr; 
	}

 nheader->Cc = NULL;
 while((addr = restore_addr(dfd)) != NULL) {
	addr->next_addr = nheader->Cc;
	nheader->Cc = addr; 
	}

 nheader->Bcc = NULL;
 while((addr = restore_addr(dfd)) != NULL) {
	addr->next_addr = nheader->Bcc;
	nheader->Bcc = addr; 
	}

 nheader->other_fields = NULL;
 while ((fld = restore_field(dfd)) != NULL) {
	fld->next_head_field = nheader->other_fields;
	nheader->other_fields = fld;
	}

 fclose(dfd);

 return nmsg;
}

struct _mail_msg
	*get_message(num, folder)
int num;
struct _mail_folder *folder;
{

int msg;
char str[255];
struct _mail_msg *message;
struct stat sb;
FILE *fmsg = NULL;
int redump;

  if (!folder)
	folder = ftemp;

  redump = 0;

 if (folder->flags & CACHED) 	{
     if ((message = restore_msg(num, folder)) == NULL)
	redump = 1;
     else {
	message->folder = folder;
	message->num = num;
	return message;
	  }
				}

 sprintf(str, "%s/%d", folder->fold_path, num);

#ifdef __FreeBSD__
  if (!(msg = open(str, O_RDONLY | O_CREAT | O_EXLOCK | O_NONBLOCK))) {
#else
  if (!(msg = open(str, O_RDONLY | O_CREAT))) {
#endif
	display_msg(MSG_WARN, NULL, "Can not open message file %s", str);
	return NULL; }

#ifndef MMAP_MSG
  if (!(fmsg = fdopen(msg, "r"))) {
	close(msg);
	fclose(fmsg);
	return NULL; }
#endif

  if (fstat(msg, &sb)) {
	display_msg(MSG_WARN, "Can not stat message", "Maybe it's locked. %s", str);
	close(msg);
#ifndef MMAP_MSG
	fclose(fmsg);
#endif
	return NULL;  }

  if (sb.st_size > MAX_MSG_LEN) {
	display_msg(MSG_WARN, "Message is too big", "len(%s) > %d Kb", str, MAX_MSG_LEN/1024);
	close(msg);
#ifndef MMAP_MSG
	fclose(fmsg);
#endif
	return NULL;  }

  message = (struct _mail_msg *)malloc(sizeof(struct _mail_msg));
  if (!message) {
	display_msg(MSG_FATAL, "malloc", "Can not allocate memory");
	close(msg);
#ifndef MMAP_MSG
	fclose(fmsg);
#endif
	return NULL; }

  local_message(message);

#ifdef MMAP_MSG
  mmlen = ((size_t)sb.st_size < MAX_HEADER_LEN) ? (size_t)sb.st_size : MAX_HEADER_LEN;
  if ((mmsg = (unsigned char *)mmap(NULL, mmlen , PROT_READ, MAP_PRIVATE, msg, (off_t)0)) == (unsigned char *)-1)
	{
	display_msg(MSG_WARN, "mmap failed on", "%s", str);
	close(msg);
	free(message);
	return NULL;
	}

  mmpos = 0;
#endif

  message->msg_len = sb.st_size;
  message->header = get_msg_header(fmsg);
  message->flags = message->header->flags;
  message->msg_body = NULL;
  message->next = NULL;
  message->mime = NULL;
  message->folder = folder;
  message->num = num;
  message->pdata = NULL;
  message->ref = NULL;
  message->refs = 0;
  message->data = MSG_DAT_EMPTY;

#ifdef MMAP_MSG
  munmap(mmsg, mmlen);
  mmsg = NULL;
  mmlen = 0;
  mmpos = 0;
#else
  fclose(fmsg);
#endif

  close(msg);

  if (redump) 
	dump_msg(message);

  return message;

}

struct _mail_msg * 
get_fwd_msg(msg)
struct _mail_msg *msg;
{
struct _mail_msg *fwd_msg;
struct _head_field *fld, *hf;
struct _mime_msg *mtext;
FILE *mfd, *tfd;
char buf[255], *p;
int i, fwdmime;

 if (!msg)
	return NULL;

 if (!msg->header)
	return NULL;

 fwd_msg = create_message(outbox);
 if (!fwd_msg)
	return NULL;

 if (!fwd_msg->header)
	return NULL;

 if (msg->header->Subject && !strncasecmp(msg->header->Subject , "FW:", 3))
   fwd_msg->header->Subject = strdup(msg->header->Subject);
 else {
	sprintf(buf, "FW: %s", msg->header->Subject ? msg->header->Subject : "");
	fwd_msg->header->Subject =  strdup(buf);
	}

 if (msg->header->To)
   add_field(fwd_msg, "Resent-To", get_full_addr_line(msg->header->To));

 if (msg->header->From)
   add_field(fwd_msg, "Resent-From", get_full_addr_line(msg->header->From));

 if ((fld = find_field(msg, "Message-Id")))
   add_field(fwd_msg, "Resent-Message-Id", fld->f_line);

 if ((fld = find_field(msg, "Date")))
   add_field(fwd_msg, "Resent-Date", fld->f_line);

/* Preserving Reply-To seems to be confusing */
/*
 if ((fld = find_field(msg, "Reply-To")))
   replace_field(fwd_msg, "Reply-To", fld->f_line);
 else {
   if (msg->header->From)
	replace_field(fwd_msg, "Reply-To", msg->header->From->addr);
	}
*/
 
 fwdmime = b_getcfg_int(conf_name, "fwdinclude", 2);
 if (fwdmime == 4) 	{
	if (attach_file(fwd_msg, msg->get_file(msg), &mailcap[MESSAGE_MAILCAP], DEFAULT_ENCODING , ATT_NODISPINFO|ATT_NOFILEINFO) == NULL)  {
		display_msg(MSG_WARN, "forward", "Failed to attach message");
		discard_message(fwd_msg);
		return NULL;
	     						    }
	return fwd_msg;
			}

 fwd_msg->flags &= ~SIGNED;

 if ((mfd = fopen(fwd_msg->get_file(fwd_msg), "w")) == NULL) {
     display_msg(MSG_WARN, "update", "Can not open file %s", fwd_msg->get_file(fwd_msg));
     return fwd_msg; }

 print_message_header(fwd_msg, mfd);
 fflush(mfd);
 fwd_msg->header->header_len = ftell(mfd);
 msg->print_body(msg, mfd);
 fflush(mfd);
 fwd_msg->msg_len = ftell(mfd);
 fclose(mfd);

 if ((hf = find_field(msg, MIME_C_TYPE)) != NULL)
	replace_field(fwd_msg, MIME_C_TYPE, hf->f_line);

 if ((hf = find_field(msg, MIME_VERSION)) != NULL)
	replace_field(fwd_msg, MIME_VERSION, hf->f_line);

 if ((hf = find_field(msg, MIME_C_ENCR)) != NULL)
	replace_field(fwd_msg, MIME_C_ENCR, hf->f_line);

 if ((hf = find_field(msg, MIME_C_DESCR)) != NULL)
	replace_field(fwd_msg, MIME_C_DESCR, hf->f_line);

 if ((hf = find_field(msg, MIME_C_ID)) != NULL)
	replace_field(fwd_msg, MIME_C_ID, hf->f_line);

 if ((hf = find_field(msg, MIME_C_LENGTH)) != NULL) 
	replace_field(fwd_msg, MIME_C_LENGTH, hf->f_line);

 if ((mtext = get_text_part(fwd_msg)) == NULL)
	return fwd_msg;

 if ((mfd = fopen(fwd_msg->get_file(fwd_msg), "r")) == NULL) {
            display_msg(MSG_WARN, "forward", "Can not open %s", fwd_msg->get_file(fwd_msg));
                return fwd_msg; }

 fseek(mfd, mtext->m_start, SEEK_SET);
   if (mtext->boundary) {
	while(fgets(buf, 255, mfd))   {
	   if (strlen(buf) <= 1)
                break;
                                      }
                        }

 strcpy(buf, get_temp_file("fwd"));

 mtext->src_info = strdup(buf);

 if ((tfd = fopen(buf, "w")) == NULL) {
            display_msg(MSG_WARN, "forward", "Can not open %s", buf);
	    fclose(mfd);
            return NULL; }

 fputc('\n', tfd);
 strcpy(buf, b_getcfg_str(conf_name, "fwdstr", "-----FW: %i-----%n%n"));
 expand_str(msg, buf);
 fputs(buf, tfd);

 switch (fwdmime) {
  default:
  break;

  case 2:
	hf = find_field(msg, "Date");
	fprintf(tfd, "Date: %s\n", hf->f_line ? hf->f_line : get_arpa_date(msg->header->rcv_time));

	print_addr(msg->header->Sender, "Sender", tfd, 0);
	print_addr(msg->header->From, "From", tfd, 0);
	print_addr(msg->header->To, "To", tfd, 0);
	if (msg->header->Subject)
		fprintf(tfd, "Subject: %s\n", msg->header->Subject);

	print_addr(msg->header->Cc, "Cc", tfd, 0);
	print_addr(msg->header->Bcc, "Bcc", tfd, 0);

	fprintf(tfd, "\n");
  break;

  case 3:
	print_message_header(msg, tfd);
  break;
						   }

 while ((ftell(mfd) < mtext->m_end) && fgets(buf, 255, mfd)) {
	if ((p = (*mtext->encoding->ce_dec)(buf, &i)) == NULL)
                continue;

	fputs(p, tfd);
							     }

 fputs("\n-------------End of forwarding message-------------------------\n", tfd);
 add_signature(fwd_msg, tfd, 0);
 fclose(tfd);

 update_mime(fwd_msg);

 return fwd_msg;
}

struct _mail_msg * 
get_vac_msg(msg, vfile)
struct _mail_msg *msg;
char *vfile;
{
struct _mail_msg *vac_msg;
struct _mail_addr *addr;
struct _head_field *mid;
FILE *vfd, *mfd;
char buf[255];

 if (!msg)
	return NULL;

 if (!msg->header)
	return NULL;

 vac_msg = create_message(outbox);
 if (!vac_msg)
	return NULL;

 if (!vac_msg->header)
	return NULL;

 if ((vfd = fopen(vfile, "r")) == NULL) {
	display_msg(MSG_WARN, "vacation", "Can not open %s", vfile);
	vac_msg->delete(vac_msg);
	return NULL;
					}

 if (msg->header->Subject && !strncasecmp(msg->header->Subject, "RE:", 3))
   sprintf(buf, "I'm on vacation (%s)", msg->header->Subject ? msg->header->Subject : "");
 else 
   sprintf(buf, "I'm on vacation (RE: %s)", msg->header->Subject ? msg->header->Subject : "");

 vac_msg->header->Subject = strdup(buf);

 if ( (mid = find_field(msg, "Reply-To")) &&
   (addr = get_address(mid->f_line)))
	vac_msg->header->To = copy_address(addr);
 else
	vac_msg->header->To = copy_address(msg->header->From);

 if ((mfd = fopen(vac_msg->get_file(vac_msg), "w")) == NULL) {
     display_msg(MSG_WARN, "update", "Can not open file %s", vac_msg->get_file(vac_msg));
     return vac_msg; }

 print_message_header(vac_msg, mfd);
 fflush(mfd);
 vac_msg->header->header_len = ftell(mfd);

 while (fgets(buf, 255, vfd))
	fputs(buf, mfd);

 fflush(mfd);
 vac_msg->msg_len = ftell(mfd);

 fclose(mfd);
 fclose(vfd);

 return vac_msg;
}

void
add_signature(msg, file, forceattach)
struct _mail_msg *msg;
FILE *file;
int forceattach;
{
FILE *sf;
char *s_file, buf[255];
char date[9];
time_t tm;
int sl = 0;
int i, sattach;

 if (file == NULL)
	return;

 if (!forceattach) {
    sattach = b_getcfg_int(conf_name, "sattach", 2);
    if ((msg->flags & SIGNED) || (sattach != 2))
	return;
		   }

 if ((s_file = get_sign_file(msg)) == NULL)
	return;

 if ((sf = fopen(s_file, "r")) == NULL) {
        display_msg(MSG_WARN, "Can not open signature file for reading", "%s", s_file);
        return;
					}

 tm = time(NULL);

 fseek(sf, 0L, SEEK_SET);

 fputc('\n', file);

 while(fgets(buf, 255, sf)) {

  if (!strchr(buf, '$'))
    fputs(buf, file);
  else {
    for (i = 0; i < strlen(buf);i++) {
	if ((buf[i] == '$') && (buf[i+1] != '\0')) {
	 switch(buf[i+1]) {
		case 't':
			strftime(date, 9, "%T", localtime(&tm));
			fputs(date, file);
		break;

		case 'd':
			strftime(date, 9, "%D", localtime(&tm));
			fputs(date, file);
		break;

		case 'm':
			fputs(b_getcfg_str(conf_name, "from", "Ghost <ghost@cemetery.com>"), file);
		break;

		case '$':
			fputc('$', file);
		break;

		default:
			fputc(buf[i], file);
			fputc(buf[i+1], file);
		break;
			 }
	  i++;
						   }
	else
	 fputc(buf[i], file);
				    }
	}

  sl++;

 if (sl > 24) 
	break;
                           }

 fclose(sf);
 msg->flags |= SIGNED;
 return;
}

void
add_signature_to_msg(msg, tfile)
struct _mail_msg *msg;
char *tfile;
{
FILE *mfd;

 if (!tfile)
	return;

 if ((mfd = fopen(tfile, "a+")) == NULL)
	return;

 add_signature(msg, mfd, 1);
 fclose(mfd);

 return;
}

void
insert_orig(tfile, msg, type)
char *tfile;
struct _mail_msg *msg;
int type;
{
int i;
FILE *fd, *mfd;
char buf[255], tmp[255];
struct _mime_msg *mtext;
char prefix[17], *p;

 if (!tfile || !msg)
	return;

 sprintf(tmp, "%s_ins", tfile);

 if ((fd = fopen(tmp, "w")) == NULL) {
	display_msg(MSG_WARN, "Can not open message file", "%s", tmp);
	return;  }

 fputc('\n', fd);
 if ((mtext = get_text_part(msg)) == NULL)
	return;

 if ((mfd = fopen(msg->get_file(msg), "r")) == NULL) {
            display_msg(MSG_WARN, "reply", "Can not open %s", msg->get_file(msg));
                return; }

 fseek(mfd, mtext->m_start, SEEK_SET);
 if (mtext->boundary) {
	while(fgets(buf, 255, mfd))   {
	   if (strlen(buf) <= 1)
                break;
                                      }
                      }

 if (type)
	fputs("-------------Original message follows----------------------\n", fd);

 strncpy(prefix, b_getcfg_str(conf_name, "prefix", ">"), 16);

 if (type == 0) {
	strcpy(buf, b_getcfg_str(conf_name, "replystr", "On %d %f wrote:%n"));
	expand_str(msg, buf);
	fputs(buf, fd);
	fputs(prefix, fd);
		}

 while ((ftell(mfd) < mtext->m_end) && fgets(buf, 255, mfd))    {
	if (type == 0)
		fputs(prefix, fd);

	if ((p = (*mtext->encoding->ce_dec)(buf, &i)) == NULL)
                continue;

	fputs(p, fd);
								}

 if (type)
	fputs("-----------------------------------------------------------\n", fd);

 fclose(mfd);
 if ((mfd = fopen(tfile, "r")) == NULL) {
	display_msg(MSG_WARN, "Can not open file", "%s", tfile);
	return;	}
 
 while(fgets(buf, 255, mfd))
	fputs(buf, fd);

 fclose(fd);
 fclose(mfd);

 if (rename(tmp, tfile) == -1) {
	display_msg(MSG_WARN, "Can not rename", "%s to %s", tmp, tfile);
	unlink(tmp);
				}

 return;
}

void
add_file_to_msg(tfile, file)
char *tfile;
char *file;
{
FILE *fd, *mfd;
char buf[255], tmp[255], *p, *ln;
struct stat sb;
int swarn, eos, len;

#define	INS_WARN_NOTEXT	0x01
#define	INS_WARN_BIGSTR	0x02

 if (!tfile || !file)
	return;

 sprintf(tmp, "%s_ins", tfile);
 swarn = 0;

 if (stat(file, &sb) == -1) {
	display_msg(MSG_WARN, "Can not access file", "%s", file);
	return;
			    }

 if (sb.st_size >= MAX_MSG_LEN) {
	display_msg(MSG_WARN, "Insert file", "File is too big, can not insert");
	return;
				}

 if ((sb.st_size > MAX_INSFILE_SIZE) &&
	!display_msg(MSG_QUEST, "File is too big , use MIME to attach big files", "Are you still want to insert it?"))
	return;

 if ((fd = fopen(file, "r")) == NULL) {
	display_msg(MSG_WARN, "Can not open file", "%s", file);
	return;	}
 
 if ((mfd = fopen(tmp, "w")) == NULL) {
	display_msg(MSG_WARN, "Can not open message file", "%s", tmp);
	return;  }

 eos = 0;
 while(fgets(buf, MAX_INSFILE_LINE, fd)) {
	if (graph_str(buf) &&
	  !(swarn & INS_WARN_NOTEXT)) 	{
	 display_msg(MSG_WARN, "insert file", "This is not a text file\nand it could be inserted incorrectly");
	 swarn |= INS_WARN_NOTEXT;
					}

	ln = buf;
npart:
	if ((p = strchr(ln, '\n')) == NULL)
		len = strlen(ln);
	else
		len = p - ln;

	if ((len + eos) > MAX_INSFILE_LINE) 	{
	  if (!(swarn & INS_WARN_BIGSTR)) {
		display_msg(MSG_WARN, "insert file", "Some lines in this file are too long\nand will be splitted");
		swarn |= INS_WARN_BIGSTR;
					  }
	  len = MAX_INSFILE_LINE - eos;
	  fwrite(ln, len, 1, mfd);
	  fputc('\n', mfd);
	  ln += len;
	  eos = 0;
	  goto npart;
				}
	else	{
		if ((p = strrchr(ln , '\n')) != NULL)
			eos += (strlen(ln) - (p - ln) - 1);
		else
			eos += len;
		fputs(ln, mfd);
		}

			    }
 fputc('\n', mfd);

 fclose(fd);
 if ((fd = fopen(tfile, "r")) == NULL) {
	display_msg(MSG_WARN, "Can not open file", "%s", tfile);
	return;	}
 
 while(fgets(buf, 255, fd))
	fputs(buf, mfd);

 fclose(fd);
 fclose(mfd);

 if (rename(tmp, tfile) == -1) {
	display_msg(MSG_WARN, "Can not rename", "%s to %s", tmp, tfile);
	unlink(tmp);
				}

 return;
}

void
read_confirm(msg)
struct _mail_msg *msg;
{
struct _mail_msg *rep_msg;
struct _head_field *hf;
FILE *mfd;
char buf[255];

 if (!(hf = find_field(msg, XFRECEIPT))) {
	if (!(hf = find_field(msg, "X-Chameleon-Return-To")))
	return; }

 if (!hf->f_line)
	return;

 rep_msg = create_message(outbox);
 if (!rep_msg)
	return;

 if (!rep_msg->header)
	return;

 rep_msg->header->To = get_address(hf->f_line);
 if (!rep_msg->header->To)
	return;

 sprintf(buf, "Receipt: %s", msg->header->Subject ? msg->header->Subject : "");
 rep_msg->header->Subject =  strdup(buf);

 if (unlink(rep_msg->get_file(rep_msg)) == -1) {
        display_msg(MSG_WARN, "Cannot delete old copy of message", "%s", rep_msg->get_file(rep_msg));
        return; }

 if ((mfd = fopen(rep_msg->get_file(rep_msg), "w")) == NULL) {
     display_msg(MSG_WARN, "update", "Can not open file %s", rep_msg->get_file(rep_msg));
     return; }

 print_message_header(rep_msg, mfd);

 fflush(mfd);
 rep_msg->header->header_len = ftell(mfd);

 fputc('\n', mfd);
 fputs("-------------Message header follows----------------------\n", mfd);
 print_message_header(msg, mfd);
 fputs("---------------------------------------------------------\n", mfd);
 fputc('\n', mfd);

 fflush(mfd);
 rep_msg->msg_len = ftell(mfd);
 fclose(mfd);

 if (send_message(rep_msg) == -1)
	display_msg(MSG_WARN, "Failed to send confirmation message", "%s", msg->header->Subject ? msg->header->Subject : "* No Subject *");

 return;
}

int
incl_in_reply(msg, addr)
struct _mail_msg *msg;
struct _mail_addr *addr;
{
int replyex;
char *replyexand, *p;
struct _mail_addr *addr1;

  if (!msg || !addr)
	return 0;

  replyex = b_getcfg_int(conf_name, "replyex", 1);
  if (replyex && msg->header->From &&
	!strcasecmp(msg->header->From->addr, addr->addr))
	return 0;

  replyexand = b_getcfg_str(conf_name, "replyexand", "");
  if (strlen(replyexand)) {
	p = strtok(replyexand, ";:, ");
	if (p && !strcasecmp(p, addr->addr))
		return 0;

	while (p && ((p = strtok(NULL, ";:, ")) != NULL)) {
	  if (!strcasecmp(p, addr->addr))
		return 0;
							  }
			  }

  addr1 = msg->header->To;
  while (addr1) {
	if (!strcasecmp(addr1->addr, addr->addr))
		return 0;
	addr1 = addr1->next_addr;
		}

  addr1 = msg->header->Cc;
  while (addr1) {
	if (!strcasecmp(addr1->addr, addr->addr))
		return 0;
	addr1 = addr1->next_addr;
		}

  addr1 = msg->header->Bcc;
  while (addr1) {
	if (!strcasecmp(addr1->addr, addr->addr))
		return 0;
	addr1 = addr1->next_addr;
		}

  return 1;
}

struct _mail_msg *
get_reply_msg(msg, to_all)
struct _mail_msg *msg;
int to_all;
{
struct _mail_msg *rep_msg;
struct _mail_addr *addr, *addr1;
struct _head_field *mid;
struct _head_field *hf;
FILE *mfd, *tfd;
int include, i;
char prefix[17];
char buf[255], *p;
struct _mime_msg *mtext;

 if (!msg)
	return NULL;

 if (!msg->header)
	return NULL;

 rep_msg = create_message(outbox);

 if (!rep_msg)
	return NULL;

 if (!rep_msg->header)
	return NULL;

 rep_msg->data = MSG_DAT_ORIG;
 rep_msg->pdata = (void *)msg;
 rep_msg->flags &= ~SIGNED;
 msg->flags |= LOCKED;

 if (msg->header->Subject && !strncasecmp(msg->header->Subject, "RE:", 3))
   rep_msg->header->Subject =  strdup(msg->header->Subject);
 else 	{
   sprintf(buf, "RE: %s", msg->header->Subject ? msg->header->Subject : "");
   rep_msg->header->Subject = strdup(buf);
	}

 if ((mid = find_field(msg, "Message-ID")))
   add_field(rep_msg, "In-Reply-To", mid->f_line);

 addr = NULL;

 if ( (mid = find_field(msg, "Reply-To")) &&
   (addr = get_address(mid->f_line)) &&
     display_msg(MSG_QUEST, "Reply to Reply-To?", "(%s)", mid->f_line)) {
   rep_msg->header->To = addr;
   addr = NULL;
   if (msg->header->From)
	rep_msg->header->Cc = copy_address(msg->header->From);
						}
 else
 if (msg->header->From) 
   rep_msg->header->To = copy_address(msg->header->From);
 else
 if (msg->header->Sender)
   rep_msg->header->To = copy_address(msg->header->Sender);
 else
   rep_msg->header->To = NULL;

 discard_address(addr);

 if (!to_all)
	goto incl_repl;

 addr = msg->header->To;

 while (addr) {
   if (!incl_in_reply(rep_msg, addr)) {
	addr = addr->next_addr;
	continue; }

   addr1 = copy_address(addr);
   if (!addr1) {
	addr = addr->next_addr;
	continue; }

   if (rep_msg->header->Cc)
	addr1->num = rep_msg->header->Cc->num + 1;

  addr1->next_addr = rep_msg->header->Cc;
  rep_msg->header->Cc = addr1;
  addr = addr->next_addr;
	}

 addr = msg->header->Cc;
 while (addr) {
   if (!incl_in_reply(rep_msg, addr)) {
	addr = addr->next_addr;
	continue; }

   addr1 = copy_address(addr);
   if (!addr1) {
	addr = addr->next_addr;
	continue; }

   if (rep_msg->header->Cc)
	addr1->num = rep_msg->header->Cc->num + 1;

  addr1->next_addr = rep_msg->header->Cc;
  rep_msg->header->Cc = addr1;
  addr = addr->next_addr;
	}

incl_repl: if ((mfd = fopen(rep_msg->get_file(rep_msg), "w")) == NULL) {
     display_msg(MSG_WARN, "update", "Can not open file %s", rep_msg->get_file(rep_msg));
     return rep_msg; }

 print_message_header(rep_msg, mfd);
 fflush(mfd);
 rep_msg->header->header_len = ftell(mfd);
 msg->print_body(msg, mfd);
 fflush(mfd);
 rep_msg->msg_len = ftell(mfd);;

 fclose(mfd);

 if ((hf = find_field(msg, MIME_C_TYPE)) != NULL)
	replace_field(rep_msg, MIME_C_TYPE, hf->f_line);

 if ((hf = find_field(msg, MIME_VERSION)) != NULL) 
	replace_field(rep_msg, MIME_VERSION, hf->f_line);

 if ((hf = find_field(msg, MIME_C_ENCR)) != NULL)
	replace_field(rep_msg, MIME_C_ENCR, hf->f_line);

 if ((hf = find_field(msg, MIME_C_DESCR)) != NULL)
	replace_field(rep_msg, MIME_C_DESCR, hf->f_line);

 if ((hf = find_field(msg, MIME_C_ID)) != NULL)
	replace_field(rep_msg, MIME_C_ID, hf->f_line);

 if ((hf = find_field(msg, MIME_C_LENGTH)) != NULL)
	replace_field(rep_msg, MIME_C_LENGTH, hf->f_line);

 if ((mtext = get_text_part(rep_msg)) == NULL)
	return rep_msg;

 if ((mfd = fopen(rep_msg->get_file(rep_msg), "r")) == NULL) {
            display_msg(MSG_WARN, "reply", "Can not open %s", rep_msg->get_file(rep_msg));
                return rep_msg; }

 fseek(mfd, mtext->m_start, SEEK_SET);
   if (mtext->boundary) {
	while(fgets(buf, 255, mfd))   {
	   if (strlen(buf) <= 1)
                break;
                                      }
                        }

 strcpy(buf, get_temp_file("rep"));
 mtext->src_info = strdup(buf);

 if ((tfd = fopen(buf, "w")) == NULL) {
            display_msg(MSG_WARN, "reply", "Can not open %s", buf);
	    fclose(mfd);
            return NULL; }

 include = b_getcfg_int(conf_name, "include", 2);
 if (include == 1) 
	goto add_sign;

 if ((include != 2) && (include != 3))
	goto add_sign;

 fputc('\n', tfd);

 if (include == 3)
	fputs("-------------Original message follows----------------------\n", tfd);

 strncpy(prefix, b_getcfg_str(conf_name, "prefix", ">"), 16);

 if (include == 2) {
	strcpy(buf, b_getcfg_str(conf_name, "replystr", "On %d %f wrote:%n"));

	expand_str(msg, buf);
	fputs(buf, tfd);

	fputs(prefix, tfd);
		}

 while ((ftell(mfd) < mtext->m_end) && fgets(buf, 255, mfd))    {
	if (include == 2)
		fputs(prefix, tfd);

	if ((p = (*mtext->encoding->ce_dec)(buf, &i)) == NULL)
                continue;

	fputs(p, tfd);
								}

 if (include == 3)
	fputs("-----------------------------------------------------------\n", tfd);

add_sign:	add_signature(rep_msg, tfd, 0);

 fclose(tfd);
 update_mime(rep_msg);

 return rep_msg;
}

int
get_msg_index(folder, msg)
struct _mail_folder *folder;
struct _mail_msg *msg;
{
int i = 0;
struct _mail_msg *msg1 = folder ? folder->messages : NULL;

 while (msg1) {
	if (msg == msg1)
		return i;
	msg1 = msg1->next;
	i++;
		}

 return -1;
}

struct _mail_msg *
get_msg_by_index(folder, index)
struct _mail_folder *folder;
int index;
{
int i = 0;
struct _mail_msg *msg1 = folder ? folder->messages : NULL;

 while (msg1) {
	if ( i == index)
		return msg1;
	msg1 = msg1->next;
	i++;  }

return NULL;
}

int
get_folder_index(folder)
struct _mail_folder *folder;
{
int i;

 if (!folder)
	return 0;

 if (!folders_num)
	return 0;

 for (i = 0; i< folders_num;i++) {
	if (mailbox[i] == folder)
		return i;
				}

 return 0;
}

struct _mail_folder *
get_folder_by_index(index)
int index;
{
 if (index >= folders_num)
	return mailbox[0];

 return mailbox[index];
}

char *
get_folder_name(folder)
struct _mail_folder *folder;
{
char *p;

 if (!folder)
	return "noname";

 if ((p = strrchr(folder->fold_path, '/')) == NULL)
	return "noname";
 p++;
 if (strlen(p) < 1)
	return "noname";

 return p;
}

void
close_all_folders()
{
int i;

   for ( i = 0; i < folders_num; i++) {
        if (mailbox[i]) {
              mailbox[i]->close(mailbox[i]);
              free(mailbox[i]);
                        }
                                      }

  ftemp->close(ftemp);
  free(ftemp);
}

char *
get_msg_file(msg)
struct _mail_msg *msg;
{
static char mfile[255];

  sprintf(mfile, "%s/%d", msg->folder ? msg->folder->fold_path : ftemp->fold_path, msg->num);

  return mfile;
}

void
local_message(msg)
struct _mail_msg *msg;
{
 msg->delete = delete_message;
 msg->print = print_message;
 msg->print_body = print_message_body;
 msg->get_text = get_message_text;
 msg->free_text = free_message_text;
 msg->get_file = get_msg_file;
 msg->update = update_message;
 msg->flags |= M_MH;
}

void
local_folder(folder)
struct _mail_folder *folder;
{
 folder->name = get_folder_name;
 folder->open = open_folder;
 folder->rescan = rescan_folder;
 folder->close = close_folder;
 folder->empty = empty_folder;
 folder->delete = delete_folder;
 folder->rename = rename_folder;
 folder->update = update_folder;
 folder->move = move_to_folder;
 folder->search = find_text;
 folder->flags |= F_MH;
}

int
open_all_folders(path)
char *path;
{
char maildir[255];
char buf[255];
DIR *dirp;
struct dirent *dp;
struct stat sb;
int i;
struct _mail_folder *fld;

  trash = NULL;
  inbox = NULL;
  outbox = NULL;
  sentm = NULL;
  draft = NULL;
  folders_num = 0;

  for (i = 0; i < MAX_FOLDERS; i++)
	mailbox[i] = NULL;

  if (!path || (strlen(path) < 1))
        sprintf(maildir, "%s/Mail", getenv("HOME")); 
  else  
        strcpy(maildir, path);

  if (!(dirp = opendir(maildir)))       {
	if (display_msg(MSG_QUEST, NULL, "Directory %s does not exists , create?", maildir)) {
        if (mkdir(maildir, 00700) == -1)
          return -1;
        if (!(dirp = opendir(maildir)))
          return -1;
	     }
	else
	  return -1;
                                }

  folders_num = 0;
  strcpy(mailbox_path, maildir);

	if ((inbox = get_folder_by_name(INBOX)) == NULL) {
	if ((inbox = rescan_folder(INBOX)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", INBOX);
		return -1;
							}
							}
	local_folder(inbox);

	if ((outbox = get_folder_by_name(OUTBOX)) == NULL) {
	if ((outbox = rescan_folder(OUTBOX)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", OUTBOX);
		return -1;
							}
							}
	local_folder(outbox);

	if ((trash = get_folder_by_name(TRASH)) == NULL) {
	if ((trash = rescan_folder(TRASH)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", TRASH);
		return -1;			}
							}
	local_folder(trash);

	if ((sentm = get_folder_by_name(SENTMAIL)) == NULL) {
	if ((sentm = rescan_folder(SENTMAIL)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", SENTMAIL);
		return -1;			}
							}
	local_folder(sentm);

	if ((draft = get_folder_by_name(DRAFT)) == NULL) {
	if ((draft = rescan_folder(DRAFT)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", DRAFT);
		return -1;			}
							}
	local_folder(draft);

	if ((template = get_folder_by_name(TEMPLATE)) == NULL) {
	if ((template = rescan_folder(TEMPLATE)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", TEMPLATE);
		return -1;			}
							}
	local_folder(template);

	if ((ftemp = get_folder_by_name(FTEMP)) == NULL) {
	if ((ftemp = rescan_folder(FTEMP)) == NULL) {
		display_msg(MSG_WARN, "INIT", "Can not scan folder %s", FTEMP);
		return -1;			}
							}
	local_folder(ftemp);
	if (open_folder(FTEMP) != ftemp) {
		display_msg(MSG_WARN, "INIT", "Can not open folder %s", FTEMP);
		return -1; }

  while ((dp = readdir(dirp)) != NULL) 	{
#if defined(SOLARIS) || defined(HP) || defined(__linux__) || defined(SGI)
        if ((strlen(dp->d_name) > MAX_FOLD_NAME_LEN) || (strlen(dp->d_name) < 2))
#else
        if ((dp->d_namlen > MAX_FOLD_NAME_LEN) || (dp->d_namlen < 2))
#endif
        	continue;

	if (!isalpha(dp->d_name[0]))
		continue;

#if defined(SOLARIS) || defined(HP) || defined(__linux__) || defined(SGI)
	for ( i = 1; i < strlen(dp->d_name); i++) {
#else
	for ( i = 1; i < dp->d_namlen; i++) {
#endif
	   if (!isgraph(dp->d_name[i]))
		goto next;
					    }	
	sprintf(buf, "%s/%s", maildir, dp->d_name);

	if (stat(buf, &sb) == -1)
		continue;

	if (!(sb.st_mode & S_IFDIR))
		continue;

	if ((fld = get_folder_by_name(dp->d_name)) != NULL)
		continue;

	if ((fld = rescan_folder(dp->d_name)) == NULL)
		continue;

	local_folder(fld);

next:	if (folders_num >= MAX_FOLDERS)
		break;

					}

	inbox->flags  |= SYSTEM;
	trash->flags  |= SYSTEM;
	outbox->flags |= SYSTEM;
	sentm->flags  |= SYSTEM;
	draft->flags  |= SYSTEM;
	template->flags  |= SYSTEM;
	ftemp->flags  |= SYSTEM;


  closedir(dirp);
  sort_folders();

  return 0;
}

struct _mail_folder *
get_folder_by_name(name)
char *name;
{
int i;

  if (!name)
	return NULL;

  if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN))
	return NULL;

  for ( i = 0; i < folders_num; i++) 	{
	if (!strcmp(mailbox[i]->name(mailbox[i]), name))
		return mailbox[i];
					}

  return NULL;
}

struct _mail_folder
	*rescan_folder(name)
char *name;
{
DIR *dirp;
struct dirent *dp;
int msg_num, ftmp;
char msg_nam[5];
mail_folder *fld;
char path[255], buf[255];
int reopen;
struct stat sb;

  if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN))
	return NULL;

  if (!strcmp(name, FTEMP))
	ftmp = 1;
  else
	ftmp = 0;

  if (!ftmp && !isalpha(name[0])) {
	display_msg(MSG_WARN, "rescan", "First character of folder's name must be letter (%s)", name);
	return NULL; }

  if (strchr(name, '/')) {
	display_msg(MSG_WARN, "rescan", "Folder's name can not contain '/' (%s)", name);
	return NULL; }

  if (strlen(mailbox_path) < 1)
	return NULL;

  sprintf(path, "%s/%s", mailbox_path, name);

  reopen = 0;
  fld = NULL;

  if (ftmp && ftemp) {
	reopen = 1;
	fld = ftemp; }
  else {
	if ((fld = get_folder_by_name(name)) != NULL)
		reopen = 1;
	}

  if (!reopen && (folders_num >= MAX_FOLDERS))  {
	display_msg(MSG_WARN, "rescan", "Can not open more then %d folders",
		MAX_FOLDERS);
	return NULL;
						}


  if (stat(path, &sb) != -1) 	{

    if (!(sb.st_mode & S_IFDIR))
   	{
	if (display_msg(MSG_QUEST, "init folder", "%s is not a directory, save it?", path)) {
		sprintf(buf, "%s.bak", path);
		if (rename(path, buf) == -1) {
		display_msg(MSG_WARN, "init folder", "Cannot rename %s", path);
		return NULL;
				}
		else
		display_msg(MSG_WARN, path, "was saved to %s", buf);
	} else
		return NULL;
   	}
				}

  if (!(dirp = opendir(path))) 	{
	if (mkdir(path, 00700) == -1)
	  return NULL;
	if (!(dirp = opendir(path)))
	  return NULL;
				}

  if (!reopen) {
  fld = (struct _mail_folder *)malloc(sizeof(mail_folder));
  if (!fld) {
	display_msg(MSG_FATAL, "malloc", "Can not allocate memory");
  	closedir(dirp);
	return NULL; }
	       }

  fld->num_msg = 0;
  fld->unread_num = 0;
  if (!reopen) {
  fld->messages = NULL;
  fld->flags = 0;
  strcpy(fld->fold_path, path);
  local_folder(fld);
		}

  while ((dp = readdir(dirp)) != NULL) {
	if (strlen(dp->d_name) > 4)
	 continue;

	msg_num = atoi(dp->d_name);
	sprintf(msg_nam, "%d", msg_num);
	if (strcmp(msg_nam, dp->d_name))
		continue;

	if (msg_num >= MAX_MSG_IN_FOLDER)
		continue;
	fld->num_msg++;
	if (fld->num_msg >= MAX_MSG_IN_FOLDER)
		break;
                   			}
  closedir(dirp);
  if (!ftmp && !reopen) {
  mailbox[folders_num] = fld;
  folders_num++;
	       }

  if (fld->num_msg > b_getcfg_int(conf_name, "mincache", 512)) {
	fld->flags |= CACHED;
	sprintf(buf, "%s/%s", fld->fold_path, CACHE_DIR);
	mkdir(buf, 00700); }

  return fld;
}

int
match_msg_text(msg, exp, where)
struct _mail_msg *msg;
void *exp;
char *where;
{
int fr;
char buf[255];
char *p, *p1, *p2;
struct _xf_rule rule;
regex_t *rx;

 strcpy(buf, where);
 p2 = buf;

#ifdef HS_REGEX
 rx = (regex_t *)exp;
 rule.rx = *rx;
#else
 strncpy(rule.tmatch, (char *)exp, 63);
 rule.tmatch[63];
#endif

do {
 p1 = p2;

 if ((p2 = strchr(p1, ',')) != NULL) 
	*(p2++) = '\0';

 if (!strcmp(p1, "To")) 	{
    if (match_addr(msg->header->To, &rule))
	return 1;
    continue;
				}
 else
 if (!strcmp(p1, "From")) 	{
    if (match_addr(msg->header->From, &rule))
	return 1;
    continue;
				}
 else
 if (!strcmp(p1, "Cc")) 	{
    if (match_addr(msg->header->Cc, &rule))
	return 1;
    continue;
				}
 else
 if (!strcmp(p1, "Bcc")) 	{
    if (match_addr(msg->header->Bcc, &rule))
	return 1;
    continue;
				}
 else
 if (!strcmp(p1, "Sender")) 	{
    if (match_addr(msg->header->Sender, &rule))
	return 1;
    continue;
				}

 fr = 0;
 if ((p = get_field_content(msg, p1, &fr)) == NULL)
	continue;

#ifdef HS_REGEX
        if (regexec((regex_t *)exp, p, 0, 0, 0) == 0) {
		free_field_content(msg, p, fr);
		return 1;
							}
#else
	if (strstr(p, (char *)exp)) 
		free_field_content(msg, p, fr);
		return 1;
							}
#endif

	free_field_content(msg, p, fr);
} while(p2 != NULL);

 return 0;
}

int
find_text(folder, text, where, flags)
struct _mail_folder *folder;
char *text;
char *where;
int flags;
{
struct _mail_msg *msg;
int opened;
int found;
int num;
char progress[5]; 
char fldname[16];

#ifdef HS_REGEX
static regex_t rx;
int regflags =  REG_NOSUB | REG_EXTENDED;

 if (!text)
	return -1;

 if (flags & S_CASE)
	regflags |= REG_ICASE;

 if (regcomp(&rx, text, regflags)) {
	display_msg(MSG_WARN, "search", "Invalid regular expression %s", text);
	regfree(&rx);
	return -1; }
#endif

 if (folder == NULL) {
#ifdef HS_REGEX
	regfree(&rx);
#endif
	return -1; }

 opened = 0;
 found = 0;
 num = 0;
 strncpy(progress, "|/-\\", 4);

 if (!(folder->flags & OPENED)) {
	if (folder->open(folder->name(folder)) == NULL) {
#ifdef HS_REGEX
		regfree(&rx);
#endif
		return -1; }

	opened = 1;
			}

 if (folder->messages == NULL) {
	folder->close(folder);
#ifdef HS_REGEX
	regfree(&rx);
#endif
	return 0; }

 strncpy(fldname, folder->name(folder), 15);
 fldname[15] = '\0';

 msg = folder->messages;
 while(msg) {
	if (abortpressed()) 	{
		found = -1;
		break;
				}

#ifdef HS_REGEX
	if (match_msg_text(msg, &rx, where ? where : "Header")) {
#else
	if (match_msg_text(msg, text, where ? where : "Header")) {
#endif
		found++;
		msg->flags |= MARKED;
			}
	else
		msg->flags &= ~MARKED;

	display_msg(MSG_STAT, NULL, "Searching in %s: %c", fldname, progress[num%4]);
	num++;
	msg = msg->next;
	}

 if (opened)
	folder->close(folder);
 else 	{
  if (found > 0) 
	folder->update(folder);
	}

#ifdef HS_REGEX
	regfree(&rx);
#endif

 return found;
}

struct _mail_folder
	*open_folder(name)
char *name;
{
DIR *dirp;
struct dirent *dp;
int i, msg_num;
char msg_nam[16];
mail_folder *fld;
mail_msg *msg, *msg1;
char buf[255];
char path[255];
int reopen, ftmp;
struct stat sb;
int prev_msg, prev_unread;

  if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN)) {
	display_msg(MSG_WARN, "open folder", "Invalid folder name %s", name);
	return NULL; }

  if (strrchr(name, '/')) {
	display_msg(MSG_WARN, "open folder", "folder name can not contain /");
	return NULL; }

  if (!strcmp(name, FTEMP))
        ftmp = 1;
  else
        ftmp = 0;

 if (!ftmp && !isalpha(name[0])) {
	display_msg(MSG_WARN, "open folder", "first character of folder name must be letter");
	return NULL; }

  if (strlen(mailbox_path) < 1)
	return NULL;

  sprintf(path, "%s/%s", mailbox_path, name);

  reopen = 0;
  fld = NULL;
  prev_msg = prev_unread = 0;

  if (ftmp && ftemp) {
	reopen = 1;
	fld = ftemp; }
  else
  for (i = 0; i < folders_num; i++)
	{
	if (mailbox[i] && !strcasecmp(mailbox[i]->fold_path, path)) 	{
	 reopen = 1;
         fld = mailbox[i];
	 if ((fld->flags & OPENED) || (fld->messages))
		fld->close(fld);
	 prev_msg = fld->num_msg;
	 prev_unread = fld->unread_num;

	 break;
									}
	}

  if (!reopen && (folders_num >= MAX_FOLDERS))  {
	display_msg(MSG_WARN, "open folder", "Can not have more then %d folders", MAX_FOLDERS);
	return NULL;
						}


  if (stat(path, &sb) != -1) 	{

    if (!(sb.st_mode & S_IFDIR))
   	{
	if (display_msg(MSG_QUEST, "open folder", "%s is not a directory, save it?", path)) {
		sprintf(buf, "%s.bak", path);
		if (rename(path, buf) == -1) {
		display_msg(MSG_WARN, "open folder", "Cannot rename %s", path);
		return NULL;
				}
		else
		display_msg(MSG_WARN, path, "was saved to %s", buf);
	} else
		return NULL;

   	}
				}

  if (!(dirp = opendir(path))) 	{
	if (display_msg(MSG_QUEST, "open folder", "%s does not exists , create?", path)) {
	if (mkdir(path, 00700) == -1)
	  return NULL;
	if (!(dirp = opendir(path)))
	  return NULL;
	} else 
	return NULL;
				}

  if (!reopen) {
  fld = (struct _mail_folder *)malloc(sizeof(mail_folder));
  if (!fld) {
	display_msg(MSG_FATAL, "malloc", "Can not allocate memory");
  	closedir(dirp);
	return NULL; }
  fld->flags = 0;
  strcpy(fld->fold_path, path);

  fld->messages = NULL;
  local_folder(fld);
	       }

  fld->num_msg = 0;
  fld->unread_num = 0;

  msg = NULL;

  while ((dp = readdir(dirp)) != NULL) {
	if (strlen(dp->d_name) > 4)
	 continue;

	msg_num = atoi(dp->d_name);
	sprintf(msg_nam, "%d", msg_num);
	if (strcmp(msg_nam, dp->d_name))
		continue;

	if (msg_num >= MAX_MSG_IN_FOLDER)
		continue;

	if (!(msg = get_message(msg_num, fld)))
		continue;

 	msg1= fld->messages;
	while(msg1) {
	  if (msg1->flags & LOCKED) {
		if (msg1->num == msg_num)
			{
			discard_message(msg);
			if (msg1->flags & UNREAD)
			   fld->unread_num++;
			fld->num_msg++;
			goto next_msg;
		    	}
		   }
	 msg1 = msg1->next;
		    }

	if (msg->flags & UNREAD)
		fld->unread_num++;
	msg->folder = fld;
	msg->next = fld->messages;
	fld->messages = msg;
	fld->num_msg++;
	if ((fld->num_msg % 5) == 1) {
		display_msg(MSG_STAT, NULL, "Rescanning %s: %d", name, fld->num_msg);
		if (abortpressed()) 	{
		  closedir(dirp);
		  fld->close(fld);
		  fld->num_msg = prev_msg;
		  fld->unread_num = prev_unread;
		  return NULL;
					}
				    }

next_msg:	if (fld->num_msg >= MAX_MSG_IN_FOLDER)
		break;
                   			}
  closedir(dirp);
  if (!ftmp && !reopen) {
  mailbox[folders_num] = fld;
  folders_num++;
	       }

  fld->flags |= OPENED;
  fld->flags &= ~SEARCH;
  if ((prev_msg != fld->num_msg) ||
	(prev_unread != fld->unread_num) ||
	!reopen)
	folder_sort &= ~FLD_SORTED;
  sort_folder(fld);
  return fld;

}

int
get_message_text(msg)
struct _mail_msg *msg;
{
int mfd;
  if (msg->msg_body)
	return 0;
 
  if ((mfd = open(msg->get_file(msg), O_RDONLY)) == -1)
	return -1;


  if ((msg->msg_body = (char *)mmap(NULL, (size_t)msg->msg_len, PROT_READ, MAP_PRIVATE, mfd, (off_t)0)) == (char *)-1) 
	{
	close(mfd);
	return -1; 
	}

  close(mfd);
  return 0;
}

void
free_message_text(msg)
struct _mail_msg *msg;
{
  if (msg->msg_body == NULL)
        return;

  munmap(msg->msg_body, (size_t)msg->msg_len);
  msg->msg_body = NULL;
  return;

}

struct _mail_addr *
copy_address(addr)
struct _mail_addr *addr;
{
struct _mail_addr *ma;

 if (!addr)
	return NULL;

 ma = (struct _mail_addr *)malloc(sizeof(struct _mail_addr));
 if (!ma) {
	display_msg(MSG_FATAL, "malloc", "malloc failed in copy address");
	return NULL;
	}

 if (addr->addr)
	ma->addr = strdup(addr->addr);
 else
	ma->addr = NULL;
 
 if (addr->name)
	ma->name = strdup(addr->name);
 else
	ma->name = NULL;

 if (addr->comment)
	ma->comment = strdup(addr->comment);
 else
	ma->comment = NULL;

 ma->next_addr = addr->next_addr;
 ma->num = addr->num;
 
 return ma;
}

struct _mail_addr *
copy_address_chain(addr)
struct _mail_addr *addr;
{
struct _mail_addr *ma, *ma1;

 ma = NULL;

 while(addr) {
	ma1 = copy_address(addr);
	ma1->next_addr = ma;
	ma = ma1;

	addr = addr->next_addr;
	}

 return ma;
}

struct _mail_msg *
copy_message(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{
struct _mail_msg *nmsg;
int cnum;
char cpath[255];
FILE *cfd;

  if (!msg || !folder)
	return NULL;

 if (folder->num_msg > MAX_MSG_IN_FOLDER) {
	display_msg(MSG_WARN, "copy", "Folder %s is full", folder->name(folder));
	return NULL; }

  if ((cnum = get_new_name(folder)) == -1) {
	display_msg(MSG_WARN, "copy", "Can not create new message in %s", folder->name(folder));
	return NULL;
	}

  sprintf(cpath, "%s/%d", folder->fold_path, cnum);

  if (!(cfd = fopen(cpath, "w"))) {
	display_msg(MSG_WARN, "copy", "Can not open %s", cpath);
	return NULL; }

  msg->print(msg, cfd, 0);
  fclose(cfd);

  folder->num_msg++;
  if (msg->flags & UNREAD)
	folder->unread_num++;

  if ((folder->flags & OPENED) || (msg->flags & LOCKED))
	{
  if (!(nmsg = get_message(cnum, folder)))
	return NULL;

  nmsg->flags = msg->flags;
  nmsg->folder = folder;
  nmsg->next = folder->messages;
  folder->messages = nmsg;
  folder->flags &= ~SORTED;
  return nmsg;
	}

  return NULL; 
}

void
discard_address(addr)
struct _mail_addr *addr;
{
struct _mail_addr *addr1;
 while(addr) {
	addr1 = addr->next_addr;
	if (addr->addr)
		free(addr->addr);

	if (addr->name)
		free(addr->name);

	if (addr->comment)
		free(addr->comment);

	free(addr);
	addr = addr1;
		}
return;
}


void
discard_message(msg)
struct _mail_msg *msg;
{
head_field *hf, *hf1;

 if (!msg)
	return;

 if (msg->msg_body)
	msg->free_text(msg);

 if (msg->header) {
 discard_address(msg->header->From);
 discard_address(msg->header->To);
 discard_address(msg->header->Sender);
 discard_address(msg->header->Cc);
 discard_address(msg->header->Bcc);

 hf = msg->header->other_fields;
 while(hf) {
	hf1 = hf->next_head_field;
	if (hf->f_line)
	 free(hf->f_line);
	free(hf);
	hf = hf1;
	   }

 if (msg->header->Subject)
	free(msg->header->Subject);

 free(msg->header);
		}

 if (msg->mime)
	discard_mime(msg->mime);

 free(msg);

 return;
}

char *
dir_path(path)
char *path;
{
static char dir[255];
char *p;

 dir[0] = '.';
 dir[1] = '\0';

 if (!path)
	return dir;

 if (strrchr(path, '/') == NULL)
	return dir;

 strcpy(dir, path);
 p = strrchr(dir, '/');
 p[0] = '\0';
 return dir;

}

char *
name_path(path)
char *path;
{
static char name[255];
char *p;

 name[0] = '.';
 name[1] = '\0';

 if (!path)
	return name;

 strcpy(name, path);

 if ((p = strrchr(path, '/')) == NULL)
	return name;
 
 p++;
 strcpy(name, p);
 return name;
}

int
rename_folder(folder, name)
struct _mail_folder *folder;
char *name;
{
char new_path[255];
struct stat sb;
int closed;

 if (!folder)
	return -1;

 closed = 0;
 if (folder->messages)  {
	closed = folder->flags;
	folder->close(folder);
	if (folder->messages) {
		display_msg(MSG_WARN, "rename", "Can not rename folder with active messages");
		if (closed & OPENED)
			folder->open(folder->name(folder));
		return -1;
				}
	closed = 1;
			}

 if ((strlen(name) < 1) || (strlen(name) > MAX_FOLD_NAME_LEN)) {
	display_msg(MSG_WARN, "rename", "Invalid folder name %s", name);
	return -1; }

 if (get_folder_by_name(name)) {
	display_msg(MSG_WARN, "rename", "folder with name %s already exists", name);
	return -1; }

 sprintf(new_path, "%s/%s", dir_path(folder->fold_path), name);

 if (stat(new_path, &sb) != -1) 	{

    if (!(sb.st_mode & S_IFDIR))
   	{
	if (unlink(new_path) == -1)
		return -1;
   	}
    else
	{
	if (rmdir(new_path) == -1)
		return -1;
	}
				}

 if (rename(folder->fold_path, new_path) == -1)
	return -1;

 strcpy(folder->fold_path, new_path);
 if (closed)
	folder->open(folder->name(folder));

 folder_sort &= ~FLD_SORTED;
 return 0;
 
}

int
delete_folder(folder)
struct _mail_folder *folder;
{
DIR *dirp;
struct dirent *dp;
char buf[255];
int i;
struct stat sb;

 if (!folder)
	return -1;

 if (folder->flags & SYSTEM) {
	display_msg(MSG_WARN, "delete", "Can not delete system folder %s", folder->name(folder));
	return -1; }

 folder->empty(folder);
 if (folder->num_msg || folder->unread_num) {
	display_msg(MSG_WARN, "delete", "Can not delete all messages in folder %s", folder->name(folder));
	return -1; }

 if (stat(folder->fold_path, &sb) != -1) {
    if (!(sb.st_mode & S_IFDIR))
   	{
	if (unlink(folder->fold_path) == -1)
		return -1;
   	}
					 }

 if (!(dirp = opendir(folder->fold_path))) 
	return -1;
 while ((dp = readdir(dirp)) != NULL)  {
	if (!strcmp(dp->d_name, "."))
		continue;

	if (!strcmp(dp->d_name, ".."))
		continue;

	sprintf(buf, "%s/%s", folder->fold_path, dp->d_name);
	if (unlink(buf) == -1) {
		closedir(dirp);
		return -1;     }
					}

 closedir(dirp);

 if (rmdir(folder->fold_path) == -1)
	return -1;

 folder_sort &= ~FLD_SORTED;

 for (i = 0; i < folders_num; i++) {
   if (mailbox[i] == folder) {
	if (folders_num)
		folders_num--;
	free(folder);
	for (; i < folders_num;i++)
	  mailbox[i] = mailbox[i+1];
	return 0;
				}
				   }

 return 0;

}

void
empty_folder(folder)
struct _mail_folder *folder;
{
DIR *dirp;
struct dirent *dp;
char buf[255];
int oflags;

 if (!folder)
	return;

 oflags = folder->flags;

 if ((folder->flags & OPENED) || (folder->messages))
	folder->close(folder);

 if (folder->messages)
	goto reop;

 if (!(dirp = opendir(folder->fold_path)))
        return;

 while ((dp = readdir(dirp)) != NULL)  {

	if (abortpressed()) {
                closedir(dirp);
		return;
			}

	if (!isdigit(dp->d_name[0]))
		continue;

        sprintf(buf, "%s/%s", folder->fold_path, dp->d_name);
        if (unlink(buf) == -1) {
                closedir(dirp);
		display_msg(MSG_WARN, "Failed to delete", "%s", buf);
                return;     }

	display_msg(MSG_STAT, NULL, "Deleting %s", dp->d_name);
                                        }

 closedir(dirp);
 folder->num_msg = folder->unread_num = 0;

reop: if (oflags & OPENED)
	folder->open(folder->name(folder));

 folder_sort &= ~FLD_SORTED;
 return;
}

void
close_folder(folder)
struct _mail_folder *folder;
{
struct _mail_msg *msg, *msg1, *msg_lock;

 if (!folder)
	return;

 folder->flags &= ~OPENED;
 folder->flags &= ~SORTED;
 folder->flags &= ~SEARCH;

 if (!folder->messages)
	return;

 folder->update(folder);

 msg = folder->messages;
 msg_lock = NULL;
 while(msg) {
	msg1 = msg->next;
	if (msg->flags & LOCKED) {
	  msg->next = msg_lock;
	  msg_lock = msg; }
	else 
	  discard_message(msg);

	msg = msg1;
	    }

 folder->messages = msg_lock;

 return;
}

/* this is ugly :-( */

int
get_new_name(folder)
struct _mail_folder *folder;
{
static char buf[255];
int i, k;

 if (!folder)
	folder = ftemp;

 for (i = 1;i <= MAX_MSG_IN_FOLDER; i += 10) {
	sprintf(buf, "%s/%d", folder->fold_path, i);
	if (access(buf, R_OK)) {

	  if (i == 1)
	      return i;

	  for (k = i-9; k < i; k++) {
	      sprintf(buf, "%s/%d", folder->fold_path, k);
	      if (access(buf, R_OK)) 
		return k;
				    }

	  return i;
				}
					}

 return -1;
}

void
update_folder(folder)
struct _mail_folder *folder;
{
struct _mail_msg *msg, *msg_nxt;
struct _mail_folder *fld;
int k;

 if (!folder)
        return;

 if (!folder->messages)
	return;

 msg = folder->messages;
 msg_nxt = folder->messages->next;
 k = 0;
 while(msg) {

	if (msg->flags & LOCKED) {
		msg->flags &= ~DELETED;
		msg->flags &= ~MOVED;
	  	msg->flags &= ~NOT_SENT;
		msg = msg->next;
		continue;
			}

        msg_nxt = msg->next;
	k++;

	if ((msg->flags & DELETED) ||
		(msg->flags & MOVED) ||
		(msg->flags & NOT_SENT)) {
	  if (abortpressed())
		return;
					  }

	 if (msg->flags & DELETED) 
		{
		if ((k%2) == 0)
		  display_msg(MSG_STAT, NULL, "Deleting %d", k);
		  msg->delete(msg);
		} 
	else 
	if (msg->flags & MOVED)
		{
	  fld = msg->folder;
	  msg->folder = folder;
	  msg->flags &= ~MOVED;
	  if (msg->folder != fld) {
		if ((k%2) == 0)
		  display_msg(MSG_STAT, NULL, "Moving %d", k);
		msg->update(msg);
		fld->move(msg, fld);
				  }
		}
	else
	if (msg->flags & NOT_SENT) 
		{
	  msg->flags &= ~NOT_SENT;
	  display_msg(MSG_STAT, NULL, "Sending %d", k);
	  msg->update(msg);
	  send_message(msg);
		}
	else
	  msg->update(msg);

        msg = msg_nxt;
            } 
 return;

}

char *
get_charset_addr_line(prefix, addr, charset)
char *prefix;
mail_addr *addr;
int *charset;
{
static char addr_line[255];

strcpy(addr_line, "unknown");

if (!addr)
	return addr_line;

if (!addr->addr)
	return addr_line;

if (addr->name && addr->comment)
	sprintf(addr_line, "%s%s <%s> (%s)", prefix ? prefix : "" , rfc1522_decode(addr->name, charset), addr->addr, rfc1522_decode(addr->comment, charset));
else
if (!addr->name && !addr->comment)
	sprintf(addr_line, "%s%s", prefix ? prefix : "" , addr->addr);
else
if (addr->comment)
	sprintf(addr_line, "%s(%s) <%s>", prefix ? prefix : "" , rfc1522_decode(addr->comment, charset), addr->addr);
else
	sprintf(addr_line, "%s%s <%s>", prefix ? prefix : "" , rfc1522_decode(addr->name, charset), addr->addr);

return addr_line;
}

char *
get_full_addr_line(addr)
mail_addr *addr;
{
static char addr_line[255];

strcpy(addr_line, "<UNKNOWN>");

if (!addr)
	return addr_line;

if (!addr->addr)
	return addr_line;

if (!addr->name && !addr->comment)
	strcpy(addr_line, addr->addr);
else
if (addr->name && addr->comment)
	sprintf(addr_line, "%s <%s> (%s)", addr->name, addr->addr, addr->comment);
else
if (addr->name)
	sprintf(addr_line, "%s <%s>", addr->name, addr->addr);
else
	sprintf(addr_line, "(%s) <%s>", addr->comment, addr->addr);
	
return addr_line;
}

char *
get_short_addr_line(addr)
mail_addr *addr;
{
 if (!addr)
	return "unknown";

 if (addr->name)
	return addr->name;

 return addr->addr;
}

char *
get_addr_line(addr)
mail_addr *addr;
{
static char addr_line[255];

strcpy(addr_line, "unknown");

if (!addr)
	return addr_line;

if (addr->name && (strlen(addr->name) > 1))
	sprintf(addr_line, "%s", addr->name);
else
	sprintf(addr_line, "%s", addr->addr);

return addr_line;
}

char *  
get_addr_addr(addr)
mail_addr *addr;
{       
static char addr_line[255];
 
strcpy(addr_line, "unknown");
   
if (!addr)
        return addr_line;

if (strlen(addr->addr) > 1)
        sprintf(addr_line, "%s", addr->addr);
else
        sprintf(addr_line, "%s", addr->name);

return addr_line;
}

void
print_addr(addr, str, file, encode)
mail_addr *addr;
char *str;
FILE* file;
int encode;
{
u_long st, st1;
u_int alen;
char *p;

if (!addr)
        return;

if (file == NULL)
	return;

st = ftell(file);

p = (encode && addr->comment) ? strdup(rfc1522_encode(addr->comment, -1, -1)) : addr->comment;

if (addr->name && addr->comment)
 fprintf(file, "%s: %s <%s> (%s)", str, encode ? rfc1522_encode(addr->name, -1, -1) : addr->name, addr->addr, p);
else
if (addr->name)
 fprintf(file, "%s: %s <%s>", str, encode ? rfc1522_encode(addr->name, -1, -1) : addr->name , addr->addr);
else
if (addr->comment)
 fprintf(file, "%s: (%s) <%s>", str, p , addr->addr);
else
 fprintf(file, "%s: %s", str, addr->addr);

addr =  addr->next_addr; 

if (!addr)
	{
	fputc('\n', file);
	if (encode && p)
		free(p);
	return;
	}

while(addr) {
  p = (encode && addr->comment) ? strdup(rfc1522_encode(addr->comment, -1, -1)) : addr->comment;
  alen = strlen(addr->addr);

  if (addr->name)
	alen += (strlen(addr->name) + 4);

  if (p)
	alen += (strlen(p) + 4);

  st1 = ftell(file);
  if ((st1 - st + alen) > 80) {
	st = st1;
	fputs(",\n ", file);
			}
  else
	fputs(", ", file);

 if (addr->name && addr->comment)
  fprintf(file, "%s <%s> (%s)", encode ? rfc1522_encode(addr->name, -1, -1) : addr->name, addr->addr, p);
 else
 if (addr->name)
  fprintf(file, "%s <%s>", encode ? rfc1522_encode(addr->name, -1, -1) : addr->name , addr->addr);
 else
 if (addr->comment)
  fprintf(file, "(%s) <%s>", p , addr->addr);
 else
  fprintf(file, "%s", addr->addr);

 addr =  addr->next_addr; 
	   }

fputc('\n', file);
if (encode && p)
	free(p);
return;
}

void chld_handler()
{
int i;
wait(&i);
}

int
send_message(msg)
struct _mail_msg *msg;
{
int stat;
char buf[255];
char id[32];
time_t lt = time(NULL);

 if (!msg)
	return -1;

 if (!msg->header)
	return -1;

 if (!msg->header->To) {
	display_msg(MSG_WARN, "send", "Must specify at least one recipient");
	return -1; }

 if (!msg->header->From) {
	display_msg(MSG_WARN, "send", "Can not send mail from Ghost!");
	return -1; }

 if (!msg->header->To->num)
	return -1;

/*
 if (msg->folder != outbox) {
	display_msg(MSG_WARN, "send", "Can send only from %s", OUTBOX);
	return -1; }
*/

 if (offline)
	return 0;

 replace_field(msg, "Sender", sender_name);

 if (msg->flags & C_DELIVERY) {
	replace_field(msg, "Return-Receipt-To", msg->header->From->addr);
	msg->flags &= ~C_DELIVERY; }

 if (msg->flags & C_READ) {
	replace_field(msg, XFRECEIPT, msg->header->From->addr);
	replace_field(msg, "X-Chameleon-Return-To", msg->header->From->addr);
	msg->flags &= ~C_READ; }

 if (msg->flags & URGENT) {
	replace_field(msg, "Priority", "urgent");
	replace_field(msg, "X-PRIORITY", "2 (High)");
				}

 sprintf(buf, "XFMail %s [%s] on %s", VERSION, PATCHLEVEL, XF_OS);
 replace_field(msg, "X-Mailer", buf);

 strftime(id, 31, "%y%m%d%H%M%S", localtime(&lt));
 sprintf(buf, "<XFMail.%s.%s>", id, msg->header->From->addr);
 replace_field(msg, "Message-ID", buf);

 sprintf(buf, "%d", (int)(msg->msg_len - msg->header->header_len));
 replace_field(msg, MIME_C_LENGTH, buf);

 switch (b_getcfg_int(conf_name, "smtpsend", 0)) {
   case 0:
	stat = sendmail_send_message(msg);
   break;

   case 1:
	stat = smtp_send_message(msg);
   break;

   case 2:
	stat = pop_send_message(msg);
   break;

   default:
	stat = sendmail_send_message(msg);
   break;
						}

 if (stat == -1)
	return -1;

 if (b_getcfg_int(conf_name, "smtpsave", 0))
	sentm->move(msg, sentm);
 else
	msg->delete(msg);

 return 0;
}

void
sendmail_exit(pinfo)
struct _proc_info *pinfo;
{
  if (pinfo && pinfo->u_data) 	{
	unlink((char *)pinfo->u_data);
	free(pinfo->u_data);
		     		}
}

int
sendmail_send_message(msg)
struct _mail_msg *msg;
{
struct _proc_info pinfo;
char mailbuf[2048];
FILE *m_tmp_fd;
int m_tmp, res;
char m_tmp_file[255];
struct _mail_addr *addr;

 strcpy(m_tmp_file, get_temp_file("send"));

 if ((m_tmp_fd = fopen(m_tmp_file, "w")) == NULL) {
	display_msg(MSG_WARN, "send", "Can not open file %s", m_tmp_file);
	return -1;
						  }	
 msg->print(msg, m_tmp_fd, 1);
 fclose(m_tmp_fd);

 if ((m_tmp = open(m_tmp_file, O_RDONLY)) < 0)  	     {
	display_msg(MSG_WARN, "send", "Can not open file %s", m_tmp_file);
	return -1;
							      }

 pinfo.wait = b_getcfg_int(conf_name,"smtpbg", 0) ? WAIT_ASYNC : WAIT_BG; 
 pinfo.u_data = strdup(m_tmp_file);
 pinfo.ul_data = 0L;
 pinfo.handle = sendmail_exit;
 pinfo.ifd = m_tmp;

 sprintf(mailbuf, "%s -i", b_getcfg_str(conf_name, "sendmail", _PATH_SENDMAIL));

 addr = msg->header->To;
 while (addr) {
	if ((strlen(addr->addr) + strlen(mailbuf) + 2) >= sizeof(mailbuf)) {
		display_msg(MSG_WARN, "send", "address list too long");
		close(m_tmp);
		return -1;
									   }
	strcat(mailbuf, " ");
	strcat(mailbuf, addr->addr);
	addr = addr->next_addr;
	     }

 addr = msg->header->Cc;
 while (addr) {
	if ((strlen(addr->addr) + strlen(mailbuf) + 2) >= sizeof(mailbuf)) {
		display_msg(MSG_WARN, "send", "address list too long");
		close(m_tmp);
		return -1;
									   }
	strcat(mailbuf, " ");
	strcat(mailbuf, addr->addr);
	addr = addr->next_addr;
	     }

 addr = msg->header->Bcc;
 while (addr) {
	if ((strlen(addr->addr) + strlen(mailbuf) + 2) >= sizeof(mailbuf)) {
		display_msg(MSG_WARN, "send", "address list too long");
		close(m_tmp);
		return -1;
									   }
	strcat(mailbuf, " ");
	strcat(mailbuf, addr->addr);
	addr = addr->next_addr;
	     }

 res = exec_child(mailbuf, &pinfo);
 if (res == -1)
	sendmail_exit(&pinfo);

 return res ? -1 : 0;
}

void
update_message_length(msg)
struct _mail_msg *msg;
{
FILE *mfd;
char buf[255];

 if (!msg)
	return;

 if ((mfd = fopen(msg->get_file(msg), "r")) == NULL)
	return;

 if (msg->header) {
	while(fgets(buf, 255, mfd)) {
	 if (strlen(buf) < 2) {
	  msg->header->header_len = ftell(mfd);
	  break; }
					}
		}

 if (fseek(mfd, 0L, SEEK_END) != -1)
	msg->msg_len = ftell(mfd);

 fclose(mfd);
 return;
}

int
set_message_text(msg, file)
struct _mail_msg *msg;
char *file;
{
FILE *mfd, *fd;
char buf[255];

 if (!msg)
	return -1;

 if (!file)
	return -1;

 if ((fd = fopen(file, "r")) == NULL) {
	display_msg(MSG_WARN, "Can not open file", "%s", file);
	return -1; }

 if (unlink(msg->get_file(msg)) == -1) {
	display_msg(MSG_WARN, "Cannot delete old copy of message", "%s", msg->get_file(msg));
	fclose(fd);
	return -1; }

 if ((mfd = fopen(msg->get_file(msg), "w")) == NULL) {
        display_msg(MSG_WARN, "update", "Can not open file %s", msg->get_file(msg));
	fclose(fd);
        return -1; }

 print_message_header(msg, mfd);

 msg->header->header_len = ftell(mfd);
 while(fgets(buf, 255, fd))
	fputs(buf, mfd);

 msg->msg_len = ftell(mfd);
 fclose(fd);
 fclose(mfd);

 return 0;
}

int
update_message_text(msg, text)
struct _mail_msg *msg;
char *text;
{
FILE *mfd;
int i;

 if (!msg)
	return -1;

 if (!text)
	return -1;

 if (unlink(msg->get_file(msg)) == -1) {
	display_msg(MSG_WARN, "Cannot delete old copy of message", "%s", msg->get_file(msg));
	return -1; }

 if ((mfd = fopen(msg->get_file(msg), "w")) == NULL) {
        display_msg(MSG_WARN, "update", "Can not open file %s", msg->get_file(msg));
        return -1; }

 print_message_header(msg, mfd);
 i = 0;

 fflush(mfd);
 msg->header->header_len = ftell(mfd);

 while(text[i] != '\0') {
	fputc(text[i], mfd);
	i++; }

 fflush(mfd);
 msg->msg_len = ftell(mfd);

 fclose(mfd);
 return 0;
}

void
update_message_status(msg)
struct _mail_msg *msg;
{
char buf[255];
u_long offset;
FILE *mfd;

 if (!msg || !msg->header)
	return;

 if ((msg->flags & 0xFFFF) == (msg->header->flags & 0xFFFF))
	return;

 if ((mfd = fopen(msg->get_file(msg), "r+")) == NULL)
	return;

 offset = 0L;
 while(fgets(buf, 255, mfd)) 	{
    if (strlen(buf) < 2)
        break;
    if (!strncmp(buf, STATUS_FIELD, strlen(STATUS_FIELD)))
	{
	offset += strlen(STATUS_FIELD);
	if (fseek(mfd, offset, SEEK_SET) == -1) {
	  fclose(mfd);
	  return; }
	fprintf(mfd, ": %04X\n", msg->flags & 0xFFFF);
	fclose(mfd);
	msg->header->flags = msg->flags;
	return;
	}
    offset = ftell(mfd);
				}

 fclose(mfd);
 msg->flags |= CHANGED;
 update_message(msg);

 return;
}

void
print_header_field(fld, file, encode)
head_field *fld;
FILE *file;
int encode;
{
char *p, *p1, c;
int nlen;

  fputs(fld->f_name, file);
  fputs(": ", file);
  p = encode ? rfc1522_encode(fld->f_line, -1, -1) : fld->f_line;
  nlen = FLD_BREAK - (strlen(fld->f_name) + 2);

  while(strlen(p) > nlen)  	{
	c = p[nlen];
	p[nlen] = '\0';
	if ( (p1 = strstr(p, "; ")) ||
		(p1 = strstr(p, ", ")) ||
		   (p1 = strrchr(p, ' ')) ) {
	 p[nlen] = c;
	 if (*p1 != ' ')
		p1++;

	 nlen = (p1 - p);
	 if (nlen <= 0)
		nlen = 1;

	 if ((nlen < (FLD_BREAK / 8)) ||
		((strlen(p) - nlen) < (FLD_BREAK / 8))   ) 	{
	 	fwrite(p, nlen, 1, file);
	 	p += nlen;
	 	nlen = FLD_BREAK;
		continue;
								}
	 fwrite(p, nlen, 1, file);
	 p1++;
	 p = p1;
  	 fputc('\n', file);
  	 fputc(' ', file);
	 nlen = FLD_BREAK - 1;
					}
	else 	{
	 p[nlen] = c;
	 fwrite(p, nlen, 1, file);
	 p += nlen;
	 nlen = FLD_BREAK;
		}
				}

  fputs(p, file);
  fputc('\n', file);

  return;
}

void
print_message_header(msg, file)
struct _mail_msg *msg;
FILE *file;
{
head_field *hf;

if (!msg)
	return;

if (msg->header) {
	hf = msg->header->other_fields;
	while(hf) {
	  print_header_field(hf, file, 0);
	  hf = hf->next_head_field;
		}
	fprintf(file, "%s: %04X\n", STATUS_FIELD, msg->flags & 0xFFFF);
	print_addr(msg->header->Sender, "Sender", file, 0);
	print_addr(msg->header->From, "From", file, 0);
	print_addr(msg->header->To, "To", file, 0);
	if (msg->header->Subject)
		fprintf(file, "Subject: %s\n", msg->header->Subject);

	print_addr(msg->header->Cc, "Cc", file, 0);
	print_addr(msg->header->Bcc, "Bcc", file, 0);

		}
	fputc('\n', file);

 return;

}

void
lpr_exit(pinfo)
struct _proc_info *pinfo;
{
  if (pinfo && pinfo->u_data) 	{
	unlink((char *)pinfo->u_data);
	free(pinfo->u_data);
		     		}
}

void
lpr_message(msg)
struct _mail_msg *msg;
{
struct _proc_info pinfo;
char m_tmp_file[255];
char buf[255];
char lp[16];

 if (!msg)
	return;

 strcpy(m_tmp_file, get_temp_file("lpr"));
 if (save_part(msg, get_text_part(msg), m_tmp_file, 1) == -1) {
	display_msg(MSG_WARN, "lpr", "Can not print message!");
	unlink(m_tmp_file);
	return; }

 strcpy(lp, b_getcfg_str(conf_name, "printer", "lp"));
#if defined(HP) || defined(SGI)
 sprintf(buf, "%s -d%s %s", b_getcfg_str(conf_name, "lpr", _PATH_LPR), lp, m_tmp_file);
#else
 sprintf(buf, "%s -P%s %s", b_getcfg_str(conf_name, "lpr", _PATH_LPR), lp, m_tmp_file);
#endif

 pinfo.wait = WAIT_ASYNC; 
 pinfo.u_data = strdup(m_tmp_file);
 pinfo.ul_data = 0L;
 pinfo.handle = lpr_exit;
 pinfo.ifd = -1;

 if (exec_child(buf, &pinfo) == -1)
	lpr_exit(&pinfo);

 return;
}

int
strip_when_send(fld)
head_field *fld;
{
int i = 0;
char *p;

 while ((p = stripheaders[i++]) != NULL)
	if (!strcasecmp(fld->f_name, p))
		return 1;

 return 0;
}

void
print_message(msg, file, send)
struct _mail_msg *msg;
FILE *file;
int send;
{
head_field *hf;

if (!msg)
	return;

if (msg->header) {
	hf = msg->header->other_fields;
	while(hf) {
	  if (!(send && strip_when_send(hf)) )
	  	print_header_field(hf, file, send);
	  hf = hf->next_head_field;
		}
	if (!send)
	fprintf(file, "%s: %04X\n", STATUS_FIELD, msg->flags & 0xFFFF);

	print_addr(msg->header->Sender, "Sender", file, send);
	print_addr(msg->header->From, "From", file, send);
	print_addr(msg->header->To, "To", file, send);
	if (msg->header->Subject)
		fprintf(file, "Subject: %s\n", send ? rfc1522_encode(msg->header->Subject, -1, -1) : msg->header->Subject);

	print_addr(msg->header->Cc, "Cc", file, send);
	print_addr(msg->header->Bcc, "Bcc", file, send);
		}

fprintf(file, "\n");

msg->print_body(msg, file);

return;
}

void
print_message_body(msg, file)
struct _mail_msg *msg;
FILE *file;
{
long len;
#if defined(HP) || defined(SGI)
char *ptr, *p;
#else
void *ptr, *p;
#endif
int body;

 if (!msg)
	return;

 if (file == NULL)
	return;

 body = 0;

 if (msg->msg_body == NULL) {
	if (msg->get_text(msg) == -1) {
		display_msg(MSG_WARN, "print", "Failed to access message");
		return;		      }
	else
		body = 1;
		     	    }

 len = msg->msg_len - msg->header->header_len;
 ptr = msg->msg_body + msg->header->header_len;

 while ((p = memchr(ptr, '\n', len)) != NULL) {
	p++;

	if (!strncmp(ptr, "From ", 5))
		fputc('>', file);

	if (fwrite(ptr, (p - ptr), 1, file) != 1) {
		display_msg(MSG_WARN, "print", "Failed to write message");
		return; }

	len -= (p - ptr);
	ptr = p;		    }

 if (len != 0) 	{
	fwrite(ptr, len, 1, file);
	fputc('\n', file);
		}

if (body)
	msg->free_text(msg);

return;
}

char *
get_temp_file(base)
char *base;
{
static char buf[255];
struct timeval tp;

 gettimeofday(&tp, NULL);
 sprintf(buf, "/tmp/xf%s%04lu%04lu.%d", base ? base : "mail", tp.tv_sec%1000, tp.tv_usec%1000 , (int)getpid());
 return buf;
}

void
update_message(msg)
struct _mail_msg *msg;
{
FILE *mfd;
char buf[255];
long h_len;

 if (!msg)
	return;

 if (!(msg->flags & CHANGED)) {
	update_message_status(msg);
	return; }

 msg->flags &= ~CHANGED;
 msg->header->flags = msg->flags;
 
 if (msg->msg_body)
	msg->free_text(msg);

 strcpy(buf, get_temp_file("upd"));

 if ((mfd = fopen(buf, "w")) == NULL) {
	display_msg(MSG_WARN, "update", "Can not open file %s", buf);
	return; }

 print_message_header(msg, mfd);
 fflush(mfd);
 h_len = ftell(mfd);

 msg->print_body(msg, mfd);

 fflush(mfd);
 msg->msg_len = ftell(mfd);
 msg->header->header_len = h_len;

 fclose(mfd);
 do_move(buf, msg->get_file(msg));
 unlink(buf);

 return;

}

int
delete_message(msg)
struct _mail_msg *msg;
{
 if (!msg)
	return -1;

 msg->flags &= ~DELETED;

 if (msg->flags & LOCKED)
	return -1;

 if (msg->flags & H_ONLY) {
	if (display_msg(MSG_QUEST, NULL, "Delete message from server?"))
		pop_delmsg_by_uidl(msg);
			  }

 if (unlink(msg->get_file(msg)) == -1) {
	display_msg(MSG_WARN, "delete", "Can not remove file %s", msg->get_file(msg));
	return -1; }

 if (msg->folder) {
   if ((msg->flags & UNREAD) && msg->folder->unread_num)
	msg->folder->unread_num--;

   if (msg->folder->num_msg)
	msg->folder->num_msg--;

   if (msg->folder->messages) {
	struct _mail_msg *msg1 = msg->folder->messages;

	if (msg == msg->folder->messages)
	  msg->folder->messages = msg->folder->messages->next;
	else {
	while(msg1) {
	 if (msg1->next == msg) {
	   msg1->next = msg->next;
	   break;		}
	 msg1 = msg1->next;
		    }
	     }
	   
			      }
		}

 if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
	( ( (folder_sort & 0x0f) == BY_UNREAD) &&
		(msg->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

  discard_message(msg);
  return 0;
}

int
move_to_folder(msg, folder)
struct _mail_msg *msg;
struct _mail_folder *folder;
{

int new;
char new_path[255];

 if (!msg || !folder)
	return -1;

 msg->flags &= ~MOVED;

 if (msg->flags & LOCKED)
	return -1;

 if (msg->folder) {
	if (msg->folder == folder)
		return 0;
		}

 if (folder->num_msg > MAX_MSG_IN_FOLDER) {
	display_msg(MSG_WARN, "move", "Folder %s is full", folder->name(folder));
	return -1; }

 if ((new = get_new_name(folder)) == -1) {
	display_msg(MSG_WARN, "move", "Folder %s is full", folder->name(folder));
	return -1; }

 msg->flags &= ~M_TEMP;
 msg->update(msg);

 if (msg->folder) {
   if ((msg->flags & UNREAD) && msg->folder->unread_num)
	msg->folder->unread_num--;

   if (msg->folder->num_msg)
	msg->folder->num_msg--;

   if (msg->folder->messages) {
	struct _mail_msg *msg1 = msg->folder->messages;
	if (msg->folder->messages == msg)
		msg->folder->messages = msg->next;
	else {
	while(msg1) {
	 if (msg1->next == msg) {
	   msg1->next = msg->next;
	   break;               }

	 msg1 = msg1->next;
		    }
	     }
			      }
		}

 sprintf(new_path, "%s/%d", folder->fold_path, new);
 do_move(msg->get_file(msg), new_path);

 msg->folder = folder;
 msg->num = new;

 folder->num_msg++;
 if (msg->flags & UNREAD)
	folder->unread_num++;

 if (folder->flags & OPENED) {
	msg->next = folder->messages;
	folder->messages = msg;
			}
 else
 	discard_message(msg);

 folder->flags &= ~SORTED;

 if ( ((folder_sort & 0x0f) == BY_MSGNUM) ||
	( ( (folder_sort & 0x0f) == BY_UNREAD) &&
		(msg->flags & UNREAD) ) )
   folder_sort &= ~FLD_SORTED;

 return 0;

}

void
convert_fields(msg)
struct _mail_msg *msg;
{
struct _head_field *fld;

 if ((fld = find_field(msg, "Priority")) != NULL) {
	if (!strncasecmp(fld->f_line, "urgent", 6)) {
	   msg->flags |= URGENT;
	   delete_field(msg, fld); }
						}

 if ((fld = find_field(msg, "X-PRIORITY")) != NULL) {
	if (atoi(fld->f_line) == 2) {
	   msg->flags |= URGENT;
	   delete_field(msg, fld); }
						}

 if ((fld = find_field(msg, "Apparently-To")) != NULL) {
	msg->header->To = get_address(fld->f_line);
	delete_field(msg, fld);
						}

 return;
}

int
inc_mail(spool, trunc)
char *spool;
int trunc;
{
char spool_file[255];
int sfd;
FILE *fsfd, *ftmp;
struct stat sb;
char buf[255], tbuf[255];
struct _mail_msg *msg;
int save_addr;
int pbuf, tnum;
int contlen, mbody;
char *p;

int num_msgs = 0;

  save_addr = b_getcfg_int(conf_name, "saveaddr", 0);

  if (strlen(spool) < 1)
	sprintf(spool_file, "%s/%s", _PATH_MAILDIR, getenv("USER"));
  else
	strcpy(spool_file, spool);

  if (lockmbox(spool_file) == -1) {
	display_msg(MSG_WARN, "inc", "Lock exists on\n%s", spool_file);
	return -1;
				  }

#if defined(SOLARIS) || defined(HP)
  if ((sfd = open(spool_file, O_RDWR)) == -1) {
#else
  if ((sfd = open(spool_file, O_RDONLY)) == -1) {
#endif
	unlockmbox();
	return 0; }

#if defined(SOLARIS) || defined(HP) || defined(_AIX)
  if (lockf(sfd, F_TLOCK, 0) == -1) {
#else
  if (flock(sfd, LOCK_EX) == -1) {
#endif
	display_msg(MSG_WARN, "inc", "Can not lock spool\nfile %s", spool_file);
	unlockmbox();
	return -1;	}

  if (fstat(sfd, &sb)) {
        close(sfd);
	unlockmbox();
        return -1;  }

  if (sb.st_size == 0) {
	close(sfd);
	unlockmbox();
	return 0; }

 if (!(fsfd = fdopen(sfd, "r"))) {
	close(sfd);
	unlockmbox();
	return -1; }

 if (!fgets(buf, 255, fsfd)) {
	display_msg(MSG_WARN, "inc", "Error reading file\n%s", spool_file);
	close(sfd);
	fclose(fsfd);
	unlockmbox();
	return -1; }

 if (strncmp(buf, "From ", 5)) {
	display_msg(MSG_WARN, "inc", "Corrupted spool file %s", spool_file);
	close(sfd);
	fclose(fsfd);
	unlockmbox();
	return -1; }

 pbuf = 0;
 contlen = 0;
 mbody = 0;

 if ((tnum = get_new_name(ftemp)) == -1) {
	display_msg(MSG_WARN, "inc", "No space in %s", FTEMP);
        close(sfd);
        fclose(fsfd);
	unlockmbox();
        return -1; }

 sprintf(tbuf, "%s/%d", ftemp->fold_path, tnum);

 if (!(ftmp = fopen(tbuf, "w"))) {
	display_msg(MSG_WARN, "inc", "Can not open file %s", tbuf);
	close(sfd);
	fclose(fsfd);
	unlockmbox();
	return -1; }

 while(fgets(buf, 255, fsfd)) 	{

  if (!contlen && !strncasecmp(buf, MIME_C_LENGTH, strlen(MIME_C_LENGTH)))
   {
    if ((p = strchr(buf, ':')) != NULL) {
	p++;
	while (*p == ' ')
		p++;
	contlen = atoi(p);
					}
   }

  if (!mbody && (buf[0] == '\n'))
	mbody = ftell(fsfd);

  if (!strncmp(buf, "From ", 5) && ((ftell(fsfd) - mbody) >= contlen) )
				{
	fclose(ftmp);
	pbuf = 0;
	contlen = 0;
	mbody = 0;
	if (!(msg = get_message(tnum, ftemp))) {
		display_msg(MSG_WARN, "inc", "Corrupted message in spool file");
		if (save_addr)
			add_msg_addr(NULL);
		close(sfd);
		fclose(fsfd);
		unlink(tbuf);
		unlockmbox();
		return -1; }

	num_msgs++;
	display_msg(MSG_STAT, NULL, "Retrieving %d (%-64s)", num_msgs, msg->header->Subject ? msg->header->Subject : "* No Subject *");

	if (abortpressed())   	{
	  if (save_addr)
		add_msg_addr(NULL);
	  close(sfd);
	  fclose(fsfd);
	  unlink(tbuf);
	  unlockmbox();
	  return -1;
				}

	msg->folder = NULL;
	set_flags_by_status(msg);
	convert_fields(msg);
	msg->flags |= CHANGED;
	add_field(msg, "X-RDate", get_arpa_date(time(NULL)));
	msg->header->rcv_time = time(NULL);

	if (save_addr)
		add_msg_addr(msg);

	update_faces(msg);

	if (apply_rule(msg, 0) == -1) {
		if (save_addr)
			add_msg_addr(NULL);
		close(sfd);
		fclose(fsfd);
		unlink(tbuf);
		unlockmbox();
		return -1;	   }

	unlink(tbuf);
 	if (!(ftmp = fopen(tbuf, "w"))) {
		if (save_addr)
			add_msg_addr(NULL);
		display_msg(MSG_WARN, "inc", "Can not open file %s", tbuf);
		close(sfd);
		fclose(fsfd);
		unlockmbox();
		return -1; }
				}

 if (pbuf) {
	 fputs("\n", ftmp);
	 pbuf = 0;
	   }

 if (buf[0] == '\n')
	pbuf = 1;
 else {
 if ( fputs(buf, ftmp) == EOF) {
	if (save_addr)
		add_msg_addr(NULL);
	close(sfd);
	fclose(fsfd);
	fclose(ftmp);
	unlink(tbuf);
	unlockmbox();
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "inc", "DISK FULL!");
	else
		display_msg(MSG_WARN, "inc", "Can not write to %s", tbuf);
	return -1;	     }
     }
				}

 close(sfd);
 fclose(fsfd);
 if (fclose(ftmp) == EOF) {
	unlink(tbuf);
	unlockmbox();
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "inc", "DISK FULL!");
	else
		display_msg(MSG_WARN, "inc", "Can not write to %s", tbuf);
	return -1;
			  }

 if ((msg = get_message(tnum, ftemp))) {
	msg->folder = NULL;
	set_flags_by_status(msg);
	convert_fields(msg);
	msg->flags |= CHANGED;
	add_field(msg, "X-RDate", get_arpa_date(time(NULL)));
	msg->header->rcv_time = time(NULL);
	num_msgs++;

	if (save_addr) {
		add_msg_addr(msg);
		add_msg_addr(NULL);
			}

	update_faces(msg);

	if (apply_rule(msg, 0) == -1) {
		unlink(tbuf);
		unlockmbox();
		return -1;	   }


				}
 else	{
	if (save_addr)
		add_msg_addr(NULL);
	display_msg(MSG_WARN, "inc", "Corrupted message in spool file");
	unlockmbox();
	return -1; }

 unlink(tbuf);

 if (trunc) 	{
 if (truncate(spool_file, 0L) == -1)
	display_msg(MSG_WARN, "Can not truncate spool file", "%s", spool_file);
		}

 unlockmbox();

 return num_msgs;
}

void
expand_str(msg, str)
struct _mail_msg *msg;
char *str;
{
char buf[255];
char *p, *p1;
struct _head_field *fld;

 if (!msg || !str)
	return;

 if ((strlen(str) < 1) || (strlen(str) > 200))
	return;

 buf[0] = '\0';
 p = str;
 p1 = buf;
 while (*p != '\0') {

  switch (*p) 	{
    case '%':
	p++;
	if ((*p == '\0') || (*p == '%')) {
	 *p1 = *p;
	 p++;
	 p1++;
	 *p1 = '\0';
	 break;
		}

	switch(*p) {
	 case 'f':
	  if (msg->header->From)
		strcpy(p1, get_short_addr_line(msg->header->From));
	  else
	  if (msg->header->Sender)
		strcpy(p1, get_short_addr_line(msg->header->Sender));
	  else
		strcpy(p1, "unknown");
	 break;

	 case 'i':
	  if ((fld = find_field(msg, "Message-ID")))
		strncpy(p1, fld->f_line, 64);
	  else
		strncpy(p1, msg->header->Subject ? msg->header->Subject : "", 64);
		p1[64] = '\0';
	 break;

	 case 's':
		strncpy(p1, msg->header->Subject ? msg->header->Subject : "* No Subject *", 64);
		p1[64] = '\0';
	 break;

	 case 't':
		strftime(p1, 48, "%T", gmtime(&msg->header->snt_time));
	 break;

	 case 'd':
		strftime(p1, 48, "%d-%b-%y", gmtime(&msg->header->snt_time));
	 break;

	 case 'n':
		strcpy(p1, "\n");
	 break;

	 default:
		sprintf(p1, "%%%c", *p);
	 break;
		   }

	p1 += strlen(p1);
	p++;
    break;

    default:
     *p1 = *p;
     p1++;
     *p1 = '\0';

     p++;
    break;
		}
	   }

 strcpy(str, buf);

 return;
}

void
set_flags_by_status(msg)
struct _mail_msg *msg;
{
struct _head_field *fld;
char *p;

 if (!msg)
	return;

 msg->flags |= UNREAD;

 if ((fld = find_field(msg, "Status")) == NULL)
	return;

 p = fld->f_line;
 while (*p != '\0') {
  switch (*p) 	{
   case 'R':
	msg->flags &= ~UNREAD;
   break;

   case 'O':
	if (p == fld->f_line)
		msg->flags |= UNREAD;
   break;

   case 'U':
	msg->flags |= UNREAD;
   break;

   default:
   break;
		}

  p++;
		    }

 delete_field(msg, fld);
 return;
}

void
set_status_by_flags(msg)
struct _mail_msg *msg;
{
struct _head_field *fld;
char nstat[4];
char *p = nstat;

 if (!msg)
	return;

 if ((fld = find_field(msg, "Status")) != NULL)
	delete_field(msg, fld);

 nstat[0] = '\0';

 if (msg->flags & UNREAD) {
	*p = 'O';
	p++;
	*p = '\0'; 	}
 else 	{
	*p = 'R';
	p++;
	*p = 'O';
	p++;
	*p = '\0';
	}
 
 if (nstat[0] != '\0')
	add_field(msg, "Status", nstat);

 return;
}

int
ishebrew(c)
u_int c;
{
  return ((c >= 224) && (c <= 250));
}

char *
hebrew_conv(str)
char *str;
{
char hbuf[255];
u_int direct;
char *p, *s;
u_int c;

 if (!str || (strlen(str) > 250))
	return NULL;

 hbuf[0] = '\0';
 p = hbuf;
 s = str;
 direct = 0;

 while (*s != '\0')   {

  c = *s & 0xff;

  if (ishebrew(c))
	direct = 1;
  else
  if (isalnum(c) ||
	(c < 32)) {
	if (direct)
		p += strlen(p);
	direct = 0;
		}

  if (direct)   {
#ifdef	SunOS
    memcpy((p + 1), p, strlen(p) + 1);
#else
    memmove((p + 1), p, strlen(p) + 1);
#endif
    *p = *s;
		}
  else  {
   *p++ = *s;
   *p = '\0';
	}

  s++;
			}
 strcpy(str, hbuf);

 return str;
}

int
lockmbox(path)
	char *path;
{
        int statfailed = 0;
 
	if (!path)
		return -1;

        if (locked)
                return 0;

        sprintf(lockname, "%s.lock", path);
        for (;; sleep(2)) {
                int fd;
                struct stat st;
                time_t now;
 
                fd = open(lockname, O_WRONLY|O_EXCL|O_CREAT, 0);
                if (fd >= 0) {
                        locked = 1;
                        close(fd);
                        return 0;
                }

		if (errno == EACCES)
			return 0;

                if (stat(lockname, &st) < 0) {
                        if (statfailed++ > 5)
                                return -1;
                        continue;
                }
                statfailed = 0;
                time(&now);
                if (now < st.st_ctime + 300)
                        continue;
                unlink(lockname);
        }
}
 
void
unlockmbox()
{
        if (!locked)
                return;

        unlink(lockname);
        locked = 0;

	return;
}

