/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: fmail.h,v 1.117 1996/06/10 16:04:13 gena Exp $
 */

#ifndef _FMAIL_H
#define _FMAIL_H

#if defined(__FreeBSD__) || defined(__linux__)
#define HS_REGEX
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#ifdef SOLARIS
#include <string.h>
#else
#include <strings.h>
#endif
#include <sys/file.h>
#ifdef _AIX
#include <time.h>
#else
#include <sys/time.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/wait.h>
#include <sys/param.h>
#include <dirent.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#if defined(sun) || defined(HP)
 extern int sys_nerr;
 extern char *sys_errlist[];
 extern int errno;
#endif

#define	MMAP_MSG

#ifdef HS_REGEX
#include <regex.h>
#endif

#if !defined(sun) && !defined(HP)
#include <paths.h>
#endif

struct _mail_msg;
struct _mail_folder;
struct _mime_msg;

extern int     do_move (char *, char *);
extern int     fastcopy (char *, char *, struct stat *);

#define MAX_FIELD_NAME_LEN	32	/* limit of field name length in mail header */
#define MAX_FIELDS_IN_HDR	64	/* maximum fields in msg header */
#define MAX_FIELD_LEN		512	/* limit of length of field */
#define	FLD_BREAK		80	/* line length limit in field */
#define MAX_FOLD_NAME_LEN	16	/* maximal length of folder name */
#define STATUS_FIELD		"XFMstatus"	/* name of status field in mail header */
#define XFRECEIPT		"X-XFmail-Return-To"	/* confirm reading field */
#define	XUIDL			"X-UIDL"	/* uid on POP server */
#define MAX_MSG_IN_FOLDER	2047	/* maximal number of messages in folder */
#define MAX_FOLDERS		127	/* maximal number of folders in mailbox */
#define	MAX_MSG_LEN		3072000	/* maximal message length (3Mb) */
#define	MAX_HEADER_LEN		(MAX_FIELDS_IN_HDR * MAX_FIELD_LEN / 4) /* limit on header length */
#define	MAX_INSFILE_SIZE	65535	/* limit on inserted file size (warning only) */
#define	MAX_INSFILE_LINE	127	/* limit on inserted file line size */

#define TRASH		"trash" 	/* trash folder */
#define INBOX		"inbox" 	/* incoming mailbox folder */
#define OUTBOX		"outbox"	/* outgoing mail folder */
#define SENTMAIL	"sent_mail"	/* mail that was successfully sent */
#define	DRAFT		"draft"		/* draft folder */
#define	TEMPLATE	"template"	/* template folder */
#define	FTEMP		".ftemp"	/* temporary folder */

#ifdef SunOS
#define _PATH_SENDMAIL	"/usr/lib/sendmail"
#define _PATH_MAILDIR	"/usr/spool/mail"
#define	_PATH_LPR	"/usr/ucb/lpr"
#endif

#ifdef SOLARIS
#define _PATH_SENDMAIL	"/usr/lib/sendmail"
#define _PATH_MAILDIR	"/usr/mail"
#define	_PATH_LPR	"/usr/ucb/lpr"
#endif

#ifdef	SGI
#define _PATH_SENDMAIL	"/usr/lib/sendmail"
#define	_PATH_LPR	"/usr/bin/lp"
#endif

#ifndef _PATH_SENDMAIL
#define _PATH_SENDMAIL	"/usr/sbin/sendmail"
#endif

#ifndef _PATH_MAILDIR
#define _PATH_MAILDIR   "/usr/spool/mail"
#endif

#ifdef	HP
#define	_PATH_LPR	"/usr/bin/lp"
#endif

#ifndef	_PATH_LPR
#define	_PATH_LPR	"/usr/bin/lpr"
#endif

#define NO_SORT		0x00

/* message sort types */
#define BY_SNT_TIME	0x01
#define BY_RCV_TIME	0x02
#define BY_SUBJECT	0x03
#define BY_AUTHOR	0x04
#define BY_FLAGS	0x05
#define BY_SIZE		0x06

/* folder sort types  */
#define	BY_NAME		0x01
#define	BY_PRIOR	0x02
#define	BY_MSGNUM	0x03
#define	BY_UNREAD	0x04

/* sort order (modifiers) */
#define MSG_ASCEND	0x10
#define MSG_DESCEND	0x20
#define MSG_THREAD	0x40
#define	FLD_SORTED	0x40

/* search options (flags) */
#define	S_CASE		0x01	/* search is case insensitive */

/* array of all mail folders */
extern struct _mail_folder *mailbox[MAX_FOLDERS];
/* number of folders */
extern int folders_num;
/* pointers to system folders */
extern struct _mail_folder *trash, *inbox, *outbox, *sentm, *draft, *template, *ftemp;
/* path of mail directory ( usually ~/Mail) */
extern char mailbox_path[255];

/* offline work */
extern u_int offline;

#ifndef __FreeBSD__
typedef void (*sig_t) (int);
#endif

/* sort criteria */
extern u_int sort_type;
extern u_int folder_sort;

/* headers to be stripped when sending mail */
extern char *stripheaders[];

/* linked address list */

typedef struct _mail_addr {
	int num;		/* number of addresses in list */
	char *addr;		/* mail address like: user@host */
	char *name;		/* name (if exists) like: John B. Smith */
	char *comment;		/* comment (if exists) like: (The WiZaRd) */
	struct _mail_addr *next_addr;	/* next address in group */
	} mail_addr;

/* structure of field in message header */

typedef struct _head_field {
	int num_fields;		/* number of fields in header */
	char f_name[MAX_FIELD_NAME_LEN];	/* field name */
	char *f_line;		/* field contents */
	struct _head_field *next_head_field;	/* pointer to the next field */
	} head_field;

/* header of mail message */

typedef struct _msg_header {
	u_int header_len;	/* header lenght in bytes */
				/* standard address fields */
	mail_addr *From;	/* From address */
	mail_addr *To;
	mail_addr *Sender;
	mail_addr *Cc;
	mail_addr *Bcc;
	char *Subject;		/* subject */
	time_t snt_time;	/* time when the message was sent */
	time_t rcv_time;	/* time when the message was received */
	int flags;		/* message status (from file) */
	head_field *other_fields;
	} msg_header;

/* mail message */

typedef int (*msg_delete) (struct _mail_msg *);              /* delete message*/
typedef void (*msg_print) (struct _mail_msg *, FILE *, int); /* print message */
typedef void (*msg_print_body) (struct _mail_msg *, FILE *); /* print body    */
typedef int (*msg_get_text) (struct _mail_msg *);   /* get message text  */
typedef void (*msg_free_text) (struct _mail_msg *); /* free message text */
typedef char *(*msg_get_file) (struct _mail_msg *); /* get message file  */
typedef void (*msg_update) (struct _mail_msg *);    /* update message    */

typedef struct _mail_msg {
	u_long	msg_len;	/* length of message (including header) */
	msg_header *header;	/* message header */
	char *msg_body;		/* message text */
	u_int	num;		/* message file number */
	int flags;		/* message status */
	/* two low bytes are permanently stored in message header */
		/* misc flags */
#define	SIGNED		0x00000001	/* message has signature attached */
#define UNREAD		0x00000002	/* unread message */
#define NOT_SENT	0x00000004	/* message waits for sending */
#define MARKED		0x00000008	/* message is marked */
#define	URGENT		0x00000010	/* urgent message	*/
#define C_DELIVERY	0x00000020	/* confirm message delivery */
#define C_READ		0x00000040	/* confirm message read */
#define	M_TEMP		0x00000080	/* temporary message */
#define	H_ONLY		0x00000100	/* header only */

	/* following flags are not stored inside the message header */
		/* message type */
#define	M_MH		0x00010000 /* local MH-style message */
#define	M_IMAP		0x00020000 /* remote (IMAP) message. Not supported yet*/
#define	M_NEWS		0x00040000 /* News article. Not supported yet*/
		/* message status */
#define LOCKED		0x00100000 /* message is locked by viewer/editor */
#define	DELETED		0x00200000 /* message marked to be removed */
#define	MOVED		0x00400000 /* message moved to another folder */
#define CHANGED		0x00800000 /* message was changed */
		/* misc flags */
#define	S_THREAD	0x01000000 /* thread start */
	struct _mail_folder *folder; /* pointer to folder of message */
	struct _mail_msg *next; /* pointer to next message in folder */
	struct _mail_msg *ref;  /* pointer to referenced message in folder */
	struct _mime_msg *mime; /* pointer to mime structure */
	int		refs;	/* depth of refererences */
	int		data;	/* type of data in pdata */
#define	MSG_DAT_EMPTY	0x00	/* pdata is empty */
#define	MSG_DAT_ORIG	0x01	/* pointer to original message */
	void		*pdata;	/* pointer to misc data */

	msg_delete	delete;
	msg_print	print;
	msg_print_body	print_body;
	msg_get_text	get_text;
	msg_free_text	free_text;
	msg_get_file	get_file;
	msg_update	update;
	} mail_msg;

/* mail folder */

typedef char * (*fld_name) (struct _mail_folder *); /* get folder name */
typedef struct _mail_folder * (*fld_open) (char *); /* open folder     */
typedef struct _mail_folder * (*fld_rescan) (char *); /* rescan folder */
typedef void (*fld_close) (struct _mail_folder *);  /* close folder    */
typedef void (*fld_empty) (struct _mail_folder *);  /* empty folder    */
typedef int (*fld_delete) (struct _mail_folder *);  /* delete folder   */
typedef void (*fld_update) (struct _mail_folder *); /* update folder   */
typedef int (*fld_rename) (struct _mail_folder *, char *); /* rename folder */
typedef int (*fld_move) (struct _mail_msg *, struct _mail_folder *); /* move message to folder */
typedef int (*fld_search) (struct _mail_folder *, char *, char *,int); /* search */

typedef struct _mail_folder {
	char fold_path[255];	/* path to folder directory */
	u_int num_msg;		/* number of messages in the folder */
	u_int unread_num;	/* number of unread messages in the folder */
	struct _mail_msg *messages;	/* pointer to messages */
	int flags;		/* folder status */
#define SYSTEM	0x01    /* system folder , can not be deleted or renamed */
#define SORTED	0x02	/* messages are sorted */
#define OPENED	0x04	/* folder is opened */
#define SEARCH	0x08	/* folder contains messages marked during search */
#define	CACHED	0x10	/* cached folder */
	/* folder location */
#define	F_MH	0x20	/* local MH-style folder */
#define	F_IMAP	0x40	/* remote (IMAP) folder. Not supported yet */
	/* folder manipulation functions */
	fld_name	name;
	fld_open	open;
	fld_rescan	rescan;
	fld_close	close;
	fld_empty	empty;
	fld_delete	delete;
	fld_update	update;
	fld_rename	rename;
	fld_move	move;
	fld_search	search;
	} mail_folder;

/* hash table entry */
typedef struct _ht {
 struct _mail_msg *msg;
 char *id;
 int next;
} ht;

/* incorporate mail rule data structure */
typedef struct _xf_rule {
	char name[16];		/* rule name */
	char fmatch[MAX_FIELD_NAME_LEN]; /* field to look in the mail header */
	char tmatch[64];	/* text to match in the field (regexp) */
	char data[64];		/* folder to move to or address to forward */
	u_int action;		/* what to do with the message */
#define R_DISCARD	0x01	/* delete it */
#define R_MOVE		0x02	/* move to folder specified in data */
#define R_FORWARD	0x03	/* forward to address specified in data */
#define R_VACATION	0x04	/* reply with vacation messages */

#define	RULE_MAXACTION	0x04
	u_int flags;		/* special flags */
#define CASE_I		0x01	/* case insensitive ( only on regexp ) */

#ifdef HS_REGEX
	regex_t rx;		/* compiled regular expression */
#endif
	} xf_rule;

/* rules functions */

/* load rules database from file */
/* returns 0 on success and -1 on error */
int load_rules( );

/* saves rules database to file */
/* returns 0 on success and -1 on error */
int save_rules( );

/* check if msg matches any rule. If yes return action type */
/* and data needed to perform it */
/* default action: move to INBOX */
u_int match_msg(struct _mail_msg *msg, char *data);

/* check if addr matches given rule */
/* returns 1 if matches and 0 if not */
u_int match_addr(mail_addr *addr, struct _xf_rule *rule);

/* check msg against all rules and act accordingly */
/* if delay > 0 all operations are delayed */
/* returns -1 on exit                              */
int apply_rule(struct _mail_msg *msg, int delay);

/* --------------- */

/* mmap stuff */
#ifdef MMAP_MSG

/* ftell analog for mmapped message */
long mmtell(FILE *stream);

/* fseek analog for mmapped message */
int mmseek(FILE *stream ,long offset, int whence);

/* fgets analog for mmapped message */
char *mmgets(char *str, size_t size, FILE *stream);

#define	msg_tell	mmtell
#define	msg_seek	mmseek
#define	msg_gets	mmgets

extern unsigned char *mmsg;     /* mmapped message */
extern long mmpos;              /* position in msg */
extern long mmlen;              /* length of mmapped block */

#else

#define	msg_tell	ftell
#define	msg_seek	fseek
#define	msg_gets	fgets

#endif

/* cache stuff */

#define	CACHE_DIR	".cache" /* directory with cached messages */

/* write message or part of it to cache file in binary form */
/* returns -1 on error and 0 on success */
int dump_field(struct _head_field *hf, FILE *dfd);
int dump_addr(struct _mail_addr *addr, FILE *dfd);
int dump_msg(struct _mail_msg *msg);

/* restore message or part of it from binary cache file */
/* returns NULL on error */
struct _head_field *restore_field(FILE *dfd);
struct _mail_addr *restore_addr(FILE *dfd);
struct _mail_msg *restore_msg(int num, struct _mail_folder *folder);

/* ----------- */

/* MIME stuff  */

/* field names */

#define	MIME_C_TYPE	"Content-Type"
#define	MIME_C_ENCR	"Content-Transfer-Encoding"
#define	MIME_C_DESCR	"Content-Description"
#define	MIME_C_ID	"Content-ID"
#define	MIME_C_LENGTH	"Content-Length"
#define	MIME_C_DISP	"Content-Disposition"
#define	MIME_VERSION	"MIME-Version"

#define	MIME_VERS_SUPP	10	/* supported MIME version */

/* encodings */
#define	CE_NONE		0x00	/* No encoding */
#define	CE_7BIT		0x01	/* 7 bit ASCII text */
#define	CE_QPRT		0x02	/* quoted-printable encoding */
#define	CE_BASE64	0x03	/* base64 encoding */
#define	CE_8BIT		0x04	/* 8 bit ASCII text */
#define	CE_BINARY	0x05	/* binary information */
#define	CE_UNSUPPORTED	0xff	/* unsupported encoding */

/* encoding function. Should be called with NULL as a string to initalize */
/* and for finish encoding , int parameter represents length of encoding */
/* buffer */
/* returns encoded ASCII string */
typedef char * (*fn_encode) (char *, int);

/* Decoding function. Should be called with NULL as a string to initialize */
/* and for finish decoding , int * parameter returns length of the decoding */
/* buffer */
/* returns decoded binary buffer */
typedef char * (*fn_decode) (char *, int *);

/* Base64 encoding/decoding */
char *base64_encode(char *str, int len);
char *base64_decode(char *str, int *len);

/* Quoted-printable encoding/decoding */
char *qprt_encode(char *str, int len);
char *qprt_decode(char *str, int *len);

/* Dumb encoding/decoding. Used for binary , unsupported and when there is */
/* no encoding at all */
char *dumb_encode(char *str, int len);
char *dumb_decode(char *str, int *len);

/* 7bit encoding/decoding */
/* Encoding strips 8's bit and breaks long lines */
/* Decoding does nothing */
char *sevenbit_encode(char *str, int len);
char *sevenbit_decode(char *str, int *len);

/* 8bit encoding/decoding */
/* Encoding breaks long lines , decoding does nothing */
char *eightbit_encode(char *str, int len);
char *eightbit_decode(char *str, int *len);

typedef struct _mime_encoding {
	u_int c_trans_enc;		/* Content-transfer-encoding */
	char *encoding_name;		/* name of the method in the ce field */
	fn_encode ce_enc;		/* encoding function */
	fn_decode ce_dec;		/* decoding function */
	u_int str_len;			/* max string length */
} mime_encoding;

/* supported character sets */
/* need to be defined in mime.c charset table */

#define	US_ASCII	0x00
#define	ISO_8859_1	0x01
#define	ISO_8859_2	0x02
#define	ISO_8859_3	0x03
#define	ISO_8859_4	0x04
#define	ISO_8859_5	0x05
#define	ISO_8859_6	0x06
#define	ISO_8859_7	0x07
#define	ISO_8859_8	0x08
#define	ISO_8859_9	0x09
#define	ISO_8859_10	0x0a
#define	KOI8R		0x0b
#define	CHAR_UNKNOWN	0xff

typedef char * (*charset_encode) (char *);
typedef char * (*charset_decode) (char *);

typedef struct _mime_charset {
	u_int charset_code;		/* Code of the charset */
	char *charset_name;		/* Name of the charset */
	char *charset_descr;		/* Description of the charset */
	charset_decode decode;		/* Decode function     */
	charset_encode encode;		/* Encode function     */
	u_int flags;			/* special flags       */
	u_int font_style;		/* font to be used with charset */
	u_int font_size;		/* font size of charset */
} mime_charset;

struct _mime_mailcap;

typedef struct _mime_msg {
	long m_start;			/* start offset */
	long m_end;			/* end offset */
	char *src_info;			/* source info (file name) */

	u_int mime_vers;		/* MIME version */

	struct _mime_mailcap  *mailcap;	/* pointer to mailcap */
	struct _mime_encoding *encoding;/* pointer to encoding structure */
	struct _mime_charset  *charset;	/* pointer to charset structure */

	char *c_id;			/* Content-ID */
	char *c_descr;			/* Contet-Description */
	long c_len;			/* Content-Length */

	struct _head_field *m_fields;	/* MIME fields */
	struct _mime_msg *mime_next;	/* next mime header (if multipart) */
	char *boundary;			/* Boundary (if multipart) */

	u_int flags;			/* MIME flags */
#define	PREAMBLE	0x01		/* multipart preamble */
#define	EPILOG		0x02		/* multipart epilog */
#define	ATTACHMENT	0x04		/* multipart attachment */
	/* modifiers */
#define	TEXT_PART	0x08		/* visible part in multipart */
#define	MSG_BODY	0x10		/* takes all message body */
} mime_msg;

typedef int (*mcap_view) (struct _mail_msg *, struct _mime_msg *);
typedef int (*mcap_print) (struct _mail_msg *, struct _mime_msg *);
typedef int (*mcap_process) (struct _mail_msg *, struct _mime_msg *);

/* Supported Content-types */

#define	CTYPE_DEFAULT		0x00	/* Always matches */
#define	CTYPE_TEXT		0x01	/* text */
#define	CTYPE_MULTIPART		0x02	/* multipart */
#define	CTYPE_MESSAGE		0x03	/* message */
#define	CTYPE_APPLICATION	0x04	/* application */
#define	CTYPE_IMAGE		0x05	/* image */
#define	CTYPE_VIDEO		0x06	/* video */
#define	CTYPE_AUDIO		0x07	/* audio */
#define	CTYPE_EXTERNAL		0xfe	/* external type */
#define	CTYPE_UNSUPPORTED	0xff	/* unsupported */

/* Supported Content-subtypes */
#define	CSUBTYPE_DEFAULT	0x00	/* Always matches */
/* text */
#define	CSUBTYPE_PLAIN		0x01	/* plain */
#define	CSUBTYPE_RTF		0x02	/* richtext */
#define	CSUBTYPE_HTML		0x03	/* richtext */
/* multiparts */
#define	CSUBTYPE_MIXED		0x04	/* mixed */
#define	CSUBTYPE_ALTERNATIVE	0x05	/* alternative */
#define	CSUBTYPE_DIGEST		0x06	/* digest */
#define	CSUBTYPE_PARALLEL	0x07	/* parallel */
#define	CSUBTYPE_REPORT		0x08	/* report */
/* message */
#define	CSUBTYPE_RFC822		0x09	/* rfc822 */
#define	CSUBTYPE_PARTIAL	0x0a	/* partial */
#define	CSUBTYPE_EXTBODY	0x0b	/* external-body */
/* application */
#define	CSUBTYPE_OSTREAM	0x0c	/* octet-stream */
#define	CSUBTYPE_PSCRIPT	0x0d	/* PostScript */
#define	CSUBTYPE_PGP		0x0e	/* PGP */
/* image */
#define	CSUBTYPE_GIF		0x0f	/* gif */
#define	CSUBTYPE_JPEG		0x10	/* jpeg */
/* audio */
#define	CSUBTYPE_BASIC		0x11	/* basic */
/* video */
#define	CSUBTYPE_MPEG		0x12	/* mpeg */

#define	CSUBTYPE_EXTERNAL	0xfe	/* external subtype */
#define	CSUBTYPE_UNSUPPORTED	0xff	/* unsupported */

typedef struct _mime_mailcap {
	u_int	type_code;		/* type code */
	char	type_text[16];		/* type name */
	u_int	subtype_code;		/* subtype code */
	char	subtype_text[16];	/* subtype name */

	mcap_view	view;		/* view function */
	mcap_print	print;		/* print function */
	mcap_process	process;	/* process function */

	char	*ext_mcap;		/* any external info */
	char	ext[5];			/* default file extension */
	u_int	encode;			/* default encoding type */
} mime_mailcap;

/* MIME variables */

extern struct _mime_mailcap  mailcap[];	      /* system mailcap */
extern struct _mime_encoding supp_encodings[]; /* supported encodings */
extern struct _mime_charset  supp_charsets[];  /* supported charsets */
extern const char sevenbit_chars[]; /* characters<32 that are allowed in 7bit */

extern u_int	def_charset;			/* default character set */
extern u_int	def_encoding;			/* default MIME encoding */

#define DEFAULT_MAILCAP         0x00		/* pointer to text/plain      */
#define MULTIPART_MAILCAP       0x04		/* pointer to multipart/mixed */
#define	MESSAGE_MAILCAP		0x0a		/* pointer to message/rfc822  */
#define	BINARY_MAILCAP		0x0e		/* pointer to application/octet-stream */

#define QPRT_ENCODING        	0x02		/* pointer to quoted-printable*/
#define	BINARY_ENCODING		0x03		/* pointer to BASE64 */
#define DEFAULT_ENCODING        0x04		/* pointer to 8bit            */
#define UNSUPPORTED_ENCODING    0x06 

#define DEFAULT_CHARSET         0x00		/* pointer to us-ascii        */
#define	BINARY_CHARSET		0x01		/* pointer to ISO8859-1 */
#define UNSUPPORTED_CHARSET     0x15

/* MIME functions */

/* extract parameter pname from specific field */
/* returns NULL if such parameter does not exists */
char *get_fld_param(struct _head_field *fld, char *pname);

/* returns version number of MIME protocol used */
u_int get_mime_version(struct _mail_msg *msg);

/* returns 1 if msg is a MIME message and 0 otherwise */
int is_mime(struct _mail_msg *msg);

/* returns 1 is msg has Content-Type field other then text/plain */
int is_mime_msg(struct _mail_msg *msg);

/* returns pointer to mailcap entry with specified type and subtype */
/* returns NULL if such entry does not exists */
struct _mime_mailcap *find_mailcap(char *type, char *subtype);

/* creates new mailcap structure and copies contents of mcap into it */
/* returns NULL is there was not enough memory */
struct _mime_mailcap *copy_mailcap(struct _mime_mailcap *mcap);

/* add mcap structure to table of supported MIME types */
void add_mailcap(struct _mime_mailcap *mcap);

/* load mime types from file */
void load_mailcap();

/* save mime types to file */
void save_mailcap();

/* returns mailcap entry for msg or for mime structure */
/* NULL if it does not exists */
struct _mime_mailcap *get_mailcap_entry(struct _mail_msg *msg, struct _mime_msg *mime);

/* free memory taken by mime structure */
void discard_mime (struct _mime_msg *mime);

/* searches for field with name str */
/* returns NULL if search failed */
struct _head_field *find_mime_field(struct _mime_msg *mime, char *str);

/* Add field to mime structure */
void add_mime_field(struct _mime_msg *mime, char *name, char *str);

/* return encoding type either from msg or from mime structure */
/* returns NULL on error or if encoding method is unknown */
struct _mime_encoding *get_mime_encoding(struct _mail_msg *msg, struct _mime_msg *mime);

/* return charset type either from msg or from mime structure */
/* returns NULL on error or if charset is unknown */
struct _mime_charset *get_mime_charset(struct _mail_msg *msg, struct _mime_msg *mime);

/* return code name of specific charset */
/* returns -1 is charset does not exists */
int get_charset_code(char *ch_name);

/* returns font style of specific charset */
/* returns -1 is charset does not exists */
int get_charset_style(char *ch_name);

/* returns font size of specific charset */
/* returns -1 is charset does not exists */
int get_charset_size(char *ch_name);

/* sets font style of specific charset */
/* returns -1 is charset does not exists */
int set_charset_style(char *ch_name, int style);

/* sets font size of specific charset */
/* returns -1 is charset does not exists */
int set_charset_size(char *ch_name, int size);

/* returns 1 if ch_name is an alias to already existing charset */
/* and 0 if not. Returns -1 if specified charset does not exists */
int is_charset_alias(char *ch_name);

/* scan message for MIME attachments */
void mime_scan(struct _mail_msg *msg);

/* debug function */
void print_mime(struct _mime_msg *mime);

/* check if str is a boundary for multipart message */
/* returns 0 if not 1 if it's usuall boundary 2 if it's final boundary */
int is_boundary(char *boundary, char *str);

/* scan one attachment in multipart */
struct _mime_msg *scan_part(char *boundary, FILE *mfd);

/* parse multipart MIME attachment */
int process_multipart(struct _mail_msg *msg, struct _mime_msg *mime);

/* return 1 if mime is a conventional unencoded text message */
int is_mime_text(struct _mime_msg *mime);

/* encode 3 or less characters using base64 encoding */
/* returns NULL on error */
/* len specified how many characters should be encoded */
char *base64_encode_3(char *triplet, int len);

/* decode 4 characters encoded using base64 encryption */
/* return NULL on error */
/* len contains how many characters was decoded */
char *base64_decode_4(char *four, int *len);

/* Save mime part of msg into file. If file is NULL function will */
/* prompt for file name. Returns -1 on error and 0 on success */
/* with_header indicates if header should be save too */
int save_part(struct _mail_msg *msg, struct _mime_msg *mime,char *file, u_int with_header);

/* view mime part using either internal or external viewer */
/* returns -1 on error and 0 on success */
int view_part(struct _mail_msg *msg, struct _mime_msg *mime);

/* get textual part of the message */
/* returns NULL on error */
struct _mime_msg *get_text_part(struct _mail_msg *msg);

/* print header of mime part */
void print_mime_header(struct _mime_msg *mime, FILE *mfd);

/* print/view text parts of mime message */
int text_view(struct _mail_msg *msg, struct _mime_msg *mime);
int text_print(struct _mail_msg *msg, struct _mime_msg *mime);

/* view rfc822 */
int msg_view(struct _mail_msg *msg, struct _mime_msg *mime);

/* view digest (and any multipart actually) */
int digest_view(struct _mail_msg *msg, struct _mime_msg *mime);

/* play audio/basic */
int mime_play(struct _mail_msg *msg, struct _mime_msg *mime);

/* Create empty MIME structure */
/* returns NULL if malloc failed */
struct _mime_msg *create_mime();

/* write mime attachment into file */
int write_part(struct _mime_msg *mime, struct _mail_msg *msg, FILE *mfd);

/* write new MIME message instead of msg */
/* returns -1 on error and 0 on success */
int update_mime(struct _mail_msg *msg);

/* assemble and view message/partial */
int assemble_partial(struct _mail_msg *msg, struct _mime_msg *mime);

/* displays and/or if possible retrieves body of message/external-body */
int view_external(struct _mail_msg *msg, struct _mime_msg *mime);

/* decode string according to rfc1522. Returns pointer to decoded */
/* string if it was successfully decoded or NULL if error  */
/* occured or no decoding was required, charset will point to character set */
/* of the message */
char *rfc1522_decode(char *string, int *charset);

/* encodes non-ASCII string according to rfc1522. If charset and/or encoding */
/* are -1 defaults are used. If error occured or string is pure ASCII        */
/* then string itself is returned.                                           */
char *rfc1522_encode(char *string, int charset, int encoding);

/* set descritpion for attachment */
void set_mime_descr(struct _mime_msg *msg, char *descr);

/* attach file to msg. mcap is attachment type, if it's NULL then   */
/* application/octet-stream is choosed. encoding is encoding type.  */
/* if -1 then base64 is used. Returns either pointer to new mime    */
/* structure in case of ATT_NOUPDATE or pointer to message mime     */
/* structure otherwise. NULL is returned if error occured           */
struct _mime_msg *attach_file(struct _mail_msg *msg, char *file, struct _mime_mailcap *mcap, int encoding, int flags);

/* attachment flags */
#define	ATT_NOFLAGS	0x00
#define	ATT_NOUPDATE	0x01	/* don't update message */
#define	ATT_NODISPINFO	0x02	/* don't include Contents-Disposition */
#define	ATT_NOFILEINFO	0x04	/* don't include information about the attached file Content-Type */

/* ----------- END of MIME stuff ------------------------------------------- */

/* removes beginning/ending spaces from str */
/* warning: str is modified */
char *rem_tr_space(char *str);

/* return name of temporary file */
char *get_temp_file(char *base);

/* replaces all non-printable characters in str to _ */
/* returns 1 if str was modified and 0 otherwise */
int graph_str(char *str);

/* get difference in seconds between GMT and local TZ */
int get_date_offt();

/* return date in ARPA format */
char *get_arpa_date(time_t t);

/* parses string with date ( mail format ) and returns time in time_t val */
time_t get_date(char *str);

/* parse mail address. returns NULL if error or linked list of addresses */
/* if ok */
struct _mail_addr *get_address(char *str);

/* prints signature file into stream file if signature attachment */
/* is enabled */
void add_signature(struct _mail_msg *msg, FILE *file, int forceattach);

/* adds signature to tfile */
void add_signature_to_msg(struct _mail_msg *msg, char *tfile);

/* adds contents of file to the end of tfile */
void add_file_to_msg(char *tfile, char *file);

/* include text of original message in reply */
/* if type == 0 each line is prefixed */
void insert_orig(char *tfile, struct _mail_msg *msg, int type);

/* uuencodes file and appends it to the end of tfile */
/* returns -1 on error and 0 on success                      */
int uuencode(char *tfile, char *file);

/* uudecodes file(s) in message msg                          */
/* returns -1 on error and 0 on success                      */
int uudecode(struct _mail_msg *msg);

/* send read confirmation message */
void read_confirm(struct _mail_msg *msg);

/* generate reply to msg , if to_all == 0 reply to sender only */
/* returns NULL on error */
struct _mail_msg *get_reply_msg(struct _mail_msg *msg , int to_all);

/* generate forward message */
/* returns NULL on error */
struct _mail_msg *get_fwd_msg(struct _mail_msg *msg);

/* generate "vacation" message from msg and vacation file */
/* returns NULL on error */
struct _mail_msg *get_vac_msg(struct _mail_msg *msg, char *vfile);

/* returns character representation of person specified in address */
char *get_addr_line(struct _mail_addr *addr);

/* returns character representation of mail address addr */
char *get_addr_addr(struct _mail_addr *addr);

/* returns full character representation of mail address addr */
char *get_full_addr_line(struct _mail_addr *addr);

/* returns short representation of addr */
char *get_short_addr_line(struct _mail_addr *addr);

/* returns full character representation of mail address addr */
/* takes care of rfc1522 encoding. Charset will return pointer */
/* to appropriate charset structure if exists */
char *get_charset_addr_line(char *prefix,struct _mail_addr *addr, int *charset);

/* parse line in mail header . Always return allocated field structure */
struct _head_field *get_field(char *str);

/* finds field with name str in header of message msg */
/* returns pointer to found field or NULL if not found */
struct _head_field *find_field(struct _mail_msg *msg, char *str);

/* returns text contents of field str in message msg */
/* returns NULL if field str does not exists */
/* returned value of fr indicates how deallocate memory */
char *get_field_content(struct _mail_msg *msg, char *str , int *fr);

/* deallocate memory allocated by get_field_content */
void free_field_content(struct _mail_msg *msg, char *p, int fr);

/* delete field in msg */
void delete_field(struct _mail_msg *msg, struct _head_field *field);

/* add field name and contents str to header of message msg */
void add_field(struct _mail_msg *msg, char *name, char *str);

/* make a copy of field */
/* return NULL on error */
struct _head_field *copy_field(struct _head_field *field);

/* replace contents of field name with str */
/* if field does not exists the function acts as add_field */
void replace_field(struct _mail_msg *msg, char *name, char *str);

/* convert special fields in msg header to flags */
void convert_fields(struct _mail_msg *msg);

/* parses whole message header. Always returns allocated mail header */ 
/* structure. mfd is open file descriptor of mail file  */
struct _msg_header *get_msg_header(FILE *mfd);

/* opens mail message num and parses it */
/* return NULL if error occured */
struct _mail_msg *get_message(int num, struct _mail_folder *folder);

/* copy msg to another folder and return pointer to it */
/* returns NULL on error */
struct _mail_msg *copy_message(struct _mail_msg *msg, struct _mail_folder *fld);

/* compare two messages. */
int compare_msgs(struct _mail_msg **msg1,struct _mail_msg **msg2);

/* compare two folders. */
int compare_folders(struct _mail_folder **folder1, struct _mail_folder **folder2);

/* sort folder by specific criteria */
void sort_folder(struct _mail_folder *folder);

/* sort mailbox's folders */
void sort_folders();

/* checks if message msg contains text/regexp */
/* returns 1 if yes and 0 if not or error occured */
/* where is comma-separated list of fields to look into */
int match_msg_text(struct _mail_msg *msg, void *exp, char *where);

/* search for text in all messages located in folder       */
/* all found messages become marked                        */
/* returns number of messages that contains specified text */
/* or -1 if error ocurred                                  */
int find_text(struct _mail_folder *folder, char *text, char *where, int flags);

/* create new message in folder. return NULL on error */
struct _mail_msg *create_message(struct _mail_folder *folder);

/* get new message number in folder */
int get_new_name(struct _mail_folder *folder);

/* returns message file name */
char *get_msg_file(struct _mail_msg *msg);

/* make folder local (MH-style) */
void local_folder(struct _mail_folder *folder);

/* make message local (MH-style) */
void local_message(struct _mail_msg *msg);

/* open mail folder name. Returns NULL if error occured */
/* reopens(rescans) folder if it's already in the folder list */
/* creates folder if it does not exists */
struct _mail_folder *open_folder(char *name);

/* rescans mail folder name. Returns NULL if error occured */
/* creates folder if it does not exists */
struct _mail_folder *rescan_folder(char *name);

/* Close folder. All memory allocated for messages is freed */
void close_folder (struct _mail_folder *folder);

/* deallocated memory taken by message msg */
void discard_message(struct _mail_msg *msg);

/* deallocated memory taken by addr */
void discard_address(struct _mail_addr *addr);

/* creates new addr structure and copies contents of addr to it */
struct _mail_addr *copy_address(struct _mail_addr *addr);

/* duplicates contents of whole address chain addr and returns pointer */
/* to it */
struct _mail_addr *copy_address_chain(struct _mail_addr *addr);

/* get text of message from disk */
/* returns 0 on success */
int get_message_text(struct _mail_msg *msg);

/* frees (unmaps) memory allocated for message text */
void free_message_text(struct _mail_msg *msg);

/* check if field fld should be stripped when sending message */
int strip_when_send(struct _head_field *fld);

/* print message with header to stream file */
/* if send - do not print STATUS and Sender fields and encode header*/
void print_message(struct _mail_msg *msg,  FILE *file, int send);

/* send message msg to printer */
void lpr_message(struct _mail_msg *msg);

/* print mail address to stream file */
void print_addr(struct _mail_addr *addr, char *str, FILE *file, int encode);

/* print message header to stream file */
void print_message_header(struct _mail_msg *msg,  FILE *file);

/* print field to stream file */
void print_header_field(struct _head_field *fld,  FILE *file, int encode);

/* print message text to stream file */
void print_message_body(struct _mail_msg *msg,  FILE *file);

/* send message msg, return 0 on success */
int send_message(struct _mail_msg *msg);

/* send message msg using sendmail, return 0 on success */
int sendmail_send_message(struct _mail_msg *msg);

/* send message msg using POP, return 0 on success */
int pop_send_message(struct _mail_msg *msg);

/* send message msg using smtp, return 0 on success */
/* and -1 on error */
int smtp_send_message(struct _mail_msg *msg);

/* send msg to the sock m_out */
/* returns -1 on error */
int smtp_message(struct _mail_msg *msg, FILE *m_out);

/* evaluates length of message and it's header and stores */
/* this values in the mail structure */
void update_message_length(struct _mail_msg *msg);

/* Update STATUS field of mail message on disk */
void update_message_status(struct _mail_msg *msg);

/* updates text of the message */
/* returns -1 on error and 0 on success */
int update_message_text(struct _mail_msg *msg , char *text);

/* places contents of file into message body */
/* returns -1 on error and 0 on success */
int set_message_text(struct _mail_msg *msg , char *file);

/* Updates and saves message on disk */
void update_message(struct _mail_msg *msg);

/* Update folder messages */
void update_folder(struct _mail_folder *folder);

/* scans all folders in a mailbox, uses default mailbox if it's not specified */
/* creates mailbox if it does not exists. initializes mailbox */
/* and folder_num variables. returns -1 on errror 0 on success  */
/* creates inbox/outbox/trash folders if does not exists */
int open_all_folders(char *mailbox);

/* closes all folders and frees all memory allocated for them */
void close_all_folders();

/* incorporate messages from spool file according to rules */
/* returns number of messages actually incorporated or -1 on error */
/* if trunc > 0 then the spool file is truncated to 0 */
int inc_mail(char *spool, int trunc);

/* incorporate messages from POP server according to rules */
/* returns number of messages actually incorporated or -1 on error */
int pop_inc();

/* retrieve message body from POP server if previosly only header */
/* was retrieved. Returns -1 on error and 0 on success */
int pop_getfull_msg(struct _mail_msg *msg);

/* delete message on the server if previously only header was retrieved */
/* Returns -1 on error and 0 on success */
int pop_delmsg_by_uidl(struct _mail_msg *msg);

/* initialize POP server connection (connect and log in) */
/* Returns -1 on error and 0 on success */
int pop_init();

/* close connection to POP server */
void pop_end();

/* connects to remote host and returns socket file descriptor       */
/* hostname - remote host name (IP or DNS name) - default localhost */
/* service - service name or port number - default pop3             */
/* protocol - protocol name - default tcp(others are not supported) */
int host_connect(char *hostname, char *service, char *protocol);

/* disconnect from host. hostname parameter is used only for log purposes */
void host_disconnect(char *hostname);

/* returns appropriate hostent for given hostname */
/* could be DNS name or IP address                */
struct hostent *gethostbystring(char *host);

/* get string from file descriptor iop */
/* result is stored in str             */
/* returns -1 on error                 */
int getline(char *str, int len, FILE *iop);

/* send string str to stream associated with file descriptor */
/* iop. Returns -1 on error.                                 */
int putline(char *str, FILE *iop);

/* returns name of given mail folder */
char *get_folder_name(struct _mail_folder *folder);

/* returns pointer to a folder with given name , NULL if such */
/* folder does not exists */
struct _mail_folder *get_folder_by_name(char *name);

/* move message to specified folder */
int move_to_folder(struct _mail_msg *msg, struct _mail_folder *folder);

/* delete message msg from memory and disk */
/* return -1 on error and 0 on success */
int delete_message(struct _mail_msg *msg);

/* delete all message in folder */
void empty_folder(struct _mail_folder *folder );

/* removes folder and all messages in it from memory and from disk */
/* returns -1 on error and 0 on success */
int delete_folder(struct _mail_folder *folder);

/* renames folder to name */
/* returns -1 on error , 0 on success */
int rename_folder(struct _mail_folder *folder , char *name);

/* extracts directory name from path */
char *dir_path(char *path);

/* extracts file name from path */
char *name_path(char *path);

/* returns index of folder in mailbox, 0 if not found */
int get_folder_index(struct _mail_folder *folder);

/* returns pointer to folder with index index */
struct _mail_folder *get_folder_by_index(int index);

/* returns index of message in folder, 0 if not found */
int get_msg_index(struct _mail_folder *folder , struct _mail_msg *msg);

/* returns pointer to message in folder folder with index index */
struct _mail_msg *get_msg_by_index(struct _mail_folder *folder, int index);

/* expands string str according to % directives and contents of message msg */
/* length of str should be sufficient to contain all the expanded string */
void expand_str(struct _mail_msg *msg, char *str);

/* set message flags according to Status field in header */
void set_flags_by_status(struct _mail_msg *msg);

/* set Status field in header according to the message flags */
void set_status_by_flags(struct _mail_msg *msg);

/* checks if c represents one of iso-8859-8 hebrew characters */
int ishebrew(u_int c);

/* converts str according to hebrew rules (right to left) */
char *hebrew_conv(char *str);

/* lock mailbox path using lock file */
/* returns -1 on error and 0 on success */
int lockmbox(char *path);

/* unlock mailbox */
void unlockmbox();

#endif /* _FMAIL_H */
