/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* Part of this software was derived from MH pop client program */

/* $Id: pop.c,v 1.60 1996/06/20 15:12:37 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <fmail.h>

#define	MAX_HOST	32
#if defined(SOLARIS) || defined(HP)
#define inaddr_copy(hp,sin) \
    memcpy((char *)&((sin)->sin_addr), (hp)->h_addr, (hp)->h_length)
#else
#define inaddr_copy(hp,sin) \
    bcopy ((hp)->h_addr, (char *)&((sin)->sin_addr), (hp)->h_length)
#endif

#define	MAX_UIDL	64

/* cache data structure */

struct _uidl {
	char uidlstr[MAX_UIDL];	/* message unique ID */
	u_int mnum;		/* message number on the server */
	u_long mlen;		/* message length as reported by LIST */
	struct _uidl *next;	/* next cache entry */
	u_int flags;		/* special flags */
#define	POP_DELETED	0x01	/* message has been deleted in thi session */
#define	POP_LENGTH	0x02	/* message length is stored in mlen */
	} uidl;

static struct _uidl *uidlcache = NULL;
static int nouidl = 0;		/* UIDL not supported by server */
static int num_msgs = -1;	/* number of messages on server */

#define	MAX_UIDS	3000	/* maxmimal number of UID's in file          */
FILE *uidfile;			/* file for storing retrieved messages' UIDs */
int uidentries;			/* number of entries in UID file             */

int sock = -1;
static FILE *pop_in = NULL;
static FILE *pop_out = NULL;
static char response[255];

struct _uidl *get_popmsg_by_uidl(char *uidl);
int get_popmsg_num();
void close_uidfile();

int
get_ipc_sock(sin)
struct sockaddr_in *sin;
{
        int             sock;
        int             sinlen;
 
        if ((sock = socket (AF_INET, SOCK_DGRAM, 0)) < 0)
        {
		display_msg(MSG_WARN, "ipc", "can not get socket");
                return -1;
        }

        sinlen = sizeof (struct sockaddr_in);
        memset ((char *) sin, '\0', sinlen);
 
        sin->sin_port = htons(0);
 
        sin->sin_family = AF_INET;
        sin->sin_addr.s_addr = INADDR_ANY;
        if (bind (sock, (struct sockaddr *) sin,
                sizeof (struct sockaddr_in)) < 0)
        {
		display_msg(MSG_WARN, "ipc", "bind failed");
                return -1;
        }

        if (getsockname (sock, (struct sockaddr *) sin, &sinlen) < 0)
        {
		display_msg(MSG_WARN, "ipc", "getsockname failed");
                return -1;
        }
        return sock;
}

struct hostent *
async_gethostbyname(s)
char *s;
{
static struct hostent hp;
static long iaddr;
static char *haddr[1];

struct sockaddr_in dnsin;
struct hostent *hp1;
int ipcsock, dnspid, len;

   if (!s)
	return NULL;

   if ((ipcsock = get_ipc_sock(&dnsin)) == -1)
	return NULL;

   signal(SIGCHLD, handler);

   if ((dnspid = (int)fork()) < 0) {
	close(ipcsock);
	display_msg(MSG_WARN, "async_gethostbyname", "can not fork");
	return NULL;
				   }

   if (dnspid != 0) 	{
	display_msg(MSG_STAT, NULL, "Resolving %s", s);
	display_msg(MSG_LOG, "connect", "Resolving %s", s);

	if (my_check_io_forms(ipcsock, 0) < 0)  {
	  close(ipcsock);
	  display_msg(MSG_STAT, NULL, "");
	  kill(dnspid, SIGKILL);
	  return NULL;
						}

	if (recvfrom(ipcsock, (char *)&iaddr, (int)sizeof(iaddr), (int)0,
		NULL, &len) != sizeof(iaddr)) {
		display_msg(MSG_WARN, "async_gethostbyname", "error reading from socket");
		close(ipcsock);
	  	display_msg(MSG_STAT, NULL, "");
		return NULL;
					      }

	close(ipcsock);
	iaddr = ntohl(iaddr);
	if (iaddr == -1)
		return NULL;

	hp.h_addr_list = haddr;
	hp.h_addr = (char *)&iaddr;
	hp.h_length = sizeof(iaddr);
	hp.h_addrtype = AF_INET;

	display_msg(MSG_STAT, NULL, "");
	return &hp;
			}

   if ((hp1 = gethostbyname(s)) != NULL)
	memcpy((char *)&iaddr, hp1->h_addr, hp1->h_length);
   else
	iaddr = -1;

   iaddr = htonl(iaddr);

   if (sendto(ipcsock,(char *)&iaddr,sizeof(iaddr),(int)0,(struct sockaddr *)&dnsin,sizeof(dnsin)) != sizeof(iaddr)) {
	close(ipcsock);
	_exit(1);
				     }

   close(ipcsock);
   _exit(0);
}

struct hostent *
gethostbystring(s)
char   *s;
{
static struct hostent hp;
static long iaddr;
static char *haddr[1];

    iaddr = inet_addr(s);
    if ((iaddr == -1) && strcmp(s, "255.255.255.255"))
        return async_gethostbyname(s);

    hp.h_addr_list = haddr;
    hp.h_addr = (char *)&iaddr;
    hp.h_length = sizeof(iaddr);
    hp.h_addrtype = AF_INET;

    return &hp;
}

void
host_disconnect(c_host)
char *c_host;
{

 if (sock == -1)
	return;

 display_msg(MSG_LOG, "connect", "Disconnected from %s", c_host);
 close(sock);
 sock = -1;
}

int
host_connect(hostname, service, protocol)
char hostname[MAX_HOST];
char *service;
char *protocol;
{
struct  servent *srv;
struct hostent *hp;
struct sockaddr_in  in_socket;
struct sockaddr_in *isock = &in_socket;

char c_host[MAX_HOST];
char c_service[10];
char c_prot[6];
char *p;
int pport, res;

 if (sock != -1) {
	display_msg(MSG_WARN, "connect", "network subsystem is busy");
	return -1;
		}

 if (!hostname)
	strcpy(c_host, "127.0.0.1");
 else
	strncpy(c_host, hostname, MAX_HOST);

 if (!service)
	strcpy(c_service, "110");
 else
	strncpy(c_service, service, 10);

 if (!protocol)
	strcpy(c_prot, "tcp");
 else
	strncpy(c_prot, protocol, 6);

 p = c_service;
 while( *p && isdigit(*p))
	p++;

 if (*p) {
   srv = getservbyname(c_service, c_prot);
   if (srv == NULL) {
	display_msg(MSG_WARN, "connect", "Unknown service %s/%s", c_service, c_prot);
	return -1;
	  }
   pport = srv->s_port;
 } else {
   pport = atoi(c_service);
   pport = htons(pport);
	}

 if ((hp = gethostbystring(c_host)) == NULL) {
	display_msg(MSG_WARN, "connect", "Unknown host %s", c_host);
	return -1;
	 }

 if ((sock = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
	display_msg(MSG_WARN, "connect", "Can not open socket");
	return -1;
	}

 if ((res = fcntl(sock, F_GETFL)) == -1) {
	display_msg(MSG_WARN, "connect", "fcntl F_GETFL failed");
	close(sock);
	sock = -1;
	return -1;
	}

 if (fcntl(sock, F_SETFL, res | O_NONBLOCK) == -1) {
	display_msg(MSG_WARN, "connect", "fcntl F_SETFL, O_NONBLOCK failed");
	close(sock);
	sock = -1;
	return -1;
	}

#if defined(SOLARIS) || defined(HP)
 memset((char *)isock, 0, sizeof *isock);
#else
 bzero((char *)isock, sizeof *isock);
#endif

 isock->sin_family = hp->h_addrtype;
 inaddr_copy(hp, isock);
 isock->sin_port = pport;

 display_msg(MSG_STAT, NULL, "Connecting to %s", c_host);
 display_msg(MSG_LOG, "connect", "Connecting to %s", c_host);

 if (connect(sock, (struct sockaddr *)isock, sizeof *isock) == -1) {
	if (errno != EINPROGRESS) {
	display_msg(MSG_WARN, "connect", "Connect to %s failed", c_host);
	display_msg(MSG_LOG, "connect", "Connect to %s failed", c_host);
 	display_msg(MSG_STAT, NULL, "");
	close(sock);
	sock = -1;
	return -1;
				  }
	}

conwait:
 if ((res = my_check_io_forms(sock, 1)) < 0) {
    if (res == -2)
	display_msg(MSG_LOG,"connect", "Connect to %s aborted by user", c_host);
    else {
	display_msg(MSG_WARN, "connect", "Connect to %s failed", c_host);
	display_msg(MSG_LOG, "connect", "Connect to %s failed", c_host);
	 }
 	display_msg(MSG_STAT, NULL, "");
	close(sock);
	sock = -1;
	return -1;
	}

 if (connect(sock, (struct sockaddr *)isock, sizeof *isock) == -1)  {
  switch (errno) {
    case EISCONN:
    break;

    case EINPROGRESS:
    case EALREADY:
	goto conwait;
    break;

    default:
	display_msg(MSG_WARN, "connect", "Connect to %s failed", c_host);
	display_msg(MSG_LOG, "connect", "Connect to %s failed", c_host);
 	display_msg(MSG_STAT, NULL, "");
	close(sock);
	sock = -1;
	return -1;
    break;
		}
								    }

 display_msg(MSG_LOG, "connect", "Connected to %s", c_host);
 display_msg(MSG_STAT, NULL, "");
 return sock;
}

int
getline (s, n, iop)
char   *s;
int     n;
FILE *iop;
{
static char lbuf[128] = "\0";
char   tbuf[128], *p, *p1;
int len, res;

    if ((len = strlen(lbuf)) > 0) {
	if ((p = strchr(lbuf, '\n')) != NULL) 	{
		p1 = p + 1;
		*p = '\0';
		p--;
		if (*p == '\r')
			*p = '\0';
		strcpy(s, lbuf);
		strcpy(tbuf, p1);
		strcpy(lbuf, tbuf);

		return 0;
						}
	strcpy(s, lbuf);
	p = s + len;
				  }
    else {
	s[0] = '\0';
	p = s;
	len = 0;
	}

    if (my_check_io_forms(fileno(iop), 0) < 0)
	return -1;

    while (len < n) {
     if ((res = read(fileno(iop), p, (n - len) > 127 ? 127 : (n - len))) == -1) {
	if (errno == EAGAIN) {
	  if (my_check_io_forms(fileno(iop), 0) < 0)
		return -1;

	  continue;
			     }
	display_msg(MSG_WARN, "recv", "connection error");
        return -1;
    				         }
     if (res == 0)
		break;

     p[res] = '\0';
     if ((p1 = strchr(p, '\n')) != NULL) {
	*p1 = '\0';
	strcpy(lbuf, p1 + 1);
	len += (p1 - p);
	p1--;
	if (*p1 == '\r')
		*p1 = '\0';

	break;
					 }
     p += res;
     len += res;
		    }

    if (len >= n)
	display_msg(MSG_LOG, "recv", "string is too long, splitting");

    if (len == 0)   {
	display_msg(MSG_WARN, "recv", "connection closed by foreign host");
        return -1;
		    }

    s[len] = '\0';

    return 0;
}

int putline(s, iop)
char   *s;
FILE * iop;
{
char sbuf[512];

    if (strlen(s) >= 510) {
	display_msg(MSG_WARN, "send", "line too long");
	return -1;
			  }

    sprintf(sbuf, "%s\r\n", s);
chwrite:
    if (my_check_io_forms(fileno(iop), 1) < 0)
	return -1;

    if (write(fileno(iop), sbuf, strlen(sbuf)) == -1) {
	if (errno == EAGAIN)
		goto chwrite;

	display_msg(MSG_WARN, "send", "connection lost");
        return -1;
					 }       

    return 0;
}

char *
pop_command(fmt, a, b, c, d)
char   *fmt,
       *a,
       *b,
       *c,
       *d;
{
static char response[512];

    fl_check_only_forms();

    (void) sprintf (response, fmt, a, b, c, d);

    if (strncasecmp(response, "pass ", 5))
	display_msg(MSG_LOG, "pop", "-> %s", response);
    else
	display_msg(MSG_LOG, "pop", "-> PASS *******");

    if (putline(response, pop_out) == -1)
        return NULL;

    response[0] = '\0';

    if (getline(response, 510, pop_in) == -1)
	return NULL;

    display_msg(MSG_LOG, "pop", "<- %s", response);

    if (response[0] == '+')
	return response;

    if (strncasecmp(fmt, "UIDL", 4) && !strncasecmp(response, "-ERR ", 4))
    	display_msg(MSG_WARN, "pop", "%s", response + 4);

    return NULL;
}

void
pop_close()
{
struct _uidl *ucache;

 host_disconnect(b_getcfg_str(conf_name, "pophost", "localhost"));

 while (uidlcache) 	{
  ucache = uidlcache;
  uidlcache = uidlcache->next;
  free(ucache);
			}

 if (pop_in != NULL)
	fclose(pop_in);

 if (pop_out != NULL)
	fclose(pop_out);

 pop_in = NULL;
 pop_out = NULL;

 signal(SIGPIPE, SIG_DFL);
}

void
pop_end()
{
pop_command("QUIT");
pop_close();
close_uidfile();
}

int
pop_init()
{
int fd1, fd2;
char buf[512];
struct passwd *pwd;
char *p;

 nouidl = 0;

 strcpy(buf, b_getcfg_str(conf_name, "pophost", "localhost"));

 (void)signal(SIGPIPE, SIG_IGN);

 fd1 = host_connect(buf, b_getcfg_str(conf_name, "poport", "110"), NULL);

 if (fd1 == -1) {
 	signal(SIGPIPE, SIG_DFL);
	return -1; }

 if ((fd2 = dup(fd1)) == -1) {
	display_msg(MSG_WARN, "pop", "dup failed");
	host_disconnect(buf);
	close(fd1);
 	signal(SIGPIPE, SIG_DFL);
	return -1;
	}

 if ( ((pop_in = fdopen(fd1, "r")) == NULL)
            || ((pop_out = fdopen(fd2, "w")) == NULL) ) {
	display_msg(MSG_WARN, "pop", "fdopen failed");
	close(fd1);
	close(fd2);
	pop_close();
	return -1;
	}

 
 if (getline(buf, 510, pop_in) == -1) {
	close(fd1);
	close(fd2);
	pop_close();
	return -1; }

 if (buf[0] != '+') {
	display_msg(MSG_WARN, "pop", "Invalid response from pop server");
	close(fd1);
	close(fd2);
	pop_close();
	return -1; }

 pwd = getpwuid(getuid());
 p = b_getcfg_str(conf_name,"popasswd", "");

 if (!supress_errors)   {
  if ((strlen(p) <= 1) || (b_getcfg_int(conf_name,"popstorpwd", 0) == 0))
	pop_account();
			}

 p = b_getcfg_str(conf_name, "popuser", pwd ? pwd->pw_name : "user");

 if (pop_command("USER %s", p) == NULL) {
	close(fd1);
	close(fd2);
	pop_close();
	return -1;
					}

 if (pop_command("PASS %s", b_getcfg_str(conf_name, "popasswd", "")))
	return 0;

 if (strncasecmp(response, "-ERR ", 4)) {
	close(fd1);
	close(fd2);
	pop_close();
	return -1;
					}

 pop_account();

 p = b_getcfg_str(conf_name, "popuser", pwd ? pwd->pw_name : "user");

 if (pop_command("USER %s", p) && pop_command("PASS %s", b_getcfg_str(conf_name, "popasswd", "")) )
	return 0;

 close(fd1);
 close(fd2);
 pop_close();
 return -1;

}

int
multiline() 
{
char buffer[512];
 
    if (getline(buffer, 510, pop_in) == -1)
        return -1;

    if (buffer[0] == '.') {
        if (buffer[1] == '\0')
            return 0;
        else
            (void)strcpy(response, buffer + 1);
    }
    else
        (void)strcpy(response, buffer);

    return 1;
}    

int
get_pop_msg(num, only_header, realen)
int num;
int only_header;
u_long *realen;
{
int mnum, resplen;
char buf[255];
char dummy[6], *p;
FILE *mfd;
int st, pbuf;
u_long rlen, acclen, tdiff;
struct timeval tv1, tv2;

 if ((mnum = get_new_name(ftemp)) == -1) {
	display_msg(MSG_WARN, "pop", "No space in %s", FTEMP);
	return -1; }

 sprintf(buf, "%s/%d", ftemp->fold_path, mnum);

 if ((mfd = fopen(buf, "w")) == NULL) {
        display_msg(MSG_WARN, "pop", "Can not open file %s", buf);
        return -1; }

 if (only_header == 1) {
 if ((p = pop_command("TOP %d 0", num)) == NULL) {
	display_msg(MSG_WARN, "pop", "Failed to retrieve header of message %d from server", num);
	fclose(mfd);
	unlink(buf);
	return -1;
					}
 } else
 if (only_header == 2) {
 if ((p = pop_command("TOP %d 999999", num)) == NULL) {
	display_msg(MSG_WARN, "pop", "Failed to retrieve message %d from server", num);
	fclose(mfd);
	unlink(buf);
	return -1;
					}
 } else  {
 if ((p = pop_command("RETR %d", num)) == NULL) {
	display_msg(MSG_WARN, "pop", "Failed to retrieve message %d from server", num);
	fclose(mfd);
	unlink(buf);
	return -1;
	}
					}

 if (realen) {
	*realen = 0;
	sscanf(p, "%s %lu", dummy, realen);
	}

 rlen = 0;
 acclen = 0;
 pbuf = 0;
 gettimeofday(&tv1, NULL);

 while ((st = multiline()) == 1) {
	resplen = strlen(response);
	rlen += resplen + 2;
	acclen += resplen + 2;
	if ((only_header != 1) && realen && (*realen > 8192)) 	{
	  if (acclen > (*realen * 0.05)) {
	    acclen = 0;
	    gettimeofday(&tv2, NULL);
	    tdiff = (tv2.tv_sec - tv1.tv_sec) * 1000000 - tv1.tv_usec + tv2.tv_usec;
 	    display_msg(MSG_STAT, NULL, "POP: retrieving message %d of %d (%d %% - %.2f kb/sec)", num, num_msgs, rlen * 100 / *realen,
		(double)(((double)rlen * 1000000)/ (double)tdiff / 1024) );
					}
								}

	if (resplen < 1) 
		pbuf = 1;
	else {
	  if (pbuf) {
		pbuf = 0;
		fputc('\n', mfd);
		    }
	  if (fputs(response, mfd) == EOF) {
		if (errno == ENOSPC)
			display_msg(MSG_WARN, "pop", "DISK FULL!");
		else
			display_msg(MSG_WARN, "pop", "Error writing %s", buf);
		fclose(mfd);
		unlink(buf);
		errno = 0;
		return -1;
					   }
	  fputc('\n', mfd);
	    }
				}

 if (fclose(mfd) == EOF) {
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "pop", "DISK FULL!");
	else
		display_msg(MSG_WARN, "pop", "Error writing %s", buf);
	fclose(mfd);
	unlink(buf);
	errno = 0;
	return -1;
			 }

 if (st == -1) {
	display_msg(MSG_WARN, "pop", "Error when retrieving message from server");
	unlink(buf);
	return -1;
		}

 return mnum;
}

int
pop_send_message(msg)
struct _mail_msg *msg;
{
char *p;

 if (!msg)
	return -1;

 if (pop_init() == -1)
	return -1;

 if ((p = pop_command("XTND XMIT")) == NULL) {
	display_msg(MSG_WARN, "Transmit command failed!", "Probably it's not supported on this POP server");
	pop_end();
	return -1; }

 if (smtp_message(msg, pop_out) == -1) {
	pop_end();
	return -1; }

 if ((p = pop_command(".")) == NULL) {
	display_msg(MSG_WARN, "POP Send", "Failed to send message");
	pop_end();
	return -1; }

 pop_end();
 return 0;
}

long
get_popmsg_len(num)
int num;
{
int st;
struct _uidl *ucache;
u_long msg_len;
u_int mnum;

 if ((uidlcache == NULL) ||
	!(uidlcache->flags & POP_LENGTH))
					{
  get_popmsg_by_uidl("");
  if (uidlcache == NULL) {
    if (!nouidl)
	return -2;

    if (pop_command("LIST %d", num) == NULL) {
	display_msg(MSG_WARN, "pop", "Can not determine message length!");
	return -2;
					     }
	mnum = -1;
	msg_len = 0L;
	sscanf(response, "%d %lu", &mnum, &msg_len);
	if ((mnum != num) || (msg_len == -1))
		return -2;

	return msg_len;

			 }

  if (pop_command("LIST") == NULL) {
	display_msg(MSG_WARN, "pop", "Can not determine message length!");
	return -2; }

  ucache = uidlcache;

  while ((st = multiline()) == 1)  {
   mnum = -1;
   msg_len = 0L;
   sscanf(response, "%d %lu", &mnum, &msg_len);

   if ((mnum == -1) || (msg_len == 0L))
	continue;

   while (ucache && (ucache->mnum != mnum))
	ucache = ucache->next;

   if (!ucache) {
	ucache = uidlcache;
	while (ucache && (ucache->mnum != mnum))
		ucache = ucache->next;
		}

   if (!ucache)
	continue;

   ucache->mlen = msg_len;
   ucache->flags |= POP_LENGTH;
   ucache = ucache->next;
				   }
					}

  ucache = uidlcache;
  while (ucache && (ucache->mnum != num))
		ucache = ucache->next;

  if (!ucache || !(ucache->flags & POP_LENGTH)) {
	display_msg(MSG_WARN, "pop", "Can not determine message length (%d)!", num);
	return -2; }

  return ucache->mlen;
}

char *
get_popmsg_uidl(num)
int num;
{
struct _uidl *ucache;

 if (uidlcache == NULL)
	get_popmsg_by_uidl("");

 if (nouidl)
	return NULL;

 ucache = uidlcache;

 while (ucache) {
	if (ucache->mnum == num)
		return ucache->uidlstr;

	ucache = ucache->next;
		}

 return NULL;
}

struct _uidl *
get_popmsg_by_uidl(uidl)
char *uidl;
{
int st;
char muidl[MAX_UIDL];
struct _uidl *ucache, *uidl1, *uidl2;

 if (uidlcache == NULL) {
  if (pop_command("UIDL") == NULL) {
	nouidl = 1;
	return NULL; }

			}
 else	{
  ucache = uidlcache;
  while (ucache) {
   if (!strcmp(ucache->uidlstr, uidl))
	return ucache;

   ucache = ucache->next;
		 }

   return NULL;
	}

 uidl2 = NULL;

 while ((st = multiline()) == 1)  {
  muidl[0] = '\0';
  st = 0;
  sscanf(response, "%d %s", &st, muidl);
  ucache = (struct _uidl *)malloc(sizeof(struct _uidl));
  ucache->mnum = st;
  ucache->mlen = 0;
  ucache->flags = 0;
  strcpy(ucache->uidlstr, muidl);
  ucache->next = NULL;

  if (!strcmp(muidl, uidl))
	uidl2 = ucache;

  if (uidlcache == NULL)
	uidlcache = ucache;
  else 	{
   uidl1 = uidlcache;
   while (uidl1->next)
	uidl1 = uidl1->next;

   uidl1->next = ucache;
	}
				  }
 return uidl2;
}

int
get_popmsg_num()
{
char *p;
char dumb[5];
int msgs_len = 0;

 if ((p = pop_command("STAT")) == NULL)
	return -1;

 sscanf(p, "%s %d %d", dumb, &num_msgs, &msgs_len);

 if (num_msgs == -1) {
	display_msg(MSG_WARN, "pop", "STAT failed");
        return -1; }

 return num_msgs;
}

int
pop_delmsg_by_uidl(msg)
struct _mail_msg *msg;
{
int mnum;
struct _head_field *fld;
struct _uidl *uidl;
int init;

 if (pop_in)
	init = 0;
 else
	init = 1;

 if (!msg)
	return -1;

 if ((fld = find_field(msg, XUIDL)) == NULL)
	return -1;

 if (init && (pop_init() == -1))
	return -1;

 if ((uidl = get_popmsg_by_uidl(fld->f_line)) == NULL) {
	if (nouidl)
		display_msg(MSG_WARN, "pop", "You can not use this feature\nsince your POP server does not support UIDL command");
	if (init)
	 pop_end();
	return -1; }

 mnum = uidl->mnum;

 if (mnum == 0) {
	if (init)
	 pop_end();
	return -1; }

 if (!(uidl->flags & POP_DELETED)) {
 display_msg(MSG_STAT, NULL, "POP: deleting message %d", mnum);
 pop_command("DELE %d", mnum);
 uidl->flags |= POP_DELETED;
				   }

 if (init)
  pop_end();

 msg->flags &= ~H_ONLY;

 return 0;
}

int
pop_getfull_msg(msg)
struct _mail_msg *msg;
{
struct _head_field *fld;
struct _uidl *uidl;
int mnum, nnum, st;
FILE *mfd, *bfd;
char buf[255];
char msgf[255];
u_long realen;

 if (!msg)
	return -1;

 if (!(msg->flags & H_ONLY))
	return -1;

 if ((fld = find_field(msg, XUIDL)) == NULL) {
	display_msg(MSG_WARN, "pop", "Message does not have %s identifier", XUIDL);
	return -1; }

 if (pop_init() == -1)
	return -1;

 if ((uidl = get_popmsg_by_uidl(fld->f_line)) == NULL) {
	if (nouidl)
	  display_msg(MSG_WARN, "pop", "You can not use thise feature\nsince your POP server does not support UIDL command");
	else
	  display_msg(MSG_WARN, "pop", "Failed to find message");
	pop_end();
	return -1; }

 mnum = uidl->mnum;

 if (mnum == 0) {
	display_msg(MSG_WARN, "pop", "Can not find message, probably it's no longer on the server");
	pop_end();
	return -1; }

 if ((nnum = get_pop_msg(mnum, 0, &realen)) == -1) {
	pop_end();
	return -1; }

 if (b_getcfg_int(conf_name, "popdelete", 0))
	pop_command("DELE %d", mnum);

 pop_end();

 sprintf(msgf, "%s/%d", ftemp->fold_path, nnum);

 if ((bfd = fopen(msgf, "r")) == NULL) {
	display_msg(MSG_WARN, "pop", "Can not open retrieved message");
	unlink(msgf);
	pop_end();
	return -1; }

 if ((mfd = fopen(msg->get_file(msg), "a")) == NULL) {
	display_msg(MSG_WARN, "pop", "Can not open message %s", msg->get_file(msg));
	fclose(bfd);
	unlink(msgf);
	pop_end();
	return -1; }

 st = 0;
 while (fgets(buf, 255, bfd)) {
  if (!st && (buf[0] == '\n')) {
	st = 1;
	continue;     }

  if (st)
	fputs(buf, mfd);
				}
 
 fflush(mfd);
 msg->msg_len = ftell(mfd);
 fclose(mfd);
 fclose(bfd);
 unlink(msgf);

 msg->flags &= ~H_ONLY;

 return 0;
}

int
if_popmsg_retr(num, xlst)
int num, xlst;
{
char *p;
int ml, st;

 if (xlst)
	p = pop_command("XTND XLST Status %d", num);
 else
	p = pop_command("TOP %d 0", num);

 if (!p) {
	display_msg(MSG_WARN, "pop", "Can not determine message status");
	return 0;
	}

 st = 0;
 while ((ml = multiline()) == 1)  {
   if ((p = strstr(response, "Status:")) != NULL) {
	p += 7;
	if (strchr(p, 'R'))
		st = 1;
						  }
				  }

 return st;
}

void
close_uidfile()
{
 if (uidfile) {
	fclose(uidfile);
	uidfile = NULL;
		}

 uidentries = -1;

 return;
}

void
open_uidfile()
{
char uidfname[255];

 close_uidfile();
 sprintf(uidfname, "%s/.popuid", b_getcfg_str(conf_name, "maildir", ""));

 if ((uidfile = fopen(uidfname, "a+")) == NULL) {
	display_msg(MSG_WARN, "Message uids will not be checked", "Can not open %s", uidfname);
	return; 
						}

 return;
}

void
count_uidfile()
{
char uidstr[MAX_UIDL + 1];

 if (uidfile == NULL)
	return;

 if (fseek(uidfile, 0L, SEEK_SET) == -1) {
	uidentries = -1;
	return; }

 uidentries = 0;
 while (fgets(uidstr, MAX_UIDL, uidfile))
	uidentries++;

 return;
}

void
truncate_uidfile()
{
char uidstr[MAX_UIDL + 1];
long uidflen, uidbufsize;
char *uidbuf;

 if (uidfile == NULL)
	return;

 uidentries = -1;

 if (fseek(uidfile, 0L, SEEK_END) == -1) {
	display_msg(MSG_WARN, "POP - truncate", "Can not seek in UID file");
	return; }

 if ((uidflen = ftell(uidfile)) == -1) {
	display_msg(MSG_WARN, "POP - truncate", "Can not determine length of UID file");
	return; }

 if (fseek(uidfile, uidflen / 2, SEEK_SET) == -1) {
	display_msg(MSG_WARN, "POP - truncate", "Can not seek in UID file");
	return; }

 if (fgets(uidstr, MAX_UIDL, uidfile) == NULL) {
	display_msg(MSG_WARN, "POP - truncate", "Failed to read from UID file");
	return; }

 uidbufsize = uidflen - ftell(uidfile);
 if ((uidbuf = (char *)malloc(uidbufsize)) == NULL) {
	display_msg(MSG_WARN, "POP - truncate", "Failed to allocate read buffer");
	return; }

 if (fread(uidbuf, uidbufsize, 1, uidfile) != 1) {
	display_msg(MSG_WARN, "POP - truncate", "Failed to read buffer from UID file");
	return; }

 if (fseek(uidfile, 0L, SEEK_SET) == -1) {
	display_msg(MSG_WARN, "POP - truncate", "Can not seek in UID file");
	return; }

 if (fwrite(uidbuf, uidbufsize, 1, uidfile) != 1) {
	display_msg(MSG_WARN, "POP - truncate", "Failed to write buffer to UID file");
	return; }

 free(uidbuf);

 fflush(uidfile);

 if (ftruncate(fileno(uidfile), uidbufsize) == -1) {
	display_msg(MSG_WARN, "POP - truncate", "Failed to truncate UID file");
	return; }

 return;
}

void
append_uidfile(uid)
char *uid;
{
 if (!uid)
	return;

 if (strlen(uid) > MAX_UIDL)
	return;

  if (uidfile == NULL)
	open_uidfile();

  if (uidfile == NULL)
	return;

  if (uidentries == -1)
	count_uidfile();

  if (uidentries >= MAX_UIDS) {
	truncate_uidfile();
	count_uidfile(); }

  if (fseek(uidfile, 0L, SEEK_END) == -1) {
	display_msg(MSG_WARN, "POP - append", "Can not seek in UID file");
	return; }

  if (fputs(uid, uidfile) == EOF) {
	display_msg(MSG_WARN, "POP - append", "Can not write to UID file");
	return; }

  fputc('\n', uidfile);
  uidentries++;

  return;
}

int
if_popmsg_uid_cached(msg)
int msg;
{
char *uid;
char uidstr[MAX_UIDL + 1];
int uide;

  if (uidfile == NULL)
	open_uidfile();

  if (uidfile == NULL)
	return -1;

  if ((uid = get_popmsg_uidl(msg)) == NULL)
	return -1;

  if (fseek(uidfile, 0L, SEEK_SET) == -1) {
	display_msg(MSG_WARN, "POP", "Can not seek in UID file");
	return -1; }

  uide = 0;
  while (fgets(uidstr, MAX_UIDL, uidfile)) {
	uide++;
	if (uidstr[strlen(uidstr) - 1] == '\n')
		uidstr[strlen(uidstr) - 1] = '\0';

	if (!strncmp(uid, uidstr, strlen(uid)))
	 return 1;
					   }

  uidentries = uide;
  return 0;
}

int
pop_inc()
{
int i, honly, uidcached;
int num_msgs, retr_num;
struct _mail_msg *msg;
int save_addr, popmaxact, popcheckretr, popgetstat, popdelete, popdmark, pop2;
long popmaxmsg;
char mclen[16];
u_long realen;
struct _head_field *fld;
int nnum = -1;

 save_addr = b_getcfg_int(conf_name, "saveaddr", 0);

 popmaxact = b_getcfg_int(conf_name, "popmaxact", 0);
 pop2 = b_getcfg_int(conf_name, "pop2", 0);
 popmaxmsg = b_getcfg_int(conf_name, "popmaxmsg", -1);
 if (popmaxmsg >= 0)
	popmaxmsg *= 1024;
 else
	popmaxmsg = MAX_MSG_LEN;

 if (popmaxmsg > MAX_MSG_LEN)
	popmaxmsg = MAX_MSG_LEN;

 popcheckretr = b_getcfg_int(conf_name, "popcheckretr", 0);
 popgetstat = b_getcfg_int(conf_name, "popgetstat", 0);

 popdelete = b_getcfg_int(conf_name, "popdelete", 0);
 popdmark = b_getcfg_int(conf_name, "popdmark", 0);
 if (popdmark)
	popdmark = 2;

 if (pop_init() == -1)
	return -1;

 if ((num_msgs = get_popmsg_num()) == -1) {
	pop_end();
	return -1; }

 if (num_msgs == 0) {
	pop_end();
	return 0; }

 retr_num = 0;

 for (i = 1; i <= num_msgs; i++) {

	honly = 0;
	uidcached = -1;

	if (abortpressed())
		break;

	if (!pop2 && popcheckretr) {
	  if (((popcheckretr == 1) || (popcheckretr == 3))
		&& (if_popmsg_retr(i, popgetstat) > 0) )
		continue;

	  if (((popcheckretr == 2) || (popcheckretr == 3))
		&& ( (uidcached = if_popmsg_uid_cached(i)) == 1) )
		continue;
			  }

	if (!pop2 && (popmaxmsg >= 0) && 
		((realen = get_popmsg_len(i)) >= popmaxmsg)) {
	  if (popmaxact > 0)
		continue;

	  if (nouidl)   {
		display_msg(MSG_WARN, "Can not retrieve message header, skipping", "Your POP server does not support UIDL command\nIt will be impossible to match header and message left on the server later");
		continue;
			}

 	  display_msg(MSG_STAT, NULL, "POP: retrieving header of message %d of %d", i, num_msgs);
	  if ((nnum = get_pop_msg(i, 1, &realen)) == -1) {
		pop_end();
		return -1; }

	  if ((msg = get_message(nnum, ftemp)) == NULL) {
		pop_end();
		return -1; }

	  msg->flags |= H_ONLY;

	  if (realen > 0) {
		sprintf(mclen, "%lu", realen);
	  	replace_field(msg, MIME_C_LENGTH, mclen);
			 }

	  honly = 1;

	  if (find_field(msg, XUIDL) == NULL)
		add_field(msg, XUIDL, get_popmsg_uidl(i));

	} else {

 	display_msg(MSG_STAT, NULL, "POP: retrieving message %d of %d", i, num_msgs);
	if ((nnum = get_pop_msg(i, popdmark, &realen)) == -1) {
		pop_end();
		return -1; }

	if ((msg = get_message(nnum, ftemp)) == NULL) {
		pop_end();
		return -1; }

		}

	if (!pop2 && ((popcheckretr == 2) || (popcheckretr == 3)) &&
		(uidcached == 0) && (!popdelete || honly)) {
	  if ((fld = find_field(msg, XUIDL)) == NULL)
		append_uidfile(get_popmsg_uidl(i));
	  else	
		append_uidfile(fld->f_line);
				  }

	set_flags_by_status(msg);
	convert_fields(msg);
	msg->folder = NULL;
	msg->flags |= CHANGED;
	add_field(msg, "X-RDate", get_arpa_date(time(NULL)));
	msg->header->rcv_time = time(NULL);

        if (save_addr)
		add_msg_addr(msg);

	update_faces(msg);

	if (apply_rule(msg, 0) == -1) {
        	if (save_addr)
			add_msg_addr(NULL);
		pop_end();
		unlink(msg->get_file(msg));
		discard_message(msg);
		return -1;	   }

	retr_num++;

	if (popdelete && !honly)
		pop_command("DELE %d", i);

						}

 if (save_addr)
	add_msg_addr(NULL);

 pop_end();
 return retr_num;
}
