/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *      
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: smtp.c,v 1.27 1996/06/09 13:46:11 gena Exp $
 */

#include <fmail.h>
#include <umail.h>

static char response[255];
FILE *smtp_in = NULL;
FILE *smtp_out = NULL;

int
smtp_command(fmt, a, b, c, d)
char   *fmt,
       *a,
       *b,
       *c,
       *d;
{
int res;
char buf[255];

 if (fmt) {
 (void)sprintf(response, fmt, a, b, c, d);

 display_msg(MSG_LOG, "smtp", "-> %s", response);
 if (putline(response, smtp_out) == -1)
	return -1;
	  }

 response[0] = '\0';
 response[3] = '-';

 while (response[3] == '-') {
 if (getline(response, sizeof(response), smtp_in) == -1)
	return -1;
			    }

 display_msg(MSG_LOG, "smtp", "<- %s", response);

 res = -1;
 sscanf(response, "%d%s", &res, buf);
 
 if (res == -1) {
	display_msg(MSG_WARN, "smtp", "%s", response);
	return -1;
		}

 return res;
}

void
smtp_close()
{

 host_disconnect(b_getcfg_str(conf_name, "smtphost", "localhost"));

 if (smtp_in != NULL)
        fclose(smtp_in);

 if (smtp_out != NULL)
        fclose(smtp_out);

 smtp_in = NULL;
 smtp_out = NULL;

 (void)signal(SIGPIPE, SIG_DFL);
}

void
smtp_end()
{
if (smtp_command("QUIT") != 221)
	display_msg(MSG_WARN, "smtp", "%s", response);

smtp_close();
}


int 
smtp_init()
{
int fd1, fd2;
char buf[255];
struct _mail_addr *fa;

 strcpy(buf, b_getcfg_str(conf_name, "smtphost", "127.0.0.1"));

 (void)signal(SIGPIPE, SIG_IGN);
    
 fd1 = host_connect(buf, b_getcfg_str(conf_name, "smtport", "25"), NULL);
    
 if (fd1 == -1) {
 	(void)signal(SIGPIPE, SIG_DFL);
        return -1; }

 if ((fd2 = dup(fd1)) == -1) {
        display_msg(MSG_WARN, "smtp", "dup failed");
        close(fd1);
 	(void)signal(SIGPIPE, SIG_DFL);
        return -1;
        }

 if ( ((smtp_in = fdopen(fd1, "r")) == NULL)
            || ((smtp_out = fdopen(fd2, "w")) == NULL) ) {
        display_msg(MSG_WARN, "smtp", "fdopen failed");
        close(fd1);
        close(fd2);
        smtp_close();
        return -1;
        }

 if (smtp_command(NULL) != 220) {
        display_msg(MSG_WARN, "smtp", "%s", response);
        close(fd1);
        close(fd2);
        smtp_close();
        return -1; }

 fa = get_address(b_getcfg_str(conf_name, "from", ""));
 if (smtp_command("HELO %s", getmyhostname() ) != 250) {
        display_msg(MSG_WARN, "smtp", "%s", response);
        close(fd1);
        close(fd2);
        smtp_close();
        return -1; }

 discard_address(fa);

return 0;
}

int
smtp_send_message(msg)
struct _mail_msg *msg;
{
struct _mail_addr *addr;

 if (smtp_init() == -1)
	return -1;

 if (smtp_command("MAIL FROM: <%s>", msg->header->From->addr ) != 250) {
        display_msg(MSG_WARN, "smtp", "%s", response);
        smtp_end();
        return -1; }

 addr = msg->header->To;
 while (addr) {
 	if (smtp_command("RCPT TO: <%s>", addr->addr ) != 250) {
        	display_msg(MSG_WARN, "smtp", "%s", response);
		smtp_end();
		return -1;				}
        addr = addr->next_addr; }

 addr = msg->header->Cc;
 while (addr) {
 	if (smtp_command("RCPT TO: <%s>", addr->addr ) != 250) {
        	display_msg(MSG_WARN, "smtp", "%s", response);
		smtp_end();
		return -1;				}
        addr = addr->next_addr; }

 addr = msg->header->Bcc;
 while (addr) {
 	if (smtp_command("RCPT TO: <%s>", addr->addr ) != 250) {
        	display_msg(MSG_WARN, "smtp", "%s", response);
		smtp_end();
		return -1;				}
        addr = addr->next_addr; }

 if (smtp_command("DATA") != 354) {
       	display_msg(MSG_WARN, "smtp", "%s", response);
	smtp_end();
	return -1;		  }

 if (smtp_message(msg, smtp_out) == -1) {
	smtp_end();
	return -1;		  }

 if (smtp_command(".") != 250) {
       	display_msg(MSG_WARN, "smtp", "%s", response);
	smtp_end();
	return -1;		  }

 smtp_end();
 return 0;
}

int
smtp_addr(addr, str, file)
mail_addr *addr;
char *str;
FILE* file;
{
u_int st, st1;
u_int alen, quote;
int len;
char buf[255];
char *p, *p1, *p2;

if (!addr)
        return -1;

if (file == NULL)
        return -1;

st = st1 = 0;

if (addr->name && strpbrk(addr->name, ".,;\'\"()<>") &&
  (*addr->name != '\'') && (*addr->name != '\"'))
	quote = 1;
else
	quote = 0;

if ((p = rfc1522_encode(addr->comment, -1, -1)) != addr->comment)
	p = p ? strdup(p) : NULL;
else
	p = NULL;

if ((p1 = rfc1522_encode(addr->name, -1, -1)) == addr->name)
	p1 = NULL;
else
	quote = 0;

if (addr->name && addr->comment)
        sprintf(buf, "%s: %s%s%s <%s> (%s)%s", str, quote ? "\"" : "" , p1 ? p1 : addr->name , quote ? "\"" : "", addr->addr, p ? p : addr->comment, addr->next_addr ? "," : "");
else
if (!addr->name && !addr->comment)
        sprintf(buf, "%s: %s%s", str, addr->addr, addr->next_addr ? "," : "");
else
if (addr->name)
        sprintf(buf, "%s: %s%s%s <%s>%s", str, quote ? "\"" : "", p1 ? p1 : addr->name, quote ? "\"" : "", addr->addr, addr->next_addr ? "," : "");
else
        sprintf(buf, "%s: (%s) <%s>%s", str, p ? p : addr->comment, addr->addr, addr->next_addr ? "," : "");

if (p)
	free(p);
p = NULL;

if (putline(buf, file) == -1)
	return -1;

if (!addr->next_addr)
	return 0;

st1 += strlen(buf);

addr =  addr->next_addr;

p2 = buf;

while(addr) {

if ((p = rfc1522_encode(addr->comment, -1, -1)) != addr->comment)
	p = p ? strdup(p) : NULL;
else
	p = NULL;

if (addr->name && strpbrk(addr->name, ".,;\'\"()<>") &&
  (*addr->name != '\'') && (*addr->name != '\"'))
	quote = 1;
else
	quote = 0;

if ((p1 = rfc1522_encode(addr->name, -1, -1)) == addr->name)
	p1 = NULL;
else
	quote = 0;

  alen = strlen(addr->addr);
  if (addr->name)
        alen += (strlen(p1 ? p1 : addr->name) + 4);

  if (addr->comment)
        alen += (strlen(p ? p : addr->comment) + 4);

  if (quote)
	alen += 2;

  if ((st1 - st + alen) > 80) {
        st = st1;
	if (putline(buf, file) == -1) {
		if (p)
			free(p);
		return -1; }
	p2 = buf;
	*p2 = '\t';
	p2++;
	*p2 = '\0';
	st1 += 3;
                        }
  else {
	*p2 = ' ';
	p2++;
	*p2 = '\0';
	st1 += 2;
	}

  if (!addr->name && !addr->comment)
	sprintf(p2, "%s%s", addr->addr, addr->next_addr ? "," : "");
  else
  if (addr->name && (p ? p : addr->comment))
        sprintf(p2, "%s%s%s <%s> (%s)%s", quote ? "\"" : "", p1 ? p1 : addr->name, quote ? "\"" : "", addr->addr, p ? p : addr->comment, addr->next_addr ? "," : "");
  else
  if (addr->name)
        sprintf(p2, "%s%s%s <%s>%s", quote ? "\"" : "", p1 ? p1 : addr->name, quote ? "\"" : "", addr->addr, addr->next_addr ? "," : "");
  else
        sprintf(p2, "(%s) <%s>%s", p ? p : addr->comment, addr->addr, addr->next_addr ? "," : "");

  len = strlen(p2);
  st1 += len;
  p2 += len;

  if (p)
	free(p);
  p = NULL;

  addr = addr->next_addr;
           }

return putline(buf, file);
}

int
smtp_header_field(fld, file)
head_field *fld;
FILE *file;
{
char buf[255];
char *p, *p1, c;
int nlen;

  sprintf(buf, "%s: ", fld->f_name);
  p = rfc1522_encode(fld->f_line, -1 , -1);
  nlen = FLD_BREAK - (strlen(fld->f_name) + 2);

  while(strlen(p) > nlen)       {
        c = p[nlen];
        p[nlen] = '\0';
        if ( (p1 = strstr(p, "; ")) ||
                (p1 = strstr(p, ", ")) ||
                   (p1 = strrchr(p, ' ')) ) {
         p[nlen] = c;
         if (*p1 != ' ')
                p1++;

         nlen = (p1 - p);
	 if (nlen <= 0)
		nlen = 1;
         if ((nlen < (FLD_BREAK / 8)) ||
                ((strlen(p) - nlen) < (FLD_BREAK / 8))   )      {
		strncat(buf, p, nlen);
                p += nlen;
                nlen = FLD_BREAK;
                continue;
                                        			}

	 strncat(buf, p, nlen);
	 if (putline(buf, file) == -1)
		return -1;

	 strcpy(buf, " ");

         p1++;
         p = p1;
         nlen = FLD_BREAK - 1;
                                        }
        else    {
         p[nlen] = c;
	 strncat(buf, p, nlen);
         p += nlen;
         nlen = FLD_BREAK;
                }
                                }

  strcat(buf, p);

  return putline(buf, file);
}

int
smtp_message(msg, m_out)  
struct _mail_msg *msg;
FILE *m_out;
{
head_field *hf;
char buf[512], *p;
FILE *mfd;

if (!msg || !m_out)
        return -1;

if (msg->header) {
        hf = msg->header->other_fields;
        while(hf) {
	  if (!strip_when_send(hf))
	    if (smtp_header_field(hf, m_out) == -1)
		return -1;
          hf = hf->next_head_field;
                }

	smtp_addr(msg->header->Sender, "Sender", m_out);
        smtp_addr(msg->header->From, "From", m_out);
        smtp_addr(msg->header->To, "To", m_out);

        if (msg->header->Subject) {
                sprintf(buf, "Subject: %s", rfc1522_encode(msg->header->Subject, -1, -1));
	  	if (putline(buf, m_out) == -1)
			return -1;
				  }

        smtp_addr(msg->header->Cc, "Cc", m_out);
        smtp_addr(msg->header->Bcc, "Bcc", m_out);

                }

if (putline("", m_out) == -1)
	return -1;

if ((mfd = fopen(msg->get_file(msg), "r")) == NULL) 	{
	display_msg(MSG_WARN, "smtp", "Can not open %s", msg->get_file(msg));
	return -1;
							}

if (fseek(mfd, msg->header->header_len, SEEK_SET) == -1) {
	display_msg(MSG_WARN, "smtp", "Can not access message body");
	fclose(mfd);
	return -1;
						}

while (fgets(buf, 255, mfd)) 	{
	if ((p = strrchr(buf, '\n')) != NULL)
		*p = '\0';
	if (strcmp(buf, "."))
		putline(buf, m_out);
	else
		putline(">.", m_out);
				}

if (ferror(mfd) && !feof(mfd)) 	{
	display_msg(MSG_WARN, "smtp", "Error reading mesage body");
	fclose(mfd);
	return -1;
				}

fclose(mfd);
return 0;
}

