/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: addr_book.c,v 1.35 1996/06/24 13:11:02 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <fmail.h>
#include <choose_folder.h>

void Address_Browse_Dbl_Call(FL_OBJECT *obj, long param);
void display_book();
void display_books();
int load_book(char *name);
int save_book();
int convert_book();
void free_book();
void edit_entry();
int get_entry_idx(struct _abook_entry *entry);
struct _abook_entry *get_entry_by_idx(int index);
struct _abook_entry *find_addr(struct _mail_addr *addr);
struct _abook_entry *add_book(struct _mail_addr *addr, int ask);

static int ready = 0;
static int modready = 0;
static int already = 0;
static int type = TO_TYPE;
static int changed = 0;
static FD_Address_Book *addr_book_obj;
static FD_addrmod *amod_obj;
static FD_aliased *alias_obj;

static struct _abook *book = NULL;
static struct _abook_entry *cur_entry = NULL;

void
add_msg_addr(msg)
struct _mail_msg *msg;
{
struct _mail_addr *addr;

 if (ready)
	return;

 if (!book ||
	strcmp(book->name, "default")) {
	if (load_book(NULL) == -1)
		return;
	    }

 if (!msg) {
	save_book();
	return; }

 if (!msg->header)
	return;

 addr = msg->header->From;
 while (addr) {
        add_book(addr, 0);
        addr = addr->next_addr; }

 addr = msg->header->To;
 while (addr) {
        add_book(addr, 0);
        addr = addr->next_addr; }

 addr = msg->header->Cc;
 while (addr) {
        add_book(addr, 0);
        addr = addr->next_addr; }

 addr = msg->header->Bcc;
 while (addr) {
        add_book(addr, 0);
        addr = addr->next_addr; }

 return;
}

struct _abook_entry *
find_addr(addr)
struct _mail_addr *addr;
{
struct _abook_entry *entry;
struct _mail_addr *addr1;

  if (!addr || !book)
	return NULL;

  entry = book->entries;
  while (entry) {
   addr1 = entry->addr;
   while (addr1) {
    if (!strcasecmp(addr1->addr, addr->addr))
	return entry;

    addr1 = addr1->next_addr;
		 }
   entry = entry->next;
		}

  return NULL;
}

struct _abook_entry *
add_book(addr, ask)
struct _mail_addr *addr;
int ask;
{
struct _abook_entry *entry;
struct _mail_addr *addr1;
int num;

 if (addr == NULL)
	return NULL;

 if (book == NULL) {
	if (load_book(NULL) == -1)
		return NULL;
	    	   }

 num = 0;
 addr1 = addr;
 while (addr1) 	{
	num++;
	addr1 = addr1->next_addr;
		}

 if ((num == 1) && find_addr(addr))   {
	if (!ask)
		return NULL;
	if (ask && !display_msg(MSG_QUEST, NULL, "\"%s\"\nalready in address book, still add?", addr->addr))
		return NULL;
			}

 if ((entry = (struct _abook_entry *)malloc(sizeof(struct _abook_entry))) 
	== NULL) {
	display_msg(MSG_FATAL, "add to book", "Malloc failed");
	return NULL;
		}

 entry->addr = copy_address_chain(addr);
 entry->description = NULL;
 entry->type = (num == 1) ? ADDR_SINGLE : ADDR_ALIAS;
 entry->next = book->entries;
 book->entries = entry;

 return entry;
}

void
free_book()
{
struct _abook_entry *entry, *entry1;

  if (book == NULL)
	return;

  entry = book->entries;
  while(entry) 	{
   if (entry->description)
	free(entry->description);
   discard_address(entry->addr);
   entry1 = entry;
   entry = entry->next;
   free(entry1);
		}

  free(book);
  book = NULL;
  cur_entry = NULL;
}

int
load_book(name)
char *name;
{
struct _abook_entry *entry, *lentry;
struct _mail_addr *addr, *laddr;
char buf[255], *p;
int num, eenum, fline;
FILE *afd;

  if (name && (strlen(name) > 15)) {
	display_msg(MSG_WARN, "address book", "Address book name too long");
	return -1;
			}

  if (changed) {
     if (display_msg(MSG_QUEST, "address book", "Do you want to save changes into the address book?"))
	save_book();
		}

  free_book();
  changed = 0;

  if (name)
	sprintf(buf, "%s/.xfbook.%s", b_getcfg_str(conf_name, "addrdir", getenv("HOME")), name);
  else
	sprintf(buf, "%s/.xfbook", b_getcfg_str(conf_name, "addrdir", getenv("HOME")));
  if ((afd = fopen(buf, "r")) == NULL) {
	if (display_msg(MSG_QUEST, "address book", "\"%s\" address book does not exists\ncreate?", name ? name : "default"))   	{
	  if ((afd = fopen(buf, "w")) == NULL)  {
		display_msg(MSG_WARN, "address book", "Failed to create address book \"%s\"", name ? name : "default");
		return -1;
						}
	  fclose(afd);
	  afd = fopen(buf, "r");
						}
	else
		return -1;
					}

  lentry = NULL;
  if ((book = (struct _abook *)malloc(sizeof(struct _abook))) == NULL) {
	display_msg(MSG_FATAL, "address book", "malloc failed\n");
	fclose(afd);
	return -1;
									 }

  book->entries = NULL;
  strcpy(book->name, name ? name : "default");
  book->flags = ABOOK_LOCAL;
  eenum = 0;
  fline = 0;

  while (fgets(buf, 255, afd))  {
   if (strncmp(buf, "@ ", 2)) {
    if (!name && (fline == 0)) {
	if (display_msg(MSG_QUEST, NULL, "Your address book is in old format, convert?")) {
	   fclose(afd);
	   if (convert_book() == -1) {
		display_msg(MSG_WARN, "conver", "Failed to convert address book");
		return -1;
				     }
	   return load_book(NULL);
	  }
	else
		return -1;
		               }
     fline = 1;
     continue;
			      }
     fline = 1;

newent:
   if ((p = strrchr(buf, '\n')) != NULL)
	*p = '\0';

   if (eenum >= MAX_ADDR_IN_BOOK) {
	display_msg(MSG_WARN, "addr book", "Address book too big\nignoring");
	break;
				}

   if ((entry = (struct _abook_entry *)malloc(sizeof(struct _abook_entry))) ==
	NULL) {
	display_msg(MSG_FATAL, "address book", "malloc failed\n");
	fclose(afd);
	return -1;
		}

   eenum++;
   entry->next = NULL;

   p = buf + 2;
   p = rem_tr_space(p);
   if (p && strlen(p))
	entry->description = strdup(p);
   else
	entry->description = NULL;

   entry->addr = NULL;
   num = 0;
   buf[0] = '\0';
   laddr = NULL;

   while (fgets(buf, 255, afd))  {
     if (buf[0] != ' ')
	break;

     if ((p = strrchr(buf, '\n')) != NULL)
	*p = '\0';

     p = rem_tr_space(buf);
     if (!strlen(p))
	continue;

     if (num >= MAX_ADDR_IN_ALIAS)
	break;

     if ((addr = get_address(p)) == NULL)
	continue;

     num++;
     addr->next_addr = NULL;
     if (laddr)
	laddr->next_addr = addr;
     else
	entry->addr = addr;
     laddr = addr;
				 }

   if (num == 0) {
     if (entry->description)
	free(entry->description);
     free(entry);
		 }
   else {
   if (lentry)
	lentry->next = entry;
   else
	book->entries = entry;
   lentry = entry;

   if (num > 1)
	entry->type = ADDR_ALIAS;
   else
	entry->type = ADDR_SINGLE;
	}

   if (!strncmp(buf, "@ ", 2))
	goto newent;
				}

  fclose(afd);
  cur_entry = book->entries;
  return 0;
}

int
save_book()
{
FILE *afd;
char buf[255];
struct _abook_entry *entry;
struct _mail_addr *addr;

  if (book == NULL)
	return 0;

  if (book->name && strlen(book->name) && strcmp(book->name, "default"))
	sprintf(buf, "%s/.xfbook.%s", b_getcfg_str(conf_name, "addrdir", getenv("HOME")), book->name);
  else
	sprintf(buf, "%s/.xfbook", b_getcfg_str(conf_name, "addrdir", getenv("HOME")));

  if ((afd = fopen(buf, "w")) == NULL)  {
	display_msg(MSG_WARN, "save book", "Can not open\n%s", buf);
	return -1;
					}

  entry = book->entries;
  while (entry) {
   fprintf(afd, "@ %s\n", entry->description ? entry->description : "");
   addr = entry->addr;
   while (addr) {
    fprintf(afd, " %s\n", get_full_addr_line(addr));
    addr = addr->next_addr;
		}
   entry = entry->next;
		}

  if (fclose(afd) == EOF) {
    if (errno == ENOSPC)
	display_msg(MSG_WARN, "save book", "DISK FULL!");
    else
	display_msg(MSG_WARN, "save book", "Failed to write %s", buf);
    return -1;
			  }

  changed = 0;
  return 0;
}

struct _abook_entry *
get_entry_by_idx(index)
int index;
{
struct _abook_entry *entry;

   if (!book || (index <= 0))
	return NULL;

   entry = book->entries;
   while (entry && --index)
	entry = entry->next;

   return entry;
}

int
get_entry_idx(entry)
struct _abook_entry *entry;
{
int index;
struct _abook_entry *entry1;

  if (entry == NULL)
	return -1;

  entry1 = book->entries;
  index = 1;
  while (entry1 && (entry1 != entry)) 	{
	entry1 = entry1->next;
	index++;
					}

  return entry1 ? index : -1;
}

int
convert_book()
{
FILE *afd, *nafd;
char buf[255], buf1[255], buf2[255];

 sprintf(buf, "%s/.xfbook", b_getcfg_str(conf_name, "addrdir", getenv("HOME")));
 sprintf(buf1, "%s_tmp", buf);

 if ((afd = fopen(buf, "r")) == NULL) {
	display_msg(MSG_WARN, "address book", "Can not open address book %s", buf);
	return -1;		      }

 if ((nafd = fopen(buf1, "w")) == NULL) {
	display_msg(MSG_WARN, "address book", "Can not open address book %s", buf1);
	return -1;		      }


 while (fgets(buf2, 255, afd))
	fprintf(nafd, "@ \n %s", buf2);

 fclose(afd);
 fclose(nafd);

 if (rename(buf1, buf) == -1)   {
	display_msg(MSG_WARN, "convert", "failed to rename %s\nto%s", buf1,buf);
	return -1;
				}

 return 0;
}

void
display_books()
{
const FL_Dirlist *dir;
int i, num, books;
char *p;

 free_book();
 books = 1;
 dir = fl_get_dirlist((const char *)b_getcfg_str(conf_name, "addrdir", getenv("HOME")), ".xfbook.*", &num, 1);
 fl_freeze_form(addr_book_obj->Address_Book);
 fl_clear_browser(addr_book_obj->Abooks_Browse);
 fl_add_browser_line(addr_book_obj->Abooks_Browse, "default");
 if (dir == NULL)
	goto extbook;
 for (i = 0; i < num; i++) {
  if (books >= MAX_BOOKS)
	break;
  if (!dir[i].name || !strlen(dir[i].name) ||
	((p = strstr(dir[i].name, ".xfbook.")) != dir[i].name) ||
	(dir[i].type != FT_FILE))
	continue;

  p = dir[i].name + 8;
  if (strlen(p) >= MAX_ABOOK_NAMELEN)
	continue;
  while (*p != '\0') {
    if (!isalnum(*p))
	break;
    p++;
		     }

  if (*p != '\0')
	continue;

  p = dir[i].name + 8;
  fl_add_browser_line(addr_book_obj->Abooks_Browse, p);
  books++;
			   }
 fl_free_dirlist((FL_Dirlist *)dir);

extbook:
 fl_unfreeze_form(addr_book_obj->Address_Book);
 load_book(NULL);
 fl_select_browser_line(addr_book_obj->Abooks_Browse, 1);
 display_book();
}

void
display_book()
{
const char *p;
struct _abook_entry *entry;
char buf[255];

 if (!book)
	return;

 fl_freeze_form(addr_book_obj->Address_Book);
 fl_clear_browser(addr_book_obj->Address_Browse);
 entry = book->entries;
 while (entry) {
  if (entry->addr) {
  switch(entry->type) 	{
   case ADDR_SINGLE:
	fl_add_browser_line(addr_book_obj->Address_Browse, get_full_addr_line(entry->addr));
   break;

   case ADDR_ALIAS:
	sprintf(buf, "[%s]", entry->description ? entry->description : get_full_addr_line(entry->addr));
	fl_add_browser_line(addr_book_obj->Address_Browse, buf);
   break;
			}
		  }

  entry = entry->next;
	}
 fl_select_browser_line(addr_book_obj->Address_Browse, 1);
 p = fl_get_browser_line(addr_book_obj->Address_Browse, 1);
 if (p) {
   fl_set_input(addr_book_obj->AB_Input, p);
   fl_set_input_selected(addr_book_obj->AB_Input, 1);
	}
 else
   fl_set_input(addr_book_obj->AB_Input, "");
 fl_unfreeze_form(addr_book_obj->Address_Book);

 cur_entry = book->entries;
}

struct _mail_addr *
addr_book(addr_type)
int *addr_type;
{
FL_OBJECT *obj;

if (ready) {
	XRaiseWindow(fl_display, addr_book_obj->Address_Book->window);
	display_msg(MSG_WARN, "addrbook", "Address book is currently in use by another window, close it and try again");
	return NULL;
	}

if (!book) {
if (load_book(NULL) == -1)
	return NULL;
	   }

ready = 1; 
changed = 0;
addr_book_obj = create_form_Address_Book();

fl_set_browser_fontsize(addr_book_obj->Address_Browse, b_getcfg_int(conf_name, "AddrBookFSize", FL_MEDIUM_SIZE));
fl_set_browser_fontstyle(addr_book_obj->Address_Browse, b_getcfg_int(conf_name, "AddrBookFStyle", FL_NORMAL_STYLE));
fl_set_object_color(addr_book_obj->Address_Browse,b_getcfg_int(conf_name, "AddrBookBgCol", FL_TOP_BCOL), b_getcfg_int(conf_name, "AddrBookFgCol", FL_YELLOW));
fl_set_browser_dblclick_callback(addr_book_obj->Address_Browse, Address_Browse_Dbl_Call, 0);

fl_set_browser_fontsize(addr_book_obj->Abooks_Browse, b_getcfg_int(conf_name, "AddrBookFSize", FL_MEDIUM_SIZE));
fl_set_browser_fontstyle(addr_book_obj->Abooks_Browse, b_getcfg_int(conf_name, "AddrBookFStyle", FL_NORMAL_STYLE));
fl_set_object_color(addr_book_obj->Abooks_Browse,b_getcfg_int(conf_name, "AddrBookBgCol", FL_TOP_BCOL), b_getcfg_int(conf_name, "AddrBookFgCol", FL_YELLOW));

fl_set_object_lsize(addr_book_obj->AB_Input, b_getcfg_int(conf_name, "AddrBookFSize", FL_MEDIUM_SIZE));
fl_set_object_lstyle(addr_book_obj->AB_Input, b_getcfg_int(conf_name, "AddrBookFStyle", FL_NORMAL_STYLE));
fl_set_input_color(addr_book_obj->AB_Input, FL_BLACK, b_getcfg_int(conf_name, "AddrBookFgCol", FL_BLUE));
fl_set_object_color(addr_book_obj->AB_Input,b_getcfg_int(conf_name, "AddrBookFgCol", FL_BLACK), b_getcfg_int(conf_name, "AddrBookBgCol", FL_BLUE));

fl_set_input_return(addr_book_obj->AB_Input, FL_RETURN_ALWAYS);

fl_set_form_minsize(addr_book_obj->Address_Book, 500, 250);
fl_set_form_maxsize(addr_book_obj->Address_Book, 600, 600);

fl_show_form(addr_book_obj->Address_Book, FL_PLACE_FREE, FL_TRANSIENT, "Address Book");
display_books();

if (addr_type) {
type = *addr_type;
switch (*addr_type) {
  case TO_TYPE:
	fl_set_button(addr_book_obj->AB_To, 1);
  break;

  case CC_TYPE:
	fl_set_button(addr_book_obj->AB_Cc, 1);
  break;

  case BCC_TYPE:
	fl_set_button(addr_book_obj->AB_Bcc, 1);
  break;

  default:
	fl_set_button(addr_book_obj->AB_To, 1);
  break;
		}
	}
else
	fl_set_button(addr_book_obj->AB_To, 1);

while (ready == 1) {
 obj = fl_do_only_forms(); 
 if (obj == addr_book_obj->AB_Ok) {
	ready = 3;
	break;
				  }
 else
 if (obj == addr_book_obj->AB_Cancel) 	{
	ready = 2;
	break;
					}
		  }

if (changed) {
 if (display_msg(MSG_QUEST, "address book", "Do you want to save changes into the address book?"))
	save_book();
	}

fl_hide_form(addr_book_obj->Address_Book);
fl_free_form(addr_book_obj->Address_Book);
free(addr_book_obj);
addr_book_obj = NULL;

 if (addr_type)
	*addr_type = type;

 if (ready == 3) {
	ready = 0;
	return cur_entry ? cur_entry->addr : NULL;
	}

ready = 0;
return NULL;
}

void AB_Add_Call(FL_OBJECT *obj, long param)
{
char *addr = (char *)fl_get_input(addr_book_obj->AB_Input);
struct _mail_addr *ma;

if (strlen(addr) < 2)
  return;

if (!(ma = get_address(addr))) {
  display_msg(MSG_WARN, "add", "Invalid address:\n%s", addr);
  return; }

discard_address(ma->next_addr);
ma->next_addr = NULL;

if (add_book(ma, 1))
	changed = 1;
discard_address(ma);
fl_set_input(addr_book_obj->AB_Input, "");
display_book();
}

void AB_Delete_Call(FL_OBJECT *obj, long param)
{
int i;
struct _abook_entry *entry, *entry1;

 for (i =1; i <= fl_get_browser_maxline(addr_book_obj->Address_Browse); i++)
 {
  if (fl_isselected_browser_line(addr_book_obj->Address_Browse, i))
  {
   if ((entry = get_entry_by_idx(i)) == NULL)
	continue;

   if (i == 1)
	book->entries = entry->next;
   else {
    if ((entry1 = get_entry_by_idx(i - 1)) == NULL)
	continue;
    entry1->next = entry->next;
	}

   discard_address(entry->addr);
   if (entry->description)
	free(entry->description);
   free(entry);
   changed = 1;
   break;
  }
 }

 display_book();
 if (i <= fl_get_browser_maxline(addr_book_obj->Address_Browse)) {
 fl_select_browser_line(addr_book_obj->Address_Browse, i);
 fl_set_browser_topline(addr_book_obj->Address_Browse, i);	 }

 return;
}

void AB_To_Call(FL_OBJECT *obj, long param)
{
type = TO_TYPE;
}

void AB_Cc_Call(FL_OBJECT *obj, long param)
{
type = CC_TYPE;
}

void AB_Bcc_Call(FL_OBJECT *obj, long param)
{
type = BCC_TYPE;
}

void AB_Save_Call(FL_OBJECT *obj, long param)
{
 save_book();
}

void Address_Browse_Dbl_Call(obj, param)
FL_OBJECT *obj;
long param;
{
fl_trigger_object(addr_book_obj->AB_Ok);
}

void Address_Browse_Call(obj, param)
FL_OBJECT *obj;
long param;
{
int num, k;
struct _abook_entry *entry;
XEvent *xev;

   if (!book)
	return;

   k = num=fl_get_browser(obj);
   xev = (XEvent *)fl_last_event();

   if ((entry = get_entry_by_idx(num)) == NULL)
	return;

   cur_entry = entry;

   if (xev->xbutton.state & Button1Mask) {
    if (entry->type == ADDR_SINGLE)
	fl_set_input(addr_book_obj->AB_Input, get_full_addr_line(entry->addr));
    else
	fl_set_input(addr_book_obj->AB_Input, entry->description ? entry->description : get_full_addr_line(entry->addr));
    fl_set_input_selected(addr_book_obj->AB_Input, 1);
    fl_set_input_cursorpos(addr_book_obj->AB_Input, 0, 0);
					  }

   return;
}

void AB_Input_Call(FL_OBJECT *obj, long param)
{
int i, sline, eline;
const char *ln;
const char *inp = fl_get_input(addr_book_obj->AB_Input);

 if (book == NULL)
	return;

 if (strlen(inp) < 1)
	return;

 eline = fl_get_browser_maxline(addr_book_obj->Address_Browse);
 if (eline <= 1)
	return;

 if (param == 1) 
	sline = fl_get_browser(addr_book_obj->Address_Browse) + 1;
 else
	sline = 1;

srch: for (i = sline; i <= eline; i++)
	{
	ln = fl_get_browser_line(addr_book_obj->Address_Browse, i);
	if (strlen(ln) < strlen(inp))
		continue;

	if (strstr(ln, inp)) {
		fl_select_browser_line(addr_book_obj->Address_Browse, i);
		fl_set_browser_topline(addr_book_obj->Address_Browse, i);
		cur_entry = get_entry_by_idx(i);
		return;	     }
	}

 if (param == 1) {
	param = 0;
	eline = sline;
	sline = 1;
	goto srch;
		 }

 return;
}

void AB_Modify_Call(FL_OBJECT *obj, long param)
{
int i;
struct _mail_addr *ma;
struct _abook_entry *entry;
char *addr = (char *)fl_get_input(addr_book_obj->AB_Input);
char buf[255];

 if (strlen(addr) < 2) {
  display_msg(MSG_WARN, "modify", "Empty input, can not modify");
  return; }

 if (!(ma = get_address(addr))) {
  display_msg(MSG_WARN, "modify", "Invalid address\n\"%s\"", addr);
  return; }
 discard_address(ma->next_addr);
 ma->next_addr = NULL;

 for (i =1; i <= fl_get_browser_maxline(addr_book_obj->Address_Browse); i++)
 {
  if (fl_isselected_browser_line(addr_book_obj->Address_Browse, i))
  {
  if ((entry = get_entry_by_idx(i)) == NULL)
	break;

  if (entry->type == ADDR_SINGLE) {
	discard_address(entry->addr);
	entry->addr = ma;
				  }
  else  {
	discard_address(ma);
	if (entry->description)
		free(entry->description);
	entry->description = strdup(addr);
	}

  changed = 1;
  if (cur_entry->type == ADDR_ALIAS)
   sprintf(buf, "[%s]", cur_entry->description ? cur_entry->description : get_full_addr_line(cur_entry->addr));
  else
   sprintf(buf, "%s", get_full_addr_line(cur_entry->addr));
  fl_replace_browser_line(addr_book_obj->Address_Browse, get_entry_idx(cur_entry), buf);

  return;
  }
 }

 display_msg(MSG_WARN, "modify", "Nothing to modify");
 discard_address(ma);
 display_book();

 return;
}

void
Abook_Browse_Call(obj, param)
FL_OBJECT *obj;
long param;
{
int num;
char *p;

    num = fl_get_browser(obj);
    if (num <= 1) {
	load_book(NULL);
	display_book();
	return;
		   }

    p = (char *)fl_get_browser_line(obj, num);
    if (load_book(p) == -1) {
	display_msg(MSG_WARN, "address book", "Can not open address book %s",p);
	load_book(NULL);
		   	    }

    display_book();
    return;
}

void
AB_Memb_Call(obj, param)
FL_OBJECT *obj;
long param;
{
 fl_deactivate_form(addr_book_obj->Address_Book);
 edit_entry();
 fl_activate_form(addr_book_obj->Address_Book);
}

void
AB_Refresh_Book_Call(obj, param)
FL_OBJECT *obj;
long param;
{
 display_books();
}

void
AB_New_Book_Call(obj, param)
FL_OBJECT *obj;
long param;
{
char *p, *p1, buf[255];
FILE *afd;

   p = (char *)fl_show_input("Enter address book name:", "");
   if (!p || (strlen(p) < 1))
	return;

   if (strlen(p) >= MAX_ABOOK_NAMELEN) {
	display_msg(MSG_WARN, "address book", "address book name can not be longer\nthen %d characters", MAX_ABOOK_NAMELEN);
	return;
					}

   for (p1 = p; *p1 != '\0'; p1++) {
    if (!isalnum(*p1))  {
	display_msg(MSG_WARN, "address book", "invalid character in address book name");
	return;
			}
				  }

   sprintf(buf, "%s/.xfbook.%s", b_getcfg_str(conf_name, "addrdir", getenv("HOME")), p);


   if ((afd = fopen(buf, "w")) == NULL) {
	display_msg(MSG_WARN, "address book", "Can not open\n%s", buf);
	return;
					}

   fclose(afd);

   fl_add_browser_line(addr_book_obj->Abooks_Browse, p);
   return;
}

void
AB_Del_Book_Call(obj, param)
FL_OBJECT *obj;
long param;
{
int num;
char *p, buf[255];

    num = fl_get_browser(addr_book_obj->Abooks_Browse);
    if (num <= 1) {
	display_msg(MSG_WARN, "address book", "You can not delete default address book");
	return;
		  }

    p = (char *)fl_get_browser_line(addr_book_obj->Abooks_Browse, num);
    if (!p || !strlen(p))
	return;

    if (!display_msg(MSG_QUEST, NULL, "Address book %s will be deleted\nAre you sure?", p))
	return;

    sprintf(buf, "%s/.xfbook.%s", b_getcfg_str(conf_name, "addrdir", getenv("HOME")), p);
    if (unlink(buf) == -1)
	display_msg(MSG_WARN, "address book", "Failed to delete address book %s", p);
    else {
	fl_delete_browser_line(addr_book_obj->Abooks_Browse, num);
	load_book(NULL);
 	fl_select_browser_line(addr_book_obj->Abooks_Browse, 1);
	display_book();
	}

    return;
}

struct _mail_addr *
mod_addr(oldadr, type)
char *oldadr;
int *type;
{
static struct _mail_addr *newadr = NULL;
char *p;

  if (newadr) {
	discard_address(newadr);
	newadr = NULL;
	}

  if (modready) {
	XRaiseWindow(fl_display, amod_obj->addrmod->window);
	display_msg(MSG_WARN, "modify address", "This dialog is currently in use by another window, close it and try again");
	return NULL;
	}
  modready = 1;
  amod_obj = create_form_addrmod();
  fl_set_input(amod_obj->AMOD_Input, oldadr ? oldadr : "");
  if (type) {
   switch(*type) {
    case TO_TYPE:
	fl_set_button(amod_obj->AMOD_To, 1);
    break;

    case CC_TYPE:
	fl_set_button(amod_obj->AMOD_Cc, 1);
    break;

    case BCC_TYPE:
	fl_set_button(amod_obj->AMOD_Bcc, 1);
    break;

    default:
	fl_set_button(amod_obj->AMOD_To, 1);
    break;
		 }
	    }
  else
	fl_set_button(amod_obj->AMOD_To, 1);

  fl_show_form(amod_obj->addrmod, FL_PLACE_CENTER, FL_TRANSIENT,"Modify Address");
adrwait:
  fl_do_only_forms(); 
  p = (char *)fl_get_input(amod_obj->AMOD_Input);
  if (p && strlen(p)) 	{
   if ((newadr = get_address(p)) == NULL)  {
	display_msg(MSG_WARN, "modify", "Invalid address, please reenter");
	goto adrwait;
					   }
			}

  if (type) {
   if (fl_get_button(amod_obj->AMOD_To))
	*type = TO_TYPE;
   else
   if (fl_get_button(amod_obj->AMOD_Cc))
	*type = CC_TYPE;
   else
   if (fl_get_button(amod_obj->AMOD_Bcc))
	*type = BCC_TYPE;
   else
	*type = TO_TYPE;
	    }

  fl_hide_form(amod_obj->addrmod);
  fl_free_form(amod_obj->addrmod);
  free(amod_obj);
  amod_obj = NULL;
  modready = 0;

  return newadr;
}

void AMOD_Call(FL_OBJECT *obj, long param)
{
}

void
display_entry()
{
struct _mail_addr *addr;

if (!cur_entry || !cur_entry->addr)
	return;
fl_freeze_form(alias_obj->aliased);
fl_set_input(alias_obj->AL_Input, get_full_addr_line(cur_entry->addr));
fl_clear_browser(alias_obj->AL_Browse);
addr = cur_entry->addr;
while(addr) {
 fl_add_browser_line(alias_obj->AL_Browse, get_full_addr_line(addr));
 addr = addr->next_addr;
	}
fl_select_browser_line(alias_obj->AL_Browse, 1);
fl_unfreeze_form(alias_obj->aliased);
}

void
edit_entry()
{
  if (!cur_entry || !cur_entry->addr)
	return;

  if (already) {
	XRaiseWindow(fl_display, alias_obj->aliased->window);
	display_msg(MSG_WARN, "alias", "This dialog is currently in use by another window, close it and try again");
	return;
	}

  already = 1;
  alias_obj = create_form_aliased();

  fl_set_browser_fontsize(alias_obj->AL_Browse, b_getcfg_int(conf_name, "AddrBookFSize", FL_MEDIUM_SIZE));
  fl_set_browser_fontstyle(alias_obj->AL_Browse, b_getcfg_int(conf_name, "AddrBookFStyle", FL_NORMAL_STYLE));
  fl_set_object_color(alias_obj->AL_Browse,b_getcfg_int(conf_name, "AddrBookBgCol", FL_TOP_BCOL), b_getcfg_int(conf_name, "AddrBookFgCol", FL_YELLOW));

  fl_set_object_lsize(alias_obj->AL_Input, b_getcfg_int(conf_name, "AddrBookFSize", FL_MEDIUM_SIZE));
  fl_set_object_lstyle(alias_obj->AL_Input, b_getcfg_int(conf_name, "AddrBookFStyle", FL_NORMAL_STYLE));
  fl_set_input_color(alias_obj->AL_Input, FL_BLACK, b_getcfg_int(conf_name, "AddrBookFgCol", FL_BLUE));
  fl_set_object_color(alias_obj->AL_Input,b_getcfg_int(conf_name, "AddrBookFgCol", FL_BLACK), b_getcfg_int(conf_name, "AddrBookBgCol", FL_BLUE));

  display_entry();
  fl_show_form(alias_obj->aliased, FL_PLACE_CENTER, FL_TRANSIENT,"Group Members");
  fl_do_only_forms();

  fl_hide_form(alias_obj->aliased);
  fl_free_form(alias_obj->aliased);
  free(alias_obj);
  alias_obj = NULL;
  already = 0;

  return;
}

void AL_Add_Call(FL_OBJECT *obj, long param)
{
char *addr = (char *)fl_get_input(alias_obj->AL_Input);
char buf[255];
struct _mail_addr *ma, *ma1;

if (strlen(addr) < 2)
  return;

if (!(ma = get_address(addr))) {
  display_msg(MSG_WARN, "add", "Invalid address:\n%s", addr);
  return; }
discard_address(ma->next_addr);
ma->next_addr = NULL;

cur_entry->type = cur_entry->addr ? ADDR_ALIAS : ADDR_SINGLE;

if (cur_entry->addr)    {
  ma1 = cur_entry->addr;
  while (ma1->next_addr)
	ma1 = ma1->next_addr;

  ma1->next_addr = ma;
			}
else 
  cur_entry->addr = ma;

changed = 1;
display_entry();
if (cur_entry->type == ADDR_ALIAS)
   sprintf(buf, "[%s]", cur_entry->description ? cur_entry->description : get_full_addr_line(cur_entry->addr));
else
   sprintf(buf, "%s", get_full_addr_line(cur_entry->addr));
fl_replace_browser_line(addr_book_obj->Address_Browse, get_entry_idx(cur_entry), buf);

return;
}

void AL_Delete_Call(FL_OBJECT *obj, long param)
{
struct _mail_addr *ma, *ma1;
char buf[255];
int num, i;

  num = fl_get_browser(alias_obj->AL_Browse);
  if (num < 1)
	return;
  if (num == 1) {
	display_msg(MSG_WARN, "delete alias", "Use address book main window\nto delete this address");
	return;
		}

  i = num - 1;
  ma = cur_entry->addr;
  while (ma && --i)
	ma = ma->next_addr;

  if ((ma == NULL) ||
	(ma->next_addr == NULL)) {
	display_msg(MSG_WARN, "delete alias", "delete failed");
	return;
		  		 }

  ma1 = ma->next_addr;
  ma->next_addr = ma1->next_addr;
  ma1->next_addr = NULL;
  discard_address(ma1);
  changed = 1;
  display_entry();

  if (!cur_entry->addr || !cur_entry->addr->next_addr)  {
	cur_entry->type = ADDR_SINGLE;
	sprintf(buf, "%s", get_full_addr_line(cur_entry->addr));
							}
  else  {
	cur_entry->type = ADDR_ALIAS;
	sprintf(buf, "[%s]", cur_entry->description ? cur_entry->description : get_full_addr_line(cur_entry->addr));
	}

  fl_replace_browser_line(addr_book_obj->Address_Browse, get_entry_idx(cur_entry), buf);
  return;
}

void AL_Call(FL_OBJECT *obj, long param)
{
}

void AL_Browse_Call(FL_OBJECT *obj, long param)
{
 fl_set_input(alias_obj->AL_Input, fl_get_browser_line(obj, fl_get_browser(obj)));
}

