/*
 * Copyright(c) 1995,1996 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: cfgfile.c,v 1.33 1996/06/12 15:42:47 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <fmail.h>

#define SECFILEMODE 00600

/*
 * This function puts string into configuration file
 */

struct _xfdef xfdefs = { -1 , -1 , -1 , -1 , -1 };

typedef struct __tagcfl {
char 	str[SLEN];
struct __tagcfl *next;
u_int	flags;
} __cfl;

static FILE *cfile = NULL;

static __cfl *__first_cfl=(__cfl *)NULL;
static char __fname[PLEN]={"\0"};
static int changed = 0;

void
cfg_lock(fn)
char *fn;
{
if (cfile != NULL) {
#if defined(SOLARIS) || defined(HP) || defined(_AIX)
	lockf(fileno(cfile), F_ULOCK, 0);
#else
	flock(fileno(cfile), LOCK_UN);
#endif
	fclose(cfile);
		}

#if defined(SOLARIS) || defined(HP) || defined(_AIX)
if ((cfile=fopen(fn,"r+"))==NULL) {
#else
if ((cfile=fopen(fn,"r"))==NULL) {
#endif
	display_msg(MSG_WARN, "Can not open", "configuration file %s", fn);
	return; }

#if defined(SOLARIS) || defined(HP) || defined(_AIX)
 if (lockf(fileno(cfile), F_TLOCK, 0) == -1) {
#else
 if (flock(fileno(cfile), LOCK_EX | LOCK_NB) == -1) {
#endif
	fprintf(stderr, "Can not lock %s\nProbably xfmail is already running\n", fn);
	exit(1);
	}
}

void
cfg_unlock()
{
 if (cfile != NULL) {
#if defined(SOLARIS) || defined(HP) || defined(_AIX)
	lockf(fileno(cfile), F_ULOCK, 0);
#else
	flock(fileno(cfile), LOCK_UN);
#endif
	fclose(cfile);
	}

 cfile = NULL;
}

int
save_cfg_file(file, nopwd)
char *file;
int nopwd;
{
FILE *cfd;
char buf[SLEN];
__cfl *tmpptr;

  if (!file)
	return -1;

  if ((cfd = fopen(file, "w")) == NULL) {
	display_msg(MSG_WARN, "save config", "Can not open %s", file);
	return -1;
					}

  tmpptr=__first_cfl;

  while(tmpptr)
  {
    if (nopwd && (tmpptr->flags & CF_UNSECURE)) {
	tmpptr=tmpptr->next;
	continue;
				     		}

    sprintf(buf,"%s\n",tmpptr->str);
    if (fputs(buf, cfd) == EOF) {
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "save config", "DISK FULL!");
	else
		display_msg(MSG_WARN, "save config", "Failed to write %s", file);
	return -1;
			        }

    tmpptr=tmpptr->next;
  }

  fclose(cfd);
  return 0;
}

/*
 * This one flushes current config 
 */

int b_savecfg(ask)
int ask;
{
FILE *f;
char buf[SLEN];
__cfl *tmpptr;

if (ask && changed) 	{
	if (!display_msg(MSG_QUEST, "Configuration has been changed", "Do you want to save it?"))
		return 0;
			}

if (__fname[0]=='\0') 
	return 1;

cfg_unlock();

if ((f=fopen(__fname,"w"))==NULL) {
	display_msg(MSG_WARN, "Can not write", "to configuration file %s", __fname);
	return -1; }

tmpptr=__first_cfl;

while(tmpptr)
  {
    if (tmpptr->flags & DONT_STORE) {
	tmpptr=tmpptr->next;
	continue; }

    sprintf(buf,"%s\n",tmpptr->str);
    if (fputs(buf,f) == EOF) {
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "save config", "DISK FULL!");
	else
		display_msg(MSG_WARN, "save config", "Failed to write %s", __fname);
	return -1;
			     }
    tmpptr=tmpptr->next;
  }

if (fclose(f) == EOF) 	{
	if (errno == ENOSPC)
		display_msg(MSG_WARN, "save config", "DISK FULL!");
	else
		display_msg(MSG_WARN, "save config", "Failed to write %s", __fname);
	return -1;
			}

changed = 0;
chmod(__fname,SECFILEMODE);
cfg_lock(__fname);
return 1;
}

/*
 * This function preloads configuration file to be
 * changed in memory
 * 0 - file does not exist.
 * 1 - ok
 * -1 , -2 and so on - errors
 */

int b_loadcfg(fn)
char *fn;
{
FILE *f;
char buf[SLEN];
int i,buflen;
__cfl *tmpptr;
__cfl *tmpptr1;

 tmpptr = __first_cfl;

 while(tmpptr) {
	tmpptr1 = tmpptr->next;
	free(tmpptr);
	tmpptr = tmpptr1;
		}

 __first_cfl=NULL;

cfg_unlock();

 if ((f=fopen(fn,"r"))==NULL)
   {
    strcpy(__fname,fn);
    return 0;
   }

while (fgets(buf,SLEN,f)!=NULL)
 {
  if (buf[0] == '#')
	continue;

  buflen=strlen(buf);
  for (i = buflen-1; i >= 0; i--)  {
	if (buf[i]=='\n') { 
	   buf[i]='\0';
	   break; }
				   }

  if (!(tmpptr=(__cfl *)malloc(sizeof(__cfl)))) {
	display_msg(MSG_FATAL, "cfg", "malloc failed");
	return -3; }

  strcpy(tmpptr->str,buf);

  tmpptr->next = __first_cfl;
  tmpptr->flags = 0;
  __first_cfl = tmpptr;

 }

strcpy(__fname,fn);
fclose(f);
cfg_lock(__fname);

return 1;
}

void
cfg_check_version()
{
__cfl *tmpptr;
__cfl *tmpptr1;

 if (__first_cfl && strcmp(VERSION, b_getcfg_str(conf_name, "xfversion", ""))) {
  if (!display_msg(MSG_QUEST, "Configuration file belongs", "to different version of XF-Mail, use it anyway?"))
	{
	tmpptr = __first_cfl;

	while(tmpptr) {
		tmpptr1 = tmpptr->next;
		free(tmpptr);
		tmpptr = tmpptr1;
		}

	__first_cfl=NULL;
	}

  b_putcfg_str(conf_name, "xfversion", VERSION);
							}

 return;
}

void
b_setcfg_flags(ln,flags)
char *ln;
int flags;
{
char tmpln[SLEN];
int tmplnlen;
__cfl *tmpptr, *newptr;

tmpptr=__first_cfl;
sprintf(tmpln,"%s=",ln);
tmplnlen=strlen(tmpln);

while(tmpptr!=NULL)
 {
   if (!strncmp(tmpln,tmpptr->str,tmplnlen)) 
     {
	tmpptr->flags = flags;
	return;
     }
tmpptr=tmpptr->next;
 }

if ((newptr=(__cfl *)malloc(sizeof(__cfl)))==NULL)  {
	display_msg(MSG_FATAL, "config", "malloc  failed!");
	return; }

  sprintf(tmpln,"%s=",ln);

  strcpy(newptr->str,tmpln);
  newptr->next=__first_cfl;

  newptr->flags = flags;
  if (!(flags & DONT_STORE) &&
	!(flags & CF_NOTCHANGED))
		changed = 1;

  __first_cfl=newptr;

return;
}

/*
 * This function puts string into configuration file
 */

int
b_delcfg_str(fn, ln)
char *fn;
char *ln;
{
char tmpln[SLEN];
int tmplnlen;
__cfl *tmpptr,*prevptr;

sprintf(tmpln,"%s=",ln);
tmplnlen=strlen(tmpln);

prevptr = tmpptr=__first_cfl;

while(tmpptr!=NULL)
 {
   if (!strncmp(tmpln,tmpptr->str,tmplnlen)) 
     {
     if (tmpptr == __first_cfl)
	__first_cfl = tmpptr->next;
     else
	prevptr->next = tmpptr->next;

     free(tmpptr);

     changed = 1;
     return 0;
     }

prevptr = tmpptr;
tmpptr=tmpptr->next;
 }

return -1;
}

int b_putcfg_str(fn,ln,b)
char *fn;
char *ln;
char *b;
{
char tmpln[SLEN];
int tmplnlen;
__cfl *tmpptr,*newptr;
int found=FALSE;

sprintf(tmpln,"%s=",ln);
tmplnlen=strlen(tmpln);

tmpptr=__first_cfl;

while(tmpptr!=NULL)
 {
   if (!strncmp(tmpln,tmpptr->str,tmplnlen)) 
     {
      char buf[255];
      sprintf(buf,"%s%s",tmpln,b);
      if (strcmp(buf, tmpptr->str)) {
	if (!(tmpptr->flags & DONT_STORE) &&
		!(tmpptr->flags & CF_NOTCHANGED))
		changed = 1;

        strcpy(tmpptr->str, buf); }

      found=TRUE;
      break;
     }
tmpptr=tmpptr->next;
 }

if (!found)
 {
  if ((newptr=(__cfl *)malloc(sizeof(__cfl)))==NULL)
	return -3;
  newptr->flags = 0;
  sprintf(newptr->str,"%s%s",tmpln,b);
  newptr->next=__first_cfl;
  __first_cfl=newptr;
  changed = 1;
 }
return 0;
}

int
cfg_exist(ln)
char *ln;
{
char tmpln[SLEN];
int tmplnlen;
__cfl *tmpptr;

sprintf(tmpln,"%s=",ln);
tmplnlen=strlen(tmpln);

tmpptr=__first_cfl;

while(tmpptr!=NULL)
 {
   if (!strncmp(tmpln,tmpptr->str,tmplnlen)) 
	return 1;

   tmpptr=tmpptr->next;
 }

return 0;
}

/*
 * This function gets configuration text string from
 * named file.
 * 
 */

char *
b_getcfg_str(fn,ln, def)
char *fn;
char *ln;
char *def;
{
static char line[255];
char tmpln[SLEN];
int tmplnlen;
__cfl *tmpptr;
int found=FALSE;

line[0] = '\0';

if (strcmp(__fname,fn))
 {
   if (b_savecfg(0)<0) return NULL; 
   if (b_loadcfg(fn)<0) return NULL;
 }

sprintf(tmpln,"%s=",ln);
tmplnlen=strlen(tmpln);

tmpptr=__first_cfl;

while(tmpptr!=NULL)
 {
   if (!strncmp(tmpln,tmpptr->str,tmplnlen)) 
     {
      strncpy(line,&(tmpptr->str[tmplnlen]),255);
      found=TRUE;
      break;
     }
tmpptr=tmpptr->next;
 }
if (!found) {
	b_putcfg_str(fn, ln, def);
	strncpy(line, def, 255); }

return line;
}


int b_putcfg_int(fn,ln,i)
char *fn; 
char *ln;
int i;
{
char buf[10];
sprintf(buf,"%d",i);
return (b_putcfg_str(fn,ln,buf));
}

int b_getcfg_int(fn,ln,def)
char *fn; 
char *ln;
int def;
{
char *buf;
char ibuf[10];
sprintf(ibuf,"%d",def);
if (!(buf=b_getcfg_str(fn,ln,ibuf)))
	return def;

return atoi(buf);
}

void
read_xfdefaults()
{
FL_IOPT cntl;
u_long mask;
char xfdef[255], *p;
FILE *xf;

  sprintf(xfdef, "%s/.xfdefaults", getenv("HOME"));
  if ((xf = fopen(xfdef, "r")) == NULL)
	return;

  mask = 0L;

  while (fgets(xfdef, 255, xf)) {
   if (xfdef[0] == '#')
	continue;

   if ((p = strchr(xfdef, '=')) == NULL)
	continue;

   *p = '\0';
   p++;
   if (!strcmp(xfdef, "Depth")) 	{
	mask |= FL_PDDepth;
	sscanf(p, "%d", &cntl.depth);
				}	
   else
   if (!strcmp(xfdef, "VisualClass")) 	{
	mask |= FL_PDClass;
	sscanf(p, "%d", &cntl.vclass);
				}	
   else
   if (!strcmp(xfdef, "DoubleBuffer")) 	{
	mask |= FL_PDDouble;
	sscanf(p, "%d", &cntl.doubleBuffer);
				}	
   else
   if (!strcmp(xfdef, "ButtonFontSize")) 	{
	mask |= FL_PDButtonFontSize;
	sscanf(p, "%d", &cntl.buttonFontSize);
				}	
   else
   if (!strcmp(xfdef, "SliderFontSize")) 	{
	mask |= FL_PDSliderFontSize;
	sscanf(p, "%d", &cntl.sliderFontSize);
				}	
   else
   if (!strcmp(xfdef, "MenuFontSize")) 	{
	mask |= FL_PDMenuFontSize;
	sscanf(p, "%d", &cntl.menuFontSize);
				}	
   else
   if (!strcmp(xfdef, "ChoiceFontSize")) 	{
	mask |= FL_PDChoiceFontSize;
	sscanf(p, "%d", &cntl.choiceFontSize);
				}	
   else
   if (!strcmp(xfdef, "BrowserFontSize")) 	{
	mask |= FL_PDBrowserFontSize;
	sscanf(p, "%d", &cntl.browserFontSize);
				}	
   else
   if (!strcmp(xfdef, "InputFontSize")) {
	mask |= FL_PDInputFontSize;
	sscanf(p, "%d", &cntl.inputFontSize);
				}	
   else
   if (!strcmp(xfdef, "LabelFontSize")) 	{
	mask |= FL_PDLabelFontSize;
	sscanf(p, "%d", &cntl.labelFontSize);
				}	
   else
   if (!strcmp(xfdef, "PopUpFontSize")) 	{
/*
	mask |= FL_PDPupFontSize;
*/
	sscanf(p, "%d", &cntl.pupFontSize);
	xfdefs.popfsize = cntl.pupFontSize;
				}	
   else
   if (!strcmp(xfdef, "PopUpFontStyle")) 	{
	sscanf(p, "%d", &cntl.pupFontStyle);
	xfdefs.popfstyle = cntl.pupFontStyle;
				}	
   else
   if (!strcmp(xfdef, "PopUpCursor"))
	sscanf(p, "%d", &xfdefs.popcursor);
   else
   if (!strcmp(xfdef, "PopUpBgCol"))
	sscanf(p, "%d", &xfdefs.popbgcol);
   else
   if (!strcmp(xfdef, "PopUpFgCol"))
	sscanf(p, "%d", &xfdefs.popfgcol);
   else
   if (!strcmp(xfdef, "PrivateMap")) 	{
	mask |= FL_PDPrivateMap;
	sscanf(p, "%d", &cntl.privateColormap);
				}	
   else
   if (!strcmp(xfdef, "SharedMap")) 	{
	mask |= FL_PDSharedMap;
	sscanf(p, "%d", &cntl.sharedColormap);
				}	
   else
   if (!strcmp(xfdef, "StandardMap")) 	{
	mask |= FL_PDStandardMap;
	sscanf(p, "%d", &cntl.standardColormap);
				}	
   else
   if (!strcmp(xfdef, "BorderWidth")) 	{
	mask |= FL_PDBorderWidth;
	sscanf(p, "%d", &cntl.borderWidth);
				}	
   else
   if (!strcmp(xfdef, "LeftScroll")) 	{
	mask |= FL_PDLeftScrollBar;
	sscanf(p, "%d", &cntl.leftScrollBar);
				}	

				}

  if (mask != 0L)
	fl_set_defaults(mask, &cntl);

  return;
}

