/*
 * Copyright(c) 1995 by Gennady B. Sorokopud (gena@NetVision.net.il)
 *                  and Ugen J. S. Antsilevich (ugen@NetVision.net.il)
 *
 * This software can be freely redistributed and modified for 
 * non-commercial purposes as long as above copyright
 * message and this permission notice appear in all
 * copies of distributed source code and included as separate file
 * in binary distribution.
 *
 * Any commercial use of this software requires author's permission.
 *
 * This software is provided "as is" without expressed or implied
 * warranty of any kind.
 * Under no circumstances is the author responsible for the proper
 * functioning of this software, nor does the author assume any
 * responsibility for damages incurred with its use.
 *
 */

/* $Id: xfmail.c,v 1.202 1996/06/30 13:49:34 gena Exp $
 */

#include <fmail.h>
#include <umail.h>
#include <X11/xpm.h>
#include <choose_folder.h>
#define	MAIN_XPM
#define	ATT_XPM
#include "pixmaps.h"

FL_FORM		*main_form = NULL;
FL_OBJECT	*b_folders,*b_messages;
FL_OBJECT	*inc_timer, *send_timer, *trash_timer, *abrt;
static FL_OBJECT	*menu_file,*menu_message,*menu_folders,*menu_send,*status;
FL_OBJECT	*menu_misc;

#define	MESG_MENU     "View (Alt+V)|Edit (Alt+E)|Delete (Alt+D)|Move to Folder (Alt+M)|Copy to Folder|Mark|Unmark|Mark Read/Unread|Print (Alt+P)|Bounce|Apply Rules|Select All|Delete from server"
#define SEND_MENU     "New (Alt+C)%l|Reply to Sender|Reply to All (Alt+R)|Forward (Alt+F)%l|Send All"
#define FOLD_MENU     "Create|Delete|Rename|Go To|Empty Trash Folder|Catch Up Unread|Search|Rescan All%l|Import|Export"

int mpup, spup;

void cb_menu_messages(FL_OBJECT *obj, long arg);
void cb_messages(FL_OBJECT *obj, long arg);
void cb_menu_folders(FL_OBJECT *obj, long arg);
void cb_menu_send(FL_OBJECT *obj, long arg);

char conf_name[255];
char sender_name[127];
struct _mail_folder *current_folder = NULL;
static struct _mail_msg *current_message = NULL;

static int current_index = 0;
static int init = 0;
static int abortbutton = 0;
static int xofft = 0, yofft = 0;
u_int busy = 0;
u_int supress_errors = 0;

Pixmap iconPix_noMail, iconPix, icon_MailEdit;

int show_msg[5] = { 2, 13, 5, 18, 100 };
int show_fld[3] = { 11, 3, 3 };

void clean_exit();

void
my_deactivate()
{
 FL_OBJECT *obj = main_form->first;

 abortbutton = 0;
 fl_deactivate_all_forms();

 fl_activate_form(main_form);

 while (obj) 	{
	if ((obj == abrt) || (obj == status)) {
	  if (obj == abrt)
		fl_activate_object(obj);
		obj = obj->next;
		continue;
				}

	fl_deactivate_object(obj);
	obj = obj->next;
		}
}

void
my_activate()
{
 FL_OBJECT *obj = main_form->first;

 fl_activate_all_forms();
 abortbutton = 0;

 while (obj) 	{
	if ((obj == abrt) || (obj == status)) {
		fl_deactivate_object(obj);
		obj = obj->next;
		continue;
				}

	fl_activate_object(obj);
	obj = obj->next;
		}
}

void
set_main_icon(hm, name)
	int	hm;
	char 	*name;
{
XWMHints *xwmh;

if ((xwmh=XAllocWMHints()) != NULL) {
    if (hm)
	xwmh->icon_pixmap = iconPix;
    else
	xwmh->icon_pixmap = iconPix_noMail;

	xwmh->flags = IconPixmapHint;
	XSetWMHints(fl_display, main_form->window, xwmh);
	XFree((char *)xwmh);
}

if ((name != NULL) && (strlen(name) > 1)) {
	fl_set_form_title(main_form, name);
	XSetIconName(fl_display, main_form->window, name);
	  }

}

void
set_cursor(form, cursor)
FL_FORM *form;
int cursor;
{

 fl_set_cursor(form->window, cursor ? cursor : XC_top_left_arrow);

 return;
}

#if __STDC__
int display_msg(int flags, const char *title, const char *fmt, ...)
#else
int display_msg(flags, title, fmt, va_list)
int flags;
char *title;
char *fmt;
va_dcl
#endif
{
va_list ap;
char buf[255];
char buf1[255];
char errstr[64];

 if (!main_form)
	return 0;

#if __STDC__
 va_start(ap, fmt);
#else
 va_start(ap);
#endif

 vsprintf(buf, fmt, ap);

#ifndef _AIX
 if ((errno > 0) && (errno < sys_nerr)) {
	strncpy(errstr, sys_errlist[errno], 63);
	errno = 0; }
#else
if (errno > 0) {
        sprintf(errstr, "System Error %d", errno);
        errno = 0; }
#endif
 else
	strcpy(errstr, "ERROR");

 switch (flags) {

	case MSG_FATAL:
	  if (supress_errors)
		return 0;

	  if (title)
		fl_show_alert(errstr, title, buf, 1);
	  else
		fl_show_alert(errstr, buf, "", 1);
	  exit(1);	
	break;

	case MSG_QUEST:
	  if (title)
		return fl_show_question(title, buf, "");
	  else
		return fl_show_question(buf, "", "");
	break;

	case MSG_WARN:
	  if (title)
		sprintf(buf1, "WARN: %s: %s", title, buf);
	  else
		sprintf(buf1, "WARN: %s", buf);

	  log_str(buf1);

	  if (supress_errors)
		return 0;

	  if (title)
		fl_show_alert(errstr, title, buf, 1);
	  else
		fl_show_alert(errstr, buf, "", 1);
	break;

	case MSG_MSG:
	  if (title)
		fl_show_message("MESSAGE", title, buf);
	  else
		fl_show_message("MESSAGE", buf, "");
	break;

	case MSG_STAT:
	  buf[200] = '\0';
	  fl_set_input(status, buf);
	break;

	case MSG_LOG:
	 if (title)
		sprintf(buf1, "%s: %s", title, buf);
	 else
		sprintf(buf1, "%s", buf);

	 log_str(buf1);
	break;

	default:
		return -1;
	break;
	}

/*
	  fl_check_only_forms();
*/
	  XFlush(fl_display);

	return 0;
}

int
msg_posthandler(FL_OBJECT *ob, int event,
                          FL_Coord mx, FL_Coord my, int key, void *raw_ev)
{
int		 i, num, x, y;
struct _mail_msg *msg;
XEvent 		 *xev;

 xev = (XEvent *)raw_ev;

 switch (event) {
  case FL_PUSH:
	num = abs(fl_get_browser(b_messages)) - 1;

	if (xev->xbutton.state & ControlMask) {
  		fl_select_browser_line(b_messages, current_index + 1);
		break; }

	if (xev->xbutton.state & ShiftMask)
			{
		if (current_index > num) {
			x = num + 1;
			y = current_index + 1;
					 }
		else    {
			x = current_index + 1;
			y =  num + 1;
			}
		fl_freeze_form(main_form);
		for (i = x; i <= y; i++)
			fl_select_browser_line(b_messages, i);
  		fl_select_browser_line(b_messages, current_index + 1);
		fl_unfreeze_form(main_form);
		break;
			}

	msg = current_message;
	current_index = num;
	current_message = get_msg_by_index(current_folder, num);
	fl_freeze_form(main_form);
	redraw_msg(msg);
	redraw_msg(current_message);
	fl_deselect_browser(b_messages);
  	fl_select_browser_line(b_messages, current_index + 1);
	fl_unfreeze_form(main_form);

	if (current_message)
		display_msg(MSG_STAT, NULL, "From: %.64s Subject: %.114s", get_full_addr_line(current_message->header->From), current_message->header->Subject ? current_message->header->Subject : "* No Subject *");

	if (key == 3) 	{
		fl_freeze_form(main_form);
		fl_dopup(mpup); 
		fl_unfreeze_form(main_form);
			}

  break;
		}

  return 0;
}

int
preemptive_handler(FL_OBJECT *ob, int event,
                          FL_Coord mx, FL_Coord my, int key, void *raw_ev)
{
static int	l = 0;
static int	dragndrop = 0;
FL_OBJECT       *obj;
int		h;
XEvent *xev;

 xev = (XEvent *)raw_ev;

 switch (event) {
  case FL_PUSH:
   if (key == 2) {
	dragndrop = 1;
	set_cursor(main_form, XC_cross);
	return FL_PREEMPT;
		 }
  break;

  case FL_MOUSE:
   obj = fl_find_last(main_form, FL_FIND_MOUSE, mx, my);

   switch (dragndrop)  {
     case 0:
	if (obj != b_folders)
		return !FL_PREEMPT;
	dragndrop = 2;
	fl_select_browser_line(b_messages, current_index + 1);
	set_cursor(main_form, XC_hand2);
	return FL_PREEMPT;
     break;

     case 1:
	if (obj == b_folders) {
	  dragndrop = 2;
	  set_cursor(main_form, XC_hand2); }

	return FL_PREEMPT;
     break;

     case 2:
	if (obj != b_folders)  {
	  dragndrop = 1;
	  set_cursor(main_form, XC_cross);  }
	else	{
	  h = b_folders->h/fl_get_browser_screenlines(b_folders);
	  l = (my - b_folders->y) / h + 1;
	  fl_select_browser_line(b_folders, l + fl_get_browser_topline(b_folders) - 1);
		}

	return FL_PREEMPT;
     break;
			}
  break;

  case FL_RELEASE:
	switch(dragndrop) {
	  case 0:
		fl_select_browser_line(b_messages, current_index + 1);
		return !FL_PREEMPT;
	  break;

	  case 1:
		set_cursor(main_form, 0); 
		ringbell();
	  break;

	  case 2:
		set_cursor(main_form, 0); 
		if (!l)
			ringbell();
		else {
			l = l + fl_get_browser_topline(b_folders) - 2;
			if (l >= folders_num)
				ringbell();
			else
				move_messages(get_folder_by_index(l));
		     }
	  break;
			  }
	dragndrop = 0;
	l = 0;
  break;
		}

  return !FL_PREEMPT;
}

char *
get_fld_line(fld)
struct _mail_folder *fld;
{
int color;
static char fldline[64];

 if (!fld)
	return NULL;

 color = 0;
 if (fld->flags & SEARCH)
	color = 8;
 else
 if (fld == current_folder)
	color = 4;
 else
 if (fld->flags & SYSTEM)
	color = 1;

 sprintf(fldline, "@C%d%-*.*s%*d%*d", color, show_fld[FOLD_LEN], show_fld[FOLD_LEN], fld->name(fld), show_fld[FNUM_LEN], fld->num_msg,  show_fld[FUNR_LEN], fld->unread_num);

 return fldline;
}

char *
get_msg_line(msg)
struct _mail_msg *msg;
{
static char msgline[255];
struct _head_field *fld;
u_long mlen;
char tbuf[18];
char size[6];
char mstat[5];
int color;

 mstat[0] = size[0] = tbuf[0] = '\0';
 color = 0;

 if (msg == current_message)
	color = 4;

 if (show_msg[STAT_LEN]) {
	mstat[2] = '\0';
	mstat[0] = mstat[1] = ' ';
	if (msg->flags & LOCKED)
		mstat[0] = 'L';
	else
	if (msg->flags & UNREAD) {
		color = 1;
		mstat[0] = 'N';   }
	else
	if (msg->flags & MARKED) {
		color = 8;
		mstat[0] = '+';   }
	if (msg->flags & MOVED)
		mstat[0] = 'M';
	else
	if (msg->flags & DELETED)
		mstat[0] = 'D';

	if (msg->flags & URGENT) 
		mstat[1] = '!';
	else
	if (msg->flags & H_ONLY) 
		mstat[1] = '/';
	else
	if (is_mime_msg(msg))
		mstat[1] = 'M';
			}

 if (show_msg[DATE_LEN])
	strftime(tbuf, 16, "%h-%d %H:%M ", localtime(&msg->header->snt_time));

 if (show_msg[SIZE_LEN]) {
	if ((msg->flags & H_ONLY) &&
		(fld = find_field(msg, MIME_C_LENGTH)) )
		mlen = atol(fld->f_line);
	else
		mlen = msg->msg_len;

	if (mlen >= 10000)
		sprintf(size, "%dK ", (int)(mlen / 1024));
	else
		sprintf(size, "%d ", (int)mlen);
			}

 if (msg->flags & S_THREAD)
  sprintf(msgline, "@C%d%*.*s%*s %-*.*s%-*.*s%-*.*s %.*s", color, msg->refs, show_msg[SUBJ_LEN], "", show_msg[STAT_LEN], mstat, show_msg[DATE_LEN], show_msg[DATE_LEN], tbuf, show_msg[SIZE_LEN], show_msg[SIZE_LEN], size, show_msg[FROM_LEN], show_msg[FROM_LEN], (current_folder == sentm) ?  get_short_addr_line(msg->header->To) : get_short_addr_line(msg->header->From), show_msg[SUBJ_LEN], msg->header->Subject ? msg->header->Subject : "");
 else
  sprintf(msgline, "@C%d%*s %-*.*s%-*.*s%-*.*s %.*s", color, show_msg[STAT_LEN], mstat, show_msg[DATE_LEN], show_msg[DATE_LEN], tbuf, show_msg[SIZE_LEN], show_msg[SIZE_LEN], size, show_msg[FROM_LEN], show_msg[FROM_LEN], (current_folder == sentm) ?  get_short_addr_line(msg->header->To) : get_short_addr_line(msg->header->From), show_msg[SUBJ_LEN], msg->header->Subject ? msg->header->Subject : "");

 return msgline;
}

void
display_fld_status(fld)
struct _mail_folder *fld;
{
char ftype[8];

 strcpy(ftype, "MH");

 if (fld->flags & F_IMAP)
	strcpy(ftype, "IMAP");
 
 display_msg(MSG_STAT, NULL, "%s: Type: %s Total messages: %d Unread: %d", fld->name(fld), ftype, fld->num_msg, fld->unread_num);
}

void
redraw_fld(fld)
struct _mail_folder *fld;
{
 if (!fld)
	return;

 if (!(folder_sort & FLD_SORTED)) {
	sort_folders();
	redraw_fld_win();
	return; }

 fl_replace_browser_line(b_folders, get_folder_index(fld) + 1, get_fld_line(fld));
 if (fld == current_folder)
	redraw_msg_win();

 display_fld_status(fld);
}

void
redraw_fld_line(fld)
struct _mail_folder *fld;
{
 if (fld) {
  fl_replace_browser_line(b_folders, get_folder_index(fld) + 1, get_fld_line(fld));
  if (fld == current_folder) {
        fl_set_menu_item_mode(menu_folders,2, (current_folder->flags & SYSTEM) ? FL_PUP_GRAY : FL_PUP_NONE);
        fl_set_menu_item_mode(menu_folders,3, (current_folder->flags & SYSTEM) ? FL_PUP_GRAY : FL_PUP_NONE);
			     }
	}
}

void
redraw_msg(msg)
struct _mail_msg *msg;
{
 if (!msg)
	return;

 if (!msg->header)
	return;

 if (msg->folder != current_folder)
	return;

 fl_replace_browser_line(b_messages, get_msg_index(current_folder, msg) + 1, get_msg_line(msg));

 return;
}

void
redraw_fld_win()
{
int i;

	fl_freeze_form(main_form);
	fl_clear_browser(b_folders);

	if (!(folder_sort & FLD_SORTED))
		sort_folders();

	for( i = 0; i < folders_num; i++)
		fl_add_browser_line(b_folders, get_fld_line(mailbox[i]));

	fl_select_browser_line(b_folders,get_folder_index(current_folder) + 1);

        if (current_folder->flags & SYSTEM)  
                i=FL_PUP_GRAY;
        else
                i=FL_PUP_NONE;

        fl_set_menu_item_mode(menu_folders,2, i);
        fl_set_menu_item_mode(menu_folders,3, i);

	display_fld_status(current_folder);

	fl_unfreeze_form(main_form);
}

void
restore_msg_win()
{
struct _mail_msg *msg;
int i;

	if (!(current_folder->flags & SORTED))
		sort_folder(current_folder);

	i = fl_get_browser_topline(b_messages);
	fl_freeze_form(main_form);
	fl_clear_browser(b_messages);

	msg = current_folder->messages;
	while (msg) {
		fl_add_browser_line(b_messages, get_msg_line(msg));
		msg = msg->next;
		}

	if (!current_message)
		current_message = current_folder->messages;

	if (current_message) {
	   current_index = get_msg_index(current_folder, current_message);
	   fl_select_browser_line(b_messages,current_index + 1);
			     }

	fl_set_browser_topline(b_messages,i);
	fl_unfreeze_form(main_form);
	display_fld_status(current_folder);
}

void
redraw_msg_win()
{
struct _mail_msg *msg;

	fl_freeze_form(main_form);
	fl_clear_browser(b_messages);

	if (!(current_folder->flags & SORTED))
		sort_folder(current_folder);

	msg = current_folder->messages;
	while (msg) {
		fl_add_browser_line(b_messages, get_msg_line(msg));
		msg = msg->next;
		}

	fl_select_browser_line(b_messages,current_index + 1);
	fl_set_browser_topline(b_messages,current_index + 1);
	fl_unfreeze_form(main_form);
	display_fld_status(current_folder);
}

void
redraw_folders()
{
struct _mail_msg *msg;
int i;

	fl_freeze_form(main_form);
	fl_clear_browser(b_folders);

	if (!(folder_sort & FLD_SORTED))
		sort_folders();

	for( i = 0; i < folders_num; i++)
		fl_add_browser_line(b_folders, get_fld_line(mailbox[i]));

	fl_select_browser_line(b_folders,get_folder_index(current_folder) + 1);

        if (current_folder->flags & SYSTEM)  
                i=FL_PUP_GRAY;
        else
                i=FL_PUP_NONE;

        fl_set_menu_item_mode(menu_folders,2, i);
        fl_set_menu_item_mode(menu_folders,3, i);

	fl_clear_browser(b_messages);

	if (!(current_folder->flags & SORTED))
		sort_folder(current_folder);

	if (current_index == 0)
		current_message = current_folder->messages;
	else
		current_message = get_msg_by_index(current_folder, current_index);

	if (!current_message) {
		current_message = current_folder->messages;
		current_index = 0; }

	msg = current_folder->messages;

	while (msg) {
		fl_add_browser_line(b_messages, get_msg_line(msg));
		msg = msg->next;
		}

	display_fld_status(current_folder);

	fl_select_browser_line(b_messages,current_index + 1);
	fl_set_browser_topline(b_messages,current_index + 1);
	fl_unfreeze_form(main_form);
}

struct _mail_msg
*delete_msg(msg)
	struct _mail_msg	*msg;
{
struct _mail_folder		*old_folder, *new_folder;
struct _mail_msg		*nxt_msg;

	if (!msg)
		return NULL;

	if (!(msg->flags & LOCKED)) {

		nxt_msg = msg->next;

		if (msg->folder)
			new_folder = old_folder = msg->folder;
		else
			new_folder = old_folder = trash;

		if (msg->folder && (msg->folder == trash)) {
			msg->flags |= DELETED;
		} else 	{
			msg->flags |= MOVED;
			msg->folder = trash; 
			new_folder = trash;
		}

		if (msg == current_message)
			current_message = nxt_msg;

		old_folder->update(old_folder);
		if (old_folder != new_folder)
			new_folder->update(new_folder);

		current_index = nxt_msg ? get_msg_index(nxt_msg->folder, nxt_msg) : 0;
		redraw_fld(trash);
		redraw_fld_line(old_folder);
		if (old_folder == current_folder)
			restore_msg_win();

		if (nxt_msg && (nxt_msg->flags & LOCKED))
			return NULL;

		return nxt_msg;
	}
	return msg;
}

struct _mail_msg
*next_msg(ur, cmsg)
	int	ur;
	struct _mail_msg 	*cmsg;
{
struct _mail_msg 	*msg;

	if (!cmsg)
		return NULL;

	msg = cmsg->next;

	if (ur) {
		while (msg) {
			if ((msg->flags & UNREAD) && !(msg->flags & LOCKED)) 
			 return msg; 
			msg = msg->next;
				}

		return NULL;
		}
	else {
		while (msg) {
			if (!(msg->flags & LOCKED)) 
			 return msg; 
			msg = msg->next;
			    }
		return NULL;
	     }

	return NULL;

}

struct _mail_msg
*prev_msg(ur, cmsg)
	int	ur;
	struct _mail_msg 	*cmsg;
{
struct _mail_msg 	*msg;
int i;
int index = get_msg_index(cmsg->folder, cmsg);

	if ((index == -1) || (index == 0))
		return NULL;

	for (i = index-1;i >= 0; i--) {
	  msg = get_msg_by_index(cmsg->folder, i);
	  if (!msg)
		return NULL;

	  if (ur) {
	  if ((msg->flags & UNREAD) && !(msg->flags & LOCKED)) 
		return msg;
		} else {
	  if (!(msg->flags & LOCKED)) 
		return msg;
			}
					}

	return NULL;

}

int
send_pup_cb(int i)
{
   cb_menu_send(NULL, i);
   return 100;
}

int
mesg_pup_cb(int i)
{
   if (i == 100)
	return i;

   cb_menu_messages(NULL, i);
   return i;
}

void
cb_messages_dbl(FL_OBJECT *obj, long arg)
{
struct _mail_msg *msg;

	if (!current_message)
		return;

	if (current_folder == draft)
		edit_msg(current_message);
	else
	if (current_folder == template) {
	  if (!display_msg(MSG_QUEST, "Template", "Create new message from template?"))
		return;

	  current_message->flags |= LOCKED;
	  if (!(msg = copy_message(current_message, outbox))) 	{
		display_msg(MSG_WARN,"template","Can not copy message");
		return;						}

	  current_message->flags &= ~LOCKED;
	  msg->flags &= ~LOCKED;
	  edit_msg(msg);
	  redraw_fld_line(outbox);
					}
	else
		view_msg(current_message);

	redraw_fld_line(current_folder);
	redraw_msg(current_message);

	return;
}

void
cb_messages(FL_OBJECT *obj, long arg)
{
/*
int num, i, line;
int x,y;
XEvent *xev;

	line = fl_get_browser(obj);
	num = abs(line) - 1;
	xev = (XEvent *)fl_last_event();

	if ((xev->type == ButtonPress) ||
		(xev->type == ButtonRelease)) 				{

	if (xev->xbutton.state & ControlMask)	{
		fl_select_browser_line(obj, current_index + 1);
		return;
						}

	if (xev->xbutton.state & ShiftMask)
			{
		if (current_index > num) {
			x = num + 1;
			y = current_index + 1;
					 }
		else    {
			x = current_index + 1;
			y =  num + 1;
			}
		fl_freeze_form(main_form);
		for (i = x; i <= y; i++)
			fl_select_browser_line(b_messages, i);
		fl_select_browser_line(obj, current_index + 1);
		fl_unfreeze_form(main_form);
		return;
			}

	if (xev && (xev->xbutton.state & Button3Mask))   {
		fl_freeze_form(main_form);
		fl_dopup(mpup); 
		fl_unfreeze_form(main_form);
		return;
							 }

									}
	return;
*/
}

void
cb_folders(FL_OBJECT *obj, long arg)
{
int		num;
struct _mail_folder *fld;

	num=fl_get_browser(obj);
	num--;

	if (!(fld = get_folder_by_index(num)))
		return;

	display_fld_status(fld);

	return;
}

void
cb_folders_dbl(FL_OBJECT *obj, long arg)
{
int		num;
struct _mail_folder *fld, *ofld;

	num=fl_get_browser(obj);
	num--;

	set_cursor(main_form, XC_watch);
	my_deactivate();
	current_folder->close(current_folder);
	fld = get_folder_by_index(num);
	if (!fld) {
		set_cursor(main_form, 0);
		my_activate();
		return;
		}
	num = fld->unread_num;
	if (!fld->open(fld->name(fld))) {
		current_folder->open(current_folder->name(current_folder));
		my_activate();
		set_cursor(main_form, 0);
		fl_select_browser_line(obj, get_folder_index(current_folder) - 1);
		display_msg(MSG_WARN, "OPEN","Can not open folder %s!", fld->name(fld));
		return;				}

	if (fld != current_folder) {
		current_index = 0;
		current_message = fld->messages;
				   }
	else
		current_message = get_msg_by_index(current_folder, current_index);

	ofld = current_folder;
	current_folder = fld;

	redraw_fld_line(ofld);
	redraw_fld_line(current_folder);
	if (ofld == fld)
		restore_msg_win();
	else
		redraw_msg_win();

	my_activate();
	set_cursor(main_form, 0);
	if (num != fld->unread_num)
		update_title();

	return;
}

void
cb_menu_misc(FL_OBJECT *obj, long arg)
{
int	num;

    num=fl_get_menu(obj);
    switch(num) {
	case 1:	
		fl_deactivate_form(main_form);
		addr_book(NULL);
		fl_activate_form(main_form);
	break;

	case 2:
		fl_deactivate_form(main_form);
		xfm_config();
		fl_activate_form(main_form);
	break;

	case 3:
		fl_deactivate_form(main_form);
		sign_edit();
		fl_activate_form(main_form);
	break;

	case 4:
		fl_deactivate_form(main_form);
		rules_conf();
		fl_activate_form(main_form);
	break;

	case 5:
		get_new_mail(0);
	break;

	case 6:
		log_win();
	break;

	case 7: /* Debug */
	break;

	default:
	break;
		}
}

void
cb_menu_help(FL_OBJECT *obj, long arg)
{
int	num;
struct _mail_msg *msg;
struct _mime_msg *mime;
char buf[255],bugtext[512];

    num=fl_get_menu(obj);
    fl_deactivate_form(main_form);
    switch(num) {
	case 1:	
		help(NULL);
	break;

	case 2:
		about(0);
	break;

	case 3:
		fl_show_message("BUG REPORT", "Please write down detailed information", " about the problem and press send");

		msg = create_message(outbox);
		strcpy(buf, "XF-Mail bug report list <xfmail-bugs@Burka.NetVision.net.il>");
		msg->header->To = get_address(buf);
		msg->header->Subject = strdup("XF-Mail bug report");
		sprintf(bugtext, "XF-Mail %s %s on %s.\nCompiled %s\n\
--------------------------Start of bug report--------------------------------\n\
\n\
--------------------------End of bug report----------------------------------\n"
, VERSION, PATCHLEVEL, XF_OS, COMPILED);

		update_message_text(msg, bugtext);

		sprintf(buf, "%s/.xfmrules", getenv("HOME"));
		if ((mime = attach_file(msg, buf, &mailcap[DEFAULT_MAILCAP], DEFAULT_ENCODING, ATT_NOUPDATE)) != NULL)
			set_mime_descr(mime, "Filtering rules file");

		sprintf(buf, "%s/.xfdefaults", getenv("HOME"));
		if ((mime = attach_file(msg, buf, &mailcap[DEFAULT_MAILCAP], DEFAULT_ENCODING, ATT_NOUPDATE)) != NULL)
			set_mime_descr(mime, "XFdefaults file");

		strcpy(buf, "/tmp/.xfmailrc");
		save_cfg_file(buf, 1);
		if ((mime = attach_file(msg, buf, &mailcap[DEFAULT_MAILCAP], DEFAULT_ENCODING, ATT_NOUPDATE)) != NULL)
			set_mime_descr(mime, "Configuration file");

		if (update_mime(msg) == -1)
			display_msg(MSG_WARN, "Bug report", "Failed to attach configuration file(s)");

		unlink(buf);

		edit_msg(msg);
		redraw_fld(outbox);
	break;

	case 4:
		if (!display_msg(MSG_QUEST, "Are you sure you want to subscribe?", "(mail will be sent to Majordomo@Burka.NetVision.net.il)"))
			break;

		msg = create_message(outbox);
		sprintf(buf, "Majordomo@Burka.NetVision.net.il");
		msg->header->To = get_address(buf);
		update_message_text(msg, "subscribe xfmail\n");
		if ((send_message(msg) == -1))
			display_msg(MSG_WARN, "subscribe", "Failed to send message to Majordomo@Burka.NetVision.net.il");
		else
			fl_show_message("MAILING LIST", "Subscribe request sent to", "Majordomo@Burka.NetVision.net.il");
		redraw_fld(outbox);
		redraw_fld(sentm);
	break;

	case 5:
		if (!display_msg(MSG_QUEST, "Are you sure you want to unsubscribe?", "(mail will be sent to Majordomo@Burka.NetVision.net.il)"))
			break;

		msg = create_message(outbox);
		sprintf(buf, "Majordomo@Burka.NetVision.net.il");
		msg->header->To = get_address(buf);
		update_message_text(msg, "unsubscribe xfmail\n");
		if ((send_message(msg) == -1))
			display_msg(MSG_WARN, "unsubscribe", "Failed to send message to Majordomo@Burka.NetVision.net.il");
		else
			fl_show_message("MAILING LIST", "Unsubscribe request sent to", "Majordomo@Burka.NetVision.net.il");
		redraw_fld(outbox);
		redraw_fld(sentm);
	break;

	default:
	break;
		}
    fl_activate_form(main_form);
}

void
cb_menu_file(FL_OBJECT *obj, long arg)
{
int	num;
struct stat sb;
const char *fname;
FILE *afd;

	if (obj)
		num=fl_get_menu(obj);
	else
		num = arg;

	switch(num) {
		case 1:
			if (!current_message)
				break;
			fl_set_fselector_title("Choose file to save as");
			if (!(fname = fl_show_file_selector("Save as","","","message")))
				break;

			if (stat(current_message->get_file(current_message), &sb) == -1) {
				display_msg(MSG_WARN, "stat", "Failed to stat %s", current_message->get_file(current_message));
				break; }
			if (fastcopy(current_message->get_file(current_message), (char *)fname, &sb) == 1)
				display_msg(MSG_WARN, "copy", "Failed to copy");
		break;

		case 2:
			if (!current_message)
				break;
			fl_set_fselector_title("Choose file to append");
			if (!(fname = fl_show_file_selector("Append to","","","")))
				break;

			if (!(afd = fopen(fname, "a"))) {
				display_msg(MSG_WARN, "add", "Can not open %s", fname);
				break; }

			current_message->print(current_message, afd, 0);
			fclose(afd);
		break;

		case 3:	
			if (b_savecfg(0)<0)
				display_msg(MSG_WARN, NULL, "Configuration Save Failed!");
		break;

		case 4:
				clean_exit();
		break;

		default:

		break;
	}
}

void
cb_button_save(FL_OBJECT *obj, long arg)
{
int header;
FILE *ffd;
struct stat sb;
const char *fname;

 if (!current_message)
	return;

 if (fl_get_button_numb(obj) == 3)
	header = 1;
 else
	header = 0;

 fl_set_fselector_title("Choose file to save as");
 if (!(fname = fl_show_file_selector(header ? "Save message with header as" : "Save message as","","","message")))
	return;

 if (header) {
 if (stat(current_message->get_file(current_message), &sb) == -1) {
	display_msg(MSG_WARN, "stat", "Failed to stat %s", current_message->get_file(current_message));
	return;
								  }
 if (fastcopy(current_message->get_file(current_message), (char *)fname, &sb) == 1)	{
	display_msg(MSG_WARN, "copy", "Failed to save message");
	return;
	}
	     }
 else	{
 if ((ffd = fopen(fname, "w")) == NULL) {
	display_msg(MSG_WARN, "save", "Can not open %s", fname);
	return;
					}
 (*current_message->print_body)(current_message, ffd);
 fclose(ffd);
	}

 return;
}

void
cb_button_abort(FL_OBJECT *obj, long arg)
{
 abortbutton = 1;
}

int
abortpressed()
{
 fl_check_forms();
 if (abortbutton) {
	abortbutton = 0;
	return 1;
	}

 return 0;
}

void
clean_exit()
{
char buf[255];

	if (!init)
		exit(0);

	close_procs();
	msg_form_shutdown();
	save_book();

	sprintf(buf, "%d %d %d %d", (xofft <= main_form->x) ? (main_form->x - xofft) : 0, (yofft <= main_form->y) ? (main_form->y - yofft) : 0, main_form->w, main_form->h);
	b_putcfg_str(conf_name, "placement", buf);
	b_putcfg_str(conf_name, "sfolder", current_folder->name(current_folder));
	b_savecfg(1);

	set_cursor(main_form, XC_watch);
	if (trash && b_getcfg_int(conf_name, "emptytrash", 1)) 
		trash->empty(trash);

	set_cursor(main_form, 0);
	close_all_folders();
	fl_hide_form(main_form);
	fl_free_form(main_form);
	fl_freepup(mpup);
	fl_freepup(spup);
#ifdef MPROF
	mprof_stop();
#endif
	exit(0);
}

void
delete_messages()
{
int num, utitle;
struct _mail_msg *msg;

		if (!current_folder || !current_folder->messages)
			return;

		if (busy)
			return;

		busy = 1;

		my_deactivate();
		set_cursor(main_form, XC_watch);
		utitle = 0;
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)   {
			if (fl_isselected_browser_line(b_messages, num)) {
			if ((msg = get_msg_by_index(current_folder, num - 1)))
				{
			if (msg == current_message)
				current_message = NULL;

			if (!(msg->flags & LOCKED)) {
			if (current_folder == trash)
				msg->flags |= DELETED;
			else 		{
				msg->flags |= MOVED;
				msg->folder = trash; 
				if (msg->flags & UNREAD)
					utitle = 1;
					}
						}
				}
								          }
									    }
		current_folder->update(current_folder);
		redraw_fld_line(current_folder);
		redraw_fld_line(trash);
		if (current_message == NULL) 	{
			current_message = get_msg_by_index(current_folder, current_index);
			if (current_message == NULL) 	{
				current_message = current_folder->messages;
				current_index = 0;
							}
						}

		restore_msg_win();

		if (utitle)
			update_title();
		set_cursor(main_form, 0);
		my_activate();
		busy = 0;

		return;
}

struct _mail_msg
*move_message(msg, fld)
	struct _mail_msg	*msg;
	struct _mail_folder	*fld;
{
struct _mail_folder		*old_folder;
int ind, i;

	if (!msg)
		return msg;

	i = (msg == current_message) ? 1 : 0;

	if (!(msg->flags & LOCKED)) {

		ind = get_msg_index(msg->folder, msg);

		old_folder = msg->folder;

		if (!fld)
			fld = choose_folder();

		if (!fld)
			return msg;

		if (fld == old_folder)
			return msg;

		fld->move(msg, fld);
		if ((fld == trash) || (old_folder == trash))
			update_title();

		msg = get_msg_by_index(old_folder, ind);
		if ((old_folder == current_folder) && i) {
			current_message = msg;
			current_index = msg ? ind : 0;
							  }

		redraw_fld_line(old_folder);
		redraw_fld_line(fld);
		if ((current_folder == fld) || (current_folder == old_folder))
			restore_msg_win();
	}
	return msg;
}

void
copy_messages()
{
int num;
struct _mail_msg *msg;
struct _mail_folder *fld;

		   if (!current_folder)
			return;

		   if (busy)
			return;

		   busy = 1;
		   fld = NULL;

		   set_cursor(main_form, XC_watch);
		   my_deactivate();	

		   for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
		{
			if (abortpressed())
				break;

			if (fl_isselected_browser_line(b_messages, num)) {
			if ((msg = get_msg_by_index(current_folder, num - 1)))
				{
				if (!fld) {
					fld = choose_folder();
					if (!fld) {
		   			  my_activate();	
		   			  set_cursor(main_form, 0);
		   			  busy = 0;
					  return; }
					  }
				if (!(msg->flags & LOCKED))  {
					display_msg(MSG_STAT, NULL, "copying %d", num);
					copy_message(msg, fld); }
				}
								      }
	    }
		   my_activate();
		   redraw_fld_line(fld);
		   if (fld == current_folder)
			redraw_msg_win();
		   update_title();
		   set_cursor(main_form, 0);
		   busy = 0;
}

void
move_messages(fld)
struct _mail_folder *fld;
{
int num, utitle;
struct _mail_msg *msg;

		   if (!current_folder)
			return;

		   if (busy)
			return;
		   busy = 1;

		   utitle = 0;
		   set_cursor(main_form, XC_watch);
		   my_deactivate();	
		   for (num = 1;num <= fl_get_browser_maxline(b_messages); num++) {
			if (fl_isselected_browser_line(b_messages, num)) {
			if ((msg = get_msg_by_index(current_folder, num - 1)))
				{
				if (msg->flags & LOCKED)
					continue;

				if (msg->flags & UNREAD)
					utitle = 1;

				if (!fld) {
					fld = choose_folder();
					fl_check_only_forms();
					if (!fld) {
					  my_activate();
		   			  set_cursor(main_form, 0);
		   			  busy = 0;
					  return; }
					}
				if (current_message == msg)
					current_message = NULL;

				msg->flags |= MOVED;
				msg->folder = fld;
				}
								      }
									    }
		   current_folder->update(current_folder);
		   redraw_fld_line(current_folder);
		   redraw_fld_line(fld);
		   if (current_message == NULL) {
			current_message = get_msg_by_index(current_folder, current_index);
			if (current_message == NULL) {
				current_message = current_folder->messages;					current_index = 0;
						     }
						}

		   restore_msg_win();	

		   if (utitle)
			update_title();
		   my_activate();
		   set_cursor(main_form, 0);
		   busy = 0;
}

void
update_title()
{
static int pres = -1;
int res, i;
char name[64];

	res = 0;

	for ( i = 0; i < folders_num; i++) {
	   if ((mailbox[i] != NULL) && (mailbox[i] != trash))
	    res += mailbox[i]->unread_num;
					    }

	if (res == pres)
		return;

	if (res) {
		sprintf(name, "%d unread", res);
		if (pres) {
			fl_set_form_title(main_form, name);
			XSetIconName(fl_display, main_form->window, name);
			  }
		else
			set_main_icon(1, name);
		}
	else	
		set_main_icon(0, "XFMail");

	pres = res;
	return;
}

void
get_new_mail(local_only)
int local_only;
{
int 	res, popres;
char spool[255];
#ifdef	USE_SOUND
char *p;
#endif

	fl_set_timer(inc_timer, b_getcfg_int(conf_name,"inctime",180));

	if (busy) 
		return; 

	if (supress_errors && is_iconized())
		goto m_count;
			
			my_deactivate();
			busy = 1;
			set_cursor(main_form, XC_watch);
			sprintf(spool, "%s/%s", _PATH_MAILDIR, getenv("USER"));
			res = 0;
			if (b_getcfg_int(conf_name, "spoolretr", 1)) {
			strcpy(spool, b_getcfg_str(conf_name, "spool", spool));
			if ((res=inc_mail(spool, 1)) < 0) {
				display_msg(MSG_WARN, "inc", "Failed to retrieve mail from spool file");
				res = 0; }
									}

			if (local_only)
				goto endretr;

			if (b_getcfg_int(conf_name, "popretr", 0)) {
			 if ((popres = pop_inc()) < 0)
				popres = 0;
			 res += popres;
								   }

endretr:
			my_activate();
			if (res) {
				redraw_fld_win();
				restore_msg_win();
#ifdef	USE_SOUND
				p = b_getcfg_str(conf_name, "play", "");
				if (p && strlen(p)) {
				  if (play_sound(p, 100) == -1)
					ringbell();
						    }
#endif	/* USE_SOUND */
				if (b_getcfg_int(conf_name, "beep", 1))
					ringbell();
				}

m_count:		update_title();

			set_cursor(main_form, 0);
			busy = 0;

			return;
}

void
cb_menu_messages(FL_OBJECT *obj, long arg)
{
int num, i;
struct _mail_msg *msg;
	
	if (arg == 0)
		num=fl_get_menu(obj);
	else
		num = arg;

/*
	fl_deactivate_form(main_form);
*/

	switch(num) {
		case 1:		/* View */
			view_msg(current_message);
			redraw_fld_line(current_folder);
			redraw_msg(current_message);
		break;

		case 2:		/* Edit */
			edit_msg(current_message);
			redraw_fld_line(current_folder);
			redraw_msg(current_message);
		break;

		case 3:		/* Delete */
			delete_messages();
		break;

		case 4:		/* Move */
			move_messages(NULL);
		break;

		case 5:		/* Copy */
			copy_messages();
		break;

		case 6:		/* Mark */
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			{
			if (fl_isselected_browser_line(b_messages, num)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, num - 1)))
			  {
			    if (!(msg->flags & LOCKED))
				msg->flags |= MARKED;
			  }
			 }
			}

		current_folder->update(current_folder);
		restore_msg_win();
		break;

		case 7:		/* Unmark */
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			{
			if (fl_isselected_browser_line(b_messages, num)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, num - 1)))
			   {
			    if (!(msg->flags & LOCKED))
				msg->flags &= ~MARKED;
			   }
			 }
			}

		current_folder->update(current_folder);
		restore_msg_win();
		break;

		case 8:		/* Mark Read/Unread */
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			{
			if (fl_isselected_browser_line(b_messages, num)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, num - 1))
				&& !(msg->flags & LOCKED)) {
			    if  (msg->flags & UNREAD) {
			  	msg->flags &= ~UNREAD;
			        if (msg->folder->unread_num)
				  msg->folder->unread_num--;
				} else 	{
			  	msg->flags |= UNREAD;
			  	msg->folder->unread_num++;
					}
							   }
			 }
			}

		current_folder->update(current_folder);
		restore_msg_win();
		redraw_fld_line(current_folder);
		update_title();
		break;

		case 9:		/* Print */
			set_cursor(main_form, XC_watch);
			lpr_message(current_message);
			set_cursor(main_form, 0);
		break;

		case 10:	/* Bounce (resend) */
			if (!current_message)
				break;

			current_message->flags |= LOCKED;
			if (!(msg = copy_message(current_message, outbox))) {
				display_msg(MSG_WARN, "bounce", "Can not copy message");
				break;	}
			current_message->flags &= ~LOCKED;
			msg->flags &= ~LOCKED;

			discard_address(msg->header->To);
			discard_address(msg->header->Cc);
			discard_address(msg->header->Bcc);

			msg->header->To = NULL;
			msg->header->Cc = NULL;
			msg->header->Bcc = NULL;

			edit_msg(msg);
			redraw_fld_line(outbox);
		break;

		case 11:	/* Apply rules */
		set_cursor(main_form, XC_watch);
		my_deactivate();
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			{
			if (fl_isselected_browser_line(b_messages, num)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, num - 1)))
				{
				if (!(msg->flags & LOCKED))
					apply_rule(msg, 1);
				}
			 }
			}

		current_folder->update(current_folder);
		my_activate();
		redraw_folders();
		set_cursor(main_form, 0);
		break;

		case 12:	/* Select all */
		fl_freeze_form(main_form);
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			fl_select_browser_line(b_messages, num);
		fl_unfreeze_form(main_form);
		break;

		case 13:	/* Delete from POP server */
		set_cursor(main_form, XC_watch);
		if (pop_init() == -1) {
			set_cursor(main_form, 0);
			break;		}

		i = 0;
		my_deactivate();
	   	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)
			{
			if (abortpressed())
				break;

			if (fl_isselected_browser_line(b_messages, num)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, num - 1)))
				{
				fl_check_only_forms();
				if (pop_delmsg_by_uidl(msg) != -1) {
					i++;
					redraw_msg(msg);
								   }
				}
			 }
			}
		my_activate();
		set_cursor(main_form, 0);
		pop_end();
		if (i)
		  display_msg(MSG_MSG, "POP", "Deleted %d messages from POP server", i);
		else
		  display_msg(MSG_WARN, "POP", "No messages were deleted");

		break;

		default:
		break;
		    }
/*
	fl_activate_form(main_form);
*/
}

int
is_iconized()
{
XWindowAttributes xattr;

   if (b_getcfg_int(conf_name, "incwhenicon", 0)) {
	xattr.map_state = IsUnmapped;
	if (XGetWindowAttributes(fl_display, main_form->window, &xattr))
		{
	 	if (xattr.map_state !=  IsUnmapped) 
			return 1;
		}
						  		    }

   return 0;
}

void
send_all()
{
struct _mail_msg *msg;
int offl, opened;

   fl_set_timer(send_timer, b_getcfg_int(conf_name,"sendtime",0));

   if (busy)
	return;

   if (supress_errors && is_iconized())
	return;

   busy = 1;
   offl = offline;
   offline = 0;
   opened = 0;

   if (!(outbox->flags & OPENED)) {
	if (outbox->open(OUTBOX) != outbox)
		{
		display_msg(MSG_WARN, "send all", "Can not open %s", OUTBOX);
   		offline = offl;
   		busy = 0;
		return;
		}
	else
		opened = 1;
				}

  set_cursor(main_form, XC_watch);
  my_deactivate();

  msg = outbox->messages;
  while(msg) {
	msg->flags |= NOT_SENT;
	msg = msg->next;
	}

  outbox->update(outbox);

  if (opened)
	outbox->close(outbox);

  my_activate();
  set_cursor(main_form, 0);

  offline = offl;

  redraw_fld(outbox);
  redraw_fld(sentm);

  busy = 0;
  return;
}

void
cb_menu_send(FL_OBJECT *obj, long arg)
{
int	num;

	if (arg == 0)
		num=fl_get_menu(obj);
	else
		num = arg;

	fl_deactivate_form(main_form);
	switch(num) {
		case 1:
			set_cursor(main_form, XC_watch);
			edit_msg(create_message(outbox));
			redraw_fld(outbox);
			set_cursor(main_form, 0);
		break;

		case 2:
			set_cursor(main_form, XC_watch);
			edit_msg(get_reply_msg(current_message, 0));
			redraw_fld(outbox);
			redraw_msg(current_message);
			set_cursor(main_form, 0);
		break;

		case 3:
			set_cursor(main_form, XC_watch);
			edit_msg(get_reply_msg(current_message, 1));
			redraw_fld(outbox);
			redraw_msg(current_message);
			set_cursor(main_form, 0);
		break;

		case 4:
			set_cursor(main_form, XC_watch);
			edit_msg(get_fwd_msg(current_message));
			redraw_fld(outbox);
			set_cursor(main_form, 0);
		break;

		case 5:
			send_all();
		break;

		default:
		break;
		  }
	fl_activate_form(main_form);
}

void
cb_menu_folders(FL_OBJECT *obj, long arg)
{
int	num,res, i;
char	*str, *p;
struct _mail_folder *fld;
struct _mail_msg *msg;
char *mbox;
FILE *mfd;

	if (busy)
		return;

	busy = 1;
	
	if (arg == 0)
		num = fl_get_menu(obj);
	else
		num = arg;

	fl_deactivate_form(main_form);
	switch(num) {
		case 1:
			str=(char *)fl_show_input("Enter new folder name:","");
			if (strlen(str) < 1)
				break;
			p = str;
			while (*p != '\0') {
			  if (!isgraph(*p)) {
				display_msg(MSG_WARN, "create", "Illegal character in folder name");
				busy = 0;
				str = NULL;
				break;
					    }
			  p++;	
					   }

			if (str == NULL)
				break;

			str=rem_tr_space(str);
			if (get_folder_by_name(str)) {
				display_msg(MSG_WARN, "create", "folder %s already exists", str);
				break;	}

			if (!(fld = open_folder(str))) {
				display_msg(MSG_WARN, "create", "Can not create folder %s", str);
				break; }

			fld->close(fld);
			redraw_fld_win();
			break;
		case 2:
			res=display_msg(MSG_QUEST, "WARNING", "Folder %s with all messages will be deleted", current_folder->name(current_folder));
			if (res) {
				set_cursor(main_form, XC_watch);
				my_deactivate();
				current_folder->delete(current_folder);
				current_folder = inbox;
				current_message = inbox->messages;
				current_index = 0;
				my_activate();
				set_cursor(main_form, 0);
				redraw_folders();
			}
			break;
		case 3:
			fl_deactivate_all_forms();
			str=(char *)fl_show_input("Enter new name:",current_folder->name(current_folder));

			p = str;
			while (*p != '\0') {
			  if (!isgraph(*p)) {
				display_msg(MSG_WARN, "create", "Illegal character in folder name");
				fl_activate_all_forms();
				busy = 0;
				break;
					    }
			  p++;	
					   }

			if (current_folder->rename(current_folder, str) == -1) {
				display_msg(MSG_WARN, "rename", "Can not rename folder %s to %s", current_folder->name(current_folder), str);
				fl_activate_all_forms();
				break;	}

			redraw_fld_win();
			fl_activate_all_forms();
			break;
		case 4:
			set_cursor(main_form, XC_watch);
			my_deactivate();
			if ((fld = choose_folder())) {
				fl_check_only_forms();
				if (fld->open(fld->name(fld)) != fld)   {
					my_activate();
					set_cursor(main_form, 0);
					break;				}

				current_folder->close(current_folder);
				current_folder = fld;
				current_message = current_folder->messages;
				current_index = 0; 
				redraw_folders(); 
						    }
			my_activate();
			set_cursor(main_form, 0);

			break;

		case 5:
			set_cursor(main_form, XC_watch);
			my_deactivate();
			trash->empty(trash);
			if (trash->num_msg) 
				display_msg(MSG_WARN, "empty trash" ,"Can not delete all messages in trash");
			redraw_fld_line(trash);
			if (current_folder == trash) {
				redraw_msg_win();
				current_index = 0;
				current_message = NULL; }
			my_activate();
			set_cursor(main_form, 0);
			display_fld_status(current_folder);
			break;

		case 6:
			set_cursor(main_form, XC_watch);
			msg = current_folder->messages;
			while (msg) {
			   if (msg->flags & UNREAD) {
				msg->flags &= ~UNREAD;
				if (current_folder->unread_num)
					current_folder->unread_num--;
				msg->update(msg);
						   }
				msg = msg->next;
				    }

			redraw_folders();
			set_cursor(main_form, 0);
			update_title();
			break;

		case 7:
			search_msgs();
			break;

		case 8:
			set_cursor(main_form, XC_watch);
			my_deactivate();
			for (i = 0; i< folders_num;i++) {
			   if (abortpressed())
				break;

			   if ((mailbox[i] != NULL) && (mailbox[i] != current_folder)) {
				if (!mailbox[i]->open(mailbox[i]->name(mailbox[i])) )
					break;
				mailbox[i]->close(mailbox[i]);
	}
							}
			redraw_folders();
			my_activate();
			set_cursor(main_form, 0);
			break;

		case 9:
			my_deactivate();
			fl_set_fselector_title("Choose mailbox file");
			mbox = (char *)fl_show_fselector("Select mailbox", getenv("HOME"), "*", "mbox");
			if (!mbox) {
				my_activate();
				break;
				   }

			set_cursor(main_form, XC_watch);
			i = inc_mail(mbox, 0);
			redraw_folders();
			set_cursor(main_form, 0);

			if (i == 0)	{
				display_msg(MSG_WARN, "Import", "Nothing to import in %s", mbox);
				my_activate();
				break;
					}

			if (i == -1)	{
				display_msg(MSG_WARN, "Import", "Failed to import from %s", mbox);
				my_activate();
				break;
					}

			display_msg(MSG_MSG, "Import", "Imported %d messages from %s", i, mbox);
			my_activate();
			break;

		case 10:
			fl_set_fselector_title("Choose mailbox file");
			mbox = (char *)fl_show_fselector("Select mailbox", getenv("HOME"), "*", "mbox");
			if (!mbox)
				break;

			if ((mfd = fopen(mbox, "a+")) == NULL) {
				display_msg(MSG_WARN, "Export", "Failed to open %s", mbox);
				break; }

			set_cursor(main_form, XC_watch);

	   		for (i = 1;i <= fl_get_browser_maxline(b_messages); i++)
			{
			if (fl_isselected_browser_line(b_messages, i)) 
			 {
			  if ((msg = get_msg_by_index(current_folder, i - 1)))
			  {
			   display_msg(MSG_STAT, NULL, "Exporting %s", msg->header->Subject ? msg->header->Subject : "* No Subject *");
			   set_status_by_flags(msg);
			   fprintf(mfd, "From %s  %s", get_addr_addr(msg->header->From), ctime(&msg->header->rcv_time) );
			   msg->print(msg, mfd, 1);
			   fputc('\n', mfd);
			  }
			 }
			}

			set_cursor(main_form, 0);
			display_msg(MSG_STAT, NULL, "");

			fclose(mfd);
			break;

		default:
			break;
	}

	busy = 0;
	fl_activate_form(main_form);
}

void
cb_button_edit(FL_OBJECT *obj, long arg)
{
	edit_msg(current_message);
	redraw_folders();
}

void
cb_button_exit(FL_OBJECT *obj, long arg)
{
	clean_exit();
}

void
cb_button_newmail(FL_OBJECT *obj, long arg)
{
	edit_msg(create_message(outbox)); 
	redraw_fld(outbox);
}

void
cb_button_print(FL_OBJECT *obj, long arg)
{
	set_cursor(main_form, XC_watch);
	lpr_message(current_message);
	set_cursor(main_form, 0);
}

void
cb_button_view(FL_OBJECT *obj, long arg)
{
	view_msg(current_message);
	redraw_fld(current_folder);
}

void
cb_button_reply(FL_OBJECT *obj, long arg)
{
	set_cursor(main_form, XC_watch);
	if (fl_get_button_numb(obj) == 3)
	 edit_msg(get_reply_msg(current_message, 0));
	else
	 edit_msg(get_reply_msg(current_message, 1));
	redraw_fld(outbox);
	redraw_msg(current_message);
	set_cursor(main_form, 0);
}

void
cb_button_forward(FL_OBJECT *obj, long arg)
{
	set_cursor(main_form, XC_watch);
	edit_msg(get_fwd_msg(current_message));
	redraw_fld(outbox);
	set_cursor(main_form, 0);
}

void
cb_button_move(FL_OBJECT *obj, long arg)
{
struct _mail_folder *fld;

	if (fl_get_button_numb(obj) == 3) {
		if ((fld = pup_choose_folder(main_form)) == NULL)
			return;
					  }
	else
		fld = NULL;

	move_messages(fld);
}

void
cb_button_retr(FL_OBJECT *obj, long arg)
{
   if (fl_get_button_numb(obj) == 3)
	get_new_mail(1);
   else
	get_new_mail(0);
}

void
cb_button_delete(FL_OBJECT *obj, long arg)
{
		delete_messages();
}

void
cb_button_send(FL_OBJECT *obj, long arg)
{
int num, offl;
struct _mail_msg *msg;
int answer;

	if (!current_folder)
		return;

	if (busy)
		return;

	answer = 0;

	offl = offline;
	if (current_folder == outbox)
		offline = 0;

	set_cursor(main_form, XC_watch);
	for (num = 1;num <= fl_get_browser_maxline(b_messages); num++)   {
	  if (fl_isselected_browser_line(b_messages, num)) {

		if (!answer && !display_msg(MSG_QUEST, "Send", "Are you sure you want to send selected message(s)?")) 	{
		   set_cursor(main_form, 0);
		   offline = offl;
		   return;		}

		answer = 1;

		if ((msg = get_msg_by_index(current_folder, num - 1)))
			msg->flags |= NOT_SENT;
		else
			continue;

		if (msg->flags & LOCKED)
			continue;

							   }
									 }

	current_folder->update(current_folder);
	offline = offl;
	redraw_folders();
	set_cursor(main_form,0);
	return;
}

void
cb_timer(FL_OBJECT *obj, long arg)
{
supress_errors = 1;
get_new_mail(0);
supress_errors = 0;
}

void
cb_send_timer(FL_OBJECT *obj, long arg)
{
supress_errors = 1;
send_all();
supress_errors = 0;
}

void
cb_trash_timer(FL_OBJECT *obj, long arg)
{
fl_set_timer(obj, b_getcfg_int(conf_name,"trashtime",0) * 60);
if (busy)
	return;

if (is_iconized())
	return;

supress_errors = 1;
busy = 1;
trash->empty(trash);
redraw_fld(trash);
if (current_folder == trash) {
	current_index = 0;
	current_message = NULL;
	redraw_msg_win();
			     }
busy = 0;
supress_errors = 0;
}
/************************************* HACKED!! ************************************************/
FL_FORM *create_main_form(void)
{
  FL_OBJECT *obj;
  FL_FORM *f_mail;
  int menubox_bottom = 2;
  int menus_y_offset = menubox_bottom + 4;
  int buttons_y_bottom = menubox_bottom + 31;
  int browserbottoms = 70;
  int browserheights = 510;

  f_mail = fl_bgn_form(FL_NO_BOX, 778, 615);
  obj = fl_add_box(FL_FLAT_BOX,0,0,778,615,"");
/*
  fl_set_object_resize(obj,FL_RESIZE_Y);
  fl_set_object_gravity(obj,NorthWestGravity,0);
*/
  obj = fl_add_browser(FL_HOLD_BROWSER,5, browserbottoms+7,180,browserheights-7,""); /* bingo! */
  fl_set_object_resize(obj,FL_RESIZE_Y);
  fl_set_object_gravity(obj,NorthWestGravity,SouthGravity);
  fl_set_browser_fontsize(obj, b_getcfg_int(conf_name, "FoldersFSize", FL_MEDIUM_SIZE));
  fl_set_browser_fontstyle(obj, b_getcfg_int(conf_name, "FoldersFStyle", FL_FIXEDBOLD_STYLE));
  fl_set_object_color(obj,b_getcfg_int(conf_name, "FoldersBgCol", FL_TOP_BCOL), b_getcfg_int(conf_name, "FoldersFgCol", FL_YELLOW));
  fl_set_object_callback(obj,cb_folders,0);
  fl_set_browser_dblclick_callback(obj, cb_folders_dbl, 0);
  b_folders=obj;

  obj = fl_add_browser(FL_MULTI_BROWSER,190,browserbottoms+7,580,browserheights-7,"");	/* bingo! */
  fl_set_object_resize(obj,FL_RESIZE_ALL);
  fl_set_object_gravity(obj,NorthWestGravity,SouthEastGravity);
  fl_set_browser_fontsize(obj, b_getcfg_int(conf_name, "MessagesFSize", FL_MEDIUM_SIZE));
  fl_set_browser_fontstyle(obj, b_getcfg_int(conf_name, "MessagesFStyle", FL_FIXED_STYLE));
  fl_set_object_color(obj,b_getcfg_int(conf_name, "MessagesBgCol", FL_TOP_BCOL), b_getcfg_int(conf_name, "MessagesFgCol", FL_YELLOW));
  fl_set_object_callback(obj,cb_messages,0);
  fl_set_browser_dblclick_callback(obj, cb_messages_dbl, 0);
  b_messages=obj;

  status = obj = fl_add_input(FL_NORMAL_INPUT,5,583,765,27,"");
  fl_set_object_lsize(obj, FL_SMALL_SIZE);
  fl_set_object_resize(obj,FL_RESIZE_X);
  fl_set_object_gravity(obj,SouthWestGravity,SouthEastGravity);
  fl_set_object_boxtype(obj,FL_DOWN_BOX);
  fl_set_object_color(obj,FL_MCOL,FL_BLUE);
  fl_set_input_color(status, FL_BLUE, FL_BLUE);
  fl_deactivate_object(status);
 
  obj = fl_add_box(FL_UP_BOX,1,menubox_bottom,777,29,"");  /* La barre de menu*/

  fl_set_object_resize(obj,FL_RESIZE_X);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,10,menus_y_offset,30,20,"File");
    menu_file=obj;
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lalign(obj,FL_ALIGN_LEFT);
    fl_set_menu(obj,"Save to File (Alt+S)|Append To File%l|Save Config%l|Exit (Alt+X)");
    fl_set_object_callback(obj,cb_menu_file,0);
    fl_set_object_shortcut(obj, "#L", 1);

    fl_set_menu_item_shortcut(obj, 1, "Ss#s");
    fl_set_menu_item_shortcut(obj, 2, "Aa#a");
    fl_set_menu_item_shortcut(obj, 3, "Cc#c");
    fl_set_menu_item_shortcut(obj, 4, "Xx#xQq#q");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,50,menus_y_offset,70,20,"Message");
    menu_message=obj;
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_menu(obj,MESG_MENU);
    fl_set_object_callback(obj,cb_menu_messages,0);
    fl_set_object_shortcut(obj, "#G", 1);

    fl_set_menu_item_shortcut(obj, 1, "Vv#v");
    fl_set_menu_item_shortcut(obj, 2, "Dd#d");
    fl_set_menu_item_shortcut(obj, 3, "Ee#e");
    fl_set_menu_item_shortcut(obj, 4, "Mm#m");
    fl_set_menu_item_shortcut(obj, 5, "Cc#c");
    fl_set_menu_item_shortcut(obj, 6, "Rr#r");
    fl_set_menu_item_shortcut(obj, 7, "Uu#u");
    fl_set_menu_item_shortcut(obj, 8, "Kk#k");
    fl_set_menu_item_shortcut(obj, 9, "Pp#p");
    fl_set_menu_item_shortcut(obj,10, "Bb#b");
    fl_set_menu_item_shortcut(obj,11, "Aa#a");
    fl_set_menu_item_shortcut(obj,12, "Ss#s");
    fl_set_menu_item_shortcut(obj,13, "Ff#f");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,120,menus_y_offset,70,20,"Folder");
    menu_folders=obj;
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_menu(obj,FOLD_MENU);
    fl_set_object_callback(obj,cb_menu_folders,0);
    fl_set_object_shortcut(obj, "#O", 1);

    fl_set_menu_item_shortcut(obj, 1, "Cc#c");
    fl_set_menu_item_shortcut(obj, 2, "Dd#d");
    fl_set_menu_item_shortcut(obj, 3, "Rr#r");
    fl_set_menu_item_shortcut(obj, 4, "Gg#g");
    fl_set_menu_item_shortcut(obj, 5, "Ee#e");
    fl_set_menu_item_shortcut(obj, 6, "Uu#u");
    fl_set_menu_item_shortcut(obj, 7, "Ss#s");
    fl_set_menu_item_shortcut(obj, 8, "Aa#a");
    fl_set_menu_item_shortcut(obj, 9, "Ii#i");
    fl_set_menu_item_shortcut(obj,10, "Xx#x");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,190,menus_y_offset,40,20,"Send");
    menu_send=obj;
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_menu(obj,SEND_MENU);
    fl_set_object_callback(obj,cb_menu_send,0);

    fl_set_menu_item_shortcut(obj, 1, "Cc#cNn#n");
    fl_set_menu_item_shortcut(obj, 2, "Ss#s");
    fl_set_menu_item_shortcut(obj, 3, "Rr#r");
    fl_set_menu_item_shortcut(obj, 4, "Ff#f");
    fl_set_menu_item_shortcut(obj, 5, "Aa#a");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,250,menus_y_offset,40,20,"Misc");
    menu_misc=obj;
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_menu(obj,"Address Book|Config|Signature|Rules|Retrieve Mail (Alt+N)|Log");
    fl_set_object_callback(obj,cb_menu_misc,0);
    fl_set_object_shortcut(obj, "#I", 1);

    fl_set_menu_item_shortcut(obj, 1, "Aa#a");
    fl_set_menu_item_shortcut(obj, 2, "Cc#c");
    fl_set_menu_item_shortcut(obj, 3, "Ss#s");
    fl_set_menu_item_shortcut(obj, 4, "Rr#r");
    fl_set_menu_item_shortcut(obj, 5, "mM#mnN#n");
    fl_set_menu_item_shortcut(obj, 6, "Ll#l");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthWestGravity,0);

  obj = fl_add_menu(FL_PULLDOWN_MENU,725,menus_y_offset,40,20,"Help");
    fl_set_object_boxtype(obj,FL_FLAT_BOX);
    fl_set_object_lalign(obj,FL_ALIGN_RIGHT);
    fl_set_menu(obj,"Contents|About|Bug report|Subscribe to mailing list|Unsubscribe from mailing list");
    fl_set_object_callback(obj,cb_menu_help,0);

    fl_set_menu_item_shortcut(obj, 1, "Cc#c");
    fl_set_menu_item_shortcut(obj, 2, "Aa#a");
    fl_set_menu_item_shortcut(obj, 3, "Bb#b");
    fl_set_menu_item_shortcut(obj, 4, "Ss#s");
    fl_set_menu_item_shortcut(obj, 5, "Uu#u");

  fl_set_object_resize(obj,FL_RESIZE_NONE);
  fl_set_object_gravity(obj,NorthGravity,EastGravity);

    /*
     * Create button pad.
     */

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,20,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,exit_xpm);
  obj->u_vdata = "Exit XF-Mail";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_exit,0);
	fl_set_button_shortcut(obj, "^QQq^XXx#Q#X", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,80,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,edit_xpm);
  obj->u_vdata = "Edit message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_edit,0);
	fl_set_button_shortcut(obj, "^EEe#E", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,125,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,newmail_xpm);
  obj->u_vdata = "Compose new";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_newmail,0);
	fl_set_button_shortcut(obj, "^CCc#C", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,170,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,view_xpm);
  obj->u_vdata = "View message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_view,0);
	fl_set_button_shortcut(obj, "^VVv#V", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,215,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,print_xpm);
  obj->u_vdata = "Print message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_print,0);
	fl_set_button_shortcut(obj, "^PPp#P", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,260,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,reply_xpm);
  obj->u_vdata = "Reply to message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_reply,0);
	fl_set_button_shortcut(obj, "^RRr#R", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,305,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,forward_xpm);
  obj->u_vdata = "Forward message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_forward,0);
	fl_set_button_shortcut(obj, "^FFf#F", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,350,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,move_xpm);
  obj->u_vdata = "Move to another folder";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_move,0);
	fl_set_button_shortcut(obj, "Mm#M", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,395,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,delete_xpm);
  obj->u_vdata = "Delete";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_delete,0);
	fl_set_button_shortcut(obj, "^DDd#D", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,440,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,retr_xpm);
  obj->u_vdata = "Check new mail";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_retr,0);
	fl_set_button_shortcut(obj, "^NNn#N", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,485,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,send_xpm);
  obj->u_vdata = "Send message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_send,0);
	fl_set_button_shortcut(obj, "^KKk#K", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,530,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,save_xpm);
  obj->u_vdata = "Save message";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_save,0);
	fl_set_button_shortcut(obj, "^SSs#S", 0);

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,575,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,abort_xpm);
  obj->u_vdata = "Abort";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_button_abort,0);
	fl_set_button_shortcut(obj, "^AAa#A", 0);
  abrt = obj;

  obj = fl_add_pixmapbutton(FL_NORMAL_BUTTON,620,buttons_y_bottom,41,41,"");
  fl_set_pixmapbutton_data(obj,help_xpm);
  obj->u_vdata = "Help";
	fl_set_object_posthandler(obj, post_handler);
	fl_set_object_resize(obj,FL_RESIZE_NONE);
	fl_set_object_gravity(obj,NorthWestGravity,0);
	fl_set_object_callback(obj,cb_help_button,0);
	fl_set_button_shortcut(obj, "^HHh#H", 0);

  inc_timer = obj = fl_add_timer(FL_HIDDEN_TIMER,0,0,1,2,"");
	fl_set_object_callback(obj,cb_timer,0);

  send_timer = obj = fl_add_timer(FL_HIDDEN_TIMER,0,0,1,2,"");
	fl_set_object_callback(obj,cb_send_timer,0);

  trash_timer = obj = fl_add_timer(FL_HIDDEN_TIMER,0,0,1,2,"");
	fl_set_object_callback(obj,cb_trash_timer,0);

  fl_end_form();

  return f_mail;
}

int
main_close(form, data)
FL_FORM *form;
void *data;
{
 XIconifyWindow(FL_FormDisplay(main_form),main_form->window,0);
 return FL_IGNORE;
}

void
init_forms()
{
	inc_init();
	choose_init();
	pop_account_init();
	search_init();
	log_init();
}

char *
getmyhostname()
{
char host_n[64], *p, **ha;
struct hostent *hp;
static char myname[127] = { '\0' };

if (strlen(myname) > 1)
	return myname;

if (cfg_exist("hostname")) 	{
 p = b_getcfg_str(conf_name, "hostname", "");
 if (p && strlen(p) && strchr(p, '.')) {
	strncpy(myname, p, 126);
	p[126] = '\0';
	return myname;
		}
				}

 if (gethostname(host_n, 63) == -1) 
	strcpy(myname, "localhost");
 else
	strcpy(myname, host_n);

 if ((hp = gethostbyname(host_n)) != NULL) {
	if (strchr(hp->h_name, '.'))
	  strcpy(myname, hp->h_name);
	else	{
	  for (ha = hp->h_aliases; *ha != NULL; ha++)
	  {
		if (strchr(*ha, '.') != NULL)
			{
			(void)strncpy(myname, *ha, 126);
			myname[126] = '\0';
			break;
                        }
	  }
		}

	 				   }

 if (!strchr(myname, '.'))
	display_msg(MSG_WARN, "Xfmail can not determine your host's canonical name!", "Make sure that your E-Mail address in Config/Misc dialog is correct\n(To get rid of this message correct: hostname=<your full hostname>\nin your ~/.xfmailrc");

 b_putcfg_str(conf_name, "hostname", myname);

 return myname;
}

int
main(argc,argv)
int argc;
char **argv;
{
int i;
int x, y, w, h;
char maildir[255];
char menu[64];
char from[255];
char gecos[32];
char user_n[9], *p;
struct passwd *pwd;

printf("XF-Mail %s %s - Xwindows e-mail client for %s\n", VERSION, PATCHLEVEL, XF_OS);
printf("(C) 1995,1996 Gennady B. Sorokopud & Ugen J. S. Antsilevich\n");
printf("Compiled at %s\n", COMPILED);

umask(077);

signal(SIGINT, clean_exit);
signal(SIGTERM, clean_exit);
signal(SIGHUP, SIG_IGN);
signal(SIGQUIT, SIG_IGN);

p = getenv("XFMAILRC");
sprintf(conf_name, "%s/%s", getenv("HOME"), p ? p : ".xfmailrc");

if (b_loadcfg(conf_name) < 0)
	display_msg(MSG_WARN, "init", "Can not load configuration file %s", conf_name);

read_xfdefaults();

if (!fl_initialize(&argc,argv, "XFmail", 0, 0)) {
	fprintf(stderr, "Failed to initialize display\n");
	exit(1);
						}

iconPix = fl_create_from_pixmapdata(fl_default_window(), icon_mail_xpm, &i, &i, NULL, &i, &i, 0);
iconPix_noMail = fl_create_from_pixmapdata(fl_default_window(), icon_nomail_xpm, &i, &i, NULL, &i, &i, 0);
icon_MailEdit = fl_create_from_pixmapdata(fl_default_window(), mail_edit, &i, &i, NULL, &i, &i, 0);

restore_fonts();
restore_charsets();

fl_set_goodies_font(FL_NORMAL_STYLE,FL_DEFAULT_SIZE);

if (xfdefs.popfsize > 0)
	fl_setpup_fontsize(xfdefs.popfsize);
if (xfdefs.popfstyle > 0)
	fl_setpup_fontstyle(xfdefs.popfstyle);
if (xfdefs.popcursor > 0)
	fl_setpup_default_cursor(xfdefs.popcursor);
if ((xfdefs.popbgcol > 0) && (xfdefs.popfgcol > 0))
	fl_setpup_color(xfdefs.popbgcol, xfdefs.popfgcol);
else
if (xfdefs.popbgcol > 0)
	fl_setpup_color(xfdefs.popbgcol, FL_BLACK);
else
if (xfdefs.popfgcol > 0)
	fl_setpup_color(FL_COL1, xfdefs.popfgcol);

main_form=create_main_form();
/*
fl_set_app_mainform(main_form);
*/
init_forms();
init_procs();

sscanf(b_getcfg_str(conf_name, "placement", "100 100 800 415"), "%d%d%d%d", &x, &y, &w, &h);
fl_set_initial_placement(main_form, x, y, w, h);
fl_set_form_minsize(main_form, 600, 300);
fl_set_form_maxsize(main_form, 950, 800);

fl_set_atclose(main_close, NULL);

fl_set_object_dblbuffer(b_folders, 1);
fl_set_object_dblbuffer(b_messages, 1);

fl_show_form(main_form, b_getcfg_int(conf_name, "starticon", 0) ? FL_PLACE_ICONIC | FL_FREE_SIZE : FL_PLACE_FREE, FL_FULLBORDER, "XFMail");
xofft = main_form->x - x;
yofft = main_form->y - y;

fl_deactivate_form(main_form);
set_cursor(main_form, XC_watch);

if (b_getcfg_int(conf_name, "startlog", 0)) 
	log_win(); 
else
	fl_set_menu_item_mode(menu_misc, 6, FL_PUP_BOX);

update_title();

sprintf(menu, "Send%%F%%t|%s", SEND_MENU);
spup = fl_defpup(main_form->window, menu, send_pup_cb);

sprintf(menu, "Message%%F%%t|%s|Send%%m", MESG_MENU);
mpup = fl_defpup(main_form->window, menu, mesg_pup_cb, spup);

fl_setpup_shadow(mpup, 0);
fl_setpup_shadow(spup, 0);
fl_setpup_softedge(mpup, 0);
fl_setpup_softedge(spup, 0);

#ifdef noneed
about(1);
#endif

init = 1;

if (cfg_exist("from"))
	cfg_check_version();

init = 0;

pwd = getpwuid(getuid());
gecos[0] = '\0';
if (pwd) {
 strncpy(gecos, pwd->pw_gecos, 31);
 strcpy(user_n, pwd->pw_name);
 for (i = 0; i < strlen(gecos); i++) {
    if (!isalpha(gecos[i]) && (gecos[i] != ' ') && (gecos[i] != '.')) {
	 gecos[i] = '\0';
	 break; }
				}
	}
else {
 strcpy(gecos, "");
 strcpy(user_n, "UNKNOWN");
	}

sprintf(sender_name, "%s@%s", user_n, getmyhostname());

/* running at the first time */
if (!cfg_exist("from")) {

if (strlen(gecos) > 1)
	sprintf(from, "%s <%s@%s>", gecos, user_n, getmyhostname());
else
	sprintf(from, "%s@%s", user_n, getmyhostname());

b_getcfg_str(conf_name, "from", from);

misc_conf();
inc_conf();
smtp_conf();
if (b_getcfg_int(conf_name, "popretr", 0))
	pop_conf();

b_putcfg_str(conf_name, "xfversion", VERSION);

b_savecfg(0);
}
else 	{
struct _mail_addr *ma;

 if (((ma = get_address(b_getcfg_str(conf_name, "from", ""))) == NULL) ||
	((p = strchr(ma->addr, '@')) == NULL) ||
	(strchr(p, '.') == NULL)) {
	display_msg(MSG_WARN, "Your E-Mail address (From field)", "is invalid, please fix it");
	misc_conf();
				   }

 if (ma)
	discard_address(ma);
	}

msg_form_init();

init = 1;

display_msg(MSG_LOG, "startup", "Opening all folders");

sort_type = b_getcfg_int(conf_name, "sort", NO_SORT | MSG_ASCEND);
folder_sort = b_getcfg_int(conf_name, "fsort", NO_SORT | MSG_ASCEND);

sprintf(maildir, "%s/Mail", getenv("HOME"));
if (open_all_folders(b_getcfg_str(conf_name, "maildir", maildir)) == -1) {
	display_msg(MSG_FATAL, "init", "initialzation failed");
	exit(1); }

display_msg(MSG_LOG, "startup", "Cleaning temp folder");
ftemp->empty(ftemp);

display_msg(MSG_LOG, "startup", "Loading rules database");
if (load_rules() == -1)
	display_msg(MSG_WARN, "init", "Failed to load rules");

display_msg(MSG_LOG, "startup", "Loading default address book");
if (load_book(NULL) == -1)
	display_msg(MSG_WARN, "init", "Failed to load address book");

display_msg(MSG_LOG, "startup", "Loading signatures");
load_signatures();

display_msg(MSG_LOG, "startup", "Loading mailcap");
load_mailcap();

show_msg[STAT_LEN] = b_getcfg_int(conf_name, "statlen", 2);
show_msg[DATE_LEN] = b_getcfg_int(conf_name, "datelen", 13);
show_msg[SIZE_LEN] = b_getcfg_int(conf_name, "sizelen", 5);
show_msg[FROM_LEN] = b_getcfg_int(conf_name, "fromlen", 18);
show_msg[SUBJ_LEN] = b_getcfg_int(conf_name, "subjlen", 100);

show_fld[FOLD_LEN] = b_getcfg_int(conf_name, "foldlen", 11);
show_fld[FNUM_LEN] = b_getcfg_int(conf_name, "fnumlen", 4);
show_fld[FUNR_LEN] = b_getcfg_int(conf_name, "funrlen", 4);

if (!(current_folder = get_folder_by_name(b_getcfg_str(conf_name, "sfolder", INBOX))))
	current_folder = inbox;
current_folder->open(current_folder->name(current_folder));
current_message = current_folder->messages;
current_index = 0;
offline = b_getcfg_int(conf_name, "smtpoffl", 0);

redraw_folders();

if (!b_getcfg_int(conf_name,"popstorpwd", 0))
	b_setcfg_flags("popasswd", DONT_STORE);

b_setcfg_flags("popasswd", CF_UNSECURE);
b_setcfg_flags("placement", CF_NOTCHANGED);
b_setcfg_flags("msgformsize", CF_NOTCHANGED);
b_setcfg_flags("sfolder", CF_NOTCHANGED);

display_msg(MSG_STAT, NULL, "Ready");
display_msg(MSG_LOG, "startup", "Done");

fl_set_timer(inc_timer, b_getcfg_int(conf_name,"inctime",180));
fl_set_timer(send_timer, b_getcfg_int(conf_name,"sendtime",0));
fl_set_timer(trash_timer, b_getcfg_int(conf_name,"trashtime",0) * 60);

fl_activate_form(main_form);
if (b_getcfg_int(conf_name,"inconstart", 0))
	get_new_mail(0);

set_cursor(main_form, 0);
fl_set_object_prehandler(b_messages,preemptive_handler);
fl_set_object_posthandler(b_messages,msg_posthandler);
my_activate();
while(1)
	fl_do_forms();

}
